/*
 * eth.c
 *
 *  Created on: 05.07.2014
 *      Author: AndreR
 */

#include "eth.h"
#include "stm32f4xx_hal.h"
#include "../constants.h"
#include "util/console.h"
#include "errors.h"
#include <string.h>

#define ETH_EVENT_QUEUE_SIZE 4

typedef struct _Eth
{
	ETH_HandleTypeDef hEth;

	uint8_t rxBuffer[ETH_RXBUFNB*ETH_RX_BUF_SIZE] __attribute__((aligned(4)));
	uint8_t txBuffer[ETH_TXBUFNB*ETH_TX_BUF_SIZE] __attribute__((aligned(4)));
	ETH_DMADescTypeDef rxDescTab[ETH_RXBUFNB] __attribute__((aligned(4)));
	ETH_DMADescTypeDef txDescTab[ETH_TXBUFNB] __attribute__((aligned(4)));

	Mailbox eventQueue;
	msg_t eventQueueData[ETH_EVENT_QUEUE_SIZE];

	uint8_t MDI;

	EthLinkCallback cbLink;
	EthRxCallback cbRx;

	EthStats stats;
} Eth;

static Eth eth;

void ETH_IRQHandler(void)
{
	CH_IRQ_PROLOGUE();

	HAL_ETH_IRQHandler(&eth.hEth);

	CH_IRQ_EPILOGUE();
}

void EXTI0_IRQHandler(void)
{
	CH_IRQ_PROLOGUE();

	EXTI->PR = GPIO_PIN_0;

	if(eth.cbLink)
		(*eth.cbLink)();

	CH_IRQ_EPILOGUE();
}

void HAL_ETH_TxCpltCallback(ETH_HandleTypeDef * heth)
{
	(void)heth;
}

void HAL_ETH_RxCpltCallback(ETH_HandleTypeDef * heth)
{
	(void)heth;

	if(eth.cbRx)
		(*eth.cbRx)();
}

void HAL_ETH_MspInit(ETH_HandleTypeDef* pEth)
{
	(void)pEth;

	GPIO_InitTypeDef gpioInit;
	gpioInit.Speed = GPIO_SPEED_FAST;
	gpioInit.Mode = GPIO_MODE_AF_PP;
	gpioInit.Pull = GPIO_NOPULL;
	gpioInit.Alternate = GPIO_AF11_ETH;
	gpioInit.Pin = GPIO_PIN_0 | GPIO_PIN_1 | GPIO_PIN_2 | GPIO_PIN_3 | GPIO_PIN_7;
	HAL_GPIO_Init(GPIOA, &gpioInit);

	gpioInit.Pin = GPIO_PIN_0 | GPIO_PIN_1 | GPIO_PIN_8 | GPIO_PIN_10 | GPIO_PIN_11 | GPIO_PIN_12 | GPIO_PIN_13;
	HAL_GPIO_Init(GPIOB, &gpioInit);

	gpioInit.Pin = GPIO_PIN_1 | GPIO_PIN_2 | GPIO_PIN_3 | GPIO_PIN_4 | GPIO_PIN_5;
	HAL_GPIO_Init(GPIOC, &gpioInit);

	gpioInit.Mode = GPIO_MODE_IT_FALLING;
	gpioInit.Pull = GPIO_PULLUP;
	gpioInit.Alternate = 0;
	gpioInit.Pin = GPIO_PIN_0;
	HAL_GPIO_Init(GPIOD, &gpioInit);

	HAL_NVIC_SetPriority(EXTI0_IRQn, IRQL_ETH_LINK, 0);
	HAL_NVIC_EnableIRQ(EXTI0_IRQn);

	__ETHMAC_CLK_ENABLE();
	__ETHMACTX_CLK_ENABLE();
	__ETHMACRX_CLK_ENABLE();

	HAL_NVIC_SetPriority(ETH_IRQn, IRQL_ETH_DATA, 0);
	HAL_NVIC_EnableIRQ(ETH_IRQn);
}

void EthInit()
{
	uint32_t regvalue;
	uint8_t mac[6] = {2, 84, 73, 71, 69, 82}; // ASCII: 02-T-I-G-E-R

	chMBInit(&eth.eventQueue, eth.eventQueueData, ETH_EVENT_QUEUE_SIZE);

	eth.hEth.Instance = ETH;
	eth.hEth.Init.AutoNegotiation = ETH_AUTONEGOTIATION_ENABLE;
	eth.hEth.Init.ChecksumMode = ETH_CHECKSUM_BY_HARDWARE;
	eth.hEth.Init.DuplexMode = ETH_MODE_FULLDUPLEX;
	eth.hEth.Init.MACAddr = mac;
	eth.hEth.Init.MediaInterface = ETH_MEDIA_INTERFACE_MII;
	eth.hEth.Init.PhyAddress = DP83848_PHY_ADDRESS;
	eth.hEth.Init.RxMode = ETH_RXINTERRUPT_MODE;
	eth.hEth.Init.Speed = ETH_SPEED_10M;

	if(HAL_ETH_Init(&eth.hEth) == HAL_OK)
		eth.hEth.LinkStatus = 1;

	HAL_StatusTypeDef halStatus = HAL_ETH_DMARxDescListInit(&eth.hEth, eth.rxDescTab, eth.rxBuffer, ETH_RXBUFNB);
	if(halStatus != HAL_OK)
		while(1);
	halStatus = HAL_ETH_DMATxDescListInit(&eth.hEth, eth.txDescTab, eth.txBuffer, ETH_TXBUFNB);
	if(halStatus != HAL_OK)
		while(1);

	HAL_ETH_ReadPHYRegister(&eth.hEth, PHY_MICR, &regvalue);

	regvalue |= (PHY_MICR_INT_EN | PHY_MICR_INT_OE);

	HAL_ETH_WritePHYRegister(&eth.hEth, PHY_MICR, regvalue );

	HAL_ETH_ReadPHYRegister(&eth.hEth, PHY_MISR, &regvalue);

	regvalue |= PHY_MISR_LINK_INT_EN;

	HAL_ETH_WritePHYRegister(&eth.hEth, PHY_MISR, regvalue);

//	HAL_ETH_WritePHYRegister(&eth.hEth, 0x04, 0x0061);	// 10MBit only
}

void EthSetCallbacks(EthLinkCallback cbLink, EthRxCallback cbRx)
{
	eth.cbLink = cbLink;
	eth.cbRx = cbRx;
}

void EthAllowMAC(uint8_t slot, uint8_t* pMac, uint8_t enable, uint8_t saFiltering)
{
	const uint32_t frReg[] = {ETH_MAC_ADDRESS0, ETH_MAC_ADDRESS1, ETH_MAC_ADDRESS2, ETH_MAC_ADDRESS3};

	if(slot > 3)
		return;

	volatile uint32_t* pMacHigh = (volatile uint32_t *)((uint32_t)(ETH_MAC_ADDR_HBASE + frReg[slot]));
	volatile uint32_t* pMacLow = (volatile uint32_t *)((uint32_t)(ETH_MAC_ADDR_LBASE + frReg[slot]));

	uint32_t hReg = ((uint32_t)pMac[5] << 8) | (uint32_t)pMac[4];
	uint32_t lReg = ((uint32_t)pMac[3] << 24) | ((uint32_t)pMac[2] << 16) | ((uint32_t)pMac[1] << 8) | pMac[0];

	if(slot == 0)
	{
		hReg |= ETH_MACA1HR_AE;	// slot 0 is the own MAC address, must always be on
	}
	else
	{
		if(enable)
			hReg |= ETH_MACA1HR_AE;

		if(saFiltering)
			hReg |= ETH_MACA1HR_SA;
	}

	*pMacHigh = hReg;
	*pMacLow = lReg;
}

int16_t EthSendEthernetFrame(uint8_t* pData, uint32_t length)
{
	volatile ETH_DMADescTypeDef *pTxDMADesc = eth.hEth.TxDesc;

	int16_t result = 0;

	for(uint32_t bytesSent = 0; bytesSent < length; bytesSent += ETH_TX_BUF_SIZE)
	{
		if(pTxDMADesc->Status & ETH_DMATXDESC_OWN)
		{
			chThdSleep(MS2ST(1));
			bytesSent -= ETH_TX_BUF_SIZE;
			result = ERROR_ETH_SEND_DELAYED;
			continue;
		}

		uint32_t sendLen = length-bytesSent;
		if(sendLen > ETH_TX_BUF_SIZE)
			sendLen = ETH_TX_BUF_SIZE;

		memcpy((void*)pTxDMADesc->Buffer1Addr, pData, sendLen);

		pData += sendLen;

		pTxDMADesc = (ETH_DMADescTypeDef*)pTxDMADesc->Buffer2NextDescAddr;
	}

	if(HAL_ETH_TransmitFrame(&eth.hEth, length) != HAL_OK)
		ConsolePrint("TransmitFrame HAL error\r\n");

	eth.stats.txFramesProcessed++;
	eth.stats.txBytesProcessed += length;

	return result;
}

uint8_t EthLinkUpdate()
{
	__IO uint32_t timeout = 0;
	uint32_t regvalue = 0;

	// Read PHY_MISR
	HAL_ETH_ReadPHYRegister(&eth.hEth, PHY_MISR, &regvalue);

	// Read PHY_SR
	HAL_ETH_ReadPHYRegister(&eth.hEth, PHY_SR, &regvalue);

	// Check whether the link is up or down
	if((regvalue & PHY_LINK_STATUS) != (uint16_t)RESET)
	{
		eth.hEth.LinkStatus = 1;

		// Restart the auto-negotiation
		if (eth.hEth.Init.AutoNegotiation != ETH_AUTONEGOTIATION_DISABLE)
		{
			// Enable Auto-Negotiation
			HAL_ETH_WritePHYRegister(&eth.hEth, PHY_BCR, PHY_AUTONEGOTIATION);

			// Wait until the auto-negotiation will be completed
			do
			{
				chThdSleep(MS2ST(10));
				timeout++;
				HAL_ETH_ReadPHYRegister(&eth.hEth, PHY_BSR, &regvalue);
			} while (!(regvalue & PHY_AUTONEGO_COMPLETE) && (timeout < PHY_READ_TO));

			if (timeout == PHY_READ_TO)
			{
				goto error;
			}

			// Read the result of the auto-negotiation
			HAL_ETH_ReadPHYRegister(&eth.hEth, PHY_SR, &regvalue);

			// Configure the MAC with the Duplex Mode fixed by the auto-negotiation process
			if ((regvalue & PHY_DUPLEX_STATUS) != (uint32_t) RESET)
			{
				// Set Ethernet duplex mode to Full-duplex following the auto-negotiation
				eth.hEth.Init.DuplexMode = ETH_MODE_FULLDUPLEX;
			}
			else
			{
				// Set Ethernet duplex mode to Half-duplex following the auto-negotiation
				eth.hEth.Init.DuplexMode = ETH_MODE_HALFDUPLEX;
			}

			// Configure the MAC with the speed fixed by the auto-negotiation process
			if (regvalue & PHY_SPEED_STATUS)
			{
				// Set Ethernet speed to 10M following the auto-negotiation
				eth.hEth.Init.Speed = ETH_SPEED_10M;
			}
			else
			{
				// Set Ethernet speed to 100M following the auto-negotiation
				eth.hEth.Init.Speed = ETH_SPEED_100M;
			}

			if(regvalue & PHY_MDI_STATUS)
				eth.MDI = 0;	// normal MDI mode
			else
				eth.MDI = 1; 	// MDI-X mode, TX/RX automatically swapped
		}
		else // AutoNegotiation Disable
		{
			error:
			// Check parameters
			assert_param(IS_ETH_SPEED(EthHandle.Init.Speed));
			assert_param(IS_ETH_DUPLEX_MODE(EthHandle.Init.DuplexMode));

			// Set MAC Speed and Duplex Mode to PHY
			HAL_ETH_WritePHYRegister(&eth.hEth, PHY_BCR, ((uint16_t) (eth.hEth.Init.DuplexMode >> 3) | (uint16_t) (eth.hEth.Init.Speed >> 1)));
		}

		// ETHERNET MAC Re-Configuration
		HAL_ETH_ConfigMAC(&eth.hEth, (ETH_MACInitTypeDef *)NULL);

		// Restart MAC interface
		HAL_ETH_Start(&eth.hEth);
	}
	else
	{
		eth.hEth.LinkStatus = 0;

		// Stop MAC interface
		HAL_ETH_Stop(&eth.hEth);
	}

	if(eth.hEth.LinkStatus)
	{
		ConsolePrint("Link UP\r\n");
		if(eth.hEth.Init.Speed == ETH_SPEED_100M)
			ConsolePrint("Speed: 100M\r\n");
		else
			ConsolePrint("Speed: 10M\r\n");

		if(eth.hEth.Init.DuplexMode == ETH_MODE_FULLDUPLEX)
			ConsolePrint("Full Duplex\r\n");
		else
			ConsolePrint("Half Duplex\r\n");

		if(eth.MDI)
			ConsolePrint("MDI mode\r\n");
		else
			ConsolePrint("MDI-X mode\r\n");
	}
	else
	{
		ConsolePrint("Link DOWN\r\n");
	}

	return eth.hEth.LinkStatus;
}

int16_t EthGetEthernetFrame(uint8_t* pData, uint32_t dataSize, uint32_t* pBytesRead)
{
	int16_t result = ERROR_ETH_EMPTY;
	if(pBytesRead)
		*pBytesRead = 0;

	while(((eth.hEth.RxDesc->Status & ETH_DMARXDESC_OWN) == (uint32_t)RESET))
	{
		HAL_ETH_GetReceivedFrame(&eth.hEth);

		if(eth.hEth.RxFrameInfos.LSRxDesc != NULL)
		{
			uint32_t length = eth.hEth.RxFrameInfos.length;
			uint32_t segments = eth.hEth.RxFrameInfos.SegCount;

			eth.hEth.RxFrameInfos.length = 0;
			eth.hEth.RxFrameInfos.SegCount = 0;
			eth.hEth.RxFrameInfos.LSRxDesc = NULL;

			volatile ETH_DMADescTypeDef *pRxDMADesc;

			if(length <= dataSize)
			{
				pRxDMADesc = eth.hEth.RxFrameInfos.FSRxDesc;

				uint32_t bytesCopied = 0;

				for(uint32_t i = 0; i < segments; i++)
				{
					uint32_t bytesToCopy = length-bytesCopied;
					if(bytesToCopy > ETH_RX_BUF_SIZE)
						bytesToCopy = ETH_RX_BUF_SIZE;

					memcpy(pData+bytesCopied, (void*)pRxDMADesc->Buffer1Addr, bytesToCopy);
					bytesCopied += bytesToCopy;

					// move on to next descriptor
					pRxDMADesc = (ETH_DMADescTypeDef *) (pRxDMADesc->Buffer2NextDescAddr);
				}

				eth.stats.rxFramesProcessed++;
				eth.stats.rxBytesProcessed += length;

				if(pBytesRead)
					*pBytesRead = length;
				result = 0;
			}
			else
			{
				result = ERROR_ETH_RX_TOO_LARGE;
			}

			// Release descriptors to DMA
			pRxDMADesc = eth.hEth.RxFrameInfos.FSRxDesc;
			// Set Own bit in Rx descriptors: gives the buffers back to DMA
			for(uint32_t i = 0; i < segments; i++)
			{
				pRxDMADesc->Status |= ETH_DMARXDESC_OWN;
				pRxDMADesc = (ETH_DMADescTypeDef *) (pRxDMADesc->Buffer2NextDescAddr);
			}

			return result;
		}
	}

	return result;
}

EthStats* EthGetStats()
{
	eth.stats.txFrames = ETH->MMCTGFCR;
	eth.stats.rxFrames = ETH->MMCRGUFCR;

	uint32_t dmamfbocr = ETH->DMAMFBOCR;
	eth.stats.rxMissedApp = (dmamfbocr & 0xFFE0000) >> 17;
	eth.stats.rxMissedCtrl = dmamfbocr & 0xFFFF;

	return &eth.stats;
}
