#include "misc.h"
#include "stm32f4xx_rcc.h"
#include "stm32f4xx_exti.h"
#include "stm32f4xx_gpio.h"
#include "stm32f4xx_iwdg.h"
#include "stm32f4xx_syscfg.h"

#include "core_cm4.h"
#include "ch.h"

#include "hal/led.h"
#include "hal/usart1.h"
#include "hal/eth.h"

#include "util/boot.h"
#include "util/console.h"
#include "util/crc.h"
#include "util/flash_fs.h"

#include "cli.h"
#include "network.h"
#include "wifi.h"
#include "vision.h"

static void NVIC_Configuration();

extern void SystemClock_Config(void);

#define portNVIC_SYSTICK_CTRL		( ( volatile unsigned long * ) 0xe000e010 )
#define portNVIC_SYSTICK_LOAD		( ( volatile unsigned long * ) 0xe000e014 )
#define portNVIC_SYSTICK_CLK		0x00000004
#define portNVIC_SYSTICK_INT		0x00000002
#define portNVIC_SYSTICK_ENABLE		0x00000001

static WORKING_AREA(waCLI, 8192);
static WORKING_AREA(waNetwork, 4096);
static WORKING_AREA(waWifi, 4096);

int main()
{
#ifdef ENV_BOOT
	if(!BootIsBootloaderSelected())
		BootJumpToApplication();

	BootSetBootloaderSelected(0);
#endif

	SystemClock_Config();

	// Enable important clocks
	RCC_AHB1PeriphClockCmd( RCC_AHB1Periph_DMA1 | RCC_AHB1Periph_DMA2, ENABLE);
	RCC_AHB1PeriphClockCmd(	RCC_AHB1Periph_GPIOA | RCC_AHB1Periph_GPIOB | RCC_AHB1Periph_GPIOC |
							RCC_AHB1Periph_GPIOD | RCC_AHB1Periph_GPIOE, ENABLE);
	RCC_APB2PeriphClockCmd(RCC_APB2Periph_SYSCFG, ENABLE);

	SYSCFG_CompensationCellCmd(ENABLE);

	SCB->SHCSR |= SCB_SHCSR_USGFAULTENA_Msk | SCB_SHCSR_BUSFAULTENA_Msk | SCB_SHCSR_MEMFAULTENA_Msk;

	*(portNVIC_SYSTICK_LOAD) = ( 168000000UL / CH_FREQUENCY ) - 1UL;
	*(portNVIC_SYSTICK_CTRL) = portNVIC_SYSTICK_CLK | portNVIC_SYSTICK_INT | portNVIC_SYSTICK_ENABLE;

	chSysInit();

	NVIC_Configuration();

	CPULoadCalibrate(10);
	chThdSleep(20);

	// --- HAL ---
	LEDInit();		// init LEDs to signal RTOS startup failures
	FlashFSInit(0x08060000, 0x20000);	// 128k
	USART1Init();
	CRCInit();

	// --- High-Level Systems ---
	CLIInit();
	VisionInit();
	EthInit();
	NetworkInit();
	WifiInit();

#ifdef ENV_BOOT
	ConsolePrint("\f--- Base Station v2 (Bootloader) ---\r\n");
#else
	ConsolePrint("\f--- Base Station v2 ---\r\n");
#endif

	chThdCreateStatic(waWifi, sizeof(waWifi), NORMALPRIO+10, WifiTask, 0);
	chThdCreateStatic(waNetwork, sizeof(waNetwork), NORMALPRIO+5, NetworkTask, 0);
	chThdCreateStatic(waCLI, sizeof(waCLI), NORMALPRIO-5, CLITask, 0);
/*
	// Configure and start Independent watchdog
	IWDG_WriteAccessCmd(IWDG_WriteAccess_Enable);
	IWDG_SetPrescaler(IWDG_Prescaler_8);
	IWDG_SetReload(2500);	// approx. 500ms
	IWDG_WriteAccessCmd(IWDG_WriteAccess_Disable);
*/

	while(1)
	{
#ifdef ENV_RUN
		LEDToggle(LED_MAIN_GREEN);
#else
		LEDToggle(LED_MAIN_RED);
#endif

		chThdSleep(MS2ST(500));
	}

	return 0;
}

static void NVIC_Configuration()
{
#ifdef ENV_BOOT
	NVIC_SetVectorTable(NVIC_VectTab_FLASH, 0x00);
#else
	NVIC_SetVectorTable(NVIC_VectTab_FLASH, BOOT_APP_OFFSET);
#endif
}
