/*
 * network.c
 *
 *  Created on: 05.01.2013
 *      Author: AndreR
 */

#include "network.h"

#include "wifi.h"
#include "constants.h"
#include "util/console.h"

#include "commands.h"

#include "network/inet.h"
#include "network/ethernet.h"
#include "network/arp.h"
#include "network/ipv4.h"
#include "network/udp.h"
#include "network/icmp.h"
#include "network/igmpv3.h"
#include "vision.h"

#include <string.h>

Network network;

#define EVENT_LINK 0
#define EVENT_RX_DONE 1
#define EVENT_BOT_DATA 0x800

int16_t NetworkSendARP(uint8_t operation, MAC targetMAC, IPv4Address targetIP);
int16_t NetworkSendICMPEchoReply(ICMPEcho* pEchoRequest, uint8_t* pData, uint16_t dataLength, MAC targetMAC, IPv4Address dstIp);

static void linkCallback()
{
	chSysLockFromIsr();
	chMBPostI(&network.eventQueue, EVENT_LINK);
	chSysUnlockFromIsr();
}

static void rxCallback()
{
	chSysLockFromIsr();
	chMBPostI(&network.eventQueue, EVENT_RX_DONE);
	chSysUnlockFromIsr();
}

static void botDataCallback(uint8_t id)
{
	chMBPost(&network.eventQueue, EVENT_BOT_DATA+id, TIME_IMMEDIATE);
}

void NetworkSaveConfig()
{
	FlashFSWrite(&network.pCfgFile, &network.my, sizeof(NetworkConfig));
}

void NetworkInit()
{
	memset(&network, 0, sizeof(Network));

	chMtxInit(&network.sendPacketMutex);
	chMBInit(&network.eventQueue, network.eventQueueData, NETWORK_EVENT_QUEUE_SIZE);

	int16_t result = FlashFSOpenOrCreate("net/cfg", 0, &network.my, sizeof(NetworkConfig), &network.pCfgFile);
	if(result == FLASH_FS_CREATED)
	{
		network.my.mac = MACSet(2, 84, 73, 71, 69, 82); // ASCII: 02-T-I-G-E-R
		network.my.ip = IPv4AddressSet(192, 168, 1, 210);
		network.my.port = 10201;

		FlashFSWrite(&network.pCfgFile, &network.my, sizeof(NetworkConfig));
	}

	network.ethernetIdent = 0;

	network.ucast.mac = MACSet(2, 0, 0, 0, 0, 0);
	network.ucast.ip = IPv4AddressSet(0, 0, 0, 0);
	network.ucast.port = 0;

	network.vision.mac = MACSet(0x01, 0x00, 0x5E, 0x05, 0x17, 0x02);
	network.vision.ip = IPv4AddressSet(224, 5, 23, 2);
	network.vision.port = 10002;

	network.igmp.respCounter = 0;
	network.igmp.respTime = 500;
	network.igmp.lastSendTime = 0;
	network.igmp.robustness = 2;

	MAC igmpMAC = MACSet(0x01, 0x00, 0x5E, 0x00, 0x00, 0x01);	// 224.0.0.1

	EthSetCallbacks(&linkCallback, &rxCallback);
	WifiSetBotDataCallback(&botDataCallback);

	EthAllowMAC(1, igmpMAC.u8, 1, 0);
	EthAllowMAC(2, network.vision.mac.u8, 1, 0);
}

static int16_t unicastUdp(uint8_t* pData, uint16_t dataLength)
{
	(void)dataLength;

	int16_t result;

	// pData should point to a BaseStationACommand packet
	PacketHeader* pHeader = (PacketHeader*)pData;
	if(pHeader->section != SECTION_BASE_STATION)
	{
		ConsolePrint("Invalid section 0x%X\r\n", pHeader->section);
		return 0;
	}

	pData += PACKET_HEADER_SIZE;
	dataLength -= PACKET_HEADER_SIZE;

	switch(pHeader->cmd)
	{
		case CMD_BASE_STATION_ACOMMAND:
		{
			BaseStationACommand* pCmd = (BaseStationACommand*)pData;

			uint16_t botId = pCmd->id;
			pData++;
			dataLength--;

			if(botId > NRF24_MAX_BOT_ID)
			{
				ConsolePrint("Invalid bot id %hu\r\n", botId);
				return 0;
			}

//			ConsolePrint("ID: %hhu, length: %hu\r\n", botId, length);

			WifiRobot* pBot = &wifi.bots[botId];
			if(!pBot->online)
			{
				ConsolePrint("Data for unconnected bot %hu\r\n", botId);
				return 1;
			}

			result = NRF24QueueSend(&pBot->queue, pData, dataLength);
			if(result)
			{
				ConsolePrint("No memory in queue %hu\r\n", botId);
				return result;
			}
		}
		break;
		case CMD_BASE_STATION_PING:
		{
			int16_t result = NetworkSendPacket(pHeader, pData, dataLength);
			if(result)
			{
				ConsolePrint("NetworkSendPacket failed 0x%hX\r\n", result);
			}
		}
		break;
		case CMD_BASE_STATION_CONFIG:
		{
			BaseStationConfig* pConfig = (BaseStationConfig*)pData;

//			WifiSetVisionRate(pConfig->maxVisionRate);
//			vision.invertPos = pConfig->invertPosition;
			WifiChangeChannel(pConfig->channel);
//			if(pConfig->maxBots > WIFI_MAX_ROBOTS)
//				pConfig->maxBots = WIFI_MAX_ROBOTS;
//			wifi.maxBots = pConfig->maxBots;
//			wifi.timeout = pConfig->timeout;

			ConsolePrint("New Vision Config\r\n");
			ConsolePrint("Vision Rate: %hu\r\n", pConfig->maxVisionRate);
//			ConsolePrint("Invert: %hu\r\n", (uint16_t)vision.invertPos);
			ConsolePrint("Channel: %hu\r\n", (uint16_t)wifi.nrf24.channel);
			ConsolePrint("Max Bots: %hu\r\n", (uint16_t)wifi.maxBots);
			ConsolePrint("Timeout: %hums\r\n", wifi.timeout);
		}
		break;
		default:
		{
			ConsolePrint("Invalid command 0x%X\r\n", pHeader->cmd);
			return 0;
		}
		break;
	}

	return 0;
}

static int16_t decodeFrame(uint8_t* pDecoded, uint16_t length)
{
	FrameHeader* pFrameHeader;

	int16_t result = EthernetDecode(pDecoded, length, &pDecoded, &length, &pFrameHeader);
	if(result)
	{
		return result;
	}

	if(pFrameHeader->dstMAC.cast.multicast == 0 && MACCompare(pFrameHeader->dstMAC, network.my.mac) == 0)
		return 0; // this should not happen, as we have a hardware MAC filter

//	ConsolePrint("A packet for us :)\r\n");

	if(pFrameHeader->type == ETHERNET_PROTOCOL_ARP)
	{
		ARPPacketIPv4* pARP;

		result = ARPDecode(pDecoded, length, &pARP);
		if(result)
			return result;

		ConsolePrint("ARP\r\n");

		if(	pARP->operation == ARP_OPERATION_REQUEST && pARP->dstProtAddr.u32 == network.my.ip.u32)
		{
			// someone wants to know our MAC address
			ConsolePrint("Replying to ARP!\r\n");

			return NetworkSendARP(ARP_OPERATION_REPLY, pARP->srcHwAddr, pARP->srcProtAddr);
		}
	}

	if(pFrameHeader->type == ETHERNET_PROTOCOL_IPV4)
	{
		IPv4Header* pIPHeader;

		result = IPv4Decode(pDecoded, length, &pDecoded, &length, &pIPHeader);
		if(result)
			return result;

		if(	pIPHeader->dstIP.u32 != network.vision.ip.u32 && pIPHeader->dstIP.u32 != network.my.ip.u32 &&
			pIPHeader->dstIP.u32 != IPv4AddressSet(224, 0, 0, 1).u32)
			return 0;	// we are not addressed

		if(pIPHeader->protocol == IPV4_PROTOCOL_UDP)
		{
			UDPHeader* pUDPHeader;

			result = UDPDecode(pDecoded, length, &pDecoded, &length, &pUDPHeader);
			if(result)
				return result;

			if(pIPHeader->dstIP.u32 == network.vision.ip.u32)
			{
				if(pUDPHeader->dstPort != network.vision.port)
				{
//					ConsolePrint("Detected non-official vision packet on port %hu\r\n", pUDPHeader->dstPort);
//					ConsolePrint("Source: %hu.%hu.%hu.%hu\r\n", (uint16_t)pIPHeader->srcIP.u8[0],
//							(uint16_t)pIPHeader->srcIP.u8[1], (uint16_t)pIPHeader->srcIP.u8[2], (uint16_t)pIPHeader->srcIP.u8[3]);
					return 0;
				}

//				ConsolePrint("Vision frame! Length: %hu\r\n", pUDPHeader->length);

				VisionInput(pDecoded, length);

				return result;
			}

			if(pUDPHeader->dstPort != network.my.port)
				return 0;

//			ConsolePrint("Valid UDP packet for us\r\n");

			// check for auth packet
			PacketHeader* pHeader = (PacketHeader*)pDecoded;
			if(pHeader->section == SECTION_BASE_STATION && pHeader->cmd == CMD_BASE_STATION_AUTH)
			{
				pDecoded += PACKET_HEADER_SIZE;

				BaseStationAuth* pAuth = (BaseStationAuth*)pDecoded;

				if(pAuth->key == 0x42424242)
				{
					network.ucast.mac = pFrameHeader->srcMAC;
					network.ucast.ip = pIPHeader->srcIP;
					network.ucast.port = pUDPHeader->srcPort;

					ConsolePrint("Valid authentication from:\r\n");
					ConsolePrint("MAC: %02hX-%02hX-%02hX-%02hX-%02hX-%02hX\r\n", (uint16_t)network.ucast.mac.u8[0], (uint16_t)network.ucast.mac.u8[1],
							(uint16_t)network.ucast.mac.u8[2], (uint16_t)network.ucast.mac.u8[3], (uint16_t)network.ucast.mac.u8[4], (uint16_t)network.ucast.mac.u8[5]);
					ConsolePrint("IP: %hu.%hu.%hu.%hu\r\n", (uint16_t)network.ucast.ip.u8[0], (uint16_t)network.ucast.ip.u8[1],
							(uint16_t)network.ucast.ip.u8[2], (uint16_t)network.ucast.ip.u8[3]);
					ConsolePrint("Port: %hd\r\n", network.ucast.port);
				}
				else
				{
					ConsolePrint("Invalid base station authentication try\r\n");
				}

				return result;
			}

			unicastUdp(pDecoded, length);

			return result;
		}

		if(pIPHeader->protocol == IPV4_PROTOCOL_ICMP)
		{
			ICMPHeader* pICMPHeader;

			result = ICMPDecode(pDecoded, length, &pDecoded, &length, &pICMPHeader);
			if(result)
				return result;

			if(pICMPHeader->type == ICMP_TYPE_ECHO_REQUEST)
			{
				ICMPEcho* pICMPEcho;

				result = ICMPDecodeEcho(pDecoded, length, &pDecoded, &length, &pICMPEcho);
				if(result)
					return result;

				ConsolePrint("ICMP echo request, ident: %hu, seq: %hu, dataLen: %hu\r\n", pICMPEcho->identifier, pICMPEcho->sequenceNumber, length);

				result = NetworkSendICMPEchoReply(pICMPEcho, pDecoded, length, pFrameHeader->srcMAC, pIPHeader->srcIP);
				if(result)
					return result;
			}
		}

		if(pIPHeader->protocol == IPV4_PROTOCOL_IGMP)
		{
			uint8_t type = 0;
			IGMPv3GetType(pDecoded, &type);

			ConsolePrint("IGMP msg. Type: %hX\r\n", (uint16_t)type);

			if(type == IGMP_MEMBERSHIP_QUERY)
			{
				IGMPv3MembershipQuery* pQuery;
				IPv4Address* pSourceList;

				result = IGMPv3DecodeMembershipQuery(pDecoded, length, &pQuery, &pSourceList);
				if(result)
					return result;

				if(pQuery->groupAddress.u32 == 0 || pQuery->groupAddress.u32 == network.vision.ip.u32)
				{
					// general query
					network.igmp.lastSendTime = chTimeNow();
					network.igmp.respCounter = 0;
					network.igmp.robustness = pQuery->QRV ? pQuery->QRV : 2;
					uint32_t respTime = IGMPv3DecodeMaxRespCode(pQuery->maxRespCode);
					network.igmp.respTime = chTimeNow() % respTime;
				}
			}

			if(type == IGMPV3_MEMBERSHIP_REPORT)
			{
			}
		}
	}

	return 0;
}

static void processWifiPacket(uint8_t id, uint32_t pktSize)
{
	int16_t result;

	// construct header
	PacketHeader header;
	header.section = SECTION_BASE_STATION;
	header.cmd = CMD_BASE_STATION_ACOMMAND;

	// construct BaseStationACommand
	BaseStationACommand* pACmd = (BaseStationACommand*)network.rxProcBuf;
	pACmd->id = id;
	pktSize++;

//	PacketHeader* pHeader = (PacketHeader*)(wifi.sendBuf+1);
//	ConsolePrint("Rx cmd: 0x%hhX 0x%hhX, len: %hu\r\n", pHeader->section, pHeader->cmd, header.len);

	result = NetworkSendPacket(&header, network.rxProcBuf, pktSize);
	if(result)
		ConsolePrint("Network send failed: 0x%hX\r\n", result);
}

msg_t NetworkTask(void* params)
{
	(void)params;

	msg_t event;
	uint32_t pktSize;

	int32_t lastStatsTime = chTimeNow();

	EthLinkUpdate();

	while (1)
	{
		if(chMBFetch(&network.eventQueue, &event, MS2ST(10)) == RDY_OK)
		{
			switch(event)
			{
				case EVENT_LINK:
				{
					if(EthLinkUpdate())
					{
						// link up
						network.igmp.respTime = 500;	// send IGMPv3 membership
					}
				}
				break;
				case EVENT_RX_DONE:
				{
					while(EthGetEthernetFrame(network.rxProcBuf, NETWORK_MAX_FRAME_SIZE, &pktSize) == 0)
						decodeFrame(network.rxProcBuf, pktSize);
				}
				break;
			}

			if(event & EVENT_BOT_DATA)
			{
				uint32_t id = event & 0xFF;

				while(NRF24QueueGetPacket(&wifi.bots[id].queue, network.rxProcBuf+1, NETWORK_MAX_FRAME_SIZE, &pktSize) == 0)
					processWifiPacket(wifi.bots[id].queue.id, pktSize);
			}
		}

		// IGMPv3 work
		if(network.igmp.respTime != 0)
		{
			if(chTimeElapsedSince(network.igmp.lastSendTime) > MS2ST(network.igmp.respTime))
			{
				network.igmp.lastSendTime = chTimeNow();

				IGMPv3GroupRecord record;
				record.auxDataLen = 0;
				record.numberOfSources = 0;
				record.recordType = IGMPV3_MODE_IS_EXCLUDE;
				record.multicastAddress.u32 = network.vision.ip.u32;

				NetworkSendIGMPv3MembershipReport(1, record._data, IGMPV3_GROUP_RECORD_SIZE);

				network.igmp.respCounter++;
				if(network.igmp.respCounter >= network.igmp.robustness)
					network.igmp.respTime = 0;
			}
		}

		if(chTimeElapsedSince(lastStatsTime) > MS2ST(500))
		{
			lastStatsTime = chTimeNow();

			EthStats ethStats = *EthGetStats();
			ethStats.rxBytesProcessed -= network.lastStats.rxBytesProcessed;
			ethStats.rxFramesProcessed -= network.lastStats.rxFramesProcessed;
			ethStats.txBytesProcessed -= network.lastStats.txBytesProcessed;
			ethStats.txFramesProcessed -= network.lastStats.txFramesProcessed;
			ethStats.rxMissedApp -= network.lastStats.rxMissedApp;
			ethStats.rxMissedCtrl -= network.lastStats.rxMissedCtrl;
			ethStats.rxFrames -= network.lastStats.rxFrames;
			ethStats.txFrames -= network.lastStats.txFrames;

			network.lastStats = *EthGetStats();

			// construct header
			PacketHeader header;
			header.section = SECTION_BASE_STATION;
			header.cmd = CMD_BASE_STATION_STATS;

			network.stats.eth.rxBytesRecv = ethStats.rxBytesProcessed;
			network.stats.eth.txBytesSent = ethStats.txBytesProcessed;
			network.stats.eth.rxFramesRecv = ethStats.rxFramesProcessed;
			network.stats.eth.txFramesSent = ethStats.txFramesProcessed;
			network.stats.eth.rxLevel = 0;
			network.stats.eth.txLevel = 0;
//			network.stats.eth.rxLevel = (ethStats.rxDmaLevel*200)/NETWORK_NUM_DESCRIPTORS;
//			network.stats.eth.txLevel = (ethStats.txDmaLevel*200)/NETWORK_NUM_DESCRIPTORS;
			network.stats.eth.rxLoss = ((ethStats.rxMissedApp+ethStats.rxMissedCtrl)*40000)/ethStats.rxFrames;
			network.stats.eth.updateRate = 1000000/wifi.runtime;

			uint8_t slotsUsed = 0;
			for(uint8_t i = 0; i <= NRF24_MAX_BOT_ID; i++)
			{
				NRF24QueueStats* pLastStats = &network.lastBotStats[i];
				NRF24QueueStats qStats = wifi.bots[i].queue.stats;
				qStats.rxBytesRecvRaw -= pLastStats->rxBytesRecvRaw;
				qStats.rxBytesRecvCOBS -= pLastStats->rxBytesRecvCOBS;
				qStats.txBytesSentRaw -= pLastStats->txBytesSentRaw;
				qStats.txBytesSentCOBS -= pLastStats->txBytesSentCOBS;
				qStats.rxPacketsRecv -= pLastStats->rxPacketsRecv;
				qStats.txPacketsSent -= pLastStats->txPacketsSent;
				qStats.rxPacketsMalformed -= pLastStats->rxPacketsMalformed;
				qStats.rxPacketsOverrun -= pLastStats->rxPacketsOverrun;
				qStats.txPacketsAcked -= pLastStats->txPacketsAcked;
				qStats.txPacketsMaxRT -= pLastStats->txPacketsMaxRT;
				qStats.txPacketsOverrun -= pLastStats->txPacketsOverrun;

				if(slotsUsed < BASE_STATION_STATS_NUM_BOTS)
				{
					if(!wifi.bots[i].online)
						continue;

					network.stats.bot[slotsUsed].botId = wifi.bots[i].queue.id;
					network.stats.bot[slotsUsed].rxBytesRecvRaw = qStats.rxBytesRecvRaw;
					network.stats.bot[slotsUsed].rxBytesRecvCOBS = qStats.rxBytesRecvCOBS;
					network.stats.bot[slotsUsed].txBytesSentRaw = qStats.txBytesSentRaw;
					network.stats.bot[slotsUsed].txBytesSentCOBS = qStats.txBytesSentCOBS;
					network.stats.bot[slotsUsed].rxPacketsRecv = qStats.rxPacketsRecv;
					network.stats.bot[slotsUsed].txPacketsSent = qStats.txPacketsSent;
					network.stats.bot[slotsUsed].linkQuality = (qStats.txPacketsAcked*40000)/(qStats.txPacketsMaxRT+qStats.txPacketsAcked);
					network.stats.bot[slotsUsed].rxLoss = (qStats.rxPacketsMalformed*200)/(qStats.rxPacketsRecv+qStats.rxPacketsMalformed);
					network.stats.bot[slotsUsed].txLoss = (qStats.txPacketsOverrun*200)/(qStats.txPacketsSent+qStats.txPacketsOverrun);
					network.stats.bot[slotsUsed].rxLevel = ((wifi.bots[i].queue.rxFifo.fifo.size-wifi.bots[i].queue.rxFifo.fifo.bytesFree)*40000)/wifi.bots[i].queue.rxFifo.fifo.size;
					network.stats.bot[slotsUsed].txLevel = ((wifi.bots[i].queue.txFifo.fifo.size-wifi.bots[i].queue.txFifo.fifo.bytesFree)*40000)/wifi.bots[i].queue.txFifo.fifo.size;
					slotsUsed++;
				}

				for(uint8_t slot = slotsUsed; slot < BASE_STATION_STATS_NUM_BOTS; slot++)
					network.stats.bot[slot].botId = 0xFF;

				network.lastBotStats[i] = wifi.bots[i].queue.stats;
			}

			NetworkSendPacket(&header, (uint8_t*)&network.stats, sizeof(BaseStationStats));
		}
	}

	return 0;
}

int16_t NetworkSendPacket(PacketHeader* pHeader, uint8_t* pData, uint16_t dataLength)
{
	if(dataLength + PACKET_HEADER_SIZE > NETWORK_MAX_FRAME_SIZE)
		return ERROR_NOT_ENOUGH_MEMORY;

	chMtxLock(&network.sendPacketMutex);	// prevent re-entering

	memcpy(network.txProcBuf, pHeader->_data, PACKET_HEADER_SIZE);
	memcpy(network.txProcBuf + PACKET_HEADER_SIZE, pData, dataLength);

	uint16_t encLength = dataLength + PACKET_HEADER_SIZE;

	uint16_t dstPort = network.ucast.port;

	IPv4Address dstIp = network.ucast.ip;

	MAC dstMAC = network.ucast.mac;

	int16_t result = UDPEncode(network.txProcBuf, NETWORK_MAX_FRAME_SIZE, encLength, &encLength, network.my.port, dstPort);
	if(result)
	{
		chMtxUnlock();
		return result;
	}

	result = IPv4Encode(network.txProcBuf, NETWORK_MAX_FRAME_SIZE, encLength, &encLength,
						network.ethernetIdent, network.my.ip, dstIp, IPV4_PROTOCOL_UDP);
	if(result)
	{
		chMtxUnlock();
		return result;
	}

	result = EthernetEncode(network.txProcBuf, NETWORK_MAX_FRAME_SIZE, encLength, &encLength,
							network.my.mac, dstMAC, ETHERNET_PROTOCOL_IPV4);
	if(result)
	{
		chMtxUnlock();
		return result;
	}

	result = EthSendEthernetFrame(network.txProcBuf, encLength);
	if(result)
	{
		chMtxUnlock();
		return result;
	}

	network.ethernetIdent++;

	chMtxUnlock();

	return 0;
}

int16_t NetworkSendARP(uint8_t operation, MAC targetMAC, IPv4Address targetIP)
{
	if(ARP_PACKET_SIZE > NETWORK_MAX_FRAME_SIZE)
		return ERROR_NOT_ENOUGH_MEMORY;

	chMtxLock(&network.sendPacketMutex);	// prevent re-entering

	uint16_t encLength;

	int16_t result = ARPEncode(	network.txProcBuf, NETWORK_MAX_FRAME_SIZE, &encLength,
								operation, network.my.mac, network.my.ip, targetMAC, targetIP);
	if(result)
	{
		chMtxUnlock();
		return result;
	}

	result = EthernetEncode(network.txProcBuf, NETWORK_MAX_FRAME_SIZE, encLength, &encLength,
							network.my.mac, targetMAC, ETHERNET_PROTOCOL_ARP);
	if(result)
	{
		chMtxUnlock();
		return result;
	}

	result = EthSendEthernetFrame(network.txProcBuf, encLength);
	if(result)
	{
		chMtxUnlock();
		return result;
	}

	chMtxUnlock();

	return 0;
}

int16_t NetworkSendICMPEchoReply(ICMPEcho* pEchoRequest, uint8_t* pData, uint16_t dataLength, MAC targetMAC, IPv4Address dstIp)
{
	if(dataLength + ICMP_ECHO_SIZE + ICMP_HEADER_SIZE > NETWORK_MAX_FRAME_SIZE)
		return ERROR_NOT_ENOUGH_MEMORY;

	chMtxLock(&network.sendPacketMutex);	// prevent re-entering

	uint16_t encLength;

	memcpy(network.txProcBuf, pData, dataLength);

	int16_t result = ICMPEncodeEcho(network.txProcBuf, NETWORK_MAX_FRAME_SIZE, dataLength, &encLength,
								pEchoRequest->identifier, pEchoRequest->sequenceNumber);
	if(result)
	{
		chMtxUnlock();
		return result;
	}

	result = ICMPEncode(network.txProcBuf, NETWORK_MAX_FRAME_SIZE, encLength, &encLength, ICMP_TYPE_ECHO_REPLY, 0);
	if(result)
	{
		chMtxUnlock();
		return result;
	}

	result = IPv4Encode(network.txProcBuf, NETWORK_MAX_FRAME_SIZE, encLength, &encLength,
						network.ethernetIdent, network.my.ip, dstIp, IPV4_PROTOCOL_ICMP);
	if(result)
	{
		chMtxUnlock();
		return result;
	}

	result = EthernetEncode(network.txProcBuf, NETWORK_MAX_FRAME_SIZE, encLength, &encLength,
							network.my.mac, targetMAC, ETHERNET_PROTOCOL_IPV4);
	if(result)
	{
		chMtxUnlock();
		return result;
	}

	result = EthSendEthernetFrame(network.txProcBuf, encLength);
	if(result)
	{
		chMtxUnlock();
		return result;
	}

	network.ethernetIdent++;

	chMtxUnlock();

	return 0;
}

int16_t NetworkSendIGMPv3MembershipReport(uint16_t numberOfGroupRecords, uint8_t* recordData, uint16_t recordDataLength)
{
	if(recordDataLength + IGMPV3_MEMBERSHIP_REPORT_SIZE > NETWORK_MAX_FRAME_SIZE)
		return ERROR_NOT_ENOUGH_MEMORY;

	ConsolePrint("Sending IGMPv3 membership report\r\n");

	chMtxLock(&network.sendPacketMutex);	// prevent re-entering

	uint16_t encLength = 0;
	int16_t result = 0;

	result = IGMPv3EncodeMembershipReport(	network.txProcBuf, NETWORK_MAX_FRAME_SIZE,
											&encLength, numberOfGroupRecords, recordData, recordDataLength);
	if(result)
	{
		chMtxUnlock();
		return result;
	}

	IPv4Address dstIP = IPv4AddressSet(224, 0, 0, 22);
	MAC dstMAC = MACSet(0x01, 0x00, 0x5E, 0x00, 0x00, 0x16);	// 224.0.0.22

	result = IPv4EncodeWithRouterAlertOption(network.txProcBuf, NETWORK_MAX_FRAME_SIZE, encLength, &encLength,
						network.ethernetIdent, network.my.ip, dstIP, IPV4_PROTOCOL_IGMP);
	if(result)
	{
		chMtxUnlock();
		return result;
	}

	result = EthernetEncode(network.txProcBuf, NETWORK_MAX_FRAME_SIZE, encLength, &encLength,
							network.my.mac, dstMAC, ETHERNET_PROTOCOL_IPV4);
	if(result)
	{
		chMtxUnlock();
		return result;
	}

	result = EthSendEthernetFrame(network.txProcBuf, encLength);
	if(result)
	{
		chMtxUnlock();
		return result;
	}

	network.ethernetIdent++;

	chMtxUnlock();

	return 0;
}

void NetworkChangePort(uint16_t port)
{
	network.my.port = port;
}

void NetworkChangeIP(IPv4Address ip)
{
	network.my.ip = ip;
}
