/*
 * wifi.c
 *
 *  Created on: 07.07.2014
 *      Author: AndreR
 */

#include "wifi.h"

#include "util/console.h"
#include "util/bit_banding.h"
#include "util/sys_time.h"
#include "stm32f4xx_gpio.h"
#include "stm32f4xx_exti.h"
#include "stm32f4xx_rcc.h"
#include "stm32f4xx_syscfg.h"
#include "stm32f4xx_tim.h"
#include "misc.h"
#include "constants.h"
#include "commands.h"

#include <string.h>

WifiGlobal wifi;

void EXTI2_IRQHandler()
{
	CH_IRQ_PROLOGUE();

	EXTI->PR = GPIO_Pin_2;

	NRF24ExIRQ(&wifi.nrf24);

	CH_IRQ_EPILOGUE();
}

void DMA1_Stream0_IRQHandler()
{
	SPISyncDmaRxInterrupt(&wifi.nrf24.spi);
}

void TIM7_IRQHandler()
{
	CH_IRQ_PROLOGUE();

	TIM7->SR = 0;
	TIM_Cmd(TIM7, DISABLE);

	chSysLockFromIsr();
	chBSemSignalI(&wifi.timerSem);
	chSysUnlockFromIsr();

	CH_IRQ_EPILOGUE();
}

static void startWifiTimeout(uint32_t microSeconds)
{
	TIM7->ARR = microSeconds;

	TIM_Cmd(TIM7, ENABLE);
}

void WifiSetVisionRate(uint16_t rateHz)
{
	uint32_t rate = rateHz;
	wifi.visionPauseTime = MS2ST(1000/rate);
}

void WifiSendPose(uint16_t botId, float x, float y, float orient)
{
	int16_t result;

	if(botId > NRF24_MAX_BOT_ID)
	{
		ConsolePrint("Invalid pose id %hu\r\n", botId);
		return;
	}

	WifiRobot* pBot = &wifi.bots[botId];
	if(!pBot->online)
		return;

	if(chTimeElapsedSince(wifi.bots[botId].lastVisionTime) < wifi.visionPauseTime)
		return;

	wifi.bots[botId].lastVisionTime = chTimeNow();

	static uint8_t data[PACKET_HEADER_SIZE+sizeof(CtrlVisionPos)+sizeof(CtrlVisionTrackedObj)];

	PacketHeader* pHeader = (PacketHeader*)data;
	pHeader->cmd = CMD_CTRL_VISION_POS;
	pHeader->section = SECTION_CTRL;

	CtrlVisionPos* pPos = (CtrlVisionPos*)(data+PACKET_HEADER_SIZE);
	pPos->pos[0] = (int16_t)x;
	pPos->pos[1] = (int16_t)y;
	pPos->pos[2] = (int16_t)(orient*1000.0f);

	result = NRF24QueueSend(&pBot->queue, data, PACKET_HEADER_SIZE+sizeof(CtrlVisionPos));
	if(result)
	{
		ConsolePrint("NRF24QueueSend for bot %hu error'd in WifiSendPose: %hd\r\n", botId, result);
		return;
	}
}

void WifiSendObjectPose(float objX, float objY, float objOrient)
{
	int16_t result;

	if(chTimeElapsedSince(wifi.lastVisionObjTime) < wifi.visionPauseTime)
		return;

	wifi.lastVisionObjTime = chTimeNow();

	static uint8_t data[PACKET_HEADER_SIZE+sizeof(CtrlVisionTrackedObj)];

	PacketHeader* pHeader = (PacketHeader*)data;
	pHeader->cmd = CMD_CTRL_VISION_TRACKED_OBJ;
	pHeader->section = SECTION_CTRL;

	CtrlVisionTrackedObj* pObj = (CtrlVisionTrackedObj*)(data+PACKET_HEADER_SIZE);
	pObj->pos[0] = (int16_t)objX;
	pObj->pos[1] = (int16_t)objY;
	pObj->pos[2] = (int16_t)(objOrient*1000.0f);

	for(uint8_t botId = 0; botId <= NRF24_MAX_BOT_ID; botId++)
	{
		WifiRobot* pBot = &wifi.bots[botId];
		if(!pBot->online)
			continue;

		result = NRF24QueueSend(&pBot->queue, data, PACKET_HEADER_SIZE+sizeof(CtrlVisionTrackedObj));
		if(result)
			ConsolePrint("NRF24QueueSend for bot %hhu error'd in WifiSendObjectPose: %hd\r\n", botId, result);
	}
}

void WifiSaveConfig()
{
	FlashFSWrite(&wifi.pCfgFile, &wifi.cfg, sizeof(WifiConfig));
}

void WifiInit()
{
	int16_t result;
	NVIC_InitTypeDef nvicInit;
	GPIO_InitTypeDef gpioInit;

	chBSemInit(&wifi.timerSem, TRUE);

	result = FlashFSOpenOrCreate("wifi/cfg", 0, &wifi.cfg, sizeof(WifiConfig), &wifi.pCfgFile);
	if(result == FLASH_FS_CREATED)
	{
		wifi.cfg.ackDelayUs = 500;
		wifi.cfg.channel = 100;
		wifi.cfg.maxNoLinkTries = 100;
		wifi.cfg.maxRt = 0;
		wifi.cfg.maxVisionRate = 30;
		wifi.cfg.speed = NRF24_SPEED_2M;

		FlashFSWrite(&wifi.pCfgFile, &wifi.cfg, sizeof(WifiConfig));
	}

	// NRF24 IRQ line
	gpioInit.GPIO_Mode = GPIO_Mode_IN;
	gpioInit.GPIO_OType = GPIO_OType_PP;
	gpioInit.GPIO_Speed = GPIO_Speed_25MHz;
	gpioInit.GPIO_PuPd = GPIO_PuPd_UP;
	gpioInit.GPIO_Pin = GPIO_Pin_2;
	GPIO_Init(GPIOD, &gpioInit);	// IRQ line

	SYSCFG_EXTILineConfig(EXTI_PortSourceGPIOD, EXTI_PinSource2);

	EXTI_InitTypeDef extiInit;
	EXTI_StructInit(&extiInit);
	extiInit.EXTI_Line = EXTI_Line2;
	extiInit.EXTI_LineCmd = ENABLE;
	extiInit.EXTI_Mode = EXTI_Mode_Interrupt;
	extiInit.EXTI_Trigger = EXTI_Trigger_Falling;	// active low
	EXTI_Init(&extiInit);

	nvicInit.NVIC_IRQChannelCmd = ENABLE;
	nvicInit.NVIC_IRQChannelPreemptionPriority = IRQL_WIFI;
	nvicInit.NVIC_IRQChannelSubPriority = 0;
	nvicInit.NVIC_IRQChannel = EXTI2_IRQn;
	NVIC_Init(&nvicInit);

	// SPI3
	RCC_APB1PeriphClockCmd(RCC_APB1Periph_SPI3, ENABLE);

	nvicInit.NVIC_IRQChannel = DMA1_Stream0_IRQn;
	nvicInit.NVIC_IRQChannelCmd = ENABLE;
	nvicInit.NVIC_IRQChannelPreemptionPriority = IRQL_SPI3;
	nvicInit.NVIC_IRQChannelSubPriority = 0x00;
	NVIC_Init(&nvicInit);

	gpioInit.GPIO_Mode = GPIO_Mode_AF;
	gpioInit.GPIO_OType = GPIO_OType_PP;
	gpioInit.GPIO_Speed = GPIO_Speed_50MHz;
	gpioInit.GPIO_PuPd = GPIO_PuPd_UP;
	gpioInit.GPIO_Pin = GPIO_Pin_10 | GPIO_Pin_11 | GPIO_Pin_12;	// SCK, MISO, MOSI
	GPIO_Init(GPIOC, &gpioInit);

	GPIO_PinAFConfig(GPIOC, GPIO_PinSource10, GPIO_AF_SPI3);
	GPIO_PinAFConfig(GPIOC, GPIO_PinSource11, GPIO_AF_SPI3);
	GPIO_PinAFConfig(GPIOC, GPIO_PinSource12, GPIO_AF_SPI3);

	// TIM7 setup
	RCC_APB1PeriphClockCmd(RCC_APB1Periph_TIM7, ENABLE);

	TIM_TimeBaseInitTypeDef timeBaseInit;
	timeBaseInit.TIM_ClockDivision = TIM_CKD_DIV1;
	timeBaseInit.TIM_CounterMode = TIM_CounterMode_Up;
	timeBaseInit.TIM_Period = 1000;	// = 1ms
	timeBaseInit.TIM_Prescaler = 83; // clock is now 1MHz
	TIM_TimeBaseInit(TIM7, &timeBaseInit);

	TIM_ARRPreloadConfig(TIM7, DISABLE);
	TIM_CCPreloadControl(TIM7, DISABLE);
	TIM_InternalClockConfig(TIM7);
	TIM_ITConfig(TIM7, TIM_IT_Update, ENABLE);

	nvicInit.NVIC_IRQChannel = TIM7_IRQn;
	nvicInit.NVIC_IRQChannelCmd = ENABLE;
	nvicInit.NVIC_IRQChannelPreemptionPriority = IRQL_NRF24_TIMER;
	nvicInit.NVIC_IRQChannelSubPriority = 0;
	NVIC_Init(&nvicInit);

	for(uint8_t i = 0; i <= NRF24_MAX_BOT_ID; i++)
	{
		WifiRobot* pBot = &wifi.bots[i];

		NRF24QueueInit(&pBot->queue, pBot->txData, WIFI_TX_SIZE, pBot->rxData, WIFI_RX_SIZE);

		wifi.bots[i].lastResponseTime = 0;
		wifi.bots[i].lastVisionTime = 0;
		wifi.bots[i].online = 0;
		wifi.bots[i].queue.id = i;
	}

	// NRF24 setup
	NRF24ExData data;

	data.spiData.csPin = GPIO_Pin_7;
	data.spiData.pCSPort = GPIOD;
	data.spiData.pRegister = SPI3;
	data.spiData.pDma = DMA1;
	data.spiData.dmaChRx = 0;
	data.spiData.dmaChTx = 7;
	data.spiData.timeoutTicks = MS2ST(100);
	data.rx = 0;
	data.channel = wifi.cfg.channel;
	data.speed = wifi.cfg.speed;
	data.maxRt = wifi.cfg.maxRt;
	data.ackDelayUs = wifi.cfg.ackDelayUs;
	data.pCEPort = GPIOD;
	data.cePin = GPIO_Pin_3;

	result = NRF24ExInit(&wifi.nrf24, &data);
	if(result)
		ConsolePrint("NRF24Init failed: 0x%02hX\r\n", result);

	wifi.scan = 0xFF;
	wifi.visionPauseTime = 30; // 30ms
	wifi.maxBots = 6;
	wifi.timeout = 1000;
	wifi.lastVisionObjTime = 0;
}

void WifiChangeChannel(uint8_t channel)
{
	NRF24ExChangeChannel(&wifi.nrf24, channel);

	wifi.cfg.channel = channel;
}

void WifiChangeSpeed(uint8_t speed)
{
	if(speed != NRF24_SPEED_250K && speed != NRF24_SPEED_1M && speed != NRF24_SPEED_2M)
		return;

	NRF24ExChangeSpeed(&wifi.nrf24, speed);

	wifi.cfg.speed = speed;
}

int16_t WifiChangeRetransmission(uint8_t maxRt, uint16_t delayUs)
{
	if(maxRt > 15)
		return 1;

	int16_t result = NRF24ExChangeRetransmission(&wifi.nrf24, maxRt, delayUs);
	if(result)
		return result;

	wifi.cfg.maxRt = maxRt;
	wifi.cfg.ackDelayUs = delayUs;

	return 0;
}

static int16_t processBot(WifiRobot* pBot)
{
	int16_t result;

	chBSemReset(&wifi.timerSem, 1);

	startWifiTimeout(994);

	NRF24ExUseQueue(&wifi.nrf24, &pBot->queue);

	result = NRF24ExTxProcessOne(&wifi.nrf24);
	if(!result)
	{
		pBot->lastResponseTime = SysTimeUSec();

		if(pBot->queue.rxFifo.numDelimiters && wifi.cbBotData)
			(*wifi.cbBotData)(pBot->queue.id);

		if(!pBot->online)
			pBot->connectCounter++;
	}
	else if(result == ERROR_NRF24EX_TIMEOUT)
	{
		pBot->connectCounter = 0;
	}
	else
	{
		ConsolePrint("NRF24ExProcess error: 0x%0X\r\n", result);
	}

	// check if new bot responded ten times, then we can define it online
	if(pBot->connectCounter == 10 && !pBot->online)
	{
		pBot->online = 1;

		ConsolePrint("Bot %hu online\r\n", (uint16_t)pBot->queue.id);
	}

	// check if bot timed out
	if(SysTimeUSec() - pBot->lastResponseTime > (wifi.timeout*1000) && pBot->online)
	{
		// no response => timeout
		pBot->online = 0;
		NRF24QueueClear(&pBot->queue);

		ConsolePrint("Bot %hu timed out.\r\n", (uint16_t)pBot->queue.id);
	}

	chBSemWait(&wifi.timerSem);

	return result;
}

void WifiSetBotDataCallback(BotDataAvailCallback cb)
{
	wifi.cbBotData = cb;
}

msg_t WifiTask(void* params)
{
	(void) params;

	ConsolePrint("WifiTask started\r\n");

	int16_t result = 0;

	uint32_t startTime = SysTimeUSec();
	uint32_t endTime;
	uint8_t checkId = 0;

	NRF24ExSetup(&wifi.nrf24);
	NRF24ExUseQueue(&wifi.nrf24, &wifi.bots[0].queue);

	while(1)
	{
		startTime = SysTimeUSec();

		// scan something?
		if(wifi.scan != 0xFF)
		{
			uint8_t scanResult;

			ConsolePrint("Scanning wifi channels...\r\n");
			ConsolePrint("     00 01 02 03 04 05 06 07 08 09 10 11 12 13 14 15\r\n000: ");

			for(uint8_t i = 0; i < 128; i++)
			{
				result = NRF24ExChangeChannel(&wifi.nrf24, i);
				if(result)
					ConsolePrint("Change channel failed: 0x%hX\r\n", result);

				result = NRF24ExCheckChannel(&wifi.nrf24, &scanResult);
				if(result)
					ConsolePrint("Check channel failed: 0x%hX\r\n", result);

				ConsolePrint("%2hhu ", scanResult);

				if((i+1)%16 == 0 && i != 0 && i != 127)
				{
					ConsolePrint("\r\n%03hhu: ", i+1);

					chThdSleep(MS2ST(100));
				}
			}

			ConsolePrint("\r\n");

			NRF24ExChangeChannel(&wifi.nrf24, wifi.cfg.channel);

			wifi.scan = 0xFF;
		}

		uint8_t numBotsOnline = 0;
		for(uint8_t i = 0; i <= NRF24_MAX_BOT_ID; i++)
		{
			WifiRobot* pBot = &wifi.bots[i];

			if(!pBot->online)
				continue;

			++numBotsOnline;

			processBot(pBot);
		}

		while(numBotsOnline < wifi.maxBots)
		{
			do
			{
				++checkId;
				checkId %= NRF24_MAX_BOT_ID+1;
			}
			while(wifi.bots[checkId].online);

			processBot(&wifi.bots[checkId]);

			++numBotsOnline;

			if(wifi.fixedRuntime == 0)	// only one check?
				break;
		}

		endTime = SysTimeUSec();

		wifi.runtime = endTime-startTime;
	}

	return 0;
}
