/*
 * cli.c
 *
 *  Created on: 19.11.2012
 *      Author: AndreR
 */

#include "cli.h"

#include "util/console.h"
#include "util/boot.h"
#include "intercom_constants.h"
#include "hal/led.h"
#include "intercom1.h"
#include "hal/kicker.h"
#include "supply.h"
#include "hal/motor.h"
#include "ctrl_6step.h"
#include "util/sys_time.h"
#include "util/flash.h"
#include "util/flash_fs.h"
#include "kick.h"
#include "cpu_load.h"

#include "stm32f30x_tim.h"

#include <string.h>

CLI cli;

void processCommand();
static void periodic(void* pUserData);

void CLIInit(EventManager* pEventManager)
{
	chMtxInit(&cli.procBufMutex);

	cli.pEventClient = EventClientInit(pEventManager, 0, &periodic, MS2ST(20), 0);

	static uint8_t rxData[CONSOLE_RX_SIZE];
	static uint8_t txData[CONSOLE_TX_SIZE];

	ConsoleData conData;
	conData.pDataTx = txData;
	conData.txSize = CONSOLE_TX_SIZE;
	conData.pDataRx = rxData;
	conData.rxSize = CONSOLE_RX_SIZE;
	conData.printSize = CONSOLE_PRINT_SIZE;
	conData.txCallback = 0;
	conData.pUserData = 0;

	ConsoleInit(conData);

	ConsoleIntercomAdapterInit(&cli.adapter, &intercom1, INTERCOM_ADDRESS_KD, INTERCOM_ADDRESS_MAIN);
}

static void periodic(void* pUserData)
{
	(void)pUserData;

	static uint8_t firstRun = 1;

	if(firstRun)
	{
		firstRun = 0;

		LEDMustOff(LED_MAIN_GREEN);

#ifdef ENV_BOOT
		ConsolePrint("--- Bootloader (KD) ---\r\n");
#else
		ConsolePrint("--- KD ---\r\n");
#endif

//		ConsolePrint("%d Tasks started\r\nHEAP usage: %d/%d Byte\r\n",
//				uxTaskGetNumberOfTasks(), configTOTAL_HEAP_SIZE-xPortGetFreeHeapSize(), configTOTAL_HEAP_SIZE);
	}

	if(ConsoleIsCmdReady())
	{
		processCommand();

		ConsoleCmdProcessed();
	}
}

void processCommand()
{
	uint16_t u16, u16b;

	if(ConsoleCmpCmd("test"))
	{
		ConsolePrint("Number: %d\r\n", 43);
	}

	if(ConsoleCmpCmd("test2"))
	{
		ConsolePrint("0123456789012345678901234567890123456789012345678901234567890123456789012345678901234567890123456789\r\n");
	}

	if(ConsoleCmpCmd("cpu"))
	{
		CPULoad load = cpuLoad;

		ConsolePrint("CPU Load: %3u.%1u%%\r\n", load.load/10, load.load%10);
		ConsolePrint("Max: %3u.%1u%%\r\n", load.maxLoad/10, load.maxLoad%10);
		ConsolePrint("incPerTick: %10u/%u\r\n", load.incPerSysTick, load.maxIncPerSysTick);

		chThdSleep(MS2ST(10));
		cpuLoad.maxLoad = 0;
	}

	if(ConsoleScanCmd("vmax %hu", &u16) == 1)
	{
		ConsolePrint("vMax: %hu\r\n", u16);

		KickerSetMaxVoltage(u16);
	}

	if(ConsoleScanCmd("chg %hu", &u16) == 1)
	{
		if(u16)
		{
			ConsolePrint("Enable charging\r\n");
			KickerEnableCharge(1);
		}
		else
		{
			ConsolePrint("Disable charging\r\n");
			KickerEnableCharge(0);
		}
	}

	if(ConsoleScanCmd("autochg %hu", &u16) == 1)
	{
		if(u16)
		{
			ConsolePrint("Enable autocharging\r\n");
			KickerEnableAutoRecharge(1);
		}
		else
		{
			ConsolePrint("Disable autocharging\r\n");
			KickerEnableAutoRecharge(0);
		}
	}

	if(ConsoleCmpCmd("autodischg"))
	{
		ConsolePrint("Enabling auto discharging\r\n");
		KickerAutoDischarge();
	}

	if(ConsoleScanCmd("kstress %hu", &u16) == 1)
	{
		ConsolePrint("Setting auto-fire stress test: %hu\r\n", u16);

		kicker.stress = ((float)u16)/1000.0f;
	}

	if(ConsoleScanCmd("dis1 %hu", &u16) == 1)
	{
		ConsolePrint("Fire1: %hu\r\n", u16);

		KickerFire1(((float)u16)/1000.0f);
	}

	if(ConsoleScanCmd("arm1 %hu", &u16) == 1)
	{
		ConsolePrint("Arm1: %hu\r\n", u16);

		KickArmIr(((float)u16)/1000.0f, KICKER_DEVICE_STRAIGHT);
	}

	if(ConsoleScanCmd("dis2 %hu", &u16) == 1)
	{
		ConsolePrint("Fire2: %hu\r\n", u16);

		KickerFire2(((float)u16)/1000.0f);
	}

	if(ConsoleScanCmd("arm2 %hu", &u16) == 1)
	{
		ConsolePrint("Arm2: %hu\r\n", u16);

		KickArmIr(((float)u16)/1000.0f, KICKER_DEVICE_CHIP);
	}

	if(ConsoleCmpCmd("boot"))
	{
		BootSetBootloaderSelected(1);
		BootReset();
	}

	if(ConsoleCmpCmd("kstat"))
	{
		ConsolePrint("vCap: %d, iChg: %d\r\n", (int32_t)kicker.vCap, (int32_t)kicker.iChg);
		ConsolePrint("State: %hhu, Auto: %hhu\r\n", kicker.state, kicker.autoCharge);
	}

	if(ConsoleCmpCmd("kadc"))
	{
		for(uint8_t i = 0; i < SPI1_ADC_SAMPLES*2; i++)
			ConsolePrint("%hu ", kicker.rxData[i]);

		ConsolePrint("\r\n");
	}

	if(ConsoleCmpCmd("iradc"))
	{
		uint16_t irData[NUM_IR_SAMPLES];
		memcpy(irData, supply.irData, sizeof(uint16_t)*NUM_IR_SAMPLES);
		ConsolePrint("IRADC: ");
		for(uint16_t i = 0; i < NUM_IR_SAMPLES; i++)
			ConsolePrint("%hu ", irData[i]);
		ConsolePrint("\r\n");
	}

	if(ConsoleCmpCmd("ir"))
	{
		ConsolePrint("IR: %hu/%hu\r\n", supply.irResult[0], supply.irResult[1]);
	}

	if(ConsoleCmpCmd("supplyadc"))
	{
		uint16_t adc[2*NUM_DATA_SAMPLES];
		memcpy(adc, supply.adcData, 2*NUM_DATA_SAMPLES*sizeof(uint16_t));
		ConsolePrint("SupplyADC: ");
		for(uint16_t i = 0; i < 2*NUM_DATA_SAMPLES; i++)
			ConsolePrint("%hu ", adc[i]);
		ConsolePrint("\r\n");
	}

	if(ConsoleCmpCmd("supply"))
	{
		ConsolePrint("Supply: %hu/%hu\r\n", supply.batResult, supply.shuntResult);
	}

	// M5
	if(ConsoleScanCmd("power %hu", &u16) == 1)
	{
		ConsolePrint("Setting drib power %hu\r\n", u16);

		MotorSetPower(pMotor, u16);
	}

	uint16_t vref, angle;
	if(ConsoleScanCmd("turn %hu", &vref) == 1)
	{
		float power = vref/1000.0f;

		MotorSetVoltageAngle(pMotor, power, 0);

		chThdSleep(MS2ST(50));

//		ConsolePrint("enc: %hu\r\n", pDrive->pTIMEncoder->CNT);

		for(float a = 0; a <= 2880; a += 5)
		{
			MotorSetVoltageAngle(pMotor, power, a);

//			angle = ((float)a)*PI/180.0;
//
//			DriveSetSpaceVector(&rear, power, angle);

			chThdSleep(MS2ST(20));
		}

//		ConsolePrint("enc: %hu\r\n", pDrive->pTIMEncoder->CNT);

		MotorSetVoltageAngle(pMotor, 0, 0);
	}

	if(ConsoleScanCmd("va %hu %hu", &vref, &angle) == 2)
	{
//		float a = ((float)angle)*PI/180.0;
		float r = ((float)vref)/1000.0f;

		MotorSetVoltageAngle(pMotor, r, angle);
	}

	if(ConsoleScanCmd("vol %hu", &vref) == 1)
	{
		ConsolePrint("Setting dribbler voltage to: %humV\r\n", vref);

		stepDribbler.setVoltage = vref*0.001f;
	}

	if(ConsoleCmpCmd("adc"))
	{
		for(uint8_t a = 0; a < 2; a++)
		{
			ConsolePrint("ADC%hhu: ", a);
			for(uint8_t i = 0; i < DRIVE_ADC_SAMPLES; i++)
			{
				ConsolePrint("%5hu ", pMotor->adc[a][i]);
			}
			ConsolePrint("\r\n");
		}
	}

	if(ConsoleCmpCmd("current"))
	{
		ConsolePrint("iA: %d, iB: %d, iC: %d\r\n", (int32_t)(pMotor->i[0]*1000), (int32_t)(pMotor->i[1]*1000), (int32_t)(pMotor->i[2]*1000));
		ConsolePrint("iAlpha: %d, iBeta: %d\r\n", (int32_t)(pMotor->iAlpha*1000), (int32_t)(pMotor->iBeta*1000));
//		ConsolePrint("iD: %d, iQ: %d\r\n", (int32_t)(drib.pOut->iD*1000), (int32_t)(drib.pOut->iQ*1000));
	}

	if(ConsoleCmpCmd("hall"))
	{
		ConsolePrint("Hall: %hu\r\n", pMotor->halPos);
	}

	if(ConsoleCmpCmd("drib"))
	{
		ConsolePrint("RPM: %d\r\nOverload: %hu\r\ndT: %d\r\n", (int32_t)pMotor->curVelocity, (uint16_t)stepDribbler.pidSpeed.overload, pMotor->commutationDt);
	}

	int32_t rpm;
	if(ConsoleScanCmd("rpm %i", &rpm) == 1)
	{
		ConsolePrint("Setting dribbler to %i RPM\r\n", rpm);

		if(rpm > DRIVE_MOVE_THRESHOLD || rpm < -DRIVE_MOVE_THRESHOLD)
			MotorSetpointInfo(pMotor, 0);
		else
			MotorSetpointInfo(pMotor, 1);

		stepDribbler.setVelocity = rpm;
	}

	int16_t t;
	if(ConsoleScanCmd("temp calib %hd", &t) == 1)
	{
		float temp = (float)t;

		MotorCalibrateCurrentForTemp(pMotor, temp);
		ConsolePrint("Front: %d, %d, %d\r\n", (int32_t)(pMotor->tCal.offset[0]*1000),
				(int32_t)(pMotor->tCal.offset[1]*1000), (int32_t)(pMotor->tCal.offset[2]*1000));
	}

	if(ConsoleCmpCmd("temp"))
	{
		ConsolePrint("Front: %3d, %3d, %3d\t(%u)\r\n", (int32_t)pMotor->t[0], (int32_t)pMotor->t[1], (int32_t)pMotor->t[2], (SysTimeUSec()-pMotor->tMeasTime)/1000);
	}

	if(ConsoleCmpCmd("flash hex"))
	{
		for(uint32_t i = 0; i < flashFS.bytesDel+flashFS.bytesUsed+48; i += 8)
		{
			uint8_t* d = (uint8_t*)(flashFS.flashAddr+i);
			ConsolePrint("0x%04X: %02X %02X %02X %02X %02X %02X %02X %02X\r\n",
					i, d[0], d[1], d[2], d[3], d[4], d[5], d[6], d[7]);
		}
	}

	if(ConsoleCmpCmd("flash list"))
	{
		FlashFSListFiles();
	}

	if(ConsoleCmpCmd("flash erase"))
	{
		ConsolePrint("Erasing flash\r\n");

		FlashErase(flashFS.flashAddr, flashFS.flashAddr+flashFS.flashSize);
	}

	if(ConsoleCmpCmd("time"))
	{
		ConsolePrint("Time: %uus\r\n", SysTimeUSec());
	}
}
