/*
 * kicker.c
 *
 *  Created on: 20.02.2014
 *      Author: AndreR
 */

#include "kicker.h"
#include "constants.h"
#include "util/console.h"

#include "stm32f30x_exti.h"
#include "stm32f30x_gpio.h"
#include "stm32f30x_misc.h"
#include "stm32f30x_syscfg.h"
#include "stm32f30x_tim.h"
#include "stm32f30x_rcc.h"
#include "stm32f30x_spi.h"
#include "stm32f30x_dma.h"

#include "intercom_constants.h"
#include "intercom1.h"


#define KICK_ERR 0x0001

#define KICK_CHG_VAL 200	// 200/720 duty cycle

#define STATE_IDLE	0
#define STATE_CHG	1
#define STATE_HYST	2
#define STATE_FIRE	3

#define EVENT_CHG_STOP		0
#define EVENT_CHG_START		1
#define EVENT_FIRE_START	2
#define EVENT_VMAX_REACHED	3
#define EVENT_BELOW_HYST	4
#define EVENT_FIRE_DONE		5

typedef union _u32u16
{
	uint32_t u32;
	uint16_t u16[2];
} u32u16_t;

Kicker kicker;

static void eventSink(void* pUserData, uint16_t event);
static void periodic(void* pUserData);

static uint16_t getChgVal()
{
	float chg = (kicker.vCap*2.5f+50.0f);
	if(chg < 50.0f)
		chg = 50.0f;
	if(chg > 600.0f)
		chg = 600.0f;

	return (uint16_t)chg;
}

static void enableChg(uint8_t enable)
{
	if(enable)
		TIM17->CCR1 = getChgVal();
	else
		TIM17->CCR1 = 0;
}

static void processEvent(uint8_t event)
{
	switch(kicker.state)
	{
		case STATE_IDLE:
		{
			switch(event)
			{
				case EVENT_CHG_START:
				{
					enableChg(1);
					kicker.state = STATE_CHG;
				}
				break;
				case EVENT_FIRE_START:
				{
					enableChg(0);
					kicker.state = STATE_FIRE;
				}
				break;
				default: break;
			}
		}
		break;
		case STATE_CHG:
		{
			switch(event)
			{
				case EVENT_CHG_STOP:
				{
					enableChg(0);
					kicker.state = STATE_IDLE;
				}
				break;
				case EVENT_FIRE_START:
				{
					enableChg(0);
					kicker.state = STATE_FIRE;
				}
				break;
				case EVENT_VMAX_REACHED:
				{
					enableChg(0);
					kicker.state = STATE_HYST;
				}
				break;
				default: break;
			}
		}
		break;
		case STATE_HYST:
		{
			switch(event)
			{
				case EVENT_CHG_STOP:
				{
					kicker.state = STATE_IDLE;
				}
				break;
				case EVENT_BELOW_HYST:
				{
					enableChg(1);
					kicker.state = STATE_CHG;
				}
				break;
				case EVENT_FIRE_START:
				{
					kicker.state = STATE_FIRE;
				}
				break;
				default: break;
			}
		}
		break;
		case STATE_FIRE:
		{
			switch(event)
			{
				case EVENT_FIRE_DONE:
				{
					if(kicker.autoCharge)
					{
						enableChg(1);
						kicker.state = STATE_CHG;
					}
					else
						kicker.state = STATE_IDLE;
				}
				break;
				default: break;
			}
		}
		break;
		default: break;
	}
}

void EXTI0_IRQHandler()
{
	CH_IRQ_PROLOGUE();

	EXTI->PR = EXTI_PR_PR0;

//	kicker.autoCharge = 0;
//	processEvent(EVENT_CHG_STOP);

	chSysLockFromIsr();
	EventClientSendI(kicker.pEventClient, KICK_ERR);
	chSysUnlockFromIsr();

	CH_IRQ_EPILOGUE();
}

void TIM4_IRQHandler()
{
	TIM4->SR = 0;

	processEvent(EVENT_FIRE_DONE);
}

void TIM1_TRG_COM_TIM17_IRQHandler()
{
	TIM17->SR = 0;

	u32u16_t sum;
	sum.u32 = kicker.rxData32[0];

	for(uint8_t i = 1; i < SPI1_ADC_SAMPLES; i++)
		sum.u32 = __UQADD16(sum.u32, kicker.rxData32[i]);

	kicker.vCap = (sum.u16[1]*0.049560546875f)/((float)SPI1_ADC_SAMPLES);
	kicker.iChg = (sum.u16[0]*0.00732421875f)/((float)SPI1_ADC_SAMPLES);

	switch(kicker.state)
	{
	case STATE_CHG:
		if(kicker.vCap > kicker.maxVoltage)
			processEvent(EVENT_VMAX_REACHED);
		else
			TIM17->CCR1 = getChgVal();
		break;
	case STATE_HYST:
		if(kicker.vCap < kicker.maxVoltage-kicker.hystVoltage)
			processEvent(EVENT_BELOW_HYST);
		break;
	}
}

void KickerInit(EventManager* pEventManager)
{
	EXTI_InitTypeDef extiInit;
	NVIC_InitTypeDef nvicInit;
	GPIO_InitTypeDef gpioInit;
	TIM_TimeBaseInitTypeDef timeBaseInit;
	TIM_OCInitTypeDef ocInit;

	// ### Init variables
	kicker.state = STATE_IDLE;
	kicker.autoCharge = 0;
	kicker.maxVoltage = 0;
	kicker.autoDischarge = 0;
	kicker.hystVoltage = 5.0f;

	for(uint16_t i = 0; i < SPI1_ADC_SAMPLES; i++)
	{
		kicker.txData[i*2+0] = 0;		// select IN1
		kicker.txData[i*2+1] = 0x800;	// select IN2
	}

	kicker.pEventClient = EventClientInit(pEventManager, &eventSink, &periodic, MS2ST(100), 0);

	// ### KICK-ERR external interrupt
	gpioInit.GPIO_Mode = GPIO_Mode_IN;
	gpioInit.GPIO_OType = GPIO_OType_PP;
	gpioInit.GPIO_Pin = GPIO_Pin_0;
	gpioInit.GPIO_PuPd = GPIO_PuPd_UP;
	gpioInit.GPIO_Speed = GPIO_Speed_Level_1;
	GPIO_Init(GPIOA, &gpioInit);

	SYSCFG_EXTILineConfig(EXTI_PortSourceGPIOA, EXTI_PinSource0);

	extiInit.EXTI_LineCmd = ENABLE;
	extiInit.EXTI_Mode = EXTI_Mode_Interrupt;
	extiInit.EXTI_Trigger = EXTI_Trigger_Falling;
	extiInit.EXTI_Line = EXTI_Line0;
	EXTI_Init(&extiInit);

	nvicInit.NVIC_IRQChannelCmd = ENABLE;
	nvicInit.NVIC_IRQChannelPreemptionPriority = IRQL_KICK_ERR;
	nvicInit.NVIC_IRQChannelSubPriority = 0;
	nvicInit.NVIC_IRQChannel = EXTI0_IRQn;
//	NVIC_Init(&nvicInit);

	// ### ADC via SPI1
	RCC_APB2PeriphClockCmd(RCC_APB2Periph_SPI1, ENABLE);

	// init GPIO
	gpioInit.GPIO_Mode = GPIO_Mode_AF;
	gpioInit.GPIO_OType = GPIO_OType_PP;
	gpioInit.GPIO_Speed = GPIO_Speed_Level_2;
	gpioInit.GPIO_PuPd = GPIO_PuPd_NOPULL;
	gpioInit.GPIO_Pin = GPIO_Pin_4 | GPIO_Pin_5 | GPIO_Pin_6 | GPIO_Pin_7;
	GPIO_Init(GPIOA, &gpioInit);

	GPIO_PinAFConfig(GPIOA, GPIO_PinSource4, GPIO_AF_5);
	GPIO_PinAFConfig(GPIOA, GPIO_PinSource5, GPIO_AF_5);
	GPIO_PinAFConfig(GPIOA, GPIO_PinSource6, GPIO_AF_5);
	GPIO_PinAFConfig(GPIOA, GPIO_PinSource7, GPIO_AF_5);

	// init SPI1
	SPI_InitTypeDef spiInit;
	spiInit.SPI_BaudRatePrescaler = SPI_BaudRatePrescaler_16;	// = 250kHz conversion rate
	spiInit.SPI_CPHA = SPI_CPHA_1Edge;
	spiInit.SPI_CPOL = SPI_CPOL_Low;
	spiInit.SPI_CRCPolynomial = 0;
	spiInit.SPI_DataSize = SPI_DataSize_16b;
	spiInit.SPI_Direction = SPI_Direction_2Lines_FullDuplex;
	spiInit.SPI_FirstBit = SPI_FirstBit_MSB;
	spiInit.SPI_Mode = SPI_Mode_Master;
	spiInit.SPI_NSS = SPI_NSS_Hard;
	SPI_Init(SPI1, &spiInit);

	SPI_SSOutputCmd(SPI1, ENABLE);
	SPI_NSSPulseModeCmd(SPI1, ENABLE);
	SPI_RxFIFOThresholdConfig(SPI1, SPI_RxFIFOThreshold_HF);

	// init DMA
	// TX
	DMA_InitTypeDef dmaInit;
	dmaInit.DMA_BufferSize = SPI1_ADC_SAMPLES*2;
	dmaInit.DMA_DIR = DMA_DIR_PeripheralDST;
	dmaInit.DMA_M2M = 0;
	dmaInit.DMA_MemoryBaseAddr = (uint32_t)&kicker.txData;
	dmaInit.DMA_MemoryDataSize = DMA_MemoryDataSize_HalfWord;
	dmaInit.DMA_MemoryInc = DMA_MemoryInc_Enable;
	dmaInit.DMA_Mode = DMA_Mode_Circular;
	dmaInit.DMA_PeripheralBaseAddr = (uint32_t)&SPI1->DR;
	dmaInit.DMA_PeripheralDataSize = DMA_PeripheralDataSize_HalfWord;
	dmaInit.DMA_PeripheralInc = DMA_PeripheralInc_Disable;
	dmaInit.DMA_Priority = DMA_Priority_Low;
	DMA_Init(DMA1_Channel3, &dmaInit);

	// RX
	dmaInit.DMA_DIR = DMA_DIR_PeripheralSRC;
	dmaInit.DMA_MemoryBaseAddr = (uint32_t)&kicker.rxData;
	DMA_Init(DMA1_Channel2, &dmaInit);

	SPI_I2S_DMACmd(SPI1, SPI_I2S_DMAReq_Rx, ENABLE);
	SPI_I2S_DMACmd(SPI1, SPI_I2S_DMAReq_Tx, ENABLE);

    // Enable the SPI peripheral
    SPI_Cmd(SPI1, ENABLE);

    // Enable the DMA channels
    DMA_Cmd(DMA1_Channel2, ENABLE);
    DMA_Cmd(DMA1_Channel3, ENABLE);

	// ### CHG output
	RCC_APB2PeriphClockCmd(RCC_APB2Periph_TIM17, ENABLE);

	timeBaseInit.TIM_ClockDivision = 0;
	timeBaseInit.TIM_CounterMode = TIM_CounterMode_Up;
	timeBaseInit.TIM_Period = 719;	// = 100kHz
	timeBaseInit.TIM_Prescaler = 0;
	timeBaseInit.TIM_RepetitionCounter = 9;	// = 10kHz
	TIM_TimeBaseInit(TIM17, &timeBaseInit);

	TIM_ITConfig(TIM17, TIM_IT_Update, ENABLE);

	nvicInit.NVIC_IRQChannelCmd = ENABLE;
	nvicInit.NVIC_IRQChannelPreemptionPriority = IRQL_KICK_CTRL;
	nvicInit.NVIC_IRQChannelSubPriority = 0;
	nvicInit.NVIC_IRQChannel = TIM1_TRG_COM_TIM17_IRQn;
	NVIC_Init(&nvicInit);

	ocInit.TIM_Pulse = 0;
	ocInit.TIM_OCMode = TIM_OCMode_PWM2;
	ocInit.TIM_OutputState = TIM_OutputState_Enable;
	ocInit.TIM_OutputNState = TIM_OutputNState_Disable;
	ocInit.TIM_OCPolarity = TIM_OCPolarity_High;
	ocInit.TIM_OCNPolarity = TIM_OCNPolarity_High;
	ocInit.TIM_OCIdleState = TIM_OCIdleState_Set;
	ocInit.TIM_OCNIdleState = TIM_OCNIdleState_Reset;
	TIM_OC1Init(TIM17, &ocInit);

	TIM_Cmd(TIM17, ENABLE);

	TIM_CtrlPWMOutputs(TIM17, ENABLE);

	gpioInit.GPIO_Mode = GPIO_Mode_AF;
	gpioInit.GPIO_OType = GPIO_OType_PP;
	gpioInit.GPIO_Pin = GPIO_Pin_5;
	gpioInit.GPIO_PuPd = GPIO_PuPd_UP;
	gpioInit.GPIO_Speed = GPIO_Speed_Level_2;
	GPIO_Init(GPIOB, &gpioInit);

	GPIO_PinAFConfig(GPIOB, GPIO_PinSource5, GPIO_AF_10);

	// ### DIS outputs
	RCC_APB1PeriphClockCmd(RCC_APB1Periph_TIM4, ENABLE);

	timeBaseInit.TIM_ClockDivision = 0;
	timeBaseInit.TIM_CounterMode = TIM_CounterMode_Up;
	timeBaseInit.TIM_Period = 0xFFFF;
	timeBaseInit.TIM_Prescaler = 21;	// max pulse width ~20ms
	timeBaseInit.TIM_RepetitionCounter = 0;
	TIM_TimeBaseInit(TIM4, &timeBaseInit);

	ocInit.TIM_Pulse = 1;
	ocInit.TIM_OCMode = TIM_OCMode_PWM1;
	ocInit.TIM_OutputState = TIM_OutputState_Enable;
	ocInit.TIM_OutputNState = TIM_OutputNState_Disable;
	ocInit.TIM_OCPolarity = TIM_OCPolarity_High;
	ocInit.TIM_OCNPolarity = TIM_OCNPolarity_High;
	ocInit.TIM_OCIdleState = TIM_OCIdleState_Reset;
	ocInit.TIM_OCNIdleState = TIM_OCNIdleState_Reset;
	TIM_OC3Init(TIM4, &ocInit);
	TIM_OC4Init(TIM4, &ocInit);

	TIM_ARRPreloadConfig(TIM4, DISABLE);
	TIM_OC3PreloadConfig(TIM4, TIM_OCPreload_Disable);
	TIM_OC4PreloadConfig(TIM4, TIM_OCPreload_Disable);

	TIM_SelectOnePulseMode(TIM4, TIM_OPMode_Single);

	TIM_ITConfig(TIM4, TIM_IT_Update, ENABLE);

	nvicInit.NVIC_IRQChannelCmd = ENABLE;
	nvicInit.NVIC_IRQChannelPreemptionPriority = IRQL_KICK_CTRL;
	nvicInit.NVIC_IRQChannelSubPriority = 0;
	nvicInit.NVIC_IRQChannel = TIM4_IRQn;
	NVIC_Init(&nvicInit);

	gpioInit.GPIO_Mode = GPIO_Mode_AF;
	gpioInit.GPIO_OType = GPIO_OType_PP;
	gpioInit.GPIO_Pin = GPIO_Pin_8 | GPIO_Pin_9;
	gpioInit.GPIO_PuPd = GPIO_PuPd_UP;
	gpioInit.GPIO_Speed = GPIO_Speed_Level_2;
	GPIO_Init(GPIOB, &gpioInit);

	GPIO_PinAFConfig(GPIOB, GPIO_PinSource8, GPIO_AF_2);
	GPIO_PinAFConfig(GPIOB, GPIO_PinSource9, GPIO_AF_2);
}

#define LOW3 (TIM_CCMR2_OC3M_2 | TIM_CCMR2_OC3M_0)
#define LOW4 (TIM_CCMR2_OC4M_2 | TIM_CCMR2_OC4M_0)
#define HIGH3 (TIM_CCMR2_OC3M_2)
#define HIGH4 (TIM_CCMR2_OC4M_2)
#define PWM3 (TIM_CCMR2_OC3M_2 | TIM_CCMR2_OC3M_1)
#define PWM4 (TIM_CCMR2_OC4M_2 | TIM_CCMR2_OC4M_1)

void KickerFire1(float duration)
{
	uint16_t arr;
	uint16_t psc = 1299;

	processEvent(EVENT_FIRE_START);

	if(duration > 1.0f)
		duration = 1.0f;

	if(duration < 0.1f)
		psc = 129;
	if(duration < 0.02f)
		psc = 25;

	arr = duration*72e6/(psc+1);

	TIM4->CCMR2 = LOW4 | PWM3;
	TIM4->PSC = psc;
	TIM4->ARR = arr;
	TIM4->CR1 |= TIM_CR1_CEN;
}

void KickerFire2(float duration)
{
	uint16_t arr;
	uint16_t psc = 1299;

	processEvent(EVENT_FIRE_START);

	if(duration > 1.0f)
		duration = 1.0f;

	if(duration < 0.1f)
		psc = 129;
	if(duration < 0.02f)
		psc = 25;

	arr = duration*84e6/(psc+1);

	TIM4->CCMR2 = LOW3 | PWM4;
	TIM4->PSC = psc;
	TIM4->ARR = arr;
	TIM4->CR1 |= TIM_CR1_CEN;
}

void KickerEnableAutoRecharge(uint8_t enable)
{
	kicker.autoCharge = enable;
}

void KickerSetMaxVoltage(float v)
{
	if(v > 190.0f)
		v = 190.0f;

	kicker.maxVoltage = v;
}

void KickerSetVoltageHysteresis(float vHyst)
{
	kicker.hystVoltage = vHyst;
}

void KickerEnableCharge(uint8_t enable)
{
	if(enable)
		processEvent(EVENT_CHG_START);
	else
		processEvent(EVENT_CHG_STOP);
}

void KickerAutoDischarge()
{
	KickerEnableCharge(0);
	KickerEnableAutoRecharge(0);
	KickerSetMaxVoltage(0);
	kicker.autoDischarge = 1;
}

static void eventSink(void* pUserData, uint16_t event)
{
	(void)pUserData;

	if(event == KICK_ERR)
	{
//		ConsolePrint("Kicker error!\r\n");
	}
}

static void periodic(void* pUserData)
{
	(void)pUserData;

	if(kicker.autoDischarge)
	{
		if(kicker.vCap < 2.0f)
			kicker.autoDischarge = 0;

		KickerFire1(0.1f/(kicker.vCap+20.0f));
	}

	if(kicker.stress != 0.0f && kicker.state == STATE_HYST)
	{
		KickerFire1(kicker.stress);
	}

	KickerLog kLog;
	kLog.vCap = kicker.vCap;
	kLog.iChg = kicker.iChg;
	kLog.state = kicker.state | (kicker.autoCharge << 8) | (kicker.autoDischarge << 12);

	IntercomHeader icHeader;
	icHeader.address = INTERCOM_ADDRESS_MAIN | INTERCOM_ADDRESS_MEDIA;
	icHeader.section = INTERCOM_SECTION_KD;
	icHeader.command = IC_KD_KICKER_LOG;

	IntercomSend(&intercom1, &icHeader, sizeof(KickerLog), (uint8_t*)&kLog);
}
