/*
 * motor.c
 *
 *  Created on: 02.02.2014
 *      Author: AndreR
 */

/**	Timer/ADC/Motor assignments
 *
 * 			Dribbler:
 * 	TIM1:	M5
 * 	ADC3:	M5B
 * 	ADC4:	M5A
 */


#pragma GCC optimize ("Ofast")

#include "hal/motor.h"
#include "stm32f30x_gpio.h"
#include "stm32f30x_tim.h"
#include "stm32f30x_rcc.h"
#include "stm32f30x_misc.h"
#include "stm32f30x_exti.h"
#include "stm32f30x_syscfg.h"
#include "stm32f30x_adc.h"
#include "stm32f30x_dma.h"
#include "arm_math.h"

#include "constants.h"
#include "ch.h"
#include "util/sys_time.h"
#include "util/flash_fs.h"
#include "util/console.h"
#include <math.h>

#define PWM_MAX_VAL 3599
#define PI_3RD 1.047197551196598f
#define MOT_ADC_TO_CURRENT 0.0029296875f
#define MOTOR_RESISTANCE 0.2015f	// [Ohm]
#define H5Mask ((1 << 13) | (1 << 14) | (1 << 15))
#define M5Pos ((GPIOC->IDR & H5Mask) >> 13)

//Motor dribbler = { .currentMapping = {{ 0,  1}, { 1,  0}, {-1, -1}} };	// M5
//Motor dribbler = { .currentMapping = {{ 1,  1}, {-1,  0}, { 0, -1}} };	// M5
Motor motor = { .currentMapping = {{ 1,  0}, { 0,  1}, {-1, -1}} };	// M5

Motor* pMotor = &motor;

#define DRIVE_NUM_CORR 72
static const uint16_t ccrCorr[] = {
0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
0, 0, 0, 27, 29, 30, 32, 33, 34, 36,
37, 38, 40, 41, 42, 43, 44, 46, 47, 48,
49, 50, 51, 52, 53, 55, 56, 57, 58, 59,
60, 61, 62, 63, 64, 65, 66, 67, 68, 69,
70, 72 };

// usage: dir = hallDirLUT[last][cur]
static const int8_t hallDirLUT[8][8] = {
	{ 0,  0,  0,  0,  0,  0,  0,  0},
	{ 0,  0,  1,  1, -1, -1,  0,  0},
	{ 0, -1,  0, -1,  1,  0,  1,  0},
	{ 0, -1,  1,  0,  0, -1,  1,  0},
	{ 0,  1, -1,  0,  0,  1, -1,  0},
	{ 0,  1,  0,  1, -1,  0, -1,  0},
	{ 0,  0, -1, -1,  1,  1,  0,  0},
	{ 0,  0,  0,  0,  0,  0,  0,  0},
};

#define U_TEMP_MEAS 1.0f

typedef struct _TempMeasConfig
{
	uint32_t ccer;
	float voltMult[3];
} TempMeasConfig;

static const TempMeasConfig tempMeasCfgs[3] = {
		{ CCER_AB, {1, 0, 0} },
		{ CCER_BC, {0, 1, 0} },
		{ CCER_CA, {0, 0, 1} },
};

static WORKING_AREA(waTempMeasDribbler, 256);

static float min3(float a, float b, float c);
static float max3(float a, float b, float c);
static void updateADC(Motor* pDrive);
static msg_t driveTempMeasTask(void* params);

/**
 * IRQs
 */
IN_CCM void TIM1_UP_TIM16_IRQHandler()
{
	CH_IRQ_PROLOGUE();

	TIM1->SR = 0;

	updateADC(pMotor);
	if(pMotor->pCtrl && pMotor->pCtrl->onCurrentUpdate)
		(*pMotor->pCtrl->onCurrentUpdate)(pMotor);

	CH_IRQ_EPILOGUE();
}

// 100Hz speed update
IN_CCM void TIM2_IRQHandler()
{
	CH_IRQ_PROLOGUE();

	TIM2->SR = 0;

	float comDt = (float)pMotor->commutationDt;
	if(pMotor->commutationDt > 100 || pMotor->commutationDt < -100)
		pMotor->curVelocity = 10e6/comDt;	// [rpm]

	if(pMotor->pCtrl && pMotor->pCtrl->onSpeedUpdate)
		(*pMotor->pCtrl->onSpeedUpdate)(pMotor);

	CH_IRQ_EPILOGUE();
}

IN_CCM void EXTI15_10_IRQHandler()
{
	CH_IRQ_PROLOGUE();

	EXTI->PR = EXTI_PR_PR13 | EXTI_PR_PR14 | EXTI_PR_PR15;

	static uint32_t lastCOMTime;
	uint8_t lastHallPos = pMotor->halPos;
	pMotor->halPos = M5Pos;
	int8_t dtMod = hallDirLUT[lastHallPos][pMotor->halPos];

	uint32_t now = SysTimeUSec();
	int32_t dt = (int32_t)(now-lastCOMTime);

	if(dtMod == 0)
	{
		if(pMotor->commutationDt >= 0)
			pMotor->commutationDt = dt;
		else
			pMotor->commutationDt = -dt;
	}
	else
	{
		pMotor->commutationDt = dtMod*dt;
	}

	lastCOMTime = now;

	if(pMotor->pCtrl && pMotor->pCtrl->onHallUpdate)
		(*pMotor->pCtrl->onHallUpdate)(pMotor);

	CH_IRQ_EPILOGUE();
}

static void initHw()
{
	GPIO_InitTypeDef gpioInit;
	TIM_TimeBaseInitTypeDef timeBaseInit;
	TIM_OCInitTypeDef ocInit;
	TIM_BDTRInitTypeDef bdtrInit;
	NVIC_InitTypeDef nvicInit;
	ADC_InitTypeDef adcInit;
	ADC_CommonInitTypeDef adcCommonInit;
	DMA_InitTypeDef dmaInit;
	EXTI_InitTypeDef extiInit;

	// data structures
	pMotor->pTIMMotor = TIM1;
	pMotor->halPos = M5Pos;
	pMotor->curBat = 12.0f;

	// ### DMA configuration (ADC)
	dmaInit.DMA_PeripheralBaseAddr = (uint32_t)&ADC3->DR;
	dmaInit.DMA_MemoryBaseAddr = (uint32_t)pMotor->adc[1];	// B
	dmaInit.DMA_DIR = DMA_DIR_PeripheralSRC;
	dmaInit.DMA_BufferSize = DRIVE_ADC_SAMPLES;
	dmaInit.DMA_PeripheralInc = DMA_PeripheralInc_Disable;
	dmaInit.DMA_MemoryInc = DMA_MemoryInc_Enable;
	dmaInit.DMA_PeripheralDataSize = DMA_PeripheralDataSize_HalfWord;
	dmaInit.DMA_MemoryDataSize = DMA_MemoryDataSize_HalfWord;
	dmaInit.DMA_Mode = DMA_Mode_Circular;
	dmaInit.DMA_Priority = DMA_Priority_Medium;
	dmaInit.DMA_M2M = DMA_M2M_Disable;
	DMA_Init(DMA2_Channel5, &dmaInit);

	dmaInit.DMA_PeripheralBaseAddr = (uint32_t)&ADC4->DR;
	dmaInit.DMA_MemoryBaseAddr = (uint32_t)pMotor->adc[0];	// A
	DMA_Init(DMA2_Channel2, &dmaInit);

	// ### ADC
	RCC_AHBPeriphClockCmd(RCC_AHBPeriph_ADC34, ENABLE);

	gpioInit.GPIO_Pin = GPIO_Pin_13 | GPIO_Pin_14;
	gpioInit.GPIO_Mode = GPIO_Mode_AN;
	gpioInit.GPIO_PuPd = GPIO_PuPd_NOPULL;
	GPIO_Init(GPIOB, &gpioInit);	// B13, B14

	adcCommonInit.ADC_Mode = ADC_Mode_Independent;
	adcCommonInit.ADC_Clock = ADC_Clock_SynClkModeDiv1;
	adcCommonInit.ADC_DMAAccessMode = ADC_DMAAccessMode_Disabled;
	adcCommonInit.ADC_DMAMode = ADC_DMAMode_Circular;
	adcCommonInit.ADC_TwoSamplingDelay = 0;
	ADC_CommonInit(ADC3, &adcCommonInit);
	ADC_CommonInit(ADC4, &adcCommonInit);

	ADC_VoltageRegulatorCmd(ADC3, ENABLE);
	ADC_VoltageRegulatorCmd(ADC4, ENABLE);
	chThdSleep(MS2ST(4));	// stabilize regulator delay

	ADC_SelectCalibrationMode(ADC3, ADC_CalibrationMode_Single);
	ADC_SelectCalibrationMode(ADC4, ADC_CalibrationMode_Single);
	ADC_StartCalibration(ADC3);
	ADC_StartCalibration(ADC4);
	chThdSleep(MS2ST(4));	// calibration delay

	pMotor->adcCal[1] = ADC_GetCalibrationValue(ADC3);
	pMotor->adcCal[0] = ADC_GetCalibrationValue(ADC4);

	ADC_StructInit(&adcInit);
	adcInit.ADC_ContinuousConvMode = ADC_ContinuousConvMode_Disable;
	adcInit.ADC_Resolution = ADC_Resolution_12b;
	adcInit.ADC_ExternalTrigConvEvent = ADC_ExternalTrigConvEvent_10;	// TIM1_TRGO2
	adcInit.ADC_ExternalTrigEventEdge = ADC_ExternalTrigEventEdge_RisingEdge;
	adcInit.ADC_DataAlign = ADC_DataAlign_Right;
	adcInit.ADC_OverrunMode = ADC_OverrunMode_Disable;
	adcInit.ADC_AutoInjMode = ADC_AutoInjec_Disable;
	adcInit.ADC_NbrOfRegChannel = DRIVE_ADC_SAMPLES;
	ADC_Init(ADC3, &adcInit);
	ADC_Init(ADC4, &adcInit);

	for(uint8_t i = 1; i <= DRIVE_ADC_SAMPLES; i++)
	{
		ADC_RegularChannelConfig(ADC3, ADC_Channel_5,  i, ADC_SampleTime_181Cycles5);
		ADC_RegularChannelConfig(ADC4, ADC_Channel_4,  i, ADC_SampleTime_181Cycles5);
	}

	// Configures the ADC DMA
	ADC_DMAConfig(ADC3, ADC_DMAMode_Circular);
	ADC_DMAConfig(ADC4, ADC_DMAMode_Circular);

	// Enable the ADC DMA
	ADC_DMACmd(ADC3, ENABLE);
	ADC_DMACmd(ADC4, ENABLE);

	ADC_Cmd(ADC3, ENABLE);
	ADC_Cmd(ADC4, ENABLE);
	chThdSleep(MS2ST(4));	// ADC ready delay

	// Enable the DMA channel
	DMA_Cmd(DMA2_Channel5, ENABLE);
	DMA_Cmd(DMA2_Channel2, ENABLE);

	// ### Motors
	nvicInit.NVIC_IRQChannelCmd = ENABLE;
	nvicInit.NVIC_IRQChannelPreemptionPriority = IRQL_MOTOR_UPDATE;
	nvicInit.NVIC_IRQChannelSubPriority = 0;
	nvicInit.NVIC_IRQChannel = TIM1_UP_TIM16_IRQn;
	NVIC_Init(&nvicInit);

	RCC_APB2PeriphClockCmd(RCC_APB2Periph_TIM1, ENABLE);
	RCC_TIMCLKConfig(RCC_TIM1CLK_PLLCLK);

	GPIOA->ODR |= GPIO_Pin_8 | GPIO_Pin_9 | GPIO_Pin_10 | GPIO_Pin_11 | GPIO_Pin_12;
	GPIOB->ODR |= GPIO_Pin_15;

	gpioInit.GPIO_Mode = GPIO_Mode_AF;
	gpioInit.GPIO_OType = GPIO_OType_PP;
	gpioInit.GPIO_PuPd = GPIO_PuPd_UP;
	gpioInit.GPIO_Speed = GPIO_Speed_2MHz;
	gpioInit.GPIO_Pin = GPIO_Pin_8 | GPIO_Pin_9 | GPIO_Pin_10 | GPIO_Pin_11 | GPIO_Pin_12;
	GPIO_Init(GPIOA, &gpioInit);

	gpioInit.GPIO_Pin = GPIO_Pin_15;
	GPIO_Init(GPIOB, &gpioInit);

	timeBaseInit.TIM_ClockDivision = 0;
	timeBaseInit.TIM_CounterMode = TIM_CounterMode_CenterAligned1;
	timeBaseInit.TIM_Period = PWM_MAX_VAL;	// = 3599 = 20kHz PWM (2*10kHz due to center aligned mode)
	timeBaseInit.TIM_Prescaler = 0;
	timeBaseInit.TIM_RepetitionCounter = 1;	// update event only every second count (e.g. downcounting) = 20kHz
	TIM_TimeBaseInit(TIM1, &timeBaseInit);

	ocInit.TIM_OCMode = TIM_OCMode_PWM2;
	ocInit.TIM_OutputState = TIM_OutputState_Enable;
	ocInit.TIM_OutputNState = TIM_OutputNState_Enable;
	ocInit.TIM_OCPolarity = TIM_OCPolarity_High;
	ocInit.TIM_OCNPolarity = TIM_OCNPolarity_High;
	ocInit.TIM_OCIdleState = TIM_OCIdleState_Set;
	ocInit.TIM_OCNIdleState = TIM_OCNIdleState_Reset;
	TIM_OC1Init(TIM1, &ocInit);
	TIM_OC2Init(TIM1, &ocInit);
	TIM_OC3Init(TIM1, &ocInit);
	ocInit.TIM_Pulse = 864;	// trigger ADC conversion 6us after update event (allow IRQ to sum up previous value)
	TIM_OC4Init(TIM1, &ocInit);

	bdtrInit.TIM_AutomaticOutput = TIM_AutomaticOutput_Enable;
	bdtrInit.TIM_Break = TIM_Break_Disable;
	bdtrInit.TIM_BreakPolarity = TIM_BreakPolarity_High;
	bdtrInit.TIM_DeadTime = 0;
	bdtrInit.TIM_LOCKLevel = TIM_LOCKLevel_OFF;
	bdtrInit.TIM_OSSIState = TIM_OSSIState_Disable;
	bdtrInit.TIM_OSSRState = TIM_OSSRState_Enable;
	TIM_BDTRConfig(TIM1, &bdtrInit);

	TIM1->EGR = TIM_EGR_UG;

	TIM_SelectOutputTrigger2(TIM1, TIM_TRGO2Source_OC4Ref);	// OC4 connected to ADC

	TIM_ITConfig(TIM1, TIM_IT_Update, ENABLE);

	TIM_Cmd(TIM1, ENABLE);

	// enable trigger input
	ADC_StartConversion(ADC3);
	ADC_StartConversion(ADC4);

	TIM_CtrlPWMOutputs(TIM1, ENABLE);

	// TIM1
	GPIO_PinAFConfig(GPIOA, GPIO_PinSource8, GPIO_AF_6);
	GPIO_PinAFConfig(GPIOA, GPIO_PinSource9, GPIO_AF_6);
	GPIO_PinAFConfig(GPIOA, GPIO_PinSource10, GPIO_AF_6);
	GPIO_PinAFConfig(GPIOA, GPIO_PinSource11, GPIO_AF_6);
	GPIO_PinAFConfig(GPIOA, GPIO_PinSource12, GPIO_AF_6);
	GPIO_PinAFConfig(GPIOB, GPIO_PinSource15, GPIO_AF_4);

	// ### Hall Sensors
	gpioInit.GPIO_Mode = GPIO_Mode_IN;
	gpioInit.GPIO_OType = GPIO_OType_PP;
	gpioInit.GPIO_PuPd = GPIO_PuPd_UP;
	gpioInit.GPIO_Speed = GPIO_Speed_2MHz;
	gpioInit.GPIO_Pin = GPIO_Pin_13 | GPIO_Pin_14 | GPIO_Pin_15;
	GPIO_Init(GPIOC, &gpioInit);

	SYSCFG_EXTILineConfig(EXTI_PortSourceGPIOC, EXTI_PinSource13);
	SYSCFG_EXTILineConfig(EXTI_PortSourceGPIOC, EXTI_PinSource14);
	SYSCFG_EXTILineConfig(EXTI_PortSourceGPIOC, EXTI_PinSource15);

	extiInit.EXTI_LineCmd = ENABLE;
	extiInit.EXTI_Mode = EXTI_Mode_Interrupt;
	extiInit.EXTI_Trigger = EXTI_Trigger_Rising_Falling;
	extiInit.EXTI_Line = EXTI_Line13;
	EXTI_Init(&extiInit);

	extiInit.EXTI_Line = EXTI_Line14;
	EXTI_Init(&extiInit);

	extiInit.EXTI_Line = EXTI_Line15;
	EXTI_Init(&extiInit);

	nvicInit.NVIC_IRQChannelCmd = ENABLE;
	nvicInit.NVIC_IRQChannelPreemptionPriority = IRQL_MOTOR_UPDATE+1;
	nvicInit.NVIC_IRQChannelSubPriority = 0;
	nvicInit.NVIC_IRQChannel = EXTI15_10_IRQn;
	NVIC_Init(&nvicInit);

	// Control timer
	RCC_APB1PeriphClockCmd(RCC_APB1Periph_TIM2, ENABLE);

	timeBaseInit.TIM_ClockDivision = 0;
	timeBaseInit.TIM_CounterMode = TIM_CounterMode_Up;
	timeBaseInit.TIM_Period = 999;	// = 10ms = 100Hz
	timeBaseInit.TIM_Prescaler = 719;	// = 10us
	timeBaseInit.TIM_RepetitionCounter = 0;
	TIM_TimeBaseInit(TIM2, &timeBaseInit);

	TIM_ITConfig(TIM2, TIM_IT_Update, ENABLE);

	nvicInit.NVIC_IRQChannelPreemptionPriority = IRQL_MOTOR_UPDATE+2;
	nvicInit.NVIC_IRQChannel = TIM2_IRQn;
	NVIC_Init(&nvicInit);

	TIM_Cmd(TIM2, ENABLE);
}

/**
 * PRIVATE functions
 */
IN_CCM static float min3(float a, float b, float c)
{
	if(a < b && a < c)
		return a;

	if(b < a && b < c)
		return b;

	return c;
}

IN_CCM static float max3(float a, float b, float c)
{
	if(a > b && a > c)
		return a;

	if(b > a && b > c)
		return b;

	return c;
}

typedef union _u32u16
{
	uint32_t u32;
	uint16_t u16[2];
} u32u16_t;

IN_CCM static void updateADC(Motor* pDrive)
{
	float i0f;
	float i1f;

	uint16_t i016;
	uint16_t i116;

	u32u16_t i1u, i0u;

	i0u.u32 = pDrive->adc32[0][0];
	i1u.u32 = pDrive->adc32[1][0];

	for(uint8_t i = 1; i < DRIVE_ADC_SAMPLES/2; i++)
	{
		i0u.u32 = __UQADD16(i0u.u32, pDrive->adc32[0][i]);
		i1u.u32 = __UQADD16(i1u.u32, pDrive->adc32[1][i]);
	}

	i016 = i0u.u16[0]+i0u.u16[1];
	i116 = i1u.u16[0]+i1u.u16[1];

	i0f = i016*(1.0f/DRIVE_ADC_SAMPLES);
	i1f = i116*(1.0f/DRIVE_ADC_SAMPLES);

	i0f = (i0f-1706.6667f)*-MOT_ADC_TO_CURRENT; // convert to real current
	i1f = (i1f-1706.6667f)*-MOT_ADC_TO_CURRENT; // convert to real current

	pDrive->i[0] = pDrive->currentMapping[0][0]*i0f + pDrive->currentMapping[0][1]*i1f;
	pDrive->i[1] = pDrive->currentMapping[1][0]*i0f + pDrive->currentMapping[1][1]*i1f;
	pDrive->i[2] = pDrive->currentMapping[2][0]*i0f + pDrive->currentMapping[2][1]*i1f;

	pDrive->iAlpha = pDrive->i[0];
	pDrive->iBeta = 0.57735026918963f*pDrive->i[0] + 1.154700538379252f*pDrive->i[1];

	if(pDrive->measSample < DRIVE_MEAS_SAMPLES)
	{
		pDrive->meas[0][pDrive->measSample] = (int16_t)(i0f*1000);
		pDrive->meas[1][pDrive->measSample] = (int16_t)(i1f*1000);
		++pDrive->measSample;
	}
}

/**
 * PUBLIC functions
 */
void MotorInit()
{
	initHw();

	chMtxInit(&pMotor->tMeasMutex);

	int16_t result = FlashFSOpenOrCreate("drib/tCal", 0, &pMotor->tCal, sizeof(DriveTempCalib), &pMotor->pTCalFile);
	if(result == FLASH_FS_CREATED)
	{
		memset(&pMotor->tCal, 0, sizeof(DriveTempCalib));
		FlashFSWrite(&pMotor->pTCalFile, &pMotor->tCal, sizeof(DriveTempCalib));
	}

	chThdCreateStatic(waTempMeasDribbler, sizeof(waTempMeasDribbler), NORMALPRIO-5, driveTempMeasTask, pMotor);

	pMotor->zeroSetpoint = 1;
}

void MotorSetController(MotorCtrl* pCtrlDribbler)
{
	// call exit actions
	if(pMotor->pCtrl && pMotor->pCtrl->onExit)
		(*pMotor->pCtrl->onExit)(pMotor);

	MotorSetPower(pMotor, 5000);

	pMotor->pCtrl = pCtrlDribbler;

	// call init actions
	if(pMotor->pCtrl && pMotor->pCtrl->onInit)
		(*pMotor->pCtrl->onInit)(pMotor);
}

void MotorSetPower(Motor* pDrive, uint16_t power)
{
	pDrive->pTIMMotor->CCR1 = power;
	pDrive->pTIMMotor->CCR2 = power;
	pDrive->pTIMMotor->CCR3 = power;
}

void MotorEnablePhase(Motor* pDrive, uint32_t ccer)
{
	uint32_t oldCCER = pDrive->pTIMMotor->CCER & (TIM_CCER_CC1P | TIM_CCER_CC1NP | TIM_CCER_CC2P | TIM_CCER_CC2NP | TIM_CCER_CC3P | TIM_CCER_CC3NP);
	pDrive->pTIMMotor->CCER = oldCCER | ccer;
}

void MotorGetSteadyPhaseCurrents(Motor* pDrive, float* i)
{
	float Vcom = (pDrive->u[0]+pDrive->u[1]+pDrive->u[2])/3;

	i[0] = (pDrive->u[0]-Vcom)/MOTOR_RESISTANCE;
	i[1] = (pDrive->u[1]-Vcom)/MOTOR_RESISTANCE;
	i[2] = (pDrive->u[2]-Vcom)/MOTOR_RESISTANCE;
}

void MotorGetTemperatures(Motor* pDrive, float* t)
{
	float r[3];
	float Vcom = (pDrive->u[0]+pDrive->u[1]+pDrive->u[2])/3;

	for(uint8_t i = 0; i < 3; i++)
	{
		r[i] = (pDrive->u[i]-Vcom)/(pDrive->i[i]+0.00001f);
		r[i] = fabsf(r[i]);
		t[i] = 25.0f+(r[i]/MOTOR_RESISTANCE-1.0f)/3.93e-3f;
	}
}

void MotorGetMeasAvg(Motor* pMotor, float* avg)
{
	float iReal[2];

	iReal[0] = 0;
	iReal[1] = 0;

	for(uint16_t n = 0; n < DRIVE_MEAS_SAMPLES; n++)
	{
		iReal[0] += pMotor->meas[0][n];
		iReal[1] += pMotor->meas[1][n];
	}

	iReal[0] *= 1.0f/(DRIVE_MEAS_SAMPLES*1000.0f);
	iReal[1] *= 1.0f/(DRIVE_MEAS_SAMPLES*1000.0f);

	avg[0] = pMotor->currentMapping[0][0]*iReal[0] + pMotor->currentMapping[0][1]*iReal[1];
	avg[1] = pMotor->currentMapping[1][0]*iReal[0] + pMotor->currentMapping[1][1]*iReal[1];
	avg[2] = pMotor->currentMapping[2][0]*iReal[0] + pMotor->currentMapping[2][1]*iReal[1];
}

IN_CCM static void setCCR(Motor* pDrive, uint16_t a, uint16_t b, uint16_t c)
{
	if(a < DRIVE_NUM_CORR)
		a = ccrCorr[a];

	if(b < DRIVE_NUM_CORR)
		b = ccrCorr[b];

	if(c < DRIVE_NUM_CORR)
		c = ccrCorr[c];

	pDrive->pTIMMotor->CCR1 = a;
	pDrive->pTIMMotor->CCR2 = b;
	pDrive->pTIMMotor->CCR3 = c;
}

IN_CCM void MotorSetPhaseVoltagesSVPWM(Motor* pDrive, float Va, float Vb, float Vc)
{
	float Vd = pDrive->curBat;
	if(Vd < 9.0f || Vd > 17.0f)
		Vd = 11.1f;

	float Vcom = Vd/2-0.5f*(max3(Va, Vb, Vc)+min3(Va, Vb, Vc));

	pDrive->u[0] = Va+Vcom;
	pDrive->u[1] = Vb+Vcom;
	pDrive->u[2] = Vc+Vcom;

	float da = 1/Vd*(Va+Vcom);
	float db = 1/Vd*(Vb+Vcom);
	float dc = 1/Vd*(Vc+Vcom);

	if(da > 1.0f)
		da = 1.0f;
	if(da < 0.0f)
		da = 0.0f;

	if(db > 1.0f)
		db = 1.0f;
	if(db < 0.0f)
		db = 0.0f;

	if(dc > 1.0f)
		dc = 1.0f;
	if(dc < 0.0f)
		dc = 0.0f;

	setCCR(pDrive, da*PWM_MAX_VAL, db*PWM_MAX_VAL, dc*PWM_MAX_VAL);
}

IN_CCM void MotorSetPhaseVoltagesDirect(Motor* pDrive, float Va, float Vb, float Vc)
{
	float Vd = pDrive->curBat;
	if(Vd < 9.0f || Vd > 17.0f)
		Vd = 11.1f;

	pDrive->u[0] = Va;
	pDrive->u[1] = Vb;
	pDrive->u[2] = Vc;

	float da = Va/Vd;
	float db = Vb/Vd;
	float dc = Vc/Vd;

	if(da > 1.0f)
		da = 1.0f;
	if(da < 0.0f)
		da = 0.0f;

	if(db > 1.0f)
		db = 1.0f;
	if(db < 0.0f)
		db = 0.0f;

	if(dc > 1.0f)
		dc = 1.0f;
	if(dc < 0.0f)
		dc = 0.0f;

	setCCR(pDrive, da*PWM_MAX_VAL, db*PWM_MAX_VAL, dc*PWM_MAX_VAL);
}

IN_CCM void MotorSetAlphaBeta(Motor* pDrive, float Valpha, float Vbeta)
{
	float Va = Valpha;
	float Vb = -0.5f*Valpha + 0.86602540378444f*Vbeta;
	float Vc = -0.5f*Valpha - 0.86602540378444f*Vbeta;

	MotorSetPhaseVoltagesSVPWM(pDrive, Va, Vb, Vc);
}

IN_CCM void MotorSetVoltageAngle(Motor* pMotor, float Vref, float angle)
{
	float sin, cos;
	while(angle > 180.0f)
		angle -= 360.0f;

	arm_sin_cos_f32(angle, &sin, &cos);

	MotorSetAlphaBeta(pMotor, cos*Vref, sin*Vref);
}

static uint8_t getCurrentForTemp(Motor* pMotor, const TempMeasConfig* pCfg, float* pI)
{
	MotorEnablePhase(pMotor, pCfg->ccer);

	uint16_t n;
	for(n = 1; n < 100; n++)
	{
		if(pMotor->pCtrl)
			break;

		float u = U_TEMP_MEAS*n*0.01f;
		MotorSetPhaseVoltagesDirect(pMotor, u*pCfg->voltMult[0], u*pCfg->voltMult[1], u*pCfg->voltMult[2]);
		chThdSleep(MS2ST(2));
	}

	if(n < 100)
		return 1;	// measurement interrupted

	chThdSleep(MS2ST(50));

	pMotor->measSample = 0;
	chThdSleep(MS2ST(20));
	MotorGetMeasAvg(pMotor, pI);

	for(uint8_t i = 0; i < 3; i++)
		pI[i] *= -1;

	if(pMotor->pCtrl)
		return 1;

	return 0;
}

static uint8_t getCombinedResistance(Motor* pMotor, float* pRcomb)
{
	float iReal[3];

	for(uint8_t i = 0; i < 3; i++)
	{
		if(getCurrentForTemp(pMotor, &tempMeasCfgs[i], iReal))
			return 1;

		iReal[i] += pMotor->tCal.offset[i];

		pRcomb[i] = U_TEMP_MEAS/iReal[i];
	}

	return 0;
}

void MotorCalibrateCurrentForTemp(Motor* pMotor, float tNow)
{
	// disable controller
	const float RNow = MOTOR_RESISTANCE*(1+3.93e-3f*(tNow-25));
	pMotor->pCtrlInactive = pMotor->pCtrl;
	pMotor->pCtrl = 0;

	float iReal[3];

	chMtxLock(&pMotor->tMeasMutex);

	for(uint16_t i = 0; i < 3; i++)
	{
		getCurrentForTemp(pMotor, &tempMeasCfgs[i], iReal);

		ConsolePrint("%hu: %d, %d, %d\r\n", i, (int32_t)(iReal[0]*1000), (int32_t)(iReal[1]*1000), (int32_t)(iReal[2]*1000));

		pMotor->tCal.offset[i] = (U_TEMP_MEAS*1.0f/(2*RNow))-iReal[i];
	}

	MotorEnablePhase(pMotor, CCER_ABC);
	MotorSetPhaseVoltagesDirect(pMotor, 0, 0, 0);

	chMtxUnlock();

	// re-enable controller
	pMotor->pCtrl = pMotor->pCtrlInactive;
	pMotor->pCtrlInactive = 0;

	FlashFSWrite(&pMotor->pTCalFile, &pMotor->tCal, sizeof(DriveTempCalib));
}


uint8_t MotorTempMeas(Motor* pMotor)
{
	float Rcomb[3]; 	// that is Rab, Rbc, Rca
	float R[3];

	// disable controller
	pMotor->pCtrlInactive = pMotor->pCtrl;
	pMotor->pCtrl = 0;

	chMtxLock(&pMotor->tMeasMutex);

	if(getCombinedResistance(pMotor, Rcomb))
		return 1;

	// check if controller is already back online, measurement has then been interrupted
	// measurement is then invalid
	if(pMotor->pCtrl)
		return 1;

	MotorEnablePhase(pMotor, CCER_ABC);
	MotorSetPhaseVoltagesDirect(pMotor, 0, 0, 0);

	chMtxUnlock();

	// re-enable controller
	pMotor->pCtrl = pMotor->pCtrlInactive;
	pMotor->pCtrlInactive = 0;

	R[2] = (Rcomb[1]-Rcomb[0]+Rcomb[1])*0.5f;
	R[1] = Rcomb[1]-R[2];
	R[0] = Rcomb[2]-R[2];

	for(uint8_t i = 0; i < 3; i++)
		pMotor->t[i] = 25.0f+(R[i]/MOTOR_RESISTANCE-1.0f)/3.93e-3f;

	pMotor->tMeasTime = SysTimeUSec();

	return 0;
}

void MotorSetpointInfo(Motor* pMotor, uint8_t zeroSetpoint)
{
	if(!zeroSetpoint && pMotor->pCtrlInactive)
	{
		pMotor->pCtrl = pMotor->pCtrlInactive;
		pMotor->pCtrlInactive = 0;
	}

	pMotor->zeroSetpoint = zeroSetpoint;
}

static msg_t driveTempMeasTask(void* params)
{
	Motor* pMotor = (Motor*)params;

	chThdSleep(S2ST(5));

	while(1)
	{
		uint8_t doMeas = 1;
		for(uint8_t i = 0; i < 100; i++)
		{
			chThdSleep(MS2ST(5));

			if(!pMotor->zeroSetpoint || pMotor->curVelocity > DRIVE_MOVE_THRESHOLD || pMotor->curVelocity < -DRIVE_MOVE_THRESHOLD)
			{
				doMeas = 0;
				break;
			}
		}

		if(doMeas == 0)
			continue;

		MotorTempMeas(pMotor);

		chThdSleep(MS2ST(5000));
	}

	return 0;
}

