/*
 * motor.h
 *
 *  Created on: 02.02.2014
 *      Author: AndreR
 */

#ifndef MOTOR_H_
#define MOTOR_H_

#include <stdint.h>

#include "stm32f30x.h"
#include "intercom_constants.h"
#include "util/flash_fs.h"
#include "ch.h"

#define IN_CCM __attribute__((section(".ccmtext")))

#define DRIVE_ADC_SAMPLES 16
#define DRIVE_MEAS_SAMPLES 100
#define MOTOR_KN 3760 // [RPM/Volt]
#define DRIVE_MOVE_THRESHOLD 100

// W3  W2  W1
//  5   5   5
#define CCER_A 0x005
#define CCER_B 0x050
#define CCER_C 0x500
#define CCER_AB (CCER_A | CCER_B)
#define CCER_BC (CCER_B | CCER_C)
#define CCER_CA (CCER_C | CCER_A)
#define CCER_ABC (CCER_A | CCER_B | CCER_C)

struct _Motor;

typedef void(*InitCtrlCb)(struct _Motor*);
typedef void(*ExitCtrlCb)(struct _Motor*);
typedef void(*CurrentUpdateCb)(struct _Motor*);
typedef void(*SpeedUpdateCb)(struct _Motor*);
typedef void(*HallUpdateCb)(struct _Motor*);

typedef struct _DriveCtrl
{
	InitCtrlCb onInit;
	ExitCtrlCb onExit;
	CurrentUpdateCb onCurrentUpdate;
	SpeedUpdateCb onSpeedUpdate;
	HallUpdateCb onHallUpdate;

	void* pUserData;
} MotorCtrl;

typedef struct _DriveTempCalib
{
	float offset[3];
} __attribute__((aligned(4),packed)) DriveTempCalib;

typedef struct _Motor
{
	// ### Setup ###
	// Motor timer
	TIM_TypeDef* pTIMMotor;

	const float currentMapping[3][2];

	// ADC calibration values
	uint32_t adcCal[2];

	// ### Measured values ###
	// ADC DMA storage area
	union
	{
		uint16_t adc[2][DRIVE_ADC_SAMPLES];
		uint32_t adc32[2][DRIVE_ADC_SAMPLES/2];
	};

	volatile int32_t commutationDt;	// deltaT between two hall IRQs
	volatile float curVelocity;	// current velocity [rounds per minute]
	float i[3];	// current on each phase [A]
	float iAlpha;	// current on stator-fixed alpha axis [A]
	float iBeta;	// current on stator-fixed beta axis [A]

	DriveTempCalib tCal;	// current sensor offset for temperature measurement
	FlashFile* pTCalFile;
	float t[3];		// temperature of each winding (measurement needs to be explicetly triggered)
	uint32_t tMeasTime;	// system time of temperature measurement
	Mutex tMeasMutex;

	uint16_t halPos;

	// ### control related ###
	MotorCtrl* pCtrl;	// current control instance
	MotorCtrl* pCtrlInactive;

	// ### output/info values ###
	float u[3];

	// ### required input ###
	float curBat;	// [V]
	volatile uint8_t zeroSetpoint;

	// ### data sampling for test purposes ###
	int16_t meas[2][DRIVE_MEAS_SAMPLES];
	uint16_t measSample;
} Motor;

extern Motor motor;
extern Motor* pMotor;

void MotorInit();
void MotorSetController(MotorCtrl* pCtrlDribbler);
void MotorSetPower(Motor* pDrive, uint16_t power);
void MotorSetPhaseVoltagesSVPWM(Motor* pDrive, float Va, float Vb, float Vc);
void MotorSetPhaseVoltagesDirect(Motor* pDrive, float Va, float Vb, float Vc);
void MotorSetAlphaBeta(Motor* pDrive, float Valpha, float Vbeta);
void MotorSetVoltageAngle(Motor* pDrive, float Vref, float angle);
void MotorGetSteadyPhaseCurrents(Motor* pDrive, float* i);
void MotorGetTemperatures(Motor* pDrive, float* t);
void MotorEnablePhase(Motor* pDrive, uint32_t ccer);
void MotorCalibrateCurrentForTemp(Motor* pMotor, float tNow);
uint8_t MotorTempMeas(Motor* pMotor);
void MotorSetpointInfo(Motor* pMotor, uint8_t zeroSetpoint);

#endif /* MOTOR_H_ */
