/*
 * kick.c
 *
 *  Created on: 21.05.2014
 *      Author: AndreR
 */

#include "kick.h"
#include "commands.h"
#include "util/fixed.h"
#include "intercom1.h"
#include "intercom_constants.h"
#include "hal/kicker.h"
#include "hal/motor.h"
#include "ctrl_6step.h"
#include "supply.h"

#define KICKER_MODE_FORCE		0
#define KICKER_MODE_ARM			1
#define KICKER_MODE_DISARM		2
#define KICKER_MODE_DRIBBLER	3

#define EVENT_BARRIER_INTERRUPTED 0
#define EVENT_DRIBBLER_SPEED_REACHED 1

#define KICKER_STATE_STANDBY		0
#define KICKER_STATE_IR_ARMED		1
#define KICKER_STATE_DRIBBLER_ARMED	2
#define KICKER_STATE_DISCHARGE		3

KickGlobal kick;

static void kickCallback(IntercomHeader* pHeader, uint16_t dataLength, uint8_t* pData, void* pUserData)
{
	(void)pUserData;
	(void)dataLength;

	if(pHeader->command != 0)
		return;

	PacketHeader* pPacketHeader = (PacketHeader*)pData;
	uint8_t* pPacketData = pData+PACKET_HEADER_SIZE;

	switch(pPacketHeader->section)
	{
		case SECTION_SYSTEM:
		{
			switch(pPacketHeader->cmd)
			{
				case CMD_SYSTEM_SET_LOGS:
				{
					SystemSetLogs* pLogs = (SystemSetLogs*)pPacketData;

					kick.logEnabled = pLogs->kicker;
				}
				break;
			}
		}
		break;
		case SECTION_KICKER:
		{
			switch(pPacketHeader->cmd)
			{
				case CMD_KICKER_CHARGE_AUTO:
				{
					KickerChargeAuto* pCharge = (KickerChargeAuto*)pPacketData;

					if(pCharge->maxCapLevel == 0)
					{
						KickDischarge();
					}
					else
					{
						KickerSetMaxVoltage(pCharge->maxCapLevel*0.1f);
						KickerEnableCharge(1);
						KickerEnableAutoRecharge(1);
					}
				}
				break;
				case CMD_KICKER_KICKV2:
				{
					KickerKickV2* pKick = (KickerKickV2*)pPacketData;

					FixedPoint142* pDuration = (FixedPoint142*)&pKick->duration;
					float time = pDuration->value;
					if(pDuration->exp == 1)
						time *= 10;
					if(pDuration->exp == 2)
						time *= 100;
					if(pDuration->exp == 3)
						time *= 1000;
					time /= 10.0f;	// now in [us]
					time *= 1e-6f;	// now in [s]

					if(pKick->mode == KICKER_MODE_FORCE)
					{
						KickFire(time, pKick->device);
					}

					if(pKick->mode == KICKER_MODE_ARM)
					{
						KickArmIr(time, pKick->device);
					}

					if(pKick->mode == KICKER_MODE_DRIBBLER)
					{
						KickArmDribbler(time, pKick->device);
					}

					if(pKick->mode == KICKER_MODE_DISARM)
					{
						KickDisarm();
					}
				}
				break;
			}
		}
		break;
		case SECTION_MOTOR:
		{
			switch(pPacketHeader->cmd)
			{
				case CMD_MOTOR_DRIBBLE:
				{
					MotorDribble* pDribble = (MotorDribble*)pPacketData;

					if(pDribble->rpm > DRIVE_MOVE_THRESHOLD || pDribble->rpm < -DRIVE_MOVE_THRESHOLD)
						MotorSetpointInfo(pMotor, 0);
					else
						MotorSetpointInfo(pMotor, 1);

					stepDribbler.setVelocity = -pDribble->rpm;
				}
				break;
				case CMD_MOTOR_SET_PID_PARAMS:
				{
					MotorSetPidParams* pParams = (MotorSetPidParams*)pPacketData;

					if(pParams->id == 4)
					{
						Ctrl6StepSetSpeedPIDParams(&stepDribbler, pParams->Kp, pParams->Ki, pParams->Kd);
					}
				}
				break;
			}
		}
		break;
	}
}

static void irCallback()
{
	kick.irDiff = supply.irResult[0] - supply.irResult[1];

	if(kick.irDiff < kick.irThreshold && kick.irFilterCounter < kick.irFilterThreshold)
		kick.irFilterCounter++;
	else
		kick.irFilterCounter = 0;

	if(kick.irFilterCounter >= kick.irFilterThreshold)
	{
		chSysLockFromIsr();
		chMBPostI(&kick.eventQueue, EVENT_BARRIER_INTERRUPTED);
		chSysUnlockFromIsr();
	}
}

void KickInit()
{
	kick.state = KICKER_STATE_STANDBY;
	kick.irFilterCounter = 0;
	kick.irFilterThreshold = 10;
	kick.irThreshold = 350;	// [mV]
	kick.logEnabled = 0;

	IntercomClient client;
	client.pUserData = 0;
	client.rxCallback = &kickCallback;
	client.section = INTERCOM_SECTION_COMMAND;

	IntercomAddClient(&intercom1, client);

	supply.irDoneCb = &irCallback;

	chMBInit(&kick.eventQueue, kick.eventQueueData, KICK_EVENT_QUEUE_SIZE);
}

msg_t KickTask(void* params)
{
	(void)params;

	msg_t event;
	msg_t result;

	uint32_t lastStatusTime = chTimeNow();

	static uint8_t statusBuf[sizeof(KickerStatusV3)+PACKET_HEADER_SIZE];

	while(1)
	{
		result = chMBFetch(&kick.eventQueue, &event, MS2ST(10));
		if(result == RDY_OK)
		{
			switch(event)
			{
				case EVENT_BARRIER_INTERRUPTED:
				{
					if(kick.state == KICKER_STATE_IR_ARMED)
					{
						if(kicker.vCap > 0.9f*kicker.maxVoltage)
							KickFire(kick.armed.time, kick.armed.device);
						else
							kick.irFilterCounter = 0;
					}
				}
				break;
				case EVENT_DRIBBLER_SPEED_REACHED:
				{
					// TODO: not completed yet
				}
				break;
			}
		}

		if(kick.state == KICKER_STATE_DRIBBLER_ARMED && fabsf(pMotor->curVelocity - kick.armed.rpm) < 250.0f)
		{
			chMBPost(&kick.eventQueue, EVENT_DRIBBLER_SPEED_REACHED, TIME_IMMEDIATE);
		}

		if(chTimeElapsedSince(lastStatusTime) > MS2ST(50) && kick.logEnabled)
		{
			lastStatusTime = chTimeNow();

			kick.latestStatus.cap = kicker.vCap*10.0f;
			kick.latestStatus.chg = kicker.iChg*1000.0f;
			kick.latestStatus.ir = kick.irDiff*10.0f;

			PacketHeader* pPacketHeader = (PacketHeader*)statusBuf;
			pPacketHeader->section = SECTION_KICKER;
			pPacketHeader->cmd = CMD_KICKER_STATUSV3;

			memcpy(statusBuf+PACKET_HEADER_SIZE, &kick.latestStatus, sizeof(KickerStatusV3));

			IntercomHeader iHeader;
			iHeader.address = INTERCOM_ADDRESS_MAIN;
			iHeader.section = INTERCOM_SECTION_NETWORK;
			iHeader.command = 0;

			IntercomSend(&intercom1, &iHeader, sizeof(statusBuf), statusBuf);
		}
	}

	return 0;
}

void KickFire(float duration, uint8_t device)
{
	if(kicker.vCap < 0.9f*kicker.maxVoltage)
		return;

	if(device == KICKER_DEVICE_STRAIGHT)
		KickerFire1(duration);

	if(device == KICKER_DEVICE_CHIP)
		KickerFire2(duration);

	kick.state = KICKER_STATE_STANDBY;
}

void KickArmIr(float fireDuration, uint8_t device)
{
	kick.armed.device = device;
	kick.armed.time = fireDuration;
	kick.state = KICKER_STATE_IR_ARMED;
	kick.irFilterCounter = 0;
}

void KickArmDribbler(float fireDuration, uint8_t device)
{
	kick.armed.device = device;
	kick.armed.time = fireDuration;
	kick.state = KICKER_STATE_IR_ARMED;	// TODO: implement correctly
	kick.irFilterCounter = 0;
}

void KickDisarm()
{
	kick.state = KICKER_STATE_STANDBY;
}

void KickDischarge()
{
	KickerAutoDischarge();
	kick.state = KICKER_STATE_STANDBY;
}
