/*
 * bootloader.c
 *
 *  Created on: 03.06.2013
 *      Author: AndreR
 */

#include "bootloader.h"

#include "util/crc.h"
#include "util/fifo_lin.h"
#include "util/log.h"
#include "util/boot.h"
#include "network.h"
#include "intercom1.h"
#include "intercom_constants.h"
#include <string.h>

#define FIFO_CMD_SIZE 1200
#define EVENT_QUEUE_SIZE 4

// events
#define EVENT_ERASE 1
#define EVENT_DATA 2

struct _Bootloader
{
	uint8_t cmdBuf[FIFO_CMD_SIZE];
	FifoLin commands;

	Mailbox eventQueue;
	msg_t eventQueueData[EVENT_QUEUE_SIZE];

	uint8_t target;	// use intercom constants
} bootloader;

static void bootloaderCb(IntercomHeader* pHeader, uint16_t dataLength, uint8_t* pData, void* pUserData)
{
	(void)pUserData;
	(void)dataLength;

	switch(pHeader->command)
	{
		case BOOTLOADER_RESPONSE:
		{
			BootloaderResponse* pResponse = (BootloaderResponse*)pData;

			PacketHeader header;
			header.section = SECTION_BOOTLOADER;
			header.cmd = CMD_BOOTLOADER_RESPONSE;

			NetworkSendPacket(&header, (uint8_t*)pResponse, sizeof(BootloaderResponse));
		}
		break;
	}
}

static void sendACK(uint32_t offset)
{
	BootloaderResponse response;
	response.type = BOOTLOADER_RESPONSE_ACK;
	response.offset = offset;

	PacketHeader header;
	header.section = SECTION_BOOTLOADER;
	header.cmd = CMD_BOOTLOADER_RESPONSE;

	NetworkSendPacket(&header, (uint8_t*)&response, sizeof(BootloaderResponse));
}

static void sendNACK(uint32_t offset)
{
	BootloaderResponse response;
	response.type = BOOTLOADER_RESPONSE_NACK;
	response.offset = offset;

	PacketHeader header;
	header.section = SECTION_BOOTLOADER;
	header.cmd = CMD_BOOTLOADER_RESPONSE;

	NetworkSendPacket(&header, (uint8_t*)&response, sizeof(BootloaderResponse));
}

void BootloaderInit()
{
	IntercomClient client;
	client.pUserData = 0;
	client.section = INTERCOM_SECTION_BOOTLOADER;
	client.rxCallback = &bootloaderCb;

	IntercomAddClient(&intercom1, client);

	bootloader.target = INTERCOM_ADDRESS_MAIN;
	chMBInit(&bootloader.eventQueue, bootloader.eventQueueData, EVENT_QUEUE_SIZE);

	FifoLinInit(&bootloader.commands, bootloader.cmdBuf, FIFO_CMD_SIZE);
}

void BootloaderStartProgramming(uint8_t processor)
{
	bootloader.target = processor;

	chMBPost(&bootloader.eventQueue, EVENT_ERASE, TIME_INFINITE);
}


void BootloaderFlashData(BootloaderData* pData)
{
	uint8_t* pCmdBuf;

	int16_t result = FifoLinReserve(&bootloader.commands, sizeof(BootloaderData), &pCmdBuf);
	if(result)
	{
		LogErrorC("Store data to FIFO failed", result);
//		sendNACK(pData->offset);
		return;
	}

	memcpy(pCmdBuf, pData, sizeof(BootloaderData));

	FifoLinCommit(&bootloader.commands, sizeof(BootloaderData));

	chMBPost(&bootloader.eventQueue, EVENT_DATA, TIME_INFINITE);
}

msg_t BootloaderTask(void* params)
{
	(void)params;
	msg_t event;
	uint32_t size;
	uint8_t* pData;

	chRegSetThreadName("Bootloader");

	while(1)
	{
		if(chMBFetch(&bootloader.eventQueue, &event, TIME_INFINITE) != RDY_OK)
			continue;

		switch(event)
		{
			case EVENT_ERASE:
			{
				if(bootloader.target == INTERCOM_ADDRESS_MAIN)
				{
					if(BootEraseProgramFlash())
					{
						LogError("Erase failed");
						sendNACK(0);
						continue;
					}

					sendACK(0);
				}
				else
				{
					IntercomHeader header;
					header.address = bootloader.target;
					header.section = INTERCOM_SECTION_BOOTLOADER;
					header.command = BOOTLOADER_ERASE;

					IntercomSend(&intercom1, &header, 0, 0);

					LogDebug("Erasing co-processor");
				}
			}
			break;
			case EVENT_DATA:
			{
				int16_t result = FifoLinGet(&bootloader.commands, &pData, &size);
				if(result)
					continue;

				BootloaderData* pBootData = (BootloaderData*)pData;

				uint32_t crc = CRC32Checksum(pBootData->data, 64);

				if(crc != pBootData->crc)
				{
					LogWarn("CRC mismatch");
					sendNACK(pBootData->offset);

					FifoLinDelete(&bootloader.commands);
					continue;
				}

				if(bootloader.target == INTERCOM_ADDRESS_MAIN)
				{
					BootProgramFlash(pBootData->data, 64, pBootData->offset);

					sendACK(pBootData->offset);
				}
				else
				{
					IntercomHeader header;
					header.address = bootloader.target;
					header.section = INTERCOM_SECTION_BOOTLOADER;
					header.command = BOOTLOADER_DATA;

					IntercomSend(&intercom1, &header, sizeof(BootloaderData), pData);
				}

				FifoLinDelete(&bootloader.commands);
			}
			break;
		}
	}

	return 0;
}
