/*
 * cli.c
 *
 *  Created on: 29.05.2010
 *      Author: AndreR
 */

#include "cli.h"

#include "util/console.h"

#include "stm32f4xx_rcc.h"

#include "hal/buzzer.h"
#include "hal/led.h"
#include "commands.h"
#include "util/console.h"
#include "intercom1.h"
#include "intercom_constants.h"
#include "util/fifo8_synced.h"
#include "util/bootloader.h"
#include "util/flash_fs.h"
#include "util/flash.h"
#include "hal/usart1.h"
#include "hal/usart6.h"
#include "util/boot.h"
#include "spi3.h"
#include "spi2.h"
#include "ctrl_vel.h"
#include "util/sys_time.h"
#include "util/log.h"
#include "cpu_load.h"
#include "ctrl/ctrl_fusion.h"
#include "ctrl.h"
#include "bootloader.h"

#include "network.h"
#include "ctrl/ctrl_fusion.h"

#include <stdio.h>
#include <string.h>

CLIGlobal cli;

static void processCommand();

#define COLOR_NONE	0
#define COLOR_RED	1
#define COLOR_GREEN	2
#define COLOR_BLUE	3
#define COLOR_YELLOW 4

#define COLOR_MEDIA COLOR_BLUE
#define COLOR_MAIN COLOR_NONE
#define COLOR_KD COLOR_RED
#define COLOR_LEFT COLOR_GREEN
#define COLOR_RIGHT COLOR_YELLOW

static void echoColor(uint8_t color)
{
	const char* pCol;

	switch(color)
	{
	case COLOR_RED:
		pCol = "\e[0;31m";
		break;
	case COLOR_GREEN:
		pCol = "\e[0;32m";
		break;
	case COLOR_BLUE:
		pCol = "\e[0;36m";
		break;
	case COLOR_YELLOW:
		pCol = "\e[0;33m";
		break;
	default:
		pCol = "\e[0m";
		break;
	}

	FifoSyncedPush(&usart1.txFifo, strlen(pCol), (uint8_t*)pCol);
}

static void echoTargetColor(uint8_t addr)
{
	switch(addr)
	{
	case INTERCOM_ADDRESS_MEDIA:
		echoColor(COLOR_MEDIA);
		break;
	case INTERCOM_ADDRESS_LEFT:
		echoColor(COLOR_LEFT);
		break;
	case INTERCOM_ADDRESS_RIGHT:
		echoColor(COLOR_RIGHT);
		break;
	case INTERCOM_ADDRESS_KD:
		echoColor(COLOR_KD);
		break;
	default:
		echoColor(COLOR_MAIN);
		break;
	}
}

// IRQ Level: USART1 IRQ
static void usart1Callback()
{
	// called on new complete packet in usart1 RX fifo
}

// IRQ Level: Task
static void consoleCallback()
{
	// called after ConsolePrint
}

// IRQ Level: Task
static void intercomCallback(IntercomHeader* pHeader, uint16_t length, uint8_t* pData, void* pUserData)
{
	(void)pUserData;

	// called after new console intercom rx
	echoTargetColor(pHeader->command);
	FifoSyncedPush(&usart1.txFifo, length, pData);
	echoTargetColor(cli.currentTarget);

	USART1StartTx();
}


void CLIInit()
{
	cli.currentTarget = INTERCOM_ADDRESS_MAIN;

	static uint8_t rxData[CONSOLE_RX_SIZE];
	static uint8_t txData[CONSOLE_TX_SIZE];

	ConsoleData conData;
	conData.pDataTx = txData;
	conData.txSize = CONSOLE_TX_SIZE;
	conData.pDataRx = rxData;
	conData.rxSize = CONSOLE_RX_SIZE;
	conData.printSize = CONSOLE_PRINT_SIZE;
	conData.txCallback = &consoleCallback;
	conData.pUserData = 0;

	ConsoleInit(conData);

	usart1.callback = &usart1Callback;

	IntercomClient client;
	client.pUserData = 0;
	client.rxCallback = &intercomCallback;
	client.section = INTERCOM_SECTION_CONSOLE;

	IntercomAddClient(&intercom1, client);
}

msg_t CLITask(void* params)
{
	(void)params;

	chRegSetThreadName("CLI");

	int16_t result = 0;
	uint16_t length;

	systime_t waitUntil = chTimeNow()+MS2ST(10);
	systime_t lastLEDTime = 0;

	LEDMustOff(LED_MAIN_GREEN | LED_MAIN_RED);

	NetworkPrint("Firmware Date: %s - %s\r\n", __DATE__, __TIME__);

	while(1)
	{
		// check user input to terminal
		while(FifoPacketSyncedNumPackets(&usart1.rxFifo) > 0)
		{
			// get packet
			result = FifoPacketSyncedPop(&usart1.rxFifo, &length, cli.procBuf, CLI_PROC_SIZE-1);
			if(result == ERROR_FIFO_PACKET_SYNCED_NOT_ENOUGH_MEMORY)
			{
				FifoPacketSyncedDrop(&usart1.rxFifo);
				continue;
			}

			if(result)
				continue;

			ConsolePrint("\r\n");

			cli.procBuf[length] = '\0';

			// target selection logic
			if(strcmp((char*)cli.procBuf, "target main") == 0)
			{
				cli.currentTarget = INTERCOM_ADDRESS_MAIN;
				echoTargetColor(cli.currentTarget);
				ConsolePrint("New target: MAIN\r\n");
				continue;
			}

			if(strcmp((char*)cli.procBuf, "target media") == 0)
			{
				cli.currentTarget = INTERCOM_ADDRESS_MEDIA;
				echoTargetColor(cli.currentTarget);
				ConsolePrint("New target: MEDIA\r\n");
				continue;
			}

			if(strcmp((char*)cli.procBuf, "target left") == 0)
			{
				cli.currentTarget = INTERCOM_ADDRESS_LEFT;
				echoTargetColor(cli.currentTarget);
				ConsolePrint("New target: LEFT\r\n");
				continue;
			}

			if(strcmp((char*)cli.procBuf, "target right") == 0)
			{
				cli.currentTarget = INTERCOM_ADDRESS_RIGHT;
				echoTargetColor(cli.currentTarget);
				ConsolePrint("New target: RIGHT\r\n");
				continue;
			}

			if(strcmp((char*)cli.procBuf, "target kd") == 0)
			{
				cli.currentTarget = INTERCOM_ADDRESS_KD;
				echoTargetColor(cli.currentTarget);
				ConsolePrint("New target: KD\r\n");
				continue;
			}

			// check current target
			if(cli.currentTarget == INTERCOM_ADDRESS_MAIN)
			{
				// process onboard
				uint8_t* pData;
				result = FifoLinReserve(&console.rxFifo, length, &pData);
				if(result)
					continue;

				memcpy(pData, cli.procBuf, length);

				FifoLinCommit(&console.rxFifo, length);
			}
			else
			{
				// forward to intercom
				IntercomHeader header;
				header.address = cli.currentTarget;
				header.command = 0;
				header.section = INTERCOM_SECTION_CONSOLE;

				IntercomSend(&intercom1, &header, length, cli.procBuf);
			}
		}

		// check data from main console to terminal
		while(console.txFifo.numPackets > 0)
		{
			echoTargetColor(INTERCOM_ADDRESS_MAIN);

			uint8_t* pData;
			uint32_t len32;
			result = FifoLinGet(&console.txFifo, &pData, &len32);
			if(result)
				continue;

			FifoSyncedPush(&usart1.txFifo, len32, pData);

			FifoLinDelete(&console.txFifo);

			echoTargetColor(cli.currentTarget);

			USART1StartTx();
		}

		if(ConsoleIsCmdReady())
		{
			processCommand();

			ConsoleCmdProcessed();
		}

		// Heartbeat LED
		if(chTimeElapsedSince(lastLEDTime) > MS2ST(500))
		{
			lastLEDTime = chTimeNow();

#ifdef ENV_BOOT
			LEDToggle(LED_MAIN_RED);
#else
			LEDToggle(LED_MAIN_GREEN);
#endif
		}

		chThdSleepUntil(waitUntil);
		waitUntil = chTimeNow()+MS2ST(10);
	}

	return 0;
}

static void sensorsCommands()
{
	uint8_t activeSensors = 0;
	if(ConsoleScanCmd("set sensors active %hhx", &activeSensors) == 1)
	{
		ctrlFusion.activeSensors = activeSensors;
		NetworkPrint("setting active sensors to %hhu", ctrlFusion.activeSensors );
		NetworkPrint("Acc: %hhu", (ctrlFusion.activeSensors & SENSOR_ACTIVE_ACC));
		NetworkPrint("Encoder: %hhu", (ctrlFusion.activeSensors & SENSOR_ACTIVE_ENCODER));
		NetworkPrint("Gyr: %hhu", (ctrlFusion.activeSensors & SENSOR_ACTIVE_GYR));
		NetworkPrint("Motor: %hhu", (ctrlFusion.activeSensors & SENSOR_ACTIVE_MOTOR));
		NetworkPrint("Vision: %hhu", (ctrlFusion.activeSensors & SENSOR_ACTIVE_VISION));
	}

	if(ConsoleCmpCmd("sensors active"))
	{
		NetworkPrint("activeSensors=%hhu", ctrlFusion.activeSensors);
		NetworkPrint("Acc: %hhu", (ctrlFusion.activeSensors & SENSOR_ACTIVE_ACC));
		NetworkPrint("Encoder: %hhu", (ctrlFusion.activeSensors & SENSOR_ACTIVE_ENCODER));
		NetworkPrint("Gyr: %hhu", (ctrlFusion.activeSensors & SENSOR_ACTIVE_GYR));
		NetworkPrint("Motor: %hhu", (ctrlFusion.activeSensors & SENSOR_ACTIVE_MOTOR));
		NetworkPrint("Vision: %hhu", (ctrlFusion.activeSensors & SENSOR_ACTIVE_VISION));
	}

	if(ConsoleCmpCmd("slippage"))
	{
		NetworkPrint("Slippage: %f", ctrl.state.slippage);
	}
}

static void controllerCommands()
{
	float f32[5];
	uint16_t u16;

	if(ConsoleCmpCmd("ctrl fusion"))
	{
		CtrlSetController(&ctrlFusionInstance);

		ConsolePrint("Controller: Fusion\r\n");
	}

	if(ConsoleCmpCmd("ctrl off"))
	{
		CtrlSetController(0);

		ConsolePrint("Controller disabled\r\n");
	}

	if(ConsoleScanCmd("structure %f %f %f %f %f", f32, f32+1, f32+2, f32+3, f32+4) == 5)
	{
		ConsolePrint("New structural parameters:\r\n");
		ConsolePrint("Front angle: %fdeg\r\n", f32[0]);
		ConsolePrint("Back angle: %fdeg\r\n", f32[1]);
		ConsolePrint("Bot radius: %fm\r\n", f32[2]);
		ConsolePrint("Wheel radius: %fm\r\n", f32[3]);
		ConsolePrint("Mass: %fkg\r\n", f32[4]);

		CtrlSetParams(f32[0], f32[1], f32[2], f32[3], f32[4]);

		// NetworkPrint("xyw2m: [ [%.2f %.2f %.2f] [%.2f %.2f %.2f] [%.2f %.2f %.2f] ]", ctrl.matXYW2Motor.pData[0+0], ctrl.matXYW2Motor.pData[0+1], ctrl.matXYW2Motor.pData[0+2], ctrl.matXYW2Motor.pData[3+0], ctrl.matXYW2Motor.pData[3+1], ctrl.matXYW2Motor.pData[3+2], ctrl.matXYW2Motor.pData[6+0], ctrl.matXYW2Motor.pData[6+1], ctrl.matXYW2Motor.pData[6+2]);
		NetworkPrint("structure: %f %f %f %f %f", f32[0], f32[1], f32[2], f32[3], f32[4]);
	}

	if(ConsoleCmpCmd("spline cubic"))
	{
		ConsolePrint("Setting test spline\r\n");

		CtrlSpline2D spline;
		spline.options = 1;
		spline.tEnd = floatToHalfFloat(5.0f);

		HermiteSpline x, y;
		HermiteSplineCreate(&x, 0, 2, 0, 0, 5);
		HermiteSplineCreate(&y, 0, 0, 0, 0, 5);

		for(uint8_t j=0;j<SPLINE_SIZE;j++) {
			spline.splines[0].a[j] = floatToHalfFloat(x.a[j]);
			spline.splines[1].a[j] = floatToHalfFloat(y.a[j]);
		}

		CtrlFusionAddSplineCmdXY(&spline);
	}
}

static void miscCommands()
{
	if(ConsoleCmpCmd("reset"))
	{
		BootReset();
	}

	if(ConsoleCmpCmd("bootloader"))
	{
		ConsolePrint("Selecting bootloader\r\n");

		// TODO: workaround, first message to media is lost
		BootloaderSelect(&intercom1);
		BootloaderSelect(&intercom1);
		BootSetBootloaderSelected(1);
	}

	if(ConsoleCmpCmd("test"))
	{
		ConsolePrint("Number: %d\r\n", 42);
	}

	float f = 21.0f;
	if(ConsoleScanCmd("float %f", &f) == 1)
	{
		ConsolePrint("Float: %f\r\n", f);
	}

	if(ConsoleCmpCmd("float"))
	{
		ConsolePrint("Float: %f\r\n", 42.0f);
	}

	if(ConsoleCmpCmd("usart6 stats"))
	{
		ConsolePrint("--- USART6 ---\r\n");
		ConsolePrint("Rx invalid: %d\r\n", usart6.stats.rxInvalid);
	}

	if(ConsoleCmpCmd("led get"))
	{
		ConsolePrint("LED State: 0x%X\r\n", LEDGetState());
	}

	if(ConsoleCmpCmd("cpu"))
	{
		ConsolePrint("CPU Load: %3u.%1u%%\r\n", cpuLoad.load/10, cpuLoad.load%10);
		ConsolePrint("Max: %3u.%1u%%\r\n", cpuLoad.maxLoad/10, cpuLoad.maxLoad%10);
		ConsolePrint("incPerTick: %10u/%u\r\n", cpuLoad.incPerSysTick, cpuLoad.maxIncPerSysTick);
		cpuLoad.maxLoad = 0;
	}

	if(ConsoleCmpCmd("tasks"))
	{
		ConsolePrint("Name            rPrio\tTime\r\n");

		for(Thread* pThread = chRegFirstThread(); pThread != 0; pThread = chRegNextThread(pThread))
		{
			ConsolePrint("%-16s%u\t%u\r\n", pThread->p_name, pThread->p_realprio, pThread->p_time);
		}
	}

	uint8_t reg;
	if(ConsoleScanCmd("nrf read %hhu", &reg) == 1)
	{
//		uint8_t value;
//		NRF24ExReadRegister(&network.nrf24, reg, &value);
//		ConsolePrint("Reg 0x%X: 0x%X\r\n", reg, value);
	}

	if(ConsoleCmpCmd("wifi stats"))
	{
		ConsolePrint("Avg: %u\r\n", network.avgRxPeriod);

//		ConsolePrint("-- TX --\r\n");
//		ConsolePrint("Bytes:   %10u/%10u\r\n", network.nrf24.stats.txBytesSent, network.nrf24.stats.txBytesLost);
//		ConsolePrint("Packets: %10u/%10u\r\n", network.nrf24.stats.txPacketsSent, network.nrf24.stats.txPacketsLost);
//		ConsolePrint("Retrans: %10u\r\n",network.nrf24.stats.retransmissions);
//		ConsolePrint("-- RX --\r\n");
//		ConsolePrint("Bytes:   %10u/%10u\r\n", network.nrf24.stats.rxBytesRecv, network.nrf24.stats.rxBytesLost);
//		ConsolePrint("Packets: %10u/%10u\r\n", network.nrf24.stats.rxPacketsRecv, network.nrf24.stats.rxPacketsLost);
	}

	if(ConsoleCmpCmd("float16"))
	{
		float x = 1.0f;
		float16_t f16 = floatToHalfFloat(x);

		float y = halfFloatToFloat(f16);

		ConsolePrint("In: %f, F16: 0x%04X, out: %f\r\n", x, f16, y);
	}

	if(ConsoleCmpCmd("net hello"))
	{
		ConsolePrint("Saying hello to network print\r\n");
		NetworkPrint("Hello World");
	}

	uint16_t u16;
	if(ConsoleScanCmd("set botid %hu", &u16) == 1)
	{
		NRF24ExChangeId(&network.nrf24, u16);
		network.config.botId = u16;

		NetworkPrint("Changed BotID to %hu", u16);
		ConsolePrint("Changed BotID to %hu\r\n", u16);
	}

	if(ConsoleScanCmd("write botid %hu", &u16) == 1)
	{
		NRF24ExChangeId(&network.nrf24, u16);
		network.config.botId = u16;

		NetworkSaveConfig();

		NetworkPrint("Changed BotID to %hu", u16);
		ConsolePrint("Changed BotID to %hu\r\n", u16);
	}

	if(ConsoleScanCmd("set channel %hu", &u16) == 1)
	{
		NRF24ExChangeChannel(&network.nrf24, u16);
		network.config.channel = u16;

		ConsolePrint("Changed channel to %hu\r\n", u16);
	}

	uint16_t speed;
	if(ConsoleScanCmd("set speed %hu", &speed) == 1)
	{
		switch(speed)
		{
			case 0: speed = NRF24_SPEED_250K; break;
			case 1: speed = NRF24_SPEED_1M; break;
			default: speed = NRF24_SPEED_2M; break;
		}

		NRF24ExChangeSpeed(&network.nrf24, speed);
		network.config.speed = speed;

		if(network.config.speed == NRF24_SPEED_250K)
			ConsolePrint("Speed: 250k\r\n");
		if(network.config.speed == NRF24_SPEED_1M)
			ConsolePrint("Speed: 1M\r\n");
		if(network.config.speed == NRF24_SPEED_2M)
			ConsolePrint("Speed: 2M\r\n");
	}

	if(ConsoleCmpCmd("config write"))
	{
		NetworkSaveConfig();
		ConsolePrint("Config saved to FLASH\r\n");
	}

	if(ConsoleCmpCmd("config"))
	{
		ConsolePrint("ID: %hu\r\n", network.config.botId);
		ConsolePrint("Channel: %hu\r\n", network.config.channel);
		if(network.config.speed == NRF24_SPEED_250K)
			ConsolePrint("Speed: 250k\r\n");
		if(network.config.speed == NRF24_SPEED_1M)
			ConsolePrint("Speed: 1M\r\n");
		if(network.config.speed == NRF24_SPEED_2M)
			ConsolePrint("Speed: 2M\r\n");
	}

	if(ConsoleCmpCmd("flash hex"))
	{
		for(uint32_t i = 0; i < flashFS.bytesDel+flashFS.bytesUsed+48; i += 8)
		{
			uint8_t* d = (uint8_t*)(flashFS.flashAddr+i);
			ConsolePrint("0x%04X: %02X %02X %02X %02X %02X %02X %02X %02X\r\n",
					i, d[0], d[1], d[2], d[3], d[4], d[5], d[6], d[7]);
		}
	}

	if(ConsoleCmpCmd("flash list"))
	{
		FlashFSListFiles();
	}

	if(ConsoleCmpCmd("flash erase"))
	{
		ConsolePrint("Erasing flash\r\n");

		FlashErase(flashFS.flashAddr, flashFS.flashAddr+flashFS.flashSize);
	}

	if(ConsoleCmpCmd("flash compact"))
	{
		FlashFSCompact();

		ConsolePrint("Flash compacted\r\n");
	}

/*	uint16_t mot;
	float iQ;
	if(ConsoleScanCmd("iq %hu %f", &mot, &iQ) == 2)
	{
		switch(mot)
		{
		case 0:
			spi3.slave[SPI3_SLAVE_RIGHT].txData.periodic.mot[MOT_FRONT].iQ = iQ;
			break;
		case 1:
			spi3.slave[SPI3_SLAVE_LEFT].txData.periodic.mot[MOT_FRONT].iQ = iQ;
			break;
		case 2:
			spi3.slave[SPI3_SLAVE_LEFT].txData.periodic.mot[MOT_REAR].iQ = iQ;
			break;
		case 3:
			spi3.slave[SPI3_SLAVE_RIGHT].txData.periodic.mot[MOT_REAR].iQ = iQ;
			break;
		}
	}
*/
	int16_t x, y, w;
	if(ConsoleScanCmd("move %hd %hd %hd", &x, &y, &w) == 3)
	{
		CtrlVelSetTarget(x/1000.0f, y/1000.0f, w/1000.0f);
	}

	if(ConsoleCmpCmd("time"))
	{
		ConsolePrint("Time: %uus\r\n", SysTimeUSec());
	}

	if(ConsoleCmpCmd("log"))
	{
		LogPrintAllRecent(4);
	}

#ifdef ENV_BOOT
	if(ConsoleCmpCmd("start bootloader"))
	{
		ConsolePrint("Starting programming on MAIN\r\n");

		BootloaderStartProgramming(INTERCOM_ADDRESS_MAIN);
	}
#endif
}

static void buzzerCommands()
{
	if(ConsoleCmpCmd("buzz on"))
	{
		ConsolePrint("Enabling buzzer\r\n");

		BuzzerEnable(1);
	}

	if(ConsoleCmpCmd("buzz off"))
	{
		ConsolePrint("Disabling buzzer\r\n");

		BuzzerEnable(0);
	}

	if(ConsoleCmpCmd("buzz"))
	{
		ConsolePrint("Buzzer test\r\n");

		BuzzerEnable(1);
		chThdSleep(MS2ST(200));
		BuzzerEnable(0);
		chThdSleep(MS2ST(200));
		BuzzerEnable(1);
		chThdSleep(MS2ST(500));
		BuzzerEnable(0);
	}
}

static void processCommand()
{
	sensorsCommands();
	miscCommands();
	buzzerCommands();
	controllerCommands();

	if(ConsoleCmpCmd("version"))
	{
		ConsolePrint("Firmware Date: %s - %s\r\n", __DATE__, __TIME__);
	}
}
