/*
 * ctrl.c
 *
 *  Created on: 06.03.2013
 *      Author: AndreR
 */

#include "ctrl.h"

#include "stm32f4xx.h"
#include "stm32f4xx_rcc.h"
#include "stm32f4xx_tim.h"
#include "misc.h"

#include "constants.h"
#include "commands.h"
#include "network.h"
#include "power.h"
#include "util/sys_time.h"
#include "spi3.h"
#include "motors.h"
#include "intercom1.h"

#include <string.h>

Ctrl ctrl;

static float pXYW2Motor[12];
static float pMotor2XYW[12];
static float pForceXYW2Motor[12];
static float pForceMotor2XYW[12];

#define CTRL_TIMER_RESOLUTION 0.0005f

void TIM1_BRK_TIM9_IRQHandler()
{
	CH_IRQ_PROLOGUE();

	TIM9->SR = 0;

	chSysLockFromIsr();
	chMBPostI(&ctrl.eventQueue, CTRL_EVENT_RUN);
	chSysUnlockFromIsr();

	CH_IRQ_EPILOGUE();
}

static void kdCallback(IntercomHeader* pHeader, uint16_t dataLength, uint8_t* pData, void* pUserData)
{
	(void)pUserData;

	if(pHeader->command == IC_KD_KICKER_LOG && dataLength >= sizeof(KickerLog))
	{
		KickerLog* pLog = (KickerLog*)pData;

		ctrl.lastKickerLog = *pLog;
	}
}

static void motorsUpdated(uint8_t source)
{
	msg_t event;

	if(source == MOT_SOURCE_RIGHT)
		event = CTRL_EVENT_UPDATE_ENC_RIGHT;
	else
		event = CTRL_EVENT_UPDATE_ENC_LEFT;

	chSysLock();
	chMBPostI(&ctrl.eventQueue, event);
	chSysUnlock();
}

static void gyrUpdated()
{
	chSysLockFromIsr();
	chMBPostI(&ctrl.eventQueue, CTRL_EVENT_UPDATE_GYR);
	chSysUnlockFromIsr();
}

static void accUpdated()
{
	chSysLockFromIsr();
	chMBPostI(&ctrl.eventQueue, CTRL_EVENT_UPDATE_ACC);
	chSysUnlockFromIsr();
}

static void powerUpdated()
{
	chSysLock();
	chMBPostI(&ctrl.eventQueue, CTRL_EVENT_UPDATE_POWER);
	chSysUnlock();
}

static void posUpdated()
{
	chSysLock();
	chMBPostI(&ctrl.eventQueue, CTRL_EVENT_UPDATE_POS);
	chSysUnlock();
}

static void objUpdated()
{
	chSysLock();
	chMBPostI(&ctrl.eventQueue, CTRL_EVENT_UPDATE_OBJ);
	chSysUnlock();
}

static void motorUpdated()
{
	chSysLockFromIsr();
	chMBPostI(&ctrl.eventQueue, CTRL_EVENT_UPDATE_MOTOR);
	chSysUnlockFromIsr();
}

static void pseudoInverse(arm_matrix_instance_f32* src, arm_matrix_instance_f32* dest)
{
	float ptransposed[12];
	arm_matrix_instance_f32 matTransposed = { .numCols = 4, .numRows = 3, .pData = ptransposed };
	float pTri[9];
	arm_matrix_instance_f32 matTri = { .numCols = 3, .numRows = 3, .pData = pTri };
	float pTriInv[9];
	arm_matrix_instance_f32 matTriInv = { .numCols = 3, .numRows = 3, .pData = pTriInv };

	arm_mat_trans_f32(src, &matTransposed);
	arm_mat_mult_f32(&matTransposed, src, &matTri);
	arm_mat_inverse_f32(&matTri, &matTriInv);
	arm_mat_mult_f32(&matTriInv, &matTransposed, dest);
}

void CtrlInit()
{
	memset(&ctrl, 0, sizeof(Ctrl));

	ctrl.controllersEnabled = 1;

	// construct matrices
	ctrl.matXYW2Motor.numCols = 3;
	ctrl.matXYW2Motor.numRows = 4;
	ctrl.matXYW2Motor.pData = pXYW2Motor;
	ctrl.matMotor2XYW.numCols = 4;
	ctrl.matMotor2XYW.numRows = 3;
	ctrl.matMotor2XYW.pData = pMotor2XYW;
	ctrl.matForceXYW2Motor.numCols = 3;
	ctrl.matForceXYW2Motor.numRows = 4;
	ctrl.matForceXYW2Motor.pData = pForceXYW2Motor;
	ctrl.matForceMotor2XYW.numCols = 4;
	ctrl.matForceMotor2XYW.numRows = 3;
	ctrl.matForceMotor2XYW.pData = pForceMotor2XYW;

	ctrl.state.matPos.numCols = 1;
	ctrl.state.matPos.numRows = 3;
	ctrl.state.matPos.pData = ctrl.state.pos;
	ctrl.state.matVel.numCols = 1;
	ctrl.state.matVel.numRows = 3;
	ctrl.state.matVel.pData = ctrl.state.vel;
	ctrl.state.matAcc.numCols = 1;
	ctrl.state.matAcc.numRows = 3;
	ctrl.state.matAcc.pData = ctrl.state.acc;

	int16_t result = FlashFSOpenOrCreate("ctrl/structure", 0, &ctrl.botParams, sizeof(BotParams), &ctrl.pBotParamsFile);
	if(result == FLASH_FS_CREATED)
	{
		// build structural data
		CtrlSetParams(30.0f, 45.0f, 0.0785f, 0.025f, 2.0f);
	}
	if(result == FLASH_FS_OPEN)
	{
		CtrlSetParams(ctrl.botParams.frontAngle, ctrl.botParams.backAngle, ctrl.botParams.botRadius,
				ctrl.botParams.wheelRadius, ctrl.botParams.mass);
	}

	// init event queue
	chMBInit(&ctrl.eventQueue, ctrl.eventQueueData, CTRL_EVENT_QUEUE_SIZE);

	// TODO: re-integrate callbacks

	// register encoder update callback
	motors.updateCb = &motorsUpdated;

/*
	// register encoder update callback
	EncSetUpdateCallback(&encUpdated);

	// register acc and gyr callbacks
	L3GSetCallback(&gyrUpdated);
	LIS3SetCallback(&accUpdated);

	// register motor current callback
	MotorIrSetCallback(&motorUpdated);
*/

	// register power callback
	PowerSetCallback(&powerUpdated);

	// register pos callback
	NetworkSetPosCallback(&posUpdated);
	NetworkSetTrackedObjCallback(&objUpdated);

	// init precision timer for accurate controller dT
	RCC_APB2PeriphClockCmd(RCC_APB2Periph_TIM9, ENABLE);

	TIM_TimeBaseInitTypeDef timeBaseInit;
	timeBaseInit.TIM_ClockDivision = TIM_CKD_DIV1;
	timeBaseInit.TIM_CounterMode = TIM_CounterMode_Up;
	timeBaseInit.TIM_Period = 1;	// TODO: verify
	timeBaseInit.TIM_Prescaler = 41999;	// = 0.5ms
	timeBaseInit.TIM_RepetitionCounter = 0;
	TIM_TimeBaseInit(TIM9, &timeBaseInit);

	TIM_ITConfig(TIM9, TIM_IT_Update, ENABLE);

	NVIC_InitTypeDef nvicInit;
	nvicInit.NVIC_IRQChannel = TIM1_BRK_TIM9_IRQn;
	nvicInit.NVIC_IRQChannelCmd = ENABLE;
	nvicInit.NVIC_IRQChannelPreemptionPriority = IRQL_CTRL_TIMER;
	nvicInit.NVIC_IRQChannelSubPriority = 0;
	NVIC_Init(&nvicInit);

	TIM_Cmd(TIM9, ENABLE);

	IntercomClient client;
	client.pUserData = 0;
	client.rxCallback = &kdCallback;
	client.section = INTERCOM_SECTION_KD;
	IntercomAddClient(&intercom1, client);
}

void CtrlSetParams(float frontAngle, float backAngle, float botRadius, float wheelRadius, float mass)
{
	BotParams newParams;
	newParams.frontAngle = frontAngle;
	newParams.backAngle = backAngle;
	newParams.botRadius = botRadius;
	newParams.wheelRadius = wheelRadius;
	newParams.mass = mass;

	if(memcmp(&ctrl.botParams, &newParams, sizeof(BotParams)) != 0)
	{
		memcpy(&ctrl.botParams, &newParams, sizeof(BotParams));

		FlashFSWrite(&ctrl.pBotParamsFile, &ctrl.botParams, sizeof(BotParams));
	}

	// TODO: inertia is approximated by evenly distributed cylinder
	ctrl.momentOfInertia = 0.5f*ctrl.botParams.mass*ctrl.botParams.botRadius*ctrl.botParams.botRadius;

	float alpha = frontAngle*PI/180.0f;
	float beta = backAngle*PI/180.0f;

	float theta[4];
	theta[0] = alpha;
	theta[1] = PI-alpha;
	theta[2] = PI+beta;
	theta[3] = 2*PI-beta;

	// construct matrix for XYW to Motor (VM) conversion
	for(uint8_t i = 0; i < 4; i++)
	{
		pXYW2Motor[(i*3)+0] = -sin(theta[i]);
		pXYW2Motor[(i*3)+1] = cos(theta[i]);
		pXYW2Motor[(i*3)+2] = botRadius;

		pForceXYW2Motor[(i*3)+0] = -sin(theta[i]);
		pForceXYW2Motor[(i*3)+1] = cos(theta[i]);
		pForceXYW2Motor[(i*3)+2] = (1.0f/botRadius);

//		pForceMotor2XYW[i] = -sin(theta[i])/mass;
//		pForceMotor2XYW[4+i] = cos(theta[i])/mass;
//		pForceMotor2XYW[8+i] = botRadius/ctrl.botParams.momentOfInertia;
	}

	pseudoInverse(&ctrl.matXYW2Motor, &ctrl.matMotor2XYW);
//	pseudoInverse(&ctrl.matForceMotor2XYW, &ctrl.matForceXYW2Motor);
	pseudoInverse(&ctrl.matForceXYW2Motor, &ctrl.matForceMotor2XYW);
}

msg_t CtrlTask(void* params)
{
	(void) params;

	msg_t event;
	float lastExecTime = SysTime();

	while(1)
	{
		if(chMBFetch(&ctrl.eventQueue, &event, TIME_INFINITE) != RDY_OK)
			continue;

		switch(event)
		{
			case CTRL_EVENT_RUN:
			{
				float now = SysTime();
				ctrl.dt = now - lastExecTime;
				lastExecTime = now;

				if(ctrl.pInstance == 0)
					continue;
				if(ctrl.pInstance->pControllerFunction == 0)
					continue;

				(*ctrl.pInstance->pControllerFunction)(&ctrl.input, &ctrl.output);

				for(uint8_t i = 0; i < 4; i++)
					motors.mot[i].set.velocity = ctrl.output.motorRPS[i];

				// send status
				if(SysTimeUSec() - ctrl.lastHfStatusTime > 50000)	// 50ms
				{
					ctrl.lastHfStatusTime = SysTimeUSec();

					SystemStatusV2 status;
					memset(&status, 0, sizeof(SystemStatusV2));

					status.kickerLevel = ctrl.lastKickerLog.vCap*10.0f;

					//TODO: re-enable
/*
					if(motorIr.atData.irLevel < motorIr.atData.irThreshold)
						status.barrierInterrupted = 1;
					else
						 status.barrierInterrupted = 0;

					if(dribbler.speedReached)
						status.bits |= 0x08;

					if(dribbler.overload)
						status.bits |= 0x10;
*/
					if(ctrl.state.updated[0])
					{
						ctrl.state.updated[0] = 0;
						status.bits |= 0x01;
						status.position[0] = (int16_t)(ctrl.state.pos[0]*1000.0f + ctrl.state.vel[0]*50.0f);
						status.position[1] = (int16_t)(ctrl.state.pos[1]*1000.0f + ctrl.state.vel[1]*50.0f);
						status.position[2] = (int16_t)(ctrl.state.pos[2]*1000.0f + ctrl.state.vel[2]*50.0f);
					}

					if(ctrl.state.updated[1])
					{
						ctrl.state.updated[1] = 0;
						status.bits |= 0x02;
						status.velocity[0] = (int16_t)(ctrl.state.vel[0]*1000.0f);
						status.velocity[1] = (int16_t)(ctrl.state.vel[1]*1000.0f);
						status.velocity[2] = (int16_t)(ctrl.state.vel[2]*1000.0f);
					}

					if(ctrl.state.updated[2])
					{
						ctrl.state.updated[2] = 0;
						status.bits |= 0x04;
						status.acceleration[0] = (int16_t)(ctrl.state.acc[0]*1000.0f);
						status.acceleration[1] = (int16_t)(ctrl.state.acc[1]*1000.0f);
						status.acceleration[2] = (int16_t)(ctrl.state.acc[2]*1000.0f);
					}

					if (ctrl.sendLog)
					{
						PacketHeader header;
						if (ctrl.sendExtLog)
						{
							header.section = SECTION_SYSTEM;
							header.cmd = CMD_SYSTEM_STATUS_EXT;

							SystemStatusExtended statusExt;
							memset(&statusExt, 0, sizeof(SystemStatusExtended));
							statusExt.kickerLevel = status.kickerLevel;
							statusExt.barrierInterrupted = status.barrierInterrupted;
							statusExt.bits = status.bits;

							for (int i = 0; i < 3; i++)
							{
								statusExt.position[i] = status.position[i];
								statusExt.velocity[i] = status.velocity[i];
								statusExt.acceleration[i] = status.acceleration[i];
								statusExt.targetPos[i] = (int16_t) (ctrl.state.setVel[i] * 100.0f);
								statusExt.targetVel[i] = (int16_t) (ctrl.state.velOutput[i] * 100.0f);
								statusExt.targetAcc[i] = (int16_t) (ctrl.state.errorVel[i] * 100.0f);
							}
							for (int i = 0; i < 4; i++)
							{
								statusExt.motorSet[i] = (int16_t) (ctrl.state.motorSet[i] * 100.0f);
								statusExt.motorCur[i] =	(int16_t) (ctrl.state.motorCur[i] * 100.0f);
								statusExt.motorOut[i] =	(int16_t) (ctrl.state.motorOut[i] * 100.0f);
							}
							NetworkSendPacket(&header, (uint8_t*) &statusExt, sizeof(SystemStatusExtended));
						}
						else
						{
							header.section = SECTION_SYSTEM;
							header.cmd = CMD_SYSTEM_STATUS_V2;
							NetworkSendPacket(&header, (uint8_t*) &status, sizeof(SystemStatusV2));
						}
					}
				}
			}
			break;
			case CTRL_EVENT_UPDATE_MOTOR:
			{
				// TODO:
/*				for(uint8_t i = 0; i < 4; i++)
				{
					ctrl.input.motor.current[i] = motorIr.atData.current[i]*0.001f;

					if(motor.M[i].ccw)
						ctrl.input.motor.current[i] *= -1;
				}

				ctrl.input.motor.updated = 1;
*/
			}
			break;
			case CTRL_EVENT_UPDATE_ENC_LEFT:
			{
				ctrl.input.enc.motorVel[MOT_FL] = motors.mot[MOT_FL].cur.velocity * (2*PI);
				ctrl.input.enc.motorVel[MOT_RL] = motors.mot[MOT_RL].cur.velocity * (2*PI);
				ctrl.input.enc.updated = 1;
			}
			break;
			case CTRL_EVENT_UPDATE_ENC_RIGHT:
			{
				ctrl.input.enc.motorVel[MOT_FR] = motors.mot[MOT_FR].cur.velocity * (2*PI);
				ctrl.input.enc.motorVel[MOT_RR] = motors.mot[MOT_RR].cur.velocity * (2*PI);
				ctrl.input.enc.updated = 1;
			}
			break;
			case CTRL_EVENT_UPDATE_GYR:
			{
//				memcpy(ctrl.input.gyr.rotVel, l3g.vel, sizeof(float)*3);
//				ctrl.input.gyr.rotVel[0] *= config.gyroScale;
//				ctrl.input.gyr.rotVel[1] *= config.gyroScale;
//				ctrl.input.gyr.rotVel[2] *= config.gyroScale;
//				ctrl.input.gyr.updated = 1;
			}
			break;
			case CTRL_EVENT_UPDATE_ACC:
			{
//				memcpy(ctrl.input.acc.linAcc, lis3.acc, sizeof(float)*3);
//				ctrl.input.acc.updated = 1;
			}
			break;
			case CTRL_EVENT_UPDATE_POS:
			{
				ctrl.input.pos.pos[0] = network.lastVisionPos[0]*0.001f;
				ctrl.input.pos.pos[1] = network.lastVisionPos[1]*0.001f;
				ctrl.input.pos.pos[2] = network.lastVisionPos[2]*0.001f;
				ctrl.input.pos.updated = 1;
			}
			break;
			case CTRL_EVENT_UPDATE_OBJ:
			{
				ctrl.input.trackedObj.pos[0] = network.lastTrackedObjPos[0]*0.001f;
				ctrl.input.trackedObj.pos[1] = network.lastTrackedObjPos[1]*0.001f;
				ctrl.input.trackedObj.pos[2] = network.lastTrackedObjPos[2]*0.001f;
				ctrl.input.trackedObj.updated = 1;
			}
			break;
			case CTRL_EVENT_UPDATE_POWER:
			{
				ctrl.input.power.voltage = power.lastLog.u1*0.001f;
				ctrl.input.power.updated = 1;
			}
			break;
			case CTRL_EVENT_SET_CTRL:
			{
				if(ctrl.pInstance)
				{
					if(ctrl.pInstance->pExit)
						(*ctrl.pInstance->pExit)();
				}

				for(uint8_t i = 0; i < 4; i++)
					motors.mot[i].set.velocity = 0;

				ctrl.pInstance = ctrl.pNewInstance;

				IntercomHeader iHeader;
				iHeader.section = INTERCOM_SECTION_MOTORS;
				iHeader.address = INTERCOM_ADDRESS_LEFT | INTERCOM_ADDRESS_RIGHT;
				iHeader.command = IC_MOTORS_CTRL_TEMP_MEAS;

				MotorsCtrlTempMeas tempMeas;
				tempMeas.enable = 0;

				if(ctrl.pInstance)
				{
					TIM9->ARR = ((uint32_t)((ctrl.pInstance->deltaT+0.0001f)/CTRL_TIMER_RESOLUTION))-1;
					TIM9->EGR = TIM_EGR_UG;

					if(ctrl.pInstance->pInit)
						(*ctrl.pInstance->pInit)();

					tempMeas.enable = 1;
				}

				IntercomSend(&intercom1, &iHeader, sizeof(MotorsCtrlTempMeas), (uint8_t*)&tempMeas);
			}
			break;
			default:
				break;
		}
	}

	return 0;
}

void CtrlSetController(CtrlInstance* pInstance)
{
	ctrl.pNewInstance = pInstance;

	chMBPost(&ctrl.eventQueue, CTRL_EVENT_SET_CTRL, TIME_INFINITE);
}

void CtrlSetLog(uint8_t enable)
{
	ctrl.sendLog = enable;
}

void CtrlSetExtLog(uint8_t enable)
{
	ctrl.sendExtLog = enable;
}
