/*
 * ctrl.h
 *
 *  Created on: 06.03.2013
 *      Author: AndreR
 */

#ifndef CTRL_H_
#define CTRL_H_

#include "stm32f4xx.h"
#include "arm_math.h"
#include "util/vector.h"
#include "intercom_constants.h"
#include "util/flash_fs.h"
#include "ch.h"

#define CTRL_MOTOR_TO_WHEEL_RATIO (15.0f/50.0f)
#define CTRL_WHEEL_TO_MOTOR_RATIO (50.0f/15.0f)
#define CTRL_MOTOR_KM (0.0255f) // motor torque constant [Nm/A]
#define CTRL_MOTOR_KN (374.0f*2.0f*PI/60.0f) // motor speed constant [rad/(s*V)]

#define CTRL_EVENT_RUN				1
#define CTRL_EVENT_UPDATE_MOTOR		2
#define CTRL_EVENT_UPDATE_ENC_LEFT	3
#define CTRL_EVENT_UPDATE_GYR		4
#define CTRL_EVENT_UPDATE_ACC		5
#define CTRL_EVENT_UPDATE_POS		6
#define CTRL_EVENT_UPDATE_POWER		7
#define CTRL_EVENT_SET_CTRL			8
#define CTRL_EVENT_UPDATE_OBJ		9
#define CTRL_EVENT_UPDATE_ENC_RIGHT	10

#define CTRL_EVENT_QUEUE_SIZE 10

typedef struct _CtrlInput
{
	struct _motor
	{
		float current[4];	// FR, FL, RL, RR [A]
		uint8_t updated;
	} motor;

	struct _enc
	{
		float motorVel[4];	// FR, FL, RL, RR angular motor velocity [rad/s]
		uint8_t updated;
	} enc;

	struct _gyr
	{
		float rotVel[3];	// around X, Y, Z [rad/s]
		uint8_t updated;
	} gyr;

	struct _acc
	{
		float linAcc[3];	// X, Y, Z [m�/s]
		uint8_t updated;
	} acc;

	struct _pos
	{
		float pos[3];		// X, Y, Z [m]
		uint8_t updated;
	} pos;

	struct _power
	{
		float voltage;
		uint8_t updated;	// battery voltage [V]
	} power;

	struct _trackedObj
	{
		float pos[3];		// X, Y, Z [m]
		uint8_t updated;
	} trackedObj;
} CtrlInput;

typedef struct _CtrlOutput
{
	float motorPWM[4];		// PWM duty cycle = avg. [V], Range -1.0f (CCW) to 1.0f (CW)
	float motorRPS[4];		// Motor rounds per second
	float pos[3]; // X, Y, W
	float vel[3]; // X, Y, W
} CtrlOutput;

typedef void(*CtrlInitFunction)();
typedef void(*CtrlFunction)(CtrlInput*, CtrlOutput*);
typedef void(*CtrlExitFunction)();

typedef struct _CtrlInstance
{
	CtrlInitFunction pInit;
	CtrlFunction pControllerFunction;
	CtrlExitFunction pExit;

	float deltaT; // in [s], min. accuracy is 0.5ms
} CtrlInstance;

// structural bot params
typedef struct _BotParams
{
	float wheelRadius;
	float frontAngle;
	float backAngle;
	float botRadius;
	float mass; // [kg]
} __attribute__((aligned(4),packed)) BotParams;

typedef struct _Ctrl
{
	CtrlInput input;
	CtrlInstance* pInstance;
	CtrlOutput output;

	CtrlInstance* pNewInstance;

	Mailbox eventQueue;
	msg_t eventQueueData[CTRL_EVENT_QUEUE_SIZE];

	float dt;

	BotParams botParams;
	float momentOfInertia;
	FlashFile* pBotParamsFile;

	arm_matrix_instance_f32 matXYW2Motor; // 4x3
	arm_matrix_instance_f32 matMotor2XYW; // 3x4

	arm_matrix_instance_f32 matForceXYW2Motor; // 4x3
	arm_matrix_instance_f32 matForceMotor2XYW; // 3x4

	struct _state
	{
		uint8_t updated[3];
		float pos[3];
		float vel[3];
		float acc[3];
		float setVel[3];
		float velOutput[3];
		float errorVel[3];
		float motorSet[4];
		float motorCur[4];
		float motorOut[4];
		float slippage[4];

		arm_matrix_instance_f32 matPos;
		arm_matrix_instance_f32 matVel;
		arm_matrix_instance_f32 matAcc;
	} state;

	uint32_t lastHfStatusTime;

	uint8_t sendLog;
	uint8_t sendExtLog;
	uint8_t controllersEnabled;

	KickerLog lastKickerLog;
} Ctrl;

extern Ctrl ctrl;

void CtrlInit();
void CtrlSetParams(float frontAngle, float backAngle, float botRadius, float wheelRadius, float mass);
msg_t CtrlTask(void* params);
void CtrlSetController(CtrlInstance* pInstance);
void CtrlSetLog(uint8_t enable);
void CtrlSetExtLog(uint8_t enable);

#endif /* CTRL_H_ */
