/*
 * ctrl_fusion.c
 *
 *  Created on: 18.03.2013
 *      Author: AndreR
 */

#include "ctrl.h"
#include "ctrl_fusion.h"

#ifndef SIMULATION
#include "util/console.h"
#include "intercom1.h"
#include "intercom_constants.h"
#include "util/angle_math.h"
#include "network.h"
#include "util/sys_time.h"
#endif

#include <string.h>

#define REACTION_THRESHOLD 0.01f
#define STALL_TIMEOUT 5000
#define DELTA_T 0.001f

static void init();
static void controller(CtrlInput* pInput, CtrlOutput* pOutput);

CtrlFusion ctrlFusion;
CtrlInstance ctrlFusionInstance = { .pInit = &init, .pExit = 0, .pControllerFunction = &controller, .deltaT = DELTA_T }; // deltaT unit: [0.5ms]

static float kfXData[KF_DATA_SIZE(3, 1, 4, 4)];
static float kfYData[KF_DATA_SIZE(3, 1, 4, 4)];
static float kfWData[KF_DATA_SIZE(3, 1, 4, 4)];

static uint8_t controlMode = 0;

static const float A[9] = { 1, DELTA_T, 0.5f * DELTA_T * DELTA_T, 0, 1, DELTA_T, 0, 0, 1, };

static const float B[3] = { 0, 0, 0 };

static const float Ex[9] = { 0.1f, 0, 0, 0, 0.1f, 0, 0, 0, 0.5f, };

static const float Ez[16] = { 10.0f, 0, 0, 0, 0, 2.0f, 0, 0, 0, 0, 1.0f, 0, 0, 0, 0, 0.1f, };

static HermiteSplineListItem listData[3][10];

static void switchControlMode(uint8_t newMode)
{
	if(newMode == controlMode)
	{
		return;
	}
	switch(controlMode)
	{
		case CONTROLLER_MODE_POS:
		case CONTROLLER_MODE_VEL:
			arm_biquad_cascade_df1_init_f32(&ctrlFusion.pt1[0], 1, ctrlFusion.pt1Coeff, ctrlFusion.pt1State[0]);
			arm_biquad_cascade_df1_init_f32(&ctrlFusion.pt1[1], 1, ctrlFusion.pt1Coeff, ctrlFusion.pt1State[1]);
			arm_biquad_cascade_df1_init_f32(&ctrlFusion.pt1[2], 1, ctrlFusion.pt1Coeff, ctrlFusion.pt1State[2]);
			break;
	}
	switch(newMode)
	{
		case CONTROLLER_MODE_SPLINE:
//			NetworkPrint("spline mode");
			break;
		case CONTROLLER_MODE_VEL:
//			NetworkPrint("vel mode");
			break;
		case CONTROLLER_MODE_POS:
//			NetworkPrint("pos mode");
			break;
	}
	controlMode = newMode;
}

static float callTime = 0;
static uint8_t motorCtrlNeeded()
{
	switch(controlMode)
	{
		case CONTROLLER_MODE_SPLINE:
			if ((ctrlFusion.splineList[0].numItems + ctrlFusion.splineList[1].numItems + ctrlFusion.splineList[2].numItems) > 0)
			{
				callTime = SysTime();
				return 1;
			}
			break;
		case CONTROLLER_MODE_VEL:
			if( fabsf(ctrlFusion.setpoint[0]) + fabsf(ctrlFusion.setpoint[1]) + fabsf(ctrlFusion.setpoint[2]) > 0.001f)
			{
				callTime = SysTime();
				return 1;
			}
			break;
		case CONTROLLER_MODE_POS:
			return 1;
	}

	if(SysTime() - callTime > 0.2f)
		return 0;

	return 1;
}

static void init()
{
	// construct state transition matrix

	// init kalman filters
	KFInit(&ctrlFusion.kfX, 3, 1, 4, kfXData);
	KFInit(&ctrlFusion.kfY, 3, 1, 4, kfYData);
	KFInit(&ctrlFusion.kfW, 3, 1, 4, kfWData);

	KFSetOrientationComponent(&ctrlFusion.kfW, 0);

	memcpy(ctrlFusion.kfX.A.pData, A, sizeof(A));
	memcpy(ctrlFusion.kfY.A.pData, A, sizeof(A));
	memcpy(ctrlFusion.kfW.A.pData, A, sizeof(A));

	memcpy(ctrlFusion.kfX.B.pData, B, sizeof(B));
	memcpy(ctrlFusion.kfY.B.pData, B, sizeof(B));
	memcpy(ctrlFusion.kfW.B.pData, B, sizeof(B));

	memcpy(ctrlFusion.kfX.Ex.pData, Ex, sizeof(Ex));
	memcpy(ctrlFusion.kfY.Ex.pData, Ex, sizeof(Ex));
	memcpy(ctrlFusion.kfW.Ex.pData, Ex, sizeof(Ex));

	memcpy(ctrlFusion.kfX.Ez.pData, Ez, sizeof(Ez));
	memcpy(ctrlFusion.kfY.Ez.pData, Ez, sizeof(Ez));
	memcpy(ctrlFusion.kfW.Ez.pData, Ez, sizeof(Ez));

	ctrlFusion.lastWakeTime = chTimeNow();
	ctrlFusion.lastReaction = ctrlFusion.lastWakeTime;
	ctrlFusion.lastPositionUpdateTime = 0;
	ctrlFusion.visionTimeout = 2.0f;
	ctrlFusion.visionLost = 1;
	ctrlFusion.visionBack = 0;
	ctrlFusion.visionBackTime = 0;
//	ctrlFusion.activeSensors = 0x17; // no motor
//	ctrlFusion.activeSensors = 0x14; // acc+gyr
	ctrlFusion.activeSensors = 0x03; // enc+vision
	ctrlFusion.theta[0] = 1.0f;
	ctrlFusion.theta[1] = 1.0f;
	ctrlFusion.theta[2] = 1.0f;
	ctrlFusion.theta[3] = 1.0f;

	// init matrices
	ctrlFusion.matCurMotor.pData = ctrlFusion.curMotor;
	ctrlFusion.matCurMotor.numCols = 1;
	ctrlFusion.matCurMotor.numRows = 4;

	ctrlFusion.matCurForce.pData = ctrlFusion.curForce;
	ctrlFusion.matCurForce.numCols = 1;
	ctrlFusion.matCurForce.numRows = 4;

	ctrlFusion.matCurAcc.pData = ctrlFusion.curAcc;
	ctrlFusion.matCurAcc.numCols = 1;
	ctrlFusion.matCurAcc.numRows = 3;

	ctrlFusion.matCurXYW.pData = ctrlFusion.curXYW;
	ctrlFusion.matCurXYW.numCols = 1;
	ctrlFusion.matCurXYW.numRows = 3;

	ctrlFusion.matGlobalOutXYW.pData = ctrlFusion.globalOutXYW;
	ctrlFusion.matGlobalOutXYW.numCols = 1;
	ctrlFusion.matGlobalOutXYW.numRows = 3;

	ctrlFusion.matOutXYW.pData = ctrlFusion.outXYW;
	ctrlFusion.matOutXYW.numCols = 1;
	ctrlFusion.matOutXYW.numRows = 3;

	ctrlFusion.matOutMotor.pData = ctrlFusion.outMotor;
	ctrlFusion.matOutMotor.numCols = 1;
	ctrlFusion.matOutMotor.numRows = 4;

	CtrlFusionSetPT1TimeConstant(0.2f);

	// init control helpers
	ctrlFusion.posCounter = 0;
	ctrlFusion.velCounter = 0;
	memset(ctrlFusion.posOutput, 0, sizeof(float) * 3);
	memset(ctrlFusion.velOutput, 0, sizeof(float) * 3);

	// init spline lists
	for(uint8_t i = 0; i < 3; i++)
	{
		HermiteSplineListInit(&ctrlFusion.splineList[i], listData[i], 10);
	}

	// init PIDs (pos and vel)
	for(uint8_t i = 0; i < 3; i++)
	{
		ctrlFusion.pidSpline[i].Kp = 0.0f;
		ctrlFusion.pidSpline[i].Ki = 0.0f;
		ctrlFusion.pidSpline[i].Kd = 0.0f;
		arm_pid_init_f32(&ctrlFusion.pidSpline[i], 1);

		ctrlFusion.pidVel[i].Kp = 0.0f;
		ctrlFusion.pidVel[i].Ki = 0.0f;
		ctrlFusion.pidVel[i].Kd = 0.0f;
		arm_pid_init_f32(&ctrlFusion.pidVel[i], 1);

		ctrlFusion.pidPos[i].Kp = 0.0f;
		ctrlFusion.pidPos[i].Ki = 0.0f;
		ctrlFusion.pidPos[i].Kd = 0.0f;
		arm_pid_init_f32(&ctrlFusion.pidPos[i], 1);
	}

	// init motor acceleration PIDs (inverted system model)
//	float Jm = 1.452e-5f; // Moment of inertia at motor
	for(uint8_t i = 0; i < 4; i++)
	{
		ctrlFusion.pidMotor[i].Kp = 0;
		ctrlFusion.pidMotor[i].Ki = 0;
		ctrlFusion.pidMotor[i].Kd = 0;
		arm_pid_init_f32(&ctrlFusion.pidMotor[i], 1);
	}
}

void CtrlFusionReset()
{
	switchControlMode(CONTROLLER_MODE_SPLINE);
	uint8_t i;
	for(i=0;i<3;i++)
	{
		arm_pid_reset_f32(&ctrlFusion.pidSpline[i]);
		arm_pid_reset_f32(&ctrlFusion.pidVel[i]);
		arm_pid_reset_f32(&ctrlFusion.pidPos[i]);
	}
	for(i=0;i<4;i++)
	{
		arm_pid_reset_f32(&ctrlFusion.pidMotor[i]);
	}
}

static void turnLocal2Global(float orientation, float localX, float localY, float* pGlobalX, float* pGlobalY)
{
	float localToGlobalAngle = -PI / 2.0f + orientation;
	float sinA = arm_sin_f32(localToGlobalAngle);
	float cosA = arm_cos_f32(localToGlobalAngle);

	// turn to global system
	*pGlobalX = localX * cosA - localY * sinA;
	*pGlobalY = localY * cosA + localX * sinA;
}

static void turnGlobal2Local(float orientation, float globalX, float globalY, float* pLocalX, float* pLocalY)
{
	float globalToLocalAngle = PI / 2.0f - orientation;
	float sinA = arm_sin_f32(globalToLocalAngle);
	float cosA = arm_cos_f32(globalToLocalAngle);

	// turn to local system
	*pLocalX = globalX * cosA - globalY * sinA;
	*pLocalY = globalY * cosA + globalX * sinA;
}

static void controller(CtrlInput* pInput, CtrlOutput* pOutput)
{
	uint8_t xyUpdateNeeded = 0;
	uint8_t wUpdateNeeded = 0;

	// clear C matrices
	arm_mat_zero_f32(&ctrlFusion.kfX.C);
	arm_mat_zero_f32(&ctrlFusion.kfY.C);
	arm_mat_zero_f32(&ctrlFusion.kfW.C);

	if(pInput->pos.updated)
	{
		ctrlFusion.lastPositionUpdateTime = SysTime();

		if(ctrlFusion.visionLost)
		{
			ctrlFusion.visionBack = 1;
			ctrlFusion.visionBackTime = chTimeNow();
			ctrlFusion.visionLost = 0;
		}
	}

	// check vision connection
	if((SysTime() - ctrlFusion.lastPositionUpdateTime) > ctrlFusion.visionTimeout)
	{
		// vision lost! stop bot. No more KF updates.
		for(uint8_t i = 0; i < 4; i++)
			pOutput->motorPWM[i] = 0;

		ctrlFusion.visionLost = 1;

//		return;
	}

	if(chTimeNow() - ctrlFusion.visionBackTime > MS2ST(1000))
	{
		ctrlFusion.visionBack = 0;
	}

	// construct C depending on updates
	if(pInput->pos.updated && (ctrlFusion.activeSensors & SENSOR_ACTIVE_VISION ))
	{
		if(!ctrlFusion.visionBack
				&& (fabsf(pInput->pos.pos[0] - ctrl.state.pos[0]) > 1.0f || fabsf(pInput->pos.pos[1] - ctrl.state.pos[1]) > 1.0f))
		{
			NetworkPrint("Input X or Y change > 1m! (%f/%f)", pInput->pos.pos[0], ctrl.state.pos[0]);
			NetworkPrint("Input X or Y change > 1m! (%f/%f)", pInput->pos.pos[1], ctrl.state.pos[1]);
		}
		xyUpdateNeeded = 1;

		MAT_ELEMENT(ctrlFusion.kfX.C, 0, 0) = 1.0f;
		MAT_ELEMENT(ctrlFusion.kfY.C, 0, 0) = 1.0f;

		MAT_ELEMENT(ctrlFusion.kfX.z, 0, 0) = pInput->pos.pos[0];
		MAT_ELEMENT(ctrlFusion.kfY.z, 0, 0) = pInput->pos.pos[1];

		if(!ctrlFusion.visionBack && fabsf(pInput->pos.pos[2] - ctrl.state.pos[2]) > 8.0f)
		{
			NetworkPrint("Input W change > 8rad! (%f/%f)", pInput->pos.pos[2], ctrl.state.pos[2]);
		}
		wUpdateNeeded = 1;

		MAT_ELEMENT(ctrlFusion.kfW.C, 0, 0) = 1.0f;

		MAT_ELEMENT(ctrlFusion.kfW.z, 0, 0) = pInput->pos.pos[2];
	}

	if(pInput->enc.updated)
	{
		// convert motor speed to local XYW system
		memcpy(ctrlFusion.curMotor, pInput->enc.motorVel, sizeof(float) * 4);

		ctrlFusion.curMotor[0] *= ctrl.botParams.wheelRadius * CTRL_MOTOR_TO_WHEEL_RATIO;
		ctrlFusion.curMotor[1] *= ctrl.botParams.wheelRadius * CTRL_MOTOR_TO_WHEEL_RATIO;
		ctrlFusion.curMotor[2] *= ctrl.botParams.wheelRadius * CTRL_MOTOR_TO_WHEEL_RATIO;
		ctrlFusion.curMotor[3] *= ctrl.botParams.wheelRadius * CTRL_MOTOR_TO_WHEEL_RATIO; // value is now speed over ground [m/s]

		arm_mat_mult_f32(&ctrl.matMotor2XYW, &ctrlFusion.matCurMotor, &ctrlFusion.matCurXYW);
		// matCurXYW now contains X and Y in [m/s] and W in [rad/s] from encoder data

		ctrlFusion.matCurXYW.pData[1] *= 2.0f/3.0f;

		if((ctrlFusion.activeSensors & SENSOR_ACTIVE_ENCODER ))
		{
			xyUpdateNeeded = 1;
			wUpdateNeeded = 1;

			MAT_ELEMENT(ctrlFusion.kfX.C, 1, 1) = 1.0f;
			MAT_ELEMENT(ctrlFusion.kfY.C, 1, 1) = 1.0f;
			MAT_ELEMENT(ctrlFusion.kfW.C, 1, 1) = 1.0f;
			MAT_ELEMENT(ctrlFusion.kfW.z, 1, 0) = ctrlFusion.matCurXYW.pData[2];	// store new encoder rotation speed around Z

			// transfer and rotation of velocity vector is done below, if needed
		}
	}

	if(pInput->acc.updated && (ctrlFusion.activeSensors & SENSOR_ACTIVE_ACC ))
	{
		xyUpdateNeeded = 1;

		MAT_ELEMENT(ctrlFusion.kfX.C, 2, 2) = 1.0f;
		MAT_ELEMENT(ctrlFusion.kfY.C, 2, 2) = 1.0f;

		// transfer and rotation of acceleration vector is done below, if needed
	}

	if(pInput->gyr.updated && (ctrlFusion.activeSensors & SENSOR_ACTIVE_GYR ))
	{
		wUpdateNeeded = 1;

		MAT_ELEMENT(ctrlFusion.kfW.C, 2, 1) = 1.0f;

		MAT_ELEMENT(ctrlFusion.kfW.z, 2, 0) = pInput->gyr.rotVel[2];	// update gyro rotation around Z axis
	}

	if(pInput->motor.updated && (ctrlFusion.activeSensors & SENSOR_ACTIVE_MOTOR ))
	{
		xyUpdateNeeded = 1;
		wUpdateNeeded = 1;

		MAT_ELEMENT(ctrlFusion.kfX.C, 3, 2) = 1.0f;
		MAT_ELEMENT(ctrlFusion.kfY.C, 3, 2) = 1.0f;
		MAT_ELEMENT(ctrlFusion.kfW.C, 3, 2) = 1.0f;

		for(uint8_t i = 0; i < 4; i++)
		{
			// force = current * motor_constant * gear_ratio * 1/wheel_radius
			// F = I * k_m * G * 1/r_rad
			// F = I * 0.0255 * 50/15 * 1/r_rad
			// F = I * 0.085 * 1/r_rad
			ctrlFusion.curForce[i] = pInput->motor.current[i] * 0.085f * 1.0f / ctrl.botParams.wheelRadius;

			if(ctrlFusion.curMotor[i] < 0)
				ctrlFusion.curForce[i] *= -1.0f;
		}

		// convert wheel forces over ground to global force and torque
		arm_mat_mult_f32(&ctrl.matForceMotor2XYW, &ctrlFusion.matCurForce, &ctrlFusion.matCurAcc);

		// convert forces and torque to (angular) acceleration
		// forces still have to be rotated to global system, done below
		ctrlFusion.curAcc[0] /= ctrl.botParams.mass;
		ctrlFusion.curAcc[1] /= ctrl.botParams.mass;
		ctrlFusion.curAcc[2] /= ctrl.momentOfInertia;

		MAT_ELEMENT(ctrlFusion.kfW.z, 3, 0) = ctrlFusion.curAcc[2];	// update angular acceleration around Z axis
	}

	pInput->pos.updated = 0;
	pInput->enc.updated = 0;
	pInput->acc.updated = 0;
	pInput->gyr.updated = 0;
	pInput->motor.updated = 0;

	ctrlFusion.kfX.A.pData[1] = ctrl.dt;
	ctrlFusion.kfX.A.pData[2] = 0.5*ctrl.dt*ctrl.dt;
	ctrlFusion.kfX.A.pData[5] = ctrl.dt;

	ctrlFusion.kfY.A.pData[1] = ctrl.dt;
	ctrlFusion.kfY.A.pData[2] = 0.5*ctrl.dt*ctrl.dt;
	ctrlFusion.kfY.A.pData[5] = ctrl.dt;

	ctrlFusion.kfW.A.pData[1] = ctrl.dt;
	ctrlFusion.kfW.A.pData[2] = 0.5*ctrl.dt*ctrl.dt;
	ctrlFusion.kfW.A.pData[5] = ctrl.dt;

	// predict W first, as it is needed to rotate X and Y to global system
	KFPredict(&ctrlFusion.kfW);
	AngleNormalizePtr(&ctrlFusion.kfW.x.pData[0]);	// normalize orientation
	if(wUpdateNeeded)
	{
		KFUpdate(&ctrlFusion.kfW);
		AngleNormalizePtr(&ctrlFusion.kfW.x.pData[0]);	// normalize orientation

		arm_mat_clip_f32(&ctrlFusion.kfW.Sigma, -100.0f, 100.0f);
	}

	// predict X,Y
	KFPredict(&ctrlFusion.kfX);
	KFPredict(&ctrlFusion.kfY);

	if(xyUpdateNeeded)
	{
		// X and Y can be updated
		turnLocal2Global(ctrlFusion.kfW.x.pData[0], ctrlFusion.matCurXYW.pData[0], ctrlFusion.matCurXYW.pData[1],
				&MAT_ELEMENT(ctrlFusion.kfX.z, 1, 0), &MAT_ELEMENT(ctrlFusion.kfY.z, 1, 0));

		turnLocal2Global(ctrlFusion.kfW.x.pData[0], pInput->acc.linAcc[0], pInput->acc.linAcc[1], &MAT_ELEMENT(ctrlFusion.kfX.z, 2, 0),
				&MAT_ELEMENT(ctrlFusion.kfY.z, 2, 0));

		turnLocal2Global(ctrlFusion.kfW.x.pData[0], ctrlFusion.curAcc[0], ctrlFusion.curAcc[1], &MAT_ELEMENT(ctrlFusion.kfX.z, 3, 0),
				&MAT_ELEMENT(ctrlFusion.kfY.z, 3, 0));

		// now update
		KFUpdate(&ctrlFusion.kfX);
		KFUpdate(&ctrlFusion.kfY);

		arm_mat_clip_f32(&ctrlFusion.kfX.Sigma, -100.0f, 100.0f);
		arm_mat_clip_f32(&ctrlFusion.kfY.Sigma, -100.0f, 100.0f);
	}

	// test plausibility
//	if(fabsf(ctrl.state.pos[0] - MAT_ELEMENT(ctrlFusion.kfX.x, 0, 0)) > 1.0f)
//	{
//		NetworkPrint("KF Update X error > 1m (%f/%f)", ctrl.state.pos[0], MAT_ELEMENT(ctrlFusion.kfX.x, 0, 0));
//	}
//	else
//	{
//		ctrl.state.pos[0] = MAT_ELEMENT(ctrlFusion.kfX.x, 0, 0);
//	}
//
//	if(fabsf(ctrl.state.pos[1] - MAT_ELEMENT(ctrlFusion.kfY.x, 0, 0)) > 1.0f)
//	{
//		NetworkPrint("KF Update Y error > 1m (%f/%f)", ctrl.state.pos[1], MAT_ELEMENT(ctrlFusion.kfY.x, 0, 0));
//	}
//	else
//	{
//		ctrl.state.pos[1] = MAT_ELEMENT(ctrlFusion.kfY.x, 0, 0);
//	}
//
//	if(fabsf(ctrl.state.pos[2] - MAT_ELEMENT(ctrlFusion.kfW.x, 0, 0)) > 1.0f)
//	{
//		NetworkPrint("KF Update W error > 1rad (%f/%f)", ctrl.state.pos[2], MAT_ELEMENT(ctrlFusion.kfW.x, 0, 0));
//	}
//	else
//	{
//		ctrl.state.pos[2] = MAT_ELEMENT(ctrlFusion.kfW.x, 0, 0);
//	}

	// update feedback state
	ctrl.state.pos[0] = MAT_ELEMENT(ctrlFusion.kfX.x, 0, 0);
	ctrl.state.pos[1] = MAT_ELEMENT(ctrlFusion.kfY.x, 0, 0);
	ctrl.state.pos[2] = MAT_ELEMENT(ctrlFusion.kfW.x, 0, 0);

	ctrl.state.vel[0] = MAT_ELEMENT(ctrlFusion.kfX.x, 1, 0);
	ctrl.state.vel[1] = MAT_ELEMENT(ctrlFusion.kfY.x, 1, 0);
	ctrl.state.vel[2] = MAT_ELEMENT(ctrlFusion.kfW.x, 1, 0);

	ctrl.state.acc[0] = MAT_ELEMENT(ctrlFusion.kfX.x, 2, 0);
	ctrl.state.acc[1] = MAT_ELEMENT(ctrlFusion.kfY.x, 2, 0);
	ctrl.state.acc[2] = MAT_ELEMENT(ctrlFusion.kfW.x, 2, 0);

	float pDDp[16];
	arm_matrix_instance_f32 matDDp = { .numCols = 4, .numRows = 4, .pData = pDDp };
	arm_mat_mult_f32(&ctrl.matXYW2Motor, &ctrl.matMotor2XYW, &matDDp);
	arm_matrix_instance_f32 matVecM1 = { .numCols = 1, .numRows = 4, .pData = ctrlFusion.curMotor };
	float pVecM2[4];
	arm_matrix_instance_f32 matVecM2 = { .numCols = 1, .numRows = 4, .pData = pVecM2 };
	arm_mat_mult_f32(&matDDp, &matVecM1, &matVecM2);

	arm_matrix_instance_f32 matSlippage = { .numCols = 1, .numRows = 4, .pData = ctrl.state.slippage };

	arm_mat_sub_f32(&matVecM1, &matVecM2, &matSlippage);

	ctrl.state.updated[0] = 1;
	ctrl.state.updated[1] = 1;
	ctrl.state.updated[2] = 1;

	for(uint8_t i = 0; i < 3; i++)
	{
		pOutput->pos[i] = ctrl.state.pos[i];
		pOutput->vel[i] = ctrl.state.vel[i];
	}

	ctrlFusion.posCounter++;
	ctrlFusion.posCounter %= 20;
	ctrlFusion.velCounter++;
	ctrlFusion.velCounter %= 4;

	switch(controlMode)
	{
		case CONTROLLER_MODE_SPLINE: // spline
			for(uint8_t i = 0; i < 3; i++)
			{
				float trajTime = (SysTime() - ctrlFusion.trajectoryStartTimeXY);
				if(i == 2)
				{
					trajTime = (SysTime() - ctrlFusion.trajectoryStartTimeW);
				}

				float partTime = 0.0f;
				HermiteSplineListItem* pItem = ctrlFusion.splineList[i].pItems;
				if(ctrlFusion.splineList[i].numItems > 0)
				{
					pItem = HermiteSplineListGetItem(&ctrlFusion.splineList[i], trajTime);
					partTime = trajTime - pItem->tStart;
				}

				if(trajTime > ctrlFusion.splineList[i].totalTime + 3.0f)
				{
					HermiteSplineListClear(&ctrlFusion.splineList[i]);
					break;
				}

				float setPos = 0;
				float errorPos = 0;
				if(ctrlFusion.splineList[i].numItems > 0)
				{
					setPos = HermiteSplineValue(&pItem->spline, partTime);
					errorPos = setPos - ctrl.state.pos[i];
					if(i == 2)
						AngleNormalizePtr(&errorPos);	// shortest rotation distance
				}

				float setVel = 0.0f;
				float errorVel = 0.0f;
				if(ctrlFusion.splineList[i].numItems > 0 && pItem->spline.tEnd >= partTime)
				{
					setVel = HermiteSplineFirstDerivative(&pItem->spline, partTime);
					errorVel = setVel - ctrl.state.vel[i];
				}

				if(ctrl.controllersEnabled)
				{
//					ctrlFusion.velOutput[i] = setVel + ctrlFusion.pidSpline[i].Kd * errorVel + ctrlFusion.pidSpline[i].Kp * errorPos;
					ctrlFusion.velOutput[i] = setVel + arm_pid_f32(&ctrlFusion.pidSpline[i], errorPos);
				}
				else
				{
					ctrlFusion.velOutput[i] = setVel;
				}
				ctrlFusion.globalOutXYW[i] = ctrlFusion.velOutput[i];

				ctrl.state.setVel[i] = setVel;
				ctrl.state.velOutput[i] = ctrlFusion.velOutput[i];
				ctrl.state.errorVel[i] = errorPos;
			}
			break;
		case CONTROLLER_MODE_VEL: // velocity
		{
			float setPoint[3];
			turnLocal2Global(ctrl.state.pos[2], ctrlFusion.setpoint[0], ctrlFusion.setpoint[1], &setPoint[0], &setPoint[1]);
			setPoint[2] = ctrlFusion.setpoint[2];
			for(uint8_t i = 0; i < 3; i++)
			{
				if(ctrl.controllersEnabled)
				{
					// PT1 on setpoint
					float setVel;
					arm_biquad_cascade_df1_f32(&ctrlFusion.pt1[i], &setPoint[i], &setVel, 1);
					// velocity control
					float errorVel = setVel - ctrl.state.vel[i];
					ctrlFusion.velOutput[i] = setVel + arm_pid_f32(&ctrlFusion.pidVel[i], errorVel);
					ctrlFusion.globalOutXYW[i] = ctrlFusion.velOutput[i];

					ctrl.state.setVel[i] = setVel;
					ctrl.state.velOutput[i] = ctrlFusion.velOutput[i];
					ctrl.state.errorVel[i] = errorVel;
				} else {
					float setVel = setPoint[i];
					float errorVel = setVel - ctrl.state.vel[i];
					ctrlFusion.velOutput[i] = setVel;
					ctrlFusion.globalOutXYW[i] = ctrlFusion.velOutput[i];

					ctrl.state.setVel[i] = setVel;
					ctrl.state.velOutput[i] = ctrlFusion.velOutput[i];
					ctrl.state.errorVel[i] = errorVel;
				}
			}
		}
			break;
		case CONTROLLER_MODE_POS: // position
		{
			float errorPosX = ctrlFusion.setpoint[0] - ctrl.state.pos[0];
			float errorPosY = ctrlFusion.setpoint[1] - ctrl.state.pos[1];
			float errorXY = sqrt(errorPosX*errorPosX+errorPosY*errorPosY);

			if(errorXY > 0.001f) {
				float setPoint = errorXY;
				if(setPoint > 4)
				{
					setPoint = 4;
				}
				float velOut;
				arm_biquad_cascade_df1_f32(&ctrlFusion.pt1[0], &setPoint, &velOut, 1);
				if(setPoint < velOut)
				{
					velOut = setPoint;
				}
				velOut = arm_pid_f32(&ctrlFusion.pidPos[0], velOut);
				ctrlFusion.velOutput[0] = errorPosX/errorXY * velOut;
				ctrlFusion.velOutput[1] = errorPosY/errorXY * velOut;
			} else {
				ctrlFusion.velOutput[0] = 0.0f;
				ctrlFusion.velOutput[1] = 0.0f;
			}

			float errorPosW = ctrlFusion.setpoint[2] - ctrl.state.pos[2];
			AngleNormalizePtr(&errorPosW);	// shortest rotation distance
			float setPointW = errorPosW;
			arm_biquad_cascade_df1_f32(&ctrlFusion.pt1[2], &setPointW, &(ctrlFusion.velOutput[2]), 1);
			if(fabsf(setPointW) < fabsf(ctrlFusion.velOutput[2]))
			{
				ctrlFusion.velOutput[2] = setPointW;
			}
			ctrlFusion.velOutput[2] = arm_pid_f32(&ctrlFusion.pidPos[2], ctrlFusion.velOutput[2]);

			ctrl.state.errorVel[0] = errorXY;
			ctrl.state.errorVel[1] = errorPosX;
			ctrl.state.errorVel[2] = errorPosW;

			for(uint8_t i = 0; i < 3; i++)
			{
				ctrlFusion.globalOutXYW[i] = ctrlFusion.velOutput[i];
				ctrl.state.setVel[i] = ctrlFusion.setpoint[i];
				ctrl.state.velOutput[i] = ctrlFusion.velOutput[i];
			}
			break;
		}
	}

	// turn back to local system
	turnGlobal2Local(ctrl.state.pos[2], ctrlFusion.globalOutXYW[0], ctrlFusion.globalOutXYW[1], &ctrlFusion.outXYW[0],
			&ctrlFusion.outXYW[1]);

	// angular acceleration is copied as-it-is
	ctrlFusion.outXYW[2] = ctrlFusion.globalOutXYW[2];

	// TODO: stall protection

	// grab current battery voltage (or guess)
	float UBattery = 15.0f;
	if(pInput->power.updated)
	{
		UBattery = pInput->power.voltage;
	}

	ctrlFusion.outXYW[1] *= 4.0f/3.0f;

	// convert to speed over ground of motors
	arm_mat_mult_f32(&ctrl.matXYW2Motor, &ctrlFusion.matOutXYW, &ctrlFusion.matOutMotor);

	if(motorCtrlNeeded())
	{
		for(uint8_t i = 0; i < 4; i++)
		{
			float outWheel = ctrlFusion.outMotor[i];;
/*			if(ctrl.controllersEnabled)
			{
				outWheel = ctrlFusion.outMotor[i] * ctrlFusion.theta[0];
				float errorVel = outWheel - ctrlFusion.curMotor[i];
				outWheel = outWheel + arm_pid_f32(&ctrlFusion.pidMotor[i], errorVel);
			} else {
				outWheel = ctrlFusion.outMotor[i];
			}
*/
			// convert speed over ground to motor speed
			pOutput->motorRPS[i] = outWheel * CTRL_WHEEL_TO_MOTOR_RATIO * 1/ctrl.botParams.wheelRadius * 1/(2*PI);
/*
			// 50/15 * speed_over_ground * 1/r_rad
			float motorSpeed = CTRL_WHEEL_TO_MOTOR_RATIO * 1 / ctrl.botParams.wheelRadius * outWheel;
			// and convert to voltage
			float outVoltage = motorSpeed / CTRL_MOTOR_KN;

			if(fabsf(pInput->motor.current[i]) > 7.0f)
			{
				NetworkPrint("Motor %hhu overcurrent %f", i, pInput->motor.current[i]);
				pOutput->motorPWM[i] = 5.0f / UBattery;
			}
			else
			{
				pOutput->motorPWM[i] = outVoltage / UBattery;
			}
*/
			ctrl.state.motorSet[i] = ctrlFusion.outMotor[i];
			ctrl.state.motorCur[i] = ctrlFusion.curMotor[i];
			ctrl.state.motorOut[i] = outWheel;
		}
	} else {
		for(uint8_t i = 0; i < 4; i++)
		{
			pOutput->motorPWM[i] = 0;
			pOutput->motorRPS[i] = 0;
			ctrl.state.motorSet[i] = ctrlFusion.outMotor[i];
			ctrl.state.motorCur[i] = ctrlFusion.curMotor[i];
			ctrl.state.motorOut[i] = 0;
			arm_pid_reset_f32(&ctrlFusion.pidMotor[i]);
			arm_biquad_cascade_df1_init_f32(&ctrlFusion.pt1[0], 1, ctrlFusion.pt1Coeff, ctrlFusion.pt1State[0]);
			arm_biquad_cascade_df1_init_f32(&ctrlFusion.pt1[1], 1, ctrlFusion.pt1Coeff, ctrlFusion.pt1State[1]);
			arm_biquad_cascade_df1_init_f32(&ctrlFusion.pt1[2], 1, ctrlFusion.pt1Coeff, ctrlFusion.pt1State[2]);
		}
	}
	ctrlFusion.lastWakeTime = chTimeNow();
}

void CtrlFusionSetStateUncertainty(float* pEx)
{
	if(ctrl.pInstance != &ctrlFusionInstance)
		return;

	for(uint16_t i = 0; i < ctrlFusion.kfX.f; i++)
	{
		MAT_ELEMENT(ctrlFusion.kfX.Ex, i, i) = pEx[i];
		MAT_ELEMENT(ctrlFusion.kfY.Ex, i, i) = pEx[i];
		MAT_ELEMENT(ctrlFusion.kfW.Ex, i, i) = pEx[i];
	}
}

void CtrlFusionSetSensorUncertainty(float* pEz)
{
	if(ctrl.pInstance != &ctrlFusionInstance)
		return;

	for(uint16_t i = 0; i < ctrlFusion.kfX.h; i++)
	{
		MAT_ELEMENT(ctrlFusion.kfX.Ez, i, i) = pEz[i];
		MAT_ELEMENT(ctrlFusion.kfY.Ez, i, i) = pEz[i];
		MAT_ELEMENT(ctrlFusion.kfW.Ez, i, i) = pEz[i];
	}
}

void CtrlFusionAddSplineCmdXY(CtrlSpline2D* pSpline)
{
	if(ctrl.pInstance != &ctrlFusionInstance)
		return;
	switchControlMode(CONTROLLER_MODE_SPLINE);

	if(pSpline->options <= 1 || ctrlFusion.splineList[0].numItems != pSpline->options-1)
	{
		HermiteSplineListClear(&ctrlFusion.splineList[0]);
		HermiteSplineListClear(&ctrlFusion.splineList[1]);

		arm_pid_reset_f32(&ctrlFusion.pidSpline[0]);
		arm_pid_reset_f32(&ctrlFusion.pidSpline[1]);

		ctrlFusion.trajectoryStartTimeXY = SysTime();

		/*		IntercomHeader iHeader;
		 iHeader.section = INTERCOM_SECTION_SENSORS;
		 iHeader.address = INTERCOM_ADDRESS_MEDIA;
		 iHeader.command = 1;

		 IntercomSend(&intercom1, &iHeader, 0, 0);
		 */
//		ConsolePrint("Splines XY cleared\r\n");
	}

	if(ctrlFusion.splineList[0].numItems == pSpline->options-1)
	{
		if(ctrlFusion.splineList[0].numItems == 0)
		{
			HermiteSplineListItem* pItemX = HermiteSplineListAllocate(&ctrlFusion.splineList[0]);
			if(!pItemX)
			{
				NetworkPrint("Could not allocate spline item");
				return;
			}

			HermiteSplineListItem* pItemY = HermiteSplineListAllocate(&ctrlFusion.splineList[1]);
			if(!pItemY)
			{
				NetworkPrint("Could not allocate spline item");
				return;
			}

			HermiteSpline* pSplineX = &pItemX->spline;
			HermiteSpline* pSplineY = &pItemY->spline;

			for(uint8_t j = 0; j < SPLINE_SIZE; j++)
			{
				pSplineX->a[j] = halfFloatToFloat(pSpline->splines[0].a[j]);
				pSplineY->a[j] = halfFloatToFloat(pSpline->splines[1].a[j]);
			}
			pSplineX->tEnd = halfFloatToFloat(pSpline->tEnd);
			pSplineY->tEnd = halfFloatToFloat(pSpline->tEnd);

			float maxVel = HermiteSpline2DMaxFirstDerivative(pSplineX, pSplineY);
			float maxAcc = HermiteSpline2DMaxSecondDerivative(pSplineX, pSplineY);

			ConsolePrint("%f %f %f\r\n", pSplineX->tEnd, maxVel, maxAcc);

			Vector2f finalPos = Vector2fCreate(HermiteSplineValue(pSplineX, pSplineX->tEnd), HermiteSplineValue(pSplineY, pSplineY->tEnd));
			Vector2f finalVel = Vector2fCreate(HermiteSplineFirstDerivative(pSplineX, pSplineX->tEnd),
					HermiteSplineFirstDerivative(pSplineY, pSplineY->tEnd));

			Vector2f initialPos = Vector2fCreate(ctrl.state.pos[0], ctrl.state.pos[1]);
			Vector2f initialVel = Vector2fCreate(ctrl.state.vel[0], ctrl.state.vel[1]);

			HermiteSpline2DGenerate(initialPos, finalPos, initialVel, finalVel, maxVel, maxAcc, pSplineX, pSplineY);

			maxVel = HermiteSpline2DMaxFirstDerivative(pSplineX, pSplineY);
			maxAcc = HermiteSpline2DMaxSecondDerivative(pSplineX, pSplineY);

			pItemX->tStart = 0;
			pItemX->tEnd = pSplineX->tEnd;

			pItemY->tStart = 0;
			pItemY->tEnd = pSplineY->tEnd;

//			ConsolePrint("pos: (%f, %f) -> (%f, %f)\r\n", initialPos.x, initialPos.y, finalPos.x, finalPos.y);
//			ConsolePrint("vel: (%f, %f) -> (%f, %f)\r\n", initialVel.x, initialVel.y, finalVel.x, finalVel.y);
//			ConsolePrint("%f %f %f\r\n", pSplineX->tEnd, maxVel, maxAcc);
		}
		else
		{
			for(uint8_t i = 0; i < 2; i++)
			{
				HermiteSplineListItem* pItem = HermiteSplineListAllocate(&ctrlFusion.splineList[i]);
				if(!pItem){
					ConsolePrint("Could not allocate spline item.\r\n");
					return;
				}

				for(uint8_t j = 0; j < SPLINE_SIZE; j++)
				{
					pItem->spline.a[j] = halfFloatToFloat(pSpline->splines[i].a[j]);
				}
				pItem->spline.tEnd = halfFloatToFloat(pSpline->tEnd);
			}
		}

		HermiteSplineListUpdate(&ctrlFusion.splineList[0]);
		HermiteSplineListUpdate(&ctrlFusion.splineList[1]);

//		ConsolePrint("Spline XY appended\r\n");
	}
}

void CtrlFusionAddSplineCmdW(CtrlSpline1D* pSpline1D)
{
	if(ctrl.pInstance != &ctrlFusionInstance)
		return;
	switchControlMode(CONTROLLER_MODE_SPLINE);

	if(pSpline1D->options <= 1 || ctrlFusion.splineList[2].numItems != pSpline1D->options-1)
	{
		HermiteSplineListClear(&ctrlFusion.splineList[2]);

		arm_pid_reset_f32(&ctrlFusion.pidSpline[2]);

		ctrlFusion.trajectoryStartTimeW = SysTime();

		ConsolePrint("Splines W cleared\r\n");
	}

	if(ctrlFusion.splineList[2].numItems == pSpline1D->options-1)
	{
		if(ctrlFusion.splineList[2].numItems == 0)
		{
			HermiteSplineListItem* pItem = HermiteSplineListAllocate(&ctrlFusion.splineList[2]);
			if(!pItem)
				return;

			HermiteSpline* pSpline = &pItem->spline;

			for(uint8_t j = 0; j < SPLINE_SIZE; j++)
			{
				pSpline->a[j] = halfFloatToFloat(pSpline1D->a[j]);
			}
			pSpline->tEnd = halfFloatToFloat(pSpline1D->tEnd);

//			float maxVel = HermiteSplineMaxFirstDerivative(pSpline);
//			float maxAcc = HermiteSplineMaxSecondDerivative(pSpline);
//
//			float finalPos = HermiteSplineValue(pSpline, pSpline->tEnd);
//			float finalVel = HermiteSplineFirstDerivative(pSpline, pSpline->tEnd);
//
//			float initialPos = ctrl.state.pos[2];
//			float initialVel = ctrl.state.vel[2];
//
//			HermiteSplineGenerate(initialPos, finalPos, initialVel, finalVel, pSpline->tEnd, maxVel, maxAcc, 1, pSpline);

			pItem->tStart = 0;
			pItem->tEnd = pSpline->tEnd;
		}
		else
		{
			HermiteSplineListItem* pItem = HermiteSplineListAllocate(&ctrlFusion.splineList[2]);
			if(!pItem)
				return;

			for(uint8_t j = 0; j < SPLINE_SIZE; j++)
			{
				pItem->spline.a[j] = halfFloatToFloat(pSpline1D->a[j]);
			}
			pItem->spline.tEnd = halfFloatToFloat(pSpline1D->tEnd);
		}

		HermiteSplineListUpdate(&ctrlFusion.splineList[2]);

//		ConsolePrint("Spline W appended\r\n");
	}
}

void CtrlFusionSetPIDParams(CtrlSetPIDParams* pParams)
{
	if(ctrl.pInstance != &ctrlFusionInstance)
		return;

	switch(pParams->type)
	{
		case CTRL_PID_PARAM_SPLINE_X:
		{
			ctrlFusion.pidSpline[0].Kp = pParams->Kp;
			ctrlFusion.pidSpline[0].Ki = pParams->Ki;
			ctrlFusion.pidSpline[0].Kd = pParams->Kd;
			arm_pid_init_f32(&ctrlFusion.pidSpline[0], 1);
		}
			break;
		case CTRL_PID_PARAM_SPLINE_Y:
		{
			ctrlFusion.pidSpline[1].Kp = pParams->Kp;
			ctrlFusion.pidSpline[1].Ki = pParams->Ki;
			ctrlFusion.pidSpline[1].Kd = pParams->Kd;
			arm_pid_init_f32(&ctrlFusion.pidSpline[1], 1);
		}
			break;
		case CTRL_PID_PARAM_SPLINE_W:
		{
			ctrlFusion.pidSpline[2].Kp = pParams->Kp;
			ctrlFusion.pidSpline[2].Ki = pParams->Ki;
			ctrlFusion.pidSpline[2].Kd = pParams->Kd;
			arm_pid_init_f32(&ctrlFusion.pidSpline[2], 1);
		}
			break;
		case CTRL_PID_PARAM_POS_X:
		{
			ctrlFusion.pidPos[0].Kp = pParams->Kp;
			ctrlFusion.pidPos[0].Ki = pParams->Ki;
			ctrlFusion.pidPos[0].Kd = pParams->Kd;
			arm_pid_init_f32(&ctrlFusion.pidPos[0], 1);
		}
			break;
		case CTRL_PID_PARAM_POS_Y:
		{
			ctrlFusion.pidPos[1].Kp = pParams->Kp;
			ctrlFusion.pidPos[1].Ki = pParams->Ki;
			ctrlFusion.pidPos[1].Kd = pParams->Kd;
			arm_pid_init_f32(&ctrlFusion.pidPos[1], 1);
		}
			break;
		case CTRL_PID_PARAM_POS_W:
		{
			ctrlFusion.pidPos[2].Kp = pParams->Kp;
			ctrlFusion.pidPos[2].Ki = pParams->Ki;
			ctrlFusion.pidPos[2].Kd = pParams->Kd;
			arm_pid_init_f32(&ctrlFusion.pidPos[2], 1);
		}
			break;
		case CTRL_PID_PARAM_VEL_X:
		{
			ctrlFusion.pidVel[0].Kp = pParams->Kp;
			ctrlFusion.pidVel[0].Ki = pParams->Ki;
			ctrlFusion.pidVel[0].Kd = pParams->Kd;
			arm_pid_init_f32(&ctrlFusion.pidVel[0], 1);
		}
			break;
		case CTRL_PID_PARAM_VEL_Y:
		{
			ctrlFusion.pidVel[1].Kp = pParams->Kp;
			ctrlFusion.pidVel[1].Ki = pParams->Ki;
			ctrlFusion.pidVel[1].Kd = pParams->Kd;
			arm_pid_init_f32(&ctrlFusion.pidVel[1], 1);
		}
			break;
		case CTRL_PID_PARAM_VEL_W:
		{
			ctrlFusion.pidVel[2].Kp = pParams->Kp;
			ctrlFusion.pidVel[2].Ki = pParams->Ki;
			ctrlFusion.pidVel[2].Kd = pParams->Kd;
			arm_pid_init_f32(&ctrlFusion.pidVel[2], 1);
		}
			break;
		case CTRL_PID_PARAM_MOTOR:
		{
			for(uint8_t i = 0; i < 4; i++)
			{
				ctrlFusion.pidMotor[i].Kp = pParams->Kp;
				ctrlFusion.pidMotor[i].Ki = pParams->Ki;
				ctrlFusion.pidMotor[i].Kd = pParams->Kd;
				arm_pid_init_f32(&ctrlFusion.pidMotor[i], 1);
			}
		}
			break;
	}
}

void CtrlFusionSetFilterParams(CtrlSetFilterParams* pParams)
{
	if(ctrl.pInstance != &ctrlFusionInstance)
		return;

	switch(pParams->type)
	{
		case CTRL_FILTER_PARAM_EX_POS:
		{
			MAT_ELEMENT(ctrlFusion.kfX.Ex, 0, 0) = pParams->params[0];
			MAT_ELEMENT(ctrlFusion.kfY.Ex, 0, 0) = pParams->params[1];
			MAT_ELEMENT(ctrlFusion.kfW.Ex, 0, 0) = pParams->params[2];
		}
			break;
		case CTRL_FILTER_PARAM_EX_VEL:
		{
			MAT_ELEMENT(ctrlFusion.kfX.Ex, 1, 1) = pParams->params[0];
			MAT_ELEMENT(ctrlFusion.kfY.Ex, 1, 1) = pParams->params[1];
			MAT_ELEMENT(ctrlFusion.kfW.Ex, 1, 1) = pParams->params[2];
		}
			break;
		case CTRL_FILTER_PARAM_EX_ACC:
		{
			MAT_ELEMENT(ctrlFusion.kfX.Ex, 2, 2) = pParams->params[0];
			MAT_ELEMENT(ctrlFusion.kfY.Ex, 2, 2) = pParams->params[1];
			MAT_ELEMENT(ctrlFusion.kfW.Ex, 2, 2) = pParams->params[2];
		}
			break;
		case CTRL_FILTER_PARAM_EZ_VISION:
		{
			MAT_ELEMENT(ctrlFusion.kfX.Ez, 0, 0) = pParams->params[0];
			MAT_ELEMENT(ctrlFusion.kfY.Ez, 0, 0) = pParams->params[1];
			MAT_ELEMENT(ctrlFusion.kfW.Ez, 0, 0) = pParams->params[2];
		}
			break;
		case CTRL_FILTER_PARAM_EZ_ENCODER:
		{
			MAT_ELEMENT(ctrlFusion.kfX.Ez, 1, 1) = pParams->params[0];
			MAT_ELEMENT(ctrlFusion.kfY.Ez, 1, 1) = pParams->params[1];
			MAT_ELEMENT(ctrlFusion.kfW.Ez, 1, 1) = pParams->params[2];
		}
			break;
		case CTRL_FILTER_PARAM_EZ_ACC_GYRO:
		{
			MAT_ELEMENT(ctrlFusion.kfX.Ez, 2, 2) = pParams->params[0];
			MAT_ELEMENT(ctrlFusion.kfY.Ez, 2, 2) = pParams->params[1];
			MAT_ELEMENT(ctrlFusion.kfW.Ez, 2, 2) = pParams->params[2];
		}
			break;
		case CTRL_FILTER_PARAM_EZ_MOTOR:
		{
			MAT_ELEMENT(ctrlFusion.kfX.Ez, 3, 3) = pParams->params[0];
			MAT_ELEMENT(ctrlFusion.kfY.Ez, 3, 3) = pParams->params[1];
			MAT_ELEMENT(ctrlFusion.kfW.Ez, 3, 3) = pParams->params[2];
		}
			break;
	}
}

void CtrlFusionSetVelTarget(float x, float y, float w)
{
	switchControlMode(CONTROLLER_MODE_VEL);
	ctrlFusion.setpoint[0] = x;
	ctrlFusion.setpoint[1] = y;
	ctrlFusion.setpoint[2] = w;

}

static uint8_t equals(float a, float b)
{
	if(fabs(a-b)<0.001)
	{
		return 1;
	}
	return 0;
}

void CtrlFusionSetPosTarget(float x, float y, float w)
{
	switchControlMode(CONTROLLER_MODE_POS);

//	if(!equals(ctrlFusion.setpoint[0],x) && !equals(ctrlFusion.setpoint[1],y))
//	{
//		arm_biquad_cascade_df1_init_f32(&ctrlFusion.pt1[0], 1, ctrlFusion.pt1Coeff, ctrlFusion.pt1State[0]);
//		arm_biquad_cascade_df1_init_f32(&ctrlFusion.pt1[1], 1, ctrlFusion.pt1Coeff, ctrlFusion.pt1State[1]);
//	}
//
//	if(!equals(ctrlFusion.setpoint[2],w))
//	{
//		arm_biquad_cascade_df1_init_f32(&ctrlFusion.pt1[2], 1, ctrlFusion.pt1Coeff, ctrlFusion.pt1State[2]);
//	}

	ctrlFusion.setpoint[0] = x;
	ctrlFusion.setpoint[1] = y;
	ctrlFusion.setpoint[2] = w;
}

void CtrlFusionSetPT1TimeConstant(float T)
{
	// init PT1 and PT1 derivative
	const float K = 1.0f;
	float a = (2 * T - DELTA_T) / (2 * T + DELTA_T);
	float b = K * DELTA_T / (2 * T + DELTA_T);
//	float c = 2*K/(2*T+DELTA_T);

	ctrlFusion.pt1Coeff[0] = b;
	ctrlFusion.pt1Coeff[1] = b;
	ctrlFusion.pt1Coeff[2] = 0.0f;
	ctrlFusion.pt1Coeff[3] = a;
	ctrlFusion.pt1Coeff[4] = 0.0f;

//	ctrlFusion.pt1DtCoeff[0] = c;
//	ctrlFusion.pt1DtCoeff[1] = -c;
//	ctrlFusion.pt1DtCoeff[2] = 0.0f;
//	ctrlFusion.pt1DtCoeff[3] = a;
//	ctrlFusion.pt1DtCoeff[4] = 0.0f;

	arm_biquad_cascade_df1_init_f32(&ctrlFusion.pt1[0], 1, ctrlFusion.pt1Coeff, ctrlFusion.pt1State[0]);
	arm_biquad_cascade_df1_init_f32(&ctrlFusion.pt1[1], 1, ctrlFusion.pt1Coeff, ctrlFusion.pt1State[1]);
	arm_biquad_cascade_df1_init_f32(&ctrlFusion.pt1[2], 1, ctrlFusion.pt1Coeff, ctrlFusion.pt1State[2]);

//	arm_biquad_cascade_df1_init_f32(&ctrlFusion.pt1Dt[0], 1, ctrlFusion.pt1DtCoeff, ctrlFusion.pt1DtState[0]);
//	arm_biquad_cascade_df1_init_f32(&ctrlFusion.pt1Dt[1], 1, ctrlFusion.pt1DtCoeff, ctrlFusion.pt1DtState[1]);
//	arm_biquad_cascade_df1_init_f32(&ctrlFusion.pt1Dt[2], 1, ctrlFusion.pt1DtCoeff, ctrlFusion.pt1DtState[2]);
}
