/*
 * ctrl_fusion_vel.c
 *
 *  Created on: 06.05.2013
 *      Author: AndreR
 */

#include "ctrl_fusion_vel.h"

#ifndef SIMULATION
#include "util/console.h"
#include "network.h"
#endif

#include <string.h>

#define REACTION_THRESHOLD 0.05f
#define STALL_TIMEOUT 5000
#define DELTA_T 0.001f

static void init();
static void controller(CtrlInput* pInput, CtrlOutput* pOutput);

CtrlFusionVel ctrlFusionVel;
CtrlInstance ctrlFusionVelInstance = {.pInit = &init, .pExit = 0, .pControllerFunction = &controller, .deltaT = DELTA_T*2000.0f}; // deltaT unit: [0.5ms]

static float kfXData[KF_DATA_SIZE(2, 1, 3, 3)];
static float kfYData[KF_DATA_SIZE(2, 1, 3, 3)];
static float kfWData[KF_DATA_SIZE(2, 1, 3, 3)];

static const float A[4] = {
	1, DELTA_T,
	0, 1,
};

static const float B[2] = {0, 0};

static const float Ex[4] = {
	0.1f, 0,
	0, 1.0f,
};

static const float Ezxy[9] = {
	0.1f, 0, 0,
	0, 10.0f, 0,
	0, 0, 1.0f,
};

static const float Ezw[9] = {
	0.1f, 0, 0,
	0, 0.5f, 0,
	0, 0, 1.0f,
};

static void setFilterParams(uint8_t type, float x, float y, float z)
{
	CtrlSetFilterParams params;
	params.params[0] = x;
	params.params[1] = y;
	params.params[2] = z;
	params.type = type;
	CtrlFusionVelSetFilterParams(&params);
}

static void setPID(uint8_t type, float p, float i, float d)
{
	CtrlSetPIDParams pidParams;
	pidParams.type = type;
	pidParams.Kp = p;
	pidParams.Ki = i;
	pidParams.Kd = d;
	CtrlFusionVelSetPIDParams(&pidParams);
}

static void init()
{
	// zero all
	memset(&ctrlFusionVel, 0, sizeof(CtrlFusionVel));

	// init kalman filters
	KFInit(&ctrlFusionVel.kfX, 2, 1, 3, kfXData);
	KFInit(&ctrlFusionVel.kfY, 2, 1, 3, kfYData);
	KFInit(&ctrlFusionVel.kfW, 2, 1, 3, kfWData);

	memcpy(ctrlFusionVel.kfX.A.pData, A, sizeof(A));
	memcpy(ctrlFusionVel.kfY.A.pData, A, sizeof(A));
	memcpy(ctrlFusionVel.kfW.A.pData, A, sizeof(A));

	memcpy(ctrlFusionVel.kfX.B.pData, B, sizeof(B));
	memcpy(ctrlFusionVel.kfY.B.pData, B, sizeof(B));
	memcpy(ctrlFusionVel.kfW.B.pData, B, sizeof(B));

	memcpy(ctrlFusionVel.kfX.Ex.pData, Ex, sizeof(Ex));
	memcpy(ctrlFusionVel.kfY.Ex.pData, Ex, sizeof(Ex));
	memcpy(ctrlFusionVel.kfW.Ex.pData, Ex, sizeof(Ex));

	memcpy(ctrlFusionVel.kfX.Ez.pData, Ezxy, sizeof(Ezxy));
	memcpy(ctrlFusionVel.kfY.Ez.pData, Ezxy, sizeof(Ezxy));
	memcpy(ctrlFusionVel.kfW.Ez.pData, Ezw, sizeof(Ezw));

#ifdef SIMULATION
	ctrlFusionVel.lastWakeTime = 10;
#else
	ctrlFusionVel.lastWakeTime = chTimeNow();
#endif
	ctrlFusionVel.lastReaction = ctrlFusionVel.lastWakeTime;

	// init matrices
	ctrlFusionVel.matCurMotor.pData = ctrlFusionVel.curMotor;
	ctrlFusionVel.matCurMotor.numCols = 1;
	ctrlFusionVel.matCurMotor.numRows = 4;

	ctrlFusionVel.matCurForce.pData = ctrlFusionVel.curForce;
	ctrlFusionVel.matCurForce.numCols = 1;
	ctrlFusionVel.matCurForce.numRows = 4;

	ctrlFusionVel.matCurAcc.pData = ctrlFusionVel.curAcc;
	ctrlFusionVel.matCurAcc.numCols = 1;
	ctrlFusionVel.matCurAcc.numRows = 3;

	ctrlFusionVel.matCurXYW.pData = ctrlFusionVel.curXYW;
	ctrlFusionVel.matCurXYW.numCols = 1;
	ctrlFusionVel.matCurXYW.numRows = 3;

	ctrlFusionVel.matOutXYW.pData = ctrlFusionVel.outXYW;
	ctrlFusionVel.matOutXYW.numCols = 1;
	ctrlFusionVel.matOutXYW.numRows = 3;

	ctrlFusionVel.matOutMotor.pData = ctrlFusionVel.outMotor;
	ctrlFusionVel.matOutMotor.numCols = 1;
	ctrlFusionVel.matOutMotor.numRows = 4;

	// init PT1 and PT1 derivative
	CtrlFusionVelSetPT1TimeConstant(0.2f);

	// init control helpers
	ctrlFusionVel.velCounter = 0;
	memset(ctrlFusionVel.velOutput, 0, sizeof(float)*3);

	// init PIDs
	float p=1,i=0.01,d=0;
	setPID(CTRL_PID_PARAM_VEL_X, p, i, d);
	setPID(CTRL_PID_PARAM_VEL_Y, p, i, d);
	setPID(CTRL_PID_PARAM_VEL_W, 1, 0, 0);

	// init filter
	setFilterParams( CTRL_FILTER_PARAM_EX_VEL, 0.01f, 0.01f, 0.01f);
	setFilterParams( CTRL_FILTER_PARAM_EX_ACC, 0.1f, 0.1f, 0.1f);
	setFilterParams( CTRL_FILTER_PARAM_EZ_ENCODER, 5e-4, 5e-4, 5e-4);
	setFilterParams( CTRL_FILTER_PARAM_EZ_ACC_GYRO, 0.4, 0.4, 0.01);
	setFilterParams( CTRL_FILTER_PARAM_EZ_MOTOR, 0.01, 0.01, 1.3);
}

static void controller(CtrlInput* pInput, CtrlOutput* pOutput)
{
	uint8_t xyUpdateNeeded = 0;
	uint8_t wUpdateNeeded = 0;

	// clear C matrices
	arm_mat_zero_f32(&ctrlFusionVel.kfX.C);
	arm_mat_zero_f32(&ctrlFusionVel.kfY.C);
	arm_mat_zero_f32(&ctrlFusionVel.kfW.C);

	// construct C depending on updates
	if(pInput->enc.updated)
	{
		xyUpdateNeeded = 1;
		wUpdateNeeded = 1;

		MAT_ELEMENT(ctrlFusionVel.kfX.C, 0, 0) = 1.0f;
		MAT_ELEMENT(ctrlFusionVel.kfY.C, 0, 0) = 1.0f;
		MAT_ELEMENT(ctrlFusionVel.kfW.C, 0, 0) = 1.0f;

		// convert motor speed to local XYW system
		memcpy(ctrlFusionVel.curMotor, pInput->enc.motorVel, sizeof(float)*4);

		ctrlFusionVel.curMotor[0] *= ctrl.botParams.wheelRadius*CTRL_MOTOR_TO_WHEEL_RATIO;
		ctrlFusionVel.curMotor[1] *= ctrl.botParams.wheelRadius*CTRL_MOTOR_TO_WHEEL_RATIO;
		ctrlFusionVel.curMotor[2] *= ctrl.botParams.wheelRadius*CTRL_MOTOR_TO_WHEEL_RATIO;
		ctrlFusionVel.curMotor[3] *= ctrl.botParams.wheelRadius*CTRL_MOTOR_TO_WHEEL_RATIO; // value is now speed over ground [m/s]

		arm_mat_mult_f32(&ctrl.matMotor2XYW, &ctrlFusionVel.matCurMotor, &ctrlFusionVel.matCurXYW);
		// matCurXYW now contains X and Y in [m/s] and W in [rad/s] from encoder data

		MAT_ELEMENT(ctrlFusionVel.kfX.z, 0, 0) = ctrlFusionVel.curXYW[0];
		MAT_ELEMENT(ctrlFusionVel.kfY.z, 0, 0) = ctrlFusionVel.curXYW[1];
		MAT_ELEMENT(ctrlFusionVel.kfW.z, 0, 0) = ctrlFusionVel.curXYW[2];
	}

//	if(pInput->acc.updated)
//	{
//		xyUpdateNeeded = 1;
//
//		MAT_ELEMENT(ctrlFusionVel.kfX.C, 1, 1) = 1.0f;
//		MAT_ELEMENT(ctrlFusionVel.kfY.C, 1, 1) = 1.0f;
//
//		MAT_ELEMENT(ctrlFusionVel.kfX.z, 1, 0) = pInput->acc.linAcc[0];
//		MAT_ELEMENT(ctrlFusionVel.kfY.z, 1, 0) = pInput->acc.linAcc[1];
//	}

//	if(pInput->gyr.updated)
//	{
//		wUpdateNeeded = 1;
//
//		MAT_ELEMENT(ctrlFusionVel.kfW.C, 1, 0) = 1.0f;
//
//		MAT_ELEMENT(ctrlFusionVel.kfW.z, 1, 0) = pInput->gyr.rotVel[2];	// update gyro rotation around Z axis
//	}

	if(pInput->motor.updated)
	{
		xyUpdateNeeded = 1;
//		wUpdateNeeded = 1;

		MAT_ELEMENT(ctrlFusionVel.kfX.C, 2, 1) = 1.0f;
		MAT_ELEMENT(ctrlFusionVel.kfY.C, 2, 1) = 1.0f;
//		MAT_ELEMENT(ctrlFusionVel.kfW.C, 2, 1) = 1.0f;

		for(uint8_t i = 0; i < 4; i++)
		{
			// force = current * motor_constant * gear_ratio * 1/wheel_radius
			// F = I * k_m * G * 1/r_rad
			// F = I * 0.0255 * 50/15 * 1/r_rad
			// F = I * 0.085 * 1/r_rad
			ctrlFusionVel.curForce[i] = pInput->motor.current[i] * 0.085f * 1.0f/ctrl.botParams.wheelRadius;
		}

		// convert wheel forces over ground to global force and torque
		arm_mat_mult_f32(&ctrl.matForceMotor2XYW, &ctrlFusionVel.matCurForce, &ctrlFusionVel.matCurAcc);

		// convert forces and torque to (angular) acceleration
		ctrlFusionVel.curAcc[0] /= ctrl.botParams.mass;
		ctrlFusionVel.curAcc[1] /= ctrl.botParams.mass;
		ctrlFusionVel.curAcc[2] /= ctrl.momentOfInertia;

		MAT_ELEMENT(ctrlFusionVel.kfX.z, 2, 0) = ctrlFusionVel.curAcc[0];
		MAT_ELEMENT(ctrlFusionVel.kfY.z, 2, 0) = ctrlFusionVel.curAcc[1];
		MAT_ELEMENT(ctrlFusionVel.kfW.z, 2, 0) = ctrlFusionVel.curAcc[2];	// update angular acceleration around Z axis
	}

	// predict W
	KFPredict(&ctrlFusionVel.kfW);
	if(wUpdateNeeded)
	{
		KFUpdate(&ctrlFusionVel.kfW);

		arm_mat_clip_f32(&ctrlFusionVel.kfW.Sigma, -100.0f, 100.0f);
	}

	// predict X,Y
	KFPredict(&ctrlFusionVel.kfX);
	KFPredict(&ctrlFusionVel.kfY);

	if(xyUpdateNeeded)
	{
		KFUpdate(&ctrlFusionVel.kfX);
		KFUpdate(&ctrlFusionVel.kfY);

		arm_mat_clip_f32(&ctrlFusionVel.kfX.Sigma, -100.0f, 100.0f);
		arm_mat_clip_f32(&ctrlFusionVel.kfY.Sigma, -100.0f, 100.0f);
	}

	// update feedback state
	ctrl.state.pos[0] = 0;
	ctrl.state.pos[1] = 0;
	ctrl.state.pos[2] = 0;

	ctrl.state.vel[0] = MAT_ELEMENT(ctrlFusionVel.kfX.x, 0, 0);
	ctrl.state.vel[1] = MAT_ELEMENT(ctrlFusionVel.kfY.x, 0, 0);
	ctrl.state.vel[2] = MAT_ELEMENT(ctrlFusionVel.kfW.x, 0, 0);

	ctrl.state.acc[0] = MAT_ELEMENT(ctrlFusionVel.kfX.x, 1, 0);
	ctrl.state.acc[1] = MAT_ELEMENT(ctrlFusionVel.kfY.x, 1, 0);
	ctrl.state.acc[2] = MAT_ELEMENT(ctrlFusionVel.kfW.x, 1, 0);

	ctrl.state.updated[0] = 0;
	ctrl.state.updated[1] = 1;
	ctrl.state.updated[2] = 1;

	// do control...

	ctrlFusionVel.velCounter++;
	ctrlFusionVel.velCounter %= 4;

	for(uint8_t i = 0; i < 3; i++)
	{
		// PT1 on setpoint
		float setVel;
//		float setAcc;
		arm_biquad_cascade_df1_f32(&ctrlFusionVel.pt1[i], &ctrlFusionVel.setpoint[i], &setVel, 1);
//		arm_biquad_cascade_df1_f32(&ctrlFusionVel.pt1Dt[i], &ctrlFusionVel.setpoint[i], &setAcc, 1);

		// velocity control at 250Hz
//		if(ctrlFusionVel.velCounter == 0)
//		{
			float errorVel = setVel - ctrl.state.vel[i];
			ctrlFusionVel.velOutput[i] = setVel + arm_pid_f32(&ctrlFusionVel.pidVel[i], errorVel);
//		}

		// acceleration control at 1000Hz
//		float errorAcc = ctrlFusionVel.velOutput[i] + setAcc - ctrl.state.acc[i];
//		ctrlFusionVel.outXYW[i] = arm_pid_f32(&ctrlFusionVel.pidAcc[i], errorAcc);
		ctrlFusionVel.outXYW[i] = ctrlFusionVel.velOutput[i]; //TODO: testing + setAcc;

		ctrl.state.setVel[i] = setVel;
		ctrl.state.velOutput[i] = ctrlFusionVel.velOutput[i];
		ctrl.state.errorVel[i] = errorVel;
	}

	// convert (angular) acceleration to forces and torques
	ctrlFusionVel.outXYW[0] *= ctrl.botParams.mass;
	ctrlFusionVel.outXYW[1] *= ctrl.botParams.mass;
	ctrlFusionVel.outXYW[2] *= ctrl.momentOfInertia;

	// grab current battery voltage (or guess)
	float UBattery = 15.0f;
	if(pInput->power.updated)
	{
		UBattery = pInput->power.voltage;
	}

	// convert to wheel forces over ground
	arm_mat_mult_f32(&ctrl.matXYW2Motor, &ctrlFusionVel.matOutXYW, &ctrlFusionVel.matOutMotor);

	for(uint8_t i = 0; i < 4; i++)
	{
		// convert speed over ground to motor speed
		// 50/15 * speed_over_ground * 1/r_rad
		float motorSpeed = CTRL_WHEEL_TO_MOTOR_RATIO * 1/ctrl.botParams.wheelRadius * ctrlFusionVel.outMotor[i];
		// and convert to voltage
		float outVoltage = motorSpeed/CTRL_MOTOR_KN;

		if(fabsf(pInput->motor.current[i]) > 7.0f)
		{
#ifdef SIMULATION
			printf("Motor %hhu overcurrent %f", i, pInput->motor.current[i]);
#else
			NetworkPrint("Motor %hhu overcurrent %f", i, pInput->motor.current[i]);
#endif
			pOutput->motorPWM[i] = 0;
		}
		else
		{
			pOutput->motorPWM[i] = outVoltage/UBattery;
		}
	}

	switch(ctrlFusionVel.logState)
	{
		case CTRL_FUSION_VEL_LOG_STEP:
		{
			pOutput->motorPWM[0] = ctrlFusionVel.testOutput;
			pOutput->motorPWM[1] = ctrlFusionVel.testOutput;
			pOutput->motorPWM[2] = ctrlFusionVel.testOutput;
			pOutput->motorPWM[3] = ctrlFusionVel.testOutput;

			ctrlFusionVel.samples[ctrlFusionVel.numSamples++] = ctrl.state.vel[2];

			if(ctrlFusionVel.numSamples >= CTRL_FUSION_VEL_NUM_SAMPLES)
				memset(pOutput->motorPWM, 0, sizeof(float)*4);
		}
		break;
		case CTRL_FUSION_VEL_LOG_ENC:
		{
			float* pCur = ctrlFusionVel.samples + ctrlFusionVel.numSamples*3;

			if(pInput->enc.updated)
			{
				pCur[0] = ctrlFusionVel.curXYW[0];
				pCur[1] = ctrlFusionVel.curXYW[1];
				pCur[2] = ctrlFusionVel.curXYW[2];
				ctrlFusionVel.numSamples++;
			}
		}
		break;
		case CTRL_FUSION_VEL_LOG_GYR:
		{
			float* pCur = ctrlFusionVel.samples + ctrlFusionVel.numSamples*3;

			if(pInput->gyr.updated)
			{
				pCur[0] = pInput->gyr.rotVel[0];
				pCur[1] = pInput->gyr.rotVel[1];
				pCur[2] = pInput->gyr.rotVel[2];
				ctrlFusionVel.numSamples++;
			}
		}
		break;
		case CTRL_FUSION_VEL_LOG_ACC:
		{
			float* pCur = ctrlFusionVel.samples + ctrlFusionVel.numSamples*3;

			if(pInput->acc.updated)
			{
				pCur[0] = pInput->acc.linAcc[0];
				pCur[1] = pInput->acc.linAcc[1];
				pCur[2] = pInput->acc.linAcc[2];
				ctrlFusionVel.numSamples++;
			}
		}
		break;
		case CTRL_FUSION_VEL_LOG_MOTOR:
		{
			float* pCur = ctrlFusionVel.samples + ctrlFusionVel.numSamples*3;

			if(pInput->motor.updated)
			{
				pCur[0] = ctrlFusionVel.curAcc[0];
				pCur[1] = ctrlFusionVel.curAcc[1];
				pCur[2] = ctrlFusionVel.curAcc[2];
				ctrlFusionVel.numSamples++;
			}
		}
		break;
	}

	if(ctrlFusionVel.numSamples >= CTRL_FUSION_VEL_NUM_SAMPLES)
	{
		ctrlFusionVel.lastLogState = ctrlFusionVel.logState;
		ctrlFusionVel.logState = CTRL_FUSION_VEL_LOG_IDLE;
		ctrlFusionVel.numSamples = 0;
	}

	if(ctrlFusionVel.velCounter == 0)
	{
//		ConsolePrint("1: %.3f, 2: %.3f, 3: %.3f, 4: %.3f\r\n", pInput->motor.current[0], pInput->motor.current[1], pInput->motor.current[2], pInput->motor.current[3]);
//		ConsolePrint("1: %.3f, 2: %.3f, 3: %.3f, 4: %.3f\r\n", pOutput->motorPWM[0], pOutput->motorPWM[1], pOutput->motorPWM[2], pOutput->motorPWM[3]);
//		ConsolePrint("X: %.3f, Y: %.3f, W: %.3f\r\n", ctrl.state.acc[0], ctrl.state.acc[1], ctrl.state.acc[2]);
//		ConsolePrint("X: %.3f, Y: %.3f, W: %.3f\r\n", ctrlFusionVel.velOutput[0], ctrlFusionVel.velOutput[1], ctrlFusionVel.velOutput[2]);
//		ConsolePrint("X: %.3f, Y: %.3f, W: %.3f\r\n", ctrlFusionVel.curAcc[0], ctrlFusionVel.curAcc[1], ctrlFusionVel.curAcc[2]);
//		ConsolePrint("X: %.3f, Y: %.3f, W: %.3f\r\n", ctrlFusionVel.globalOutXYW[0], ctrlFusionVel.globalOutXYW[1], ctrlFusionVel.globalOutXYW[2]);
//		ConsolePrint("X: %.3f, Y: %.3f, W: %.3f\r\n", ctrlFusionVel.outXYW[0], ctrlFusionVel.outXYW[1], ctrlFusionVel.outXYW[2]);
	}

	// stall protection
	for(int j = 0; j < 3; j++)
	{
		if(ctrl.state.vel[j] < -REACTION_THRESHOLD || ctrl.state.vel[j] > REACTION_THRESHOLD)
		{
			ctrlFusionVel.lastReaction = ctrlFusionVel.lastWakeTime;
//			ConsolePrint("%f %d \n\r", ctrl.state.vel[j], j);
		}
	}

	if(ctrlFusionVel.lastWakeTime - ctrlFusionVel.lastReaction > MS2ST(STALL_TIMEOUT))
	{
		for(uint8_t i = 0; i < 3; i++)
		{
			ctrlFusionVel.setpoint[i] = 0.0f;
			arm_pid_reset_f32(&ctrlFusionVel.pidVel[i]);
			pOutput->motorPWM[i] = 0.0f;
		}

		ctrlFusionVel.lastReaction = ctrlFusionVel.lastWakeTime;
	}

	// zero updated input states
	pInput->enc.updated = 0;
	pInput->acc.updated = 0;
	pInput->gyr.updated = 0;
	pInput->motor.updated = 0;

//	memset(pOutput->motorPWM, 0, sizeof(float)*4);
#ifdef SIMULATION
	ctrlFusionVel.lastWakeTime = 10;
#else
	ctrlFusionVel.lastWakeTime = chTimeNow();
#endif
}

void CtrlFusionVelTestStep(float step)
{
	ctrlFusionVel.testOutput = step;
	ctrlFusionVel.logState = CTRL_FUSION_VEL_LOG_STEP;
}

void CtrlFusionVelEnableLog(uint8_t logType)
{
	ctrlFusionVel.logState = logType;
}

void CtrlFusionVelGatherData()
{
#ifndef SIMULATION
	switch(ctrlFusionVel.lastLogState)
	{
		case CTRL_FUSION_VEL_LOG_STEP:
		{
			for(uint32_t i = 0; i < CTRL_FUSION_VEL_NUM_SAMPLES; i++)
			{
#ifndef SIMULATION
				NetworkPrint("%f", ctrlFusionVel.samples[i]);
				ConsolePrint("%f\r\n", ctrlFusionVel.samples[i]);
				chThdSleep(MS2ST(10));
#endif
			}
		}
		break;
		case CTRL_FUSION_VEL_LOG_ENC:
		case CTRL_FUSION_VEL_LOG_ACC:
		case CTRL_FUSION_VEL_LOG_GYR:
		case CTRL_FUSION_VEL_LOG_MOTOR:
		{
			for(uint32_t i = 0; i < CTRL_FUSION_VEL_NUM_SAMPLES; i++)
			{
#ifndef SIMULATION
				float* pCur = ctrlFusionVel.samples + i*3;
				NetworkPrint("%.5f, %.5f, %.5f", pCur[0], pCur[1], pCur[2]);
				ConsolePrint("%.5f, %.5f, %.5f\r\n", pCur[0], pCur[1], pCur[2]);
				chThdSleep(MS2ST(10));
#endif
			}
		}
		break;
	}
#endif
}

void CtrlFusionVelSetVelTarget(float x, float y, float w)
{
	ctrlFusionVel.setpoint[0] = x;
	ctrlFusionVel.setpoint[1] = y;
	ctrlFusionVel.setpoint[2] = w;
}

void CtrlFusionVelSetStateUncertainty(float* pEx)
{
	for(uint16_t i = 0; i < ctrlFusionVel.kfX.f; i++)
	{
		MAT_ELEMENT(ctrlFusionVel.kfX.Ex, i, i) = pEx[i];
		MAT_ELEMENT(ctrlFusionVel.kfY.Ex, i, i) = pEx[i];
		MAT_ELEMENT(ctrlFusionVel.kfW.Ex, i, i) = pEx[i];
	}
}

void CtrlFusionVelSetSensorUncertainty(float* pEz)
{
	for(uint16_t i = 0; i < ctrlFusionVel.kfX.h; i++)
	{
		MAT_ELEMENT(ctrlFusionVel.kfX.Ez, i, i) = pEz[i];
		MAT_ELEMENT(ctrlFusionVel.kfY.Ez, i, i) = pEz[i];
		MAT_ELEMENT(ctrlFusionVel.kfW.Ez, i, i) = pEz[i];
	}
}

void CtrlFusionVelSetPIDParams(CtrlSetPIDParams* pParams)
{
#ifndef SIMULATION
	ConsolePrint("Rcv PID params. Type: %hhu\r\n", pParams->type);
#endif

	switch(pParams->type)
	{
		case CTRL_PID_PARAM_POS_X:
		{
		}
		break;
		case CTRL_PID_PARAM_POS_Y:
		{
		}
		break;
		case CTRL_PID_PARAM_POS_W:
		{
		}
		break;
		case CTRL_PID_PARAM_VEL_X:
		{
			ctrlFusionVel.pidVel[0].Kp = pParams->Kp;
			ctrlFusionVel.pidVel[0].Ki = pParams->Ki;
			ctrlFusionVel.pidVel[0].Kd = pParams->Kd;
			arm_pid_init_f32(&ctrlFusionVel.pidVel[0], 1);
		}
		break;
		case CTRL_PID_PARAM_VEL_Y:
		{
			ctrlFusionVel.pidVel[1].Kp = pParams->Kp;
			ctrlFusionVel.pidVel[1].Ki = pParams->Ki;
			ctrlFusionVel.pidVel[1].Kd = pParams->Kd;
			arm_pid_init_f32(&ctrlFusionVel.pidVel[1], 1);
		}
		break;
		case CTRL_PID_PARAM_VEL_W:
		{
			ctrlFusionVel.pidVel[2].Kp = pParams->Kp;
			ctrlFusionVel.pidVel[2].Ki = pParams->Ki;
			ctrlFusionVel.pidVel[2].Kd = pParams->Kd;
			arm_pid_init_f32(&ctrlFusionVel.pidVel[2], 1);
		}
		break;

		// TODO: add acceleration stuff
	}
}

void CtrlFusionVelSetFilterParams(CtrlSetFilterParams* pParams)
{
#ifndef SIMULATION
	ConsolePrint("Rcv filter params. Type: %hhu\r\n", pParams->type);
#endif

	switch(pParams->type)
	{
		case CTRL_FILTER_PARAM_EX_VEL:
		{
			MAT_ELEMENT(ctrlFusionVel.kfX.Ex, 0, 0) = pParams->params[0];
			MAT_ELEMENT(ctrlFusionVel.kfY.Ex, 0, 0) = pParams->params[1];
			MAT_ELEMENT(ctrlFusionVel.kfW.Ex, 0, 0) = pParams->params[2];
		}
		break;
		case CTRL_FILTER_PARAM_EX_ACC:
		{
			MAT_ELEMENT(ctrlFusionVel.kfX.Ex, 1, 1) = pParams->params[0];
			MAT_ELEMENT(ctrlFusionVel.kfY.Ex, 1, 1) = pParams->params[1];
			MAT_ELEMENT(ctrlFusionVel.kfW.Ex, 1, 1) = pParams->params[2];
		}
		break;
		case CTRL_FILTER_PARAM_EZ_ENCODER:
		{
			MAT_ELEMENT(ctrlFusionVel.kfX.Ez, 0, 0) = pParams->params[0];
			MAT_ELEMENT(ctrlFusionVel.kfY.Ez, 0, 0) = pParams->params[1];
			MAT_ELEMENT(ctrlFusionVel.kfW.Ez, 0, 0) = pParams->params[2];
		}
		break;
		case CTRL_FILTER_PARAM_EZ_ACC_GYRO:
		{
			MAT_ELEMENT(ctrlFusionVel.kfX.Ez, 1, 1) = pParams->params[0];
			MAT_ELEMENT(ctrlFusionVel.kfY.Ez, 1, 1) = pParams->params[1];
			MAT_ELEMENT(ctrlFusionVel.kfW.Ez, 1, 1) = pParams->params[2];
		}
		break;
		case CTRL_FILTER_PARAM_EZ_MOTOR:
		{
			MAT_ELEMENT(ctrlFusionVel.kfX.Ez, 2, 2) = pParams->params[0];
			MAT_ELEMENT(ctrlFusionVel.kfY.Ez, 2, 2) = pParams->params[1];
			MAT_ELEMENT(ctrlFusionVel.kfW.Ez, 2, 2) = pParams->params[2];
		}
		break;
	}
}

void CtrlFusionVelSetPT1TimeConstant(float T)
{
	// init PT1 and PT1 derivative
	const float K = 1.0f;
	float a = (2*T-DELTA_T)/(2*T+DELTA_T);
	float b = K*DELTA_T/(2*T+DELTA_T);
	float c = 2*K/(2*T+DELTA_T);

	ctrlFusionVel.pt1Coeff[0] = b;
	ctrlFusionVel.pt1Coeff[1] = b;
	ctrlFusionVel.pt1Coeff[2] = 0.0f;
	ctrlFusionVel.pt1Coeff[3] = a;
	ctrlFusionVel.pt1Coeff[4] = 0.0f;

	ctrlFusionVel.pt1DtCoeff[0] = c;
	ctrlFusionVel.pt1DtCoeff[1] = -c;
	ctrlFusionVel.pt1DtCoeff[2] = 0.0f;
	ctrlFusionVel.pt1DtCoeff[3] = a;
	ctrlFusionVel.pt1DtCoeff[4] = 0.0f;

	arm_biquad_cascade_df1_init_f32(&ctrlFusionVel.pt1[0], 1, ctrlFusionVel.pt1Coeff, ctrlFusionVel.pt1State[0]);
	arm_biquad_cascade_df1_init_f32(&ctrlFusionVel.pt1[1], 1, ctrlFusionVel.pt1Coeff, ctrlFusionVel.pt1State[1]);
	arm_biquad_cascade_df1_init_f32(&ctrlFusionVel.pt1[2], 1, ctrlFusionVel.pt1Coeff, ctrlFusionVel.pt1State[2]);

	arm_biquad_cascade_df1_init_f32(&ctrlFusionVel.pt1Dt[0], 1, ctrlFusionVel.pt1DtCoeff, ctrlFusionVel.pt1DtState[0]);
	arm_biquad_cascade_df1_init_f32(&ctrlFusionVel.pt1Dt[1], 1, ctrlFusionVel.pt1DtCoeff, ctrlFusionVel.pt1DtState[1]);
	arm_biquad_cascade_df1_init_f32(&ctrlFusionVel.pt1Dt[2], 1, ctrlFusionVel.pt1DtCoeff, ctrlFusionVel.pt1DtState[2]);
}

void CtrlFusionVelNetPrintInput()
{
#ifndef SIMULATION
	NetworkPrint("Enc: %.3f, %.3f, %.3f, %.3f", ctrl.input.enc.motorVel[0], ctrl.input.enc.motorVel[1], ctrl.input.enc.motorVel[2], ctrl.input.enc.motorVel[3]);
	chThdSleep(MS2ST(10));
	NetworkPrint("Acc/Gyr: %.3f, %.3f, %.3f", ctrl.input.acc.linAcc[0], ctrl.input.acc.linAcc[1], ctrl.input.gyr.rotVel[2]);
	chThdSleep(MS2ST(10));
	NetworkPrint("Motor: %.3f, %.3f, %.3f, %.3f", ctrl.input.motor.current[0], ctrl.input.motor.current[1], ctrl.input.motor.current[2], ctrl.input.motor.current[3]);
#endif
}
