#include "misc.h"
#include "stm32f4xx_rcc.h"
#include "stm32f4xx_exti.h"
#include "stm32f4xx_gpio.h"
#include "stm32f4xx_iwdg.h"
#include "stm32f4xx_syscfg.h"

#include "core_cm4.h"
#include "ch.h"

#include "hal/mco.h"
#include "hal/led.h"
#include "hal/buzzer.h"
#include "hal/exti.h"
#include "hal/usart1.h"
#include "hal/usart2.h"
#include "hal/usart6.h"
#include "hal/hw_sync1.h"

#include "util/boot.h"
#include "util/console.h"
#include "util/crc.h"
#include "util/event.h"
#include "util/flash_fs.h"

#include "cli.h"
#include "network.h"
#include "spi2.h"
#include "spi3.h"
#include "intercom1.h"
#include "power.h"
#include "bootloader.h"
#include "motors.h"
#include "ctrl.h"
#include "ctrl_vel.h"
#include "ui.h"

static void NVIC_Configuration();

extern void SystemClock_Config(void);

#define portNVIC_SYSTICK_CTRL		( ( volatile unsigned long * ) 0xe000e010 )
#define portNVIC_SYSTICK_LOAD		( ( volatile unsigned long * ) 0xe000e014 )
#define portNVIC_SYSTICK_CLK		0x00000004
#define portNVIC_SYSTICK_INT		0x00000002
#define portNVIC_SYSTICK_ENABLE		0x00000001

static WORKING_AREA(waUSART2, 512);
static WORKING_AREA(waSPI3, 512);
static WORKING_AREA(waSPI2, 512);
static WORKING_AREA(waUSART6, 512);
static WORKING_AREA(waMotors, 4096);
static WORKING_AREA(waLow, 4096);
static WORKING_AREA(waCLI, 8192);
static WORKING_AREA(waNetwork, 4096);
static WORKING_AREA(waPower, 512);
static WORKING_AREA(waUI, 512);
static WORKING_AREA(waCTRL, 8192);
static WORKING_AREA(waBootloader, 512);

#define EVENTS_LOW_MSGS 10
static msg_t lowBuf[EVENTS_LOW_MSGS];
static EventManager eventsLow;

int main()
{
#ifdef ENV_BOOT
	if(!BootIsBootloaderSelected())
		BootJumpToApplication();

	BootSetBootloaderSelected(0);
#endif

	SystemClock_Config();

	// Enable important clocks
	RCC_AHB1PeriphClockCmd( RCC_AHB1Periph_DMA1 | RCC_AHB1Periph_DMA2, ENABLE);
	RCC_AHB1PeriphClockCmd(	RCC_AHB1Periph_GPIOA | RCC_AHB1Periph_GPIOB | RCC_AHB1Periph_GPIOC |
							RCC_AHB1Periph_GPIOD, ENABLE);
	RCC_APB2PeriphClockCmd(RCC_APB2Periph_SYSCFG, ENABLE);

	SYSCFG_CompensationCellCmd(ENABLE);

	SCB->SHCSR |= SCB_SHCSR_USGFAULTENA_Msk | SCB_SHCSR_BUSFAULTENA_Msk | SCB_SHCSR_MEMFAULTENA_Msk;

	*(portNVIC_SYSTICK_LOAD) = ( 168000000UL / CH_FREQUENCY ) - 1UL;
	*(portNVIC_SYSTICK_CTRL) = portNVIC_SYSTICK_CLK | portNVIC_SYSTICK_INT | portNVIC_SYSTICK_ENABLE;

	chSysInit();

	NVIC_Configuration();

	CPULoadCalibrate(10);
	chThdSleep(2000);	// TODO:quickfix

	EventManagerInit(&eventsLow, lowBuf, EVENTS_LOW_MSGS);

	// --- HAL ---
	LEDInit();		// init LEDs to signal RTOS startup failures
	HWSync1Init();	// Init sync mechanism for timed distributed startup
	PowerInit();
	PowerEnable5VSupply(1);
	FlashFSInit(0x08060000, 0x20000);	// 128k
	BuzzerInit();
	BuzzerEnable(1);
	chThdSleep(MS2ST(200));
	BuzzerEnable(0);
	ExtiInit();
	USART1Init();
	USART2Init();
	USART6Init();
	SPI2Init();
	SPI3Init();
	CRCInit();

	// --- High-Level Systems ---
	Intercom1Init(&eventsLow);
	CLIInit();
	NetworkInit();
	BootloaderInit();

	PowerAddIntercomClient();

#ifdef ENV_RUN
	MotorsInit();
	UIInit();
#endif

	CtrlInit();
#ifdef ENV_RUN
	CtrlVelInit();
#endif

	// now start motor left/right and kicker processor
	MCOInit();

#ifdef ENV_RUN
	chThdCreateStatic(waCTRL, sizeof(waCTRL), NORMALPRIO+5, CtrlTask, 0);
	chThdCreateStatic(waMotors, sizeof(waMotors), NORMALPRIO-5, MotorsTask, 0);
#endif
	chThdCreateStatic(waNetwork, sizeof(waNetwork), NORMALPRIO+1, NetworkTask, 0);
	chThdCreateStatic(waUSART2, sizeof(waUSART2), NORMALPRIO, USARTTask, &usart2);
	chThdCreateStatic(waUSART6, sizeof(waUSART6), NORMALPRIO, USARTTask, &usart6);
	chThdCreateStatic(waSPI3, sizeof(waSPI3), NORMALPRIO, SPI3Task, 0);
	chThdCreateStatic(waLow, sizeof(waLow), NORMALPRIO-1, EventManagerTask, &eventsLow);
	chThdCreateStatic(waCLI, sizeof(waCLI), NORMALPRIO-2, CLITask, 0);
	chThdCreateStatic(waSPI2, sizeof(waSPI2), NORMALPRIO-3, SPI2Task, 0);
	chThdCreateStatic(waPower, sizeof(waPower), NORMALPRIO-4, PowerTask, 0);
	chThdCreateStatic(waBootloader, sizeof(waBootloader), NORMALPRIO-6, BootloaderTask, 0);

#ifdef ENV_BOOT
	ConsolePrint("\f--- Bootloader (MAIN) ---\r\n");
#else
	ConsolePrint("\f--- Main ---\r\n");
#endif

	HWSyncBarrier(&hwSync1);

	chThdCreateStatic(waUI, sizeof(waUI), NORMALPRIO-10, UITask, 0);

/*
	// Configure and start Independent watchdog
	IWDG_WriteAccessCmd(IWDG_WriteAccess_Enable);
	IWDG_SetPrescaler(IWDG_Prescaler_8);
	IWDG_SetReload(2500);	// approx. 500ms
	IWDG_WriteAccessCmd(IWDG_WriteAccess_Disable);
*/

	chThdSleep(TIME_INFINITE);

	return 0;
}

static void NVIC_Configuration()
{
#ifdef ENV_BOOT
	NVIC_SetVectorTable(NVIC_VectTab_FLASH, 0x00);
#else
	NVIC_SetVectorTable(NVIC_VectTab_FLASH, BOOT_APP_OFFSET);
#endif
}
