/*
 * motors.c
 *
 *  Created on: 02.02.2014
 *      Author: AndreR
 */

#include "stm32f4xx_gpio.h"
#include "stm32f4xx_syscfg.h"
#include "stm32f4xx_exti.h"
#include "misc.h"

#include "motors.h"
#include "spi3.h"
#include "hal/exti.h"
#include "constants.h"
#include "util/console.h"
#include "intercom1.h"
#include "intercom_constants.h"
#include "power.h"

#include "ch.h"

Motors motors;

#define EVENT_MASK_MOTOR_FAULT 0x0F

static void sendFaultEvent(uint8_t event)
{
	msg_t msg = event;

	chSysLockFromIsr();
	chMBPostI(&motors.eventQueue, msg);
	chSysUnlockFromIsr();
}

// Motor 2
void EXTI1_IRQHandler()
{
	CH_IRQ_PROLOGUE();

	EXTI->PR = EXTI_PR_PR1;

	sendFaultEvent(1);

	CH_IRQ_EPILOGUE();
}

// Motor 3
void EXTI2_IRQHandler()
{
	CH_IRQ_PROLOGUE();

	EXTI->PR = EXTI_PR_PR2;

	sendFaultEvent(2);

	CH_IRQ_EPILOGUE();
}

static void faultM1()
{
	sendFaultEvent(0);
}

static void faultM4()
{
	sendFaultEvent(3);
}

static void faultM5()
{
	sendFaultEvent(4);
}

static void spi3AboutToSend(SPI3Slave* pSlave, uint8_t slaveID)
{
	(void)pSlave;

	IntercomHeader iHeader;
	iHeader.command = IC_MOTORS_MAIN2MOT;
	iHeader.section = INTERCOM_SECTION_MOTORS;

	MotorsMain2Mot main2Mot;
	main2Mot.bat = power.supplyLog.bat;

	if(slaveID == SPI3_SLAVE_RIGHT)
	{
		iHeader.address = INTERCOM_ADDRESS_RIGHT;
		main2Mot.mot[MOT_FRONT].velocity = motors.mot[MOT_FR].set.velocity;
		main2Mot.mot[MOT_REAR].velocity = motors.mot[MOT_RR].set.velocity;
	}
	else
	{
		iHeader.address = INTERCOM_ADDRESS_LEFT;
		main2Mot.mot[MOT_FRONT].velocity = motors.mot[MOT_FL].set.velocity;
		main2Mot.mot[MOT_REAR].velocity = motors.mot[MOT_RL].set.velocity;
	}

	IntercomSend(&intercom1, &iHeader, sizeof(main2Mot), (uint8_t*)&main2Mot);
}

static void motorsCallback(IntercomHeader* pHeader, uint16_t dataLength, uint8_t* pData, void* pUserData)
{
	(void)pUserData;

	if(pHeader->command == IC_MOTORS_MOT2MAIN && dataLength >= sizeof(MotorsMot2Main))
	{
		MotorsMot2Main* pMot2Main = (MotorsMot2Main*)pData;

		if(pMot2Main->source == MOT_SOURCE_RIGHT)
		{
			motors.mot[MOT_FR].cur.velocity = pMot2Main->mot[MOT_FRONT].velocity;
			motors.mot[MOT_RR].cur.velocity = pMot2Main->mot[MOT_REAR].velocity;
		}
		else
		{
			motors.mot[MOT_FL].cur.velocity = pMot2Main->mot[MOT_FRONT].velocity;
			motors.mot[MOT_RL].cur.velocity = pMot2Main->mot[MOT_REAR].velocity;
		}

		if(motors.updateCb)
			(*motors.updateCb)(pMot2Main->source);
	}
}

void MotorsInit()
{
	GPIO_InitTypeDef gpioInit;
	EXTI_InitTypeDef extiInit;
	NVIC_InitTypeDef nvicInit;

	chMBInit(&motors.eventQueue, motors.eventQueueBuf, MOTORS_EVENT_QUEUE_SIZE);

	gpioInit.GPIO_Mode = GPIO_Mode_IN;
	gpioInit.GPIO_OType = GPIO_OType_PP;
	gpioInit.GPIO_PuPd = GPIO_PuPd_UP;
	gpioInit.GPIO_Speed = GPIO_Speed_2MHz;
	gpioInit.GPIO_Pin = GPIO_Pin_10 | GPIO_Pin_12;
	GPIO_Init(GPIOA, &gpioInit);

	gpioInit.GPIO_Pin = GPIO_Pin_1 | GPIO_Pin_2 | GPIO_Pin_8;
	GPIO_Init(GPIOC, &gpioInit);

	SYSCFG_EXTILineConfig(EXTI_PortSourceGPIOA, EXTI_PinSource10);
	SYSCFG_EXTILineConfig(EXTI_PortSourceGPIOA, EXTI_PinSource12);
	SYSCFG_EXTILineConfig(EXTI_PortSourceGPIOC, EXTI_PinSource1);
	SYSCFG_EXTILineConfig(EXTI_PortSourceGPIOC, EXTI_PinSource2);
	SYSCFG_EXTILineConfig(EXTI_PortSourceGPIOC, EXTI_PinSource8);

	extiInit.EXTI_Line = EXTI_Line10 | EXTI_Line12 | EXTI_Line1 | EXTI_Line2 | EXTI_Line8;
	extiInit.EXTI_LineCmd = ENABLE;
	extiInit.EXTI_Mode = EXTI_Mode_Interrupt;
	extiInit.EXTI_Trigger = EXTI_Trigger_Falling;
	EXTI_Init(&extiInit);

	nvicInit.NVIC_IRQChannelCmd = ENABLE;
	nvicInit.NVIC_IRQChannelPreemptionPriority = IRQL_MOTOR_FAULT;
	nvicInit.NVIC_IRQChannelSubPriority = 0;
	nvicInit.NVIC_IRQChannel = EXTI1_IRQn;
	NVIC_Init(&nvicInit);

	nvicInit.NVIC_IRQChannel = EXTI2_IRQn;
	NVIC_Init(&nvicInit);

	ExtiSetCallback(&faultM1, 12);
	ExtiSetCallback(&faultM4, 8);
	ExtiSetCallback(&faultM5, 10);

	IntercomClient client;
	client.pUserData = 0;
	client.rxCallback = &motorsCallback;
	client.section = INTERCOM_SECTION_MOTORS;

	IntercomAddClient(&intercom1, client);

	spi3.aboutToSendCb = &spi3AboutToSend;
}

msg_t MotorsTask(void* params)
{
	(void)params;

	chRegSetThreadName("Motors");

	msg_t event;

	while(1)
	{
		chMBFetch(&motors.eventQueue, &event, TIME_INFINITE);

		if(event & EVENT_MASK_MOTOR_FAULT)
			ConsolePrint("Motor %u fault!\r\n", event);
	}

	return 0;
}
