/*
 * dect.c
 *
 *  Created on: 01.07.2010
 *      Author: AndreR
 */

#include "network.h"

#include "stm32f4xx.h"
#include "misc.h"
#include "stm32f4xx_rcc.h"
#include "stm32f4xx_gpio.h"
#include "stm32f4xx_usart.h"
#include "stm32f4xx_exti.h"
#include "stm32f4xx_syscfg.h"
#include "stm32f4xx_tim.h"
#include "constants.h"

#include "util/bit_banding.h"
#include "util/console.h"
#include "errors.h"
#include "commands.h"
#include "util/float16_t.h"
#include "intercom_constants.h"
#include "intercom1.h"
#include "bootloader.h"
#include "util/boot.h"
#include "ctrl_vel.h"
#include "util/sys_time.h"
#include "ctrl/ctrl_fusion.h"
#include "power.h"
#include "util/bootloader.h"
#include "util/flash_fs.h"
#include "intercom1.h"

#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <stdio.h>

Network network;

static uint8_t processPacket(PacketHeader* pHeader, uint8_t* pBuf, uint16_t dataLength);

void DMA2_Stream0_IRQHandler()
{
	SPISyncDmaRxInterrupt(&network.nrf24.spi);
}

void EXTI4_IRQHandler()
{
	CH_IRQ_PROLOGUE();

	EXTI->PR = (1 << 4);

	NRF24ExIRQ(&network.nrf24);

	CH_IRQ_EPILOGUE();
}

void TIM7_IRQHandler()
{
	CH_IRQ_PROLOGUE();

	TIM7->SR = 0;
	TIM_Cmd(TIM7, DISABLE);

	chSysLockFromIsr();
	chBSemSignalI(&network.timerSem);
	chSysUnlockFromIsr();

	CH_IRQ_EPILOGUE();
}

static void startNRF24Timeout(uint32_t microSeconds)
{
	TIM7->ARR = microSeconds;

	TIM_Cmd(TIM7, ENABLE);
}

static void networkIntercomCallback(IntercomHeader* pHeader, uint16_t dataLength, uint8_t* pData, void* pUserData)
{
	(void)pUserData;

	if(pHeader->command != 0)
		return;

	NetworkSendPacket((PacketHeader*)pData, pData+PACKET_HEADER_SIZE, dataLength-PACKET_HEADER_SIZE);
}

void NetworkInit()
{
	// data setup
	chBSemInit(&network.timerSem, TRUE);
	chMtxInit(&network.printMutex);
	chMtxInit(&network.sendPacketMutex);

	int16_t result = FlashFSOpenOrCreate("net/config", 0, &network.config, sizeof(ConfigNetwork), &network.pConfigFile);
	if(result == FLASH_FS_CREATED)
	{
		network.config.botId = 31;
		network.config.channel = 100;
		network.config.speed = 0x08;	// 2MBit/s

		FlashFSWrite(&network.pConfigFile, &network.config, sizeof(ConfigNetwork));
	}

	network.posCb = 0;
	network.avgRxPeriod = 0;
	network.lastUSec = 0;

	// NRF24 IRQ line setup
	GPIO_InitTypeDef gpioInit;
	gpioInit.GPIO_Mode = GPIO_Mode_IN;
	gpioInit.GPIO_PuPd = GPIO_PuPd_UP;
	gpioInit.GPIO_Speed = GPIO_Speed_25MHz;
	gpioInit.GPIO_OType = GPIO_OType_PP;
	gpioInit.GPIO_Pin = GPIO_Pin_4;
	GPIO_Init(GPIOC, &gpioInit);

	SYSCFG_EXTILineConfig(EXTI_PortSourceGPIOC, EXTI_PinSource4);

	EXTI_InitTypeDef extiInit;
	EXTI_StructInit(&extiInit);
	extiInit.EXTI_Line = EXTI_Line4;
	extiInit.EXTI_LineCmd = ENABLE;
	extiInit.EXTI_Mode = EXTI_Mode_Interrupt;
	extiInit.EXTI_Trigger = EXTI_Trigger_Falling;
	EXTI_Init(&extiInit);

	NVIC_InitTypeDef nvicInit;
	nvicInit.NVIC_IRQChannel = EXTI4_IRQn;
	nvicInit.NVIC_IRQChannelCmd = ENABLE;
	nvicInit.NVIC_IRQChannelPreemptionPriority = IRQL_SPI1;
	nvicInit.NVIC_IRQChannelSubPriority = 0;
	NVIC_Init(&nvicInit);

	// TIM7 setup
	RCC_APB1PeriphClockCmd(RCC_APB1Periph_TIM7, ENABLE);

	TIM_TimeBaseInitTypeDef timeBaseInit;
	timeBaseInit.TIM_ClockDivision = TIM_CKD_DIV1;
	timeBaseInit.TIM_CounterMode = TIM_CounterMode_Up;
	timeBaseInit.TIM_Period = 1000;	// = 1ms
	timeBaseInit.TIM_Prescaler = 83; // clock is now 1MHz
//	timeBaseInit.TIM_Prescaler = 41; // clock is now 1MHz
	TIM_TimeBaseInit(TIM7, &timeBaseInit);

	TIM_ARRPreloadConfig(TIM7, DISABLE);
	TIM_CCPreloadControl(TIM7, DISABLE);
	TIM_InternalClockConfig(TIM7);
	TIM_ITConfig(TIM7, TIM_IT_Update, ENABLE);

	nvicInit.NVIC_IRQChannel = TIM7_IRQn;
	nvicInit.NVIC_IRQChannelCmd = ENABLE;
	nvicInit.NVIC_IRQChannelPreemptionPriority = IRQL_NRF24_TIMER;
	nvicInit.NVIC_IRQChannelSubPriority = 0;
	NVIC_Init(&nvicInit);

	// SPI1 init
	RCC_APB2PeriphClockCmd(RCC_APB2Periph_SPI1, ENABLE);

	nvicInit.NVIC_IRQChannel = DMA2_Stream0_IRQn;
	nvicInit.NVIC_IRQChannelCmd = ENABLE;
	nvicInit.NVIC_IRQChannelPreemptionPriority = IRQL_SPI1;
	nvicInit.NVIC_IRQChannelSubPriority = 0x00;
	NVIC_Init(&nvicInit);

	gpioInit.GPIO_Mode = GPIO_Mode_AF;
	gpioInit.GPIO_OType = GPIO_OType_PP;
	gpioInit.GPIO_PuPd = GPIO_PuPd_UP;
	gpioInit.GPIO_Pin = GPIO_Pin_5 | GPIO_Pin_6 | GPIO_Pin_7;	// SCK, MISO, MOSI
	gpioInit.GPIO_Speed = GPIO_Speed_50MHz;
	GPIO_Init(GPIOA, &gpioInit);

	GPIO_PinAFConfig(GPIOA, GPIO_PinSource5, GPIO_AF_SPI1);
	GPIO_PinAFConfig(GPIOA, GPIO_PinSource6, GPIO_AF_SPI1);
	GPIO_PinAFConfig(GPIOA, GPIO_PinSource7, GPIO_AF_SPI1);

	// NRF24 setup
	NRF24ExData data;

	static uint8_t rxData[NRF24_RX_SIZE];
	static uint8_t txData[NRF24_TX_SIZE];

	NRF24QueueInit(&network.queue, txData, NRF24_TX_SIZE, rxData, NRF24_RX_SIZE);
	network.queue.id = network.config.botId;

	data.spiData.csPin = GPIO_Pin_2;
	data.spiData.pCSPort = GPIOB;
	data.spiData.pRegister = SPI1;
	data.spiData.pDma = DMA2;
	data.spiData.dmaChRx = 0;
	data.spiData.dmaChTx = 3;
	data.spiData.timeoutTicks = MS2ST(1000);
	data.rx = 1;
	data.channel = network.config.channel;
	data.speed = network.config.speed;
	data.maxRt = 0;
	data.ackDelayUs = 0;
	data.pCEPort = GPIOA;
	data.cePin = GPIO_Pin_4;

	NRF24ExInit(&network.nrf24, &data);

	IntercomClient client;
	client.pUserData = 0;
	client.rxCallback = &networkIntercomCallback;
	client.section = INTERCOM_SECTION_NETWORK;

	IntercomAddClient(&intercom1, client);
}

void NetworkSaveConfig()
{
	FlashFSWrite(&network.pConfigFile, &network.config, sizeof(ConfigNetwork));
}

msg_t NetworkTask(void* params)
{
	(void)params;

	uint32_t len;
	int16_t result;
	uint32_t numDelimiters = 0;
	uint32_t i;

	NRF24ExSetup(&network.nrf24);
	NRF24ExUseQueue(&network.nrf24, &network.queue);

	while(1)
	{
		result = NRF24ExWaitForIRQ(&network.nrf24, MS2ST(100));
		if(result && result != ERROR_NRF24EX_TIMEOUT)
			ConsolePrint("NRF24ExWaitForIRQ result: 0x%02X\r\n", result);

		const uint32_t alpha = 96;
		uint32_t tNow = SysTimeUSec();
		uint32_t tPeriod = tNow - network.lastUSec;
		network.lastUSec = tNow;
		network.avgRxPeriod = (alpha*network.avgRxPeriod + (100-alpha)*tPeriod)/100;

		uint32_t readDelay = (network.avgRxPeriod*20)/100;
		uint32_t writeDelay = (network.avgRxPeriod*20)/100;

		chBSemReset(&network.timerSem, TRUE);
		startNRF24Timeout(readDelay);
		if(chBSemWaitTimeout(&network.timerSem, MS2ST(100)) != RDY_OK)
			ConsolePrint("RxDownDelay error\r\n");

		result = NRF24ExRxDownload(&network.nrf24);
		if(result)
			ConsolePrint("NRF24ExRxDownload result: 0x%02X\r\n", result);

		chBSemReset(&network.timerSem, TRUE);
		startNRF24Timeout(writeDelay);


		numDelimiters = network.queue.rxFifo.numDelimiters;
		for(i=0;i<numDelimiters;i++)
		{
			result = NRF24QueueGetPacket(&network.queue, network.rxProcessingBuf, NETWORK_RX_PROCESSING_SIZE, &len);
			if(result)
			{
				ConsolePrint("NRF24QueueGetPacket failed: 0x%0X\r\n", result);
				break;
			}

			if(len > PACKET_HEADER_SIZE)
			{
				PacketHeader* pHeader = (PacketHeader*)network.rxProcessingBuf;
				uint8_t* pData = network.rxProcessingBuf+PACKET_HEADER_SIZE;

				processPacket(pHeader, pData, len-PACKET_HEADER_SIZE);
			}
		}

		if(chBSemWaitTimeout(&network.timerSem, MS2ST(100)) != RDY_OK)
			ConsolePrint("RxUpDelay error\r\n");

		result = NRF24ExRxUpload(&network.nrf24);
		if(result)
			ConsolePrint("NRF24ExRxUpload result: 0x%02X\r\n", result);
	}

	return 0;
}

void NetworkSetPosCallback(PosCallback cb)
{
	network.posCb = cb;
}

void NetworkSetTrackedObjCallback(TrackedObjCallback cb)
{
	network.trackedObjCb = cb;
}

int16_t NetworkSendPacket(PacketHeader* pHeader, uint8_t* pData, uint16_t dataLength)
{
	if(dataLength + PACKET_HEADER_SIZE > NETWORK_TX_PROCESSING_SIZE)
		return ERROR_NOT_ENOUGH_MEMORY;

	chMtxLock(&network.sendPacketMutex);	// prevent re-entering

	memcpy(network.txProcessingBuf, pHeader->_data, PACKET_HEADER_SIZE);
	memcpy(network.txProcessingBuf + PACKET_HEADER_SIZE, pData, dataLength);

	// data to send is now in network.txProcessingBuf
	int16_t result = NRF24QueueSend(&network.queue, network.txProcessingBuf, dataLength+PACKET_HEADER_SIZE);
	if(result)
	{
//		ConsolePrint("NRF24QueueSend error'd: %hx\r\n", result);
	}
	if(result == ERROR_FIFO_FULL)
	{
		FifoDelimClear(&network.queue.txFifo);
//		ConsolePrint("txFifo cleared.\r\n");
	}

	chMtxUnlock();

	return 0;
}

int32_t NetworkPrint(const char* fmt, ...)
{
	va_list args;
	int32_t bytesWritten;

	chMtxLock(&network.printMutex);
	{
		va_start(args, fmt);
		bytesWritten = vsnprintf((char*)(network.netPrint+sizeof(SystemConsoleCommand)), sizeof(network.netPrint)-sizeof(SystemConsoleCommand), fmt, args);
		va_end(args);

		if(bytesWritten > sizeof(network.netPrint)-sizeof(SystemConsoleCommand))
			bytesWritten = sizeof(network.netPrint)-sizeof(SystemConsoleCommand);

		SystemConsolePrint* pConPrint = (SystemConsolePrint*)network.netPrint;
		pConPrint->source = INTERCOM_ADDRESS_MAIN;

		PacketHeader header;
		header.cmd = CMD_SYSTEM_CONSOLE_PRINT;
		header.section = SECTION_SYSTEM;

		NetworkSendPacket(&header, (uint8_t*)&network.netPrint, bytesWritten+sizeof(SystemConsoleCommand));
	}
	chMtxUnlock();

	return bytesWritten;
}


// this is the main network processing function
static uint8_t processPacket(PacketHeader* pHeader, uint8_t* pBuf, uint16_t dataLength)
{
//	ConsolePrint("Sec/Cmd: %hhu, %hhu\r\n", pHeader->section, pHeader->cmd);

	switch(pHeader->section)
	{
		case SECTION_SYSTEM:
		{
			switch(pHeader->cmd)
			{
				case CMD_SYSTEM_PING:
				{
					SystemPing* pPing = (SystemPing*)pBuf;

					SystemPong pong;
					pong.id = pPing->id;

					PacketHeader header;
					header.section = SECTION_SYSTEM;
					header.cmd = CMD_SYSTEM_PONG;

					NetworkSendPacket(&header, pBuf, dataLength);
//					SystemPing* pPing = (SystemPing*)pBuf;
//
//					SystemPong pong;
//					pong.id = pPing->id;
//
//					NetworkSendPacket(&header, (uint8_t*)&pong, sizeof(SystemPong));

					return 1;
				}
				break;
				case CMD_SYSTEM_BIG_PING:
				{
					SystemBigPing* pPing = (SystemBigPing*)pBuf;

					SystemPong pong;
					pong.id = pPing->id;

					PacketHeader header;
					header.section = SECTION_SYSTEM;
					header.cmd = CMD_SYSTEM_PONG;

					NetworkSendPacket(&header, (uint8_t*)&pong, sizeof(SystemPong));

					return 1;
				}
				break;
				case CMD_SYSTEM_SET_LOGS:
				{
					SystemSetLogs* pLogs = (SystemSetLogs*)pBuf;

//					for(uint8_t i = 0; i < 4; i++)
//						CtrlMotorSetMotorLog(i, pLogs->motor[i]);

//					DribblerSetLog(pLogs->motor[4]);
//					KickerSetLogging(pLogs->kicker);
					PowerSetLogging(pLogs->power);
					CtrlSetLog(pLogs->movement);
					CtrlSetExtLog(pLogs->extMovement);

					// old controller
//					CtrlMotorSetLog(pLogs->movement);

					IntercomHeader iHeader;
					iHeader.address = INTERCOM_ADDRESS_KD;
					iHeader.section = INTERCOM_SECTION_COMMAND;
					iHeader.command = 0;

					IntercomSend(&intercom1, &iHeader, dataLength+PACKET_HEADER_SIZE, (uint8_t*)pHeader);

					return 1;
				}
				break;
				case CMD_SYSTEM_CONSOLE_COMMAND:
				{
					SystemConsoleCommand* pCmd = (SystemConsoleCommand*)pBuf;

					if(pCmd->target == INTERCOM_ADDRESS_MAIN)
					{
						// process onboard
						uint8_t* pData;
						uint32_t length = dataLength-sizeof(SystemConsoleCommand);
						int16_t result = FifoLinReserve(&console.rxFifo, length, &pData);
						if(result)
							return 0;

						memcpy(pData, pBuf+sizeof(SystemConsoleCommand), length);

						FifoLinCommit(&console.rxFifo, length);
					}
					else
					{
						// forward to intercom
						IntercomHeader header;
						header.address = pCmd->target;
						header.command = 0;
						header.section = INTERCOM_SECTION_CONSOLE;

						IntercomSend(&intercom1, &header, dataLength-sizeof(SystemConsoleCommand), pBuf+sizeof(SystemConsoleCommand));
					}
				}
				break;
			}
		}
		break;
#ifdef ENV_RUN
		case SECTION_BOOTLOADER:
		{
			switch(pHeader->cmd)
			{
				case CMD_BOOTLOADER_COMMAND:
				{
					BootloaderCommand* pCmd = (BootloaderCommand*)pBuf;

					switch(pCmd->cmd)
					{
						case BOOTLOADER_CMD_MODE_QUERY:
						{
							BootloaderResponse response;
							response.type = BOOTLOADER_RESPONSE_MODE_NORMAL;
							response.offset = 0;

							PacketHeader header;
							header.section = SECTION_BOOTLOADER;
							header.cmd = CMD_BOOTLOADER_RESPONSE;

							NetworkSendPacket(&header, (uint8_t*)&response, sizeof(BootloaderResponse));
						}
						break;
						case BOOTLOADER_CMD_ENTER:
						{
							BootloaderSelect(&intercom1);
							BootloaderSelect(&intercom1);
							BootSetBootloaderSelected(1);
							chThdSleep(MS2ST(500));
							BootReset();
						}
						break;
					}
				}
				break;
			}
		}
		break;
#else
		case SECTION_BOOTLOADER:
		{
			switch(pHeader->cmd)
			{
				case CMD_BOOTLOADER_COMMAND:
				{
					BootloaderCommand* pCmd = (BootloaderCommand*)pBuf;

					ConsolePrint("Bootloader CMD: %hu\r\n", (uint16_t)pCmd->cmd);

					switch(pCmd->cmd)
					{
						case BOOTLOADER_CMD_MODE_QUERY:
						{
							BootloaderResponse response;
							response.type = BOOTLOADER_RESPONSE_MODE_BOOTLOADER;
							response.offset = 0;

							PacketHeader header;
							header.section = SECTION_BOOTLOADER;
							header.cmd = CMD_BOOTLOADER_RESPONSE;

							NetworkSendPacket(&header, (uint8_t*)&response, sizeof(BootloaderResponse));
						}
						break;
						case BOOTLOADER_CMD_EXIT:
						{
							BootReset();
						}
						break;
						case BOOTLOADER_CMD_ERASE_MAIN:
						{
							BootloaderStartProgramming(INTERCOM_ADDRESS_MAIN);
						}
						break;
						case BOOTLOADER_CMD_ERASE_MEDIA:
						{
							BootloaderStartProgramming(INTERCOM_ADDRESS_MEDIA);
						}
						break;
						case BOOTLOADER_CMD_ERASE_KD:
						{
							BootloaderStartProgramming(INTERCOM_ADDRESS_KD);
						}
						break;
						case BOOTLOADER_CMD_ERASE_LEFT:
						{
							BootloaderStartProgramming(INTERCOM_ADDRESS_LEFT);
						}
						break;
						case BOOTLOADER_CMD_ERASE_RIGHT:
						{
							BootloaderStartProgramming(INTERCOM_ADDRESS_RIGHT);
						}
						break;
					}
				}
				break;
				case CMD_BOOTLOADER_DATA:
				{
					BootloaderData* pData = (BootloaderData*)pBuf;

					BootloaderFlashData(pData);
				}
				break;
			}
		}
		break;
#endif
		case SECTION_DECT:
		{
		}
		break;
#ifdef ENV_RUN
		case SECTION_KICKER:
		{
			IntercomHeader iHeader;
			iHeader.address = INTERCOM_ADDRESS_KD;
			iHeader.section = INTERCOM_SECTION_COMMAND;
			iHeader.command = 0;

			IntercomSend(&intercom1, &iHeader, dataLength+PACKET_HEADER_SIZE, (uint8_t*)pHeader);
/*
			switch(pHeader->cmd)
			{
				case CMD_KICKER_CHARGE_AUTO:
				case CMD_KICKER_KICKV2:
				{
					KickerKickV2* pKick = (KickerKickV2*)pBuf;

					FixedPoint142* pDuration = (FixedPoint142*)&pKick->duration;
					float time = pDuration->value;
					if(pDuration->exp == 1)
						time *= 10;
					if(pDuration->exp == 2)
						time *= 100;
					if(pDuration->exp == 3)
						time *= 1000;
					time /= 10.0f;

					if(pKick->mode == KICKER_MODE_FORCE)
					{
						KickerFire(time, pKick->device);
					}

					if(pKick->mode == KICKER_MODE_ARM)
					{
						KickerArmIr(time, pKick->device);
					}

					if(pKick->mode == KICKER_MODE_DRIBBLER)
					{
						KickerArmDribbler(time, pKick->device);
					}

					if(pKick->mode == KICKER_MODE_DISARM)
					{
						KickerDisarm();
					}

					return 1;
				}
				break;
				case CMD_KICKER_CHARGE_MANUAL:
				{
					KickerChargeManual* pCharge = (KickerChargeManual*)pBuf;

					KickerLoadManual(pCharge->tOnTicks, pCharge->tOffTicks, pCharge->duration);

					return 1;
				}
				break;
				case CMD_KICKER_CHARGE_AUTO:
				{
					KickerChargeAuto* pCharge = (KickerChargeAuto*)pBuf;

					KickerLoadAuto(pCharge->maxCapLevel);

					return 1;
				}
				break;
			}*/
		}
		break;
		case SECTION_MOTOR:
		{
			switch(pHeader->cmd)
			{
				case CMD_MOTOR_MOVE:
				{
					MotorMove* pMove = (MotorMove*)pBuf;

//					SkillReset();
//					CtrlMotorSetTarget(pMove->vx*0.001f, pMove->vy*0.001f, pMove->vw*0.001f);
//					CtrlFusionVelSetVelTarget(pMove->vx*0.001f, pMove->vy*0.001f, pMove->vw*0.001f);
					CtrlFusionSetVelTarget(pMove->vx*0.001f, pMove->vy*0.001f, pMove->vw*0.001f);

					return 1;
				}
				break;
				case CMD_MOTOR_MOVE_V2:
				{
					MotorMoveV2* pMove = (MotorMoveV2*)pBuf;

//					SkillReset();
//					CtrlMotorSetTarget(pMove->vx*0.001f, pMove->vy*0.001f, pMove->vw*0.001f);
//					CtrlFusionVelSetVelTarget(pMove->vx*0.001f, pMove->vy*0.001f, pMove->vw*0.001f);
					CtrlFusionSetVelTarget(pMove->vx*0.001f, pMove->vy*0.001f, pMove->vw*0.001f);

					return 1;
				}
				break;
				case CMD_MOTOR_DRIBBLE:
				{
					IntercomHeader iHeader;
					iHeader.address = INTERCOM_ADDRESS_KD;
					iHeader.section = INTERCOM_SECTION_COMMAND;
					iHeader.command = 0;

					IntercomSend(&intercom1, &iHeader, dataLength+PACKET_HEADER_SIZE, (uint8_t*)pHeader);

//					MotorDribble* pDribble = (MotorDribble*)pBuf;

//					DribblerSetTarget(pDribble->rpm);

					return 1;
				}
				break;
				case CMD_MOTOR_SET_PID_PARAMS:
				{
					MotorSetPidParams* pParams = (MotorSetPidParams*)pBuf;

					if(pParams->id == 4)
					{
						IntercomHeader iHeader;
						iHeader.address = INTERCOM_ADDRESS_KD;
						iHeader.section = INTERCOM_SECTION_COMMAND;
						iHeader.command = 0;

						IntercomSend(&intercom1, &iHeader, dataLength+PACKET_HEADER_SIZE, (uint8_t*)pHeader);

//						DribblerSetParams(pParams->Kp*0.0001f, pParams->Ki*0.0001f, pParams->Kd*0.0001f);
					}
/*					else
					{
						CtrlMotorSetParams(pParams->id, pParams->Kp*0.0001f, pParams->Ki*0.0001f, pParams->Kd*0.0001f);
					}
*/
					return 1;
				}
				break;
				case CMD_MOTOR_SET_MODE:
				{
//					MotorSetMode* pMode = (MotorSetMode*)pBuf;

					ConsolePrint("MotorSetMode not supported\r\n");

					// TODO: supported any more?
//					MovementSetMode(pMode->mode);
//					MovementSetLog(pMode->id, pMode->log);

					return 1;
				}
				break;
				case CMD_MOTOR_SET_MANUAL:
				{
//					MotorSetManual* pManual = (MotorSetManual*)pBuf;

					// TODO: support?
					// MovementSetManual(pManual->id, pManual->power);

					return 1;
				}
				break;
				case CMD_MOTOR_SET_PID_SP:
				{
					MotorSetPidSp* pSp = (MotorSetPidSp*)pBuf;

//					if(pSp->id == 4)
//						DribblerSetTarget(pSp->setpoint);

					// TODO: support
					// MovementSetPidSetpoint(pSp->id, pSp->setpoint);

					return 1;
				}
				break;
				case CMD_MOTOR_SET_PARAMS:
				{
					MotorSetParams* pParams = (MotorSetParams*)pBuf;

					// TODO: mode support?
					// MovementSetMode(pParams->mode);
/*
					DribblerSetParams(pParams->Kp[4]*0.0001f, pParams->Ki[4]*0.0001f, pParams->Kd[4]*0.0001f);

					for(uint8_t i = 0; i < 4; i++)
					{
						CtrlMotorSetParams(i, pParams->Kp[i]*0.0001f, pParams->Ki[i]*0.0001f, pParams->Kd[i]*0.0001f);
					}
*/
					return 1;
				}
				break;
			}
		}
		break;
		case SECTION_MULTICAST:
		{
		}
		break;
		case SECTION_MOVEMENT:
		{
			switch(pHeader->cmd)
			{
				case CMD_MOVEMENT_SET_PARAMS:
				{
				}
				break;
			}
		}
		break;
		case SECTION_CTRL:
		{
			switch(pHeader->cmd)
			{
				case CMD_CTRL_VISION_POS:
				{
					CtrlVisionPos* pPos = (CtrlVisionPos*)pBuf;

					memcpy(network.lastVisionPos, pPos->pos, sizeof(int16_t)*3);

					if(network.posCb)
						(*network.posCb)();
				}
				break;
				case CMD_CTRL_VISION_TRACKED_OBJ:
				{
					CtrlVisionTrackedObj* pPos = (CtrlVisionTrackedObj*)pBuf;

					memcpy(network.lastTrackedObjPos, pPos->pos, sizeof(int16_t)*3);

					if(network.trackedObjCb)
						(*network.trackedObjCb)();
				}
				break;
				case CMD_CTRL_SPLINE_2D:
				{
					CtrlSpline2D* pSpline = (CtrlSpline2D*)pBuf;

//					SkillReset();
					CtrlFusionAddSplineCmdXY(pSpline);
				}
				break;
				case CMD_CTRL_SPLINE_1D:
				{
					CtrlSpline1D* pSpline = (CtrlSpline1D*)pBuf;

//					SkillReset();
					CtrlFusionAddSplineCmdW(pSpline);
				}
				break;
				case CMD_CTRL_SET_FILTER_PARAMS:
				{
					CtrlSetFilterParams* pParams = (CtrlSetFilterParams*)pBuf;

					CtrlFusionSetFilterParams(pParams);
//					CtrlFusionVelSetFilterParams(pParams);
				}
				break;
				case CMD_CTRL_SET_PID_PARAMS:
				{
					CtrlSetPIDParams* pParams = (CtrlSetPIDParams*)pBuf;

					CtrlFusionSetPIDParams(pParams);
//					CtrlFusionVelSetPIDParams(pParams);
				}
				break;
				case CMD_CTRL_SET_CONTROLLER_TYPE:
				{
					CtrlSetControllerType* pCtrl = (CtrlSetControllerType*)pBuf;

//					SkillReset();
					switch(pCtrl->controllerId)
					{
						case CTRL_TYPE_FUSION_VEL:
						{
//							CtrlSetController(&ctrlFusionVelInstance);
						}
						break;
						case CTRL_TYPE_FUSION:
						{
							CtrlSetController(&ctrlFusionInstance);
						}
						break;
						default:
						{
							CtrlSetController(0);
						}
						break;
					}
					ConsolePrint("Controller changed to %u\r\n", pCtrl->controllerId);
				}
				break;
				case CMD_CTRL_RESET:
				{
//					SkillReset();
					CtrlFusionReset();
				}
				break;
			}
		}
		break;
		case SECTION_SKILL:
		{
			switch(pHeader->cmd)
			{
				case CMD_SKILL_KEEPER:
				{
//					SkillKeeperCommand* pSkillParams = (SkillKeeperCommand*)pBuf;
//					skillKeeperInstance.pInitFunction(pSkillParams);
//					SkillSet(&skillKeeperInstance);
				}
				break;
				case CMD_SKILL_SHOOTER:
				{
//					SkillShooterCommand* pSkillParams = (SkillShooterCommand*)pBuf;
//					skillShooterInstance.pInitFunction(pSkillParams);
//					SkillSet(&skillShooterInstance);
				}
				break;
				case CMD_SKILL_POSITIONING:
				{
					SkillPositioningCommand* pSkillParams = (SkillPositioningCommand*)pBuf;
					CtrlFusionSetPosTarget(pSkillParams->destination[0]/1000.0f,
							pSkillParams->destination[1]/1000.0f,
							pSkillParams->destination[2]/1000.0f);
//					SkillPositioningCommand* pSkillParams = (SkillPositioningCommand*)pBuf;
//					skillPositioningInstance.pInitFunction(pSkillParams);
//					SkillSet(&skillPositioningInstance);
				}
				break;
				case CMD_SKILL_CALIBRATE_MOTOR:
				{
//					SkillCalibrateMotorCommand* pSkillParams = (SkillCalibrateMotorCommand*)pBuf;
//					skillCalibrateMotorInstance.pInitFunction(pSkillParams);
//					SkillSet(&skillCalibrateMotorInstance);
				}
				break;
			}
		}
#endif
/*		case SECTION_MOTOR:
		{
			switch(pHeader->cmd)
			{
				case CMD_MOTOR_MOVE:
				{
					MotorMove* pMove = (MotorMove*)pBuf;

//					SkillReset();
//					CtrlMotorSetTarget(pMove->vx*0.001f, pMove->vy*0.001f, pMove->vw*0.001f);
//					CtrlFusionVelSetVelTarget(pMove->vx*0.001f, pMove->vy*0.001f, pMove->vw*0.001f);
//					CtrlFusionSetVelTarget(pMove->vx*0.001f, pMove->vy*0.001f, pMove->vw*0.001f);

					CtrlVelSetTarget(pMove->vx*0.001f, pMove->vy*0.001f, pMove->vw*0.001f);

					return 1;
				}
				break;
				case CMD_MOTOR_MOVE_V2:
				{
					MotorMoveV2* pMove = (MotorMoveV2*)pBuf;

//					SkillReset();
//					CtrlMotorSetTarget(pMove->vx*0.001f, pMove->vy*0.001f, pMove->vw*0.001f);
//					CtrlFusionVelSetVelTarget(pMove->vx*0.001f, pMove->vy*0.001f, pMove->vw*0.001f);
//					CtrlFusionSetVelTarget(pMove->vx*0.001f, pMove->vy*0.001f, pMove->vw*0.001f);

					CtrlVelSetTarget(pMove->vx*0.001f, pMove->vy*0.001f, pMove->vw*0.001f);

					return 1;
				}
				break;
			}
		}*/
	}

	return 0;
}
