/*
 * power.c
 *
 * Power control and overcurrent watchdog.
 *
 *  Created on: 25.10.2010
 *      Author: AndreR
 */

#include "power.h"

#include "stm32f4xx_gpio.h"

#include "commands.h"
#include "network.h"
#include "intercom1.h"
#include "spi3.h"
#include "hal/buzzer.h"
#include "ctrl.h"
#include "util/console.h"

#define POWER_5V_ENABLE (1 << 11)	// GPIOA

PowerGlobal power;

static void supplyCallback(IntercomHeader* pHeader, uint16_t dataLength, uint8_t* pData, void* pUserData)
{
	(void)pUserData;

	if(pHeader->command == 0 && dataLength >= sizeof(SupplyPowerLog))
	{
		SupplyPowerLog* pLog = (SupplyPowerLog*)pData;

		memcpy(&power.supplyLog, pLog, sizeof(SupplyPowerLog));

		power.lastLog.u1 = pLog->bat*1000.0f;
		power.lastLog.iVCC = pLog->cur*1000.0f;

		if(power.cb)
			(*power.cb)();
	}
}

void PowerInit()
{
	GPIO_InitTypeDef gpioInit;

	gpioInit.GPIO_Mode = GPIO_Mode_OUT;
	gpioInit.GPIO_Pin = POWER_5V_ENABLE;
	gpioInit.GPIO_Speed = GPIO_Speed_2MHz;
	gpioInit.GPIO_OType = GPIO_OType_PP;
	GPIO_Init(GPIOA, &gpioInit);

	power.logEnabled = 1;
}

void PowerAddIntercomClient()
{
	IntercomClient client;
	client.pUserData = 0;
	client.rxCallback = &supplyCallback;
	client.section = INTERCOM_SECTION_SUPPLY;

	IntercomAddClient(&intercom1, client);
}

void PowerEnable5VSupply(uint8_t enable)
{
	if(enable)
		GPIOA->BSRRL = POWER_5V_ENABLE;
	else
		GPIOA->BSRRH = POWER_5V_ENABLE;
}

uint8_t PowerIs5VSupplyEnabled()
{
	if(GPIOA->ODR & POWER_5V_ENABLE)
		return 0;
	else
		return 1;
}

void PowerSetCallback(PowerUpdatedCallback cb)
{
	power.cb = cb;
}

void PowerSetLogging(uint8_t enable)
{
	power.logEnabled = enable;
}

msg_t PowerTask(void* params)
{
	(void)params;
	PacketHeader header;

	chRegSetThreadName("Power");

	uint32_t lowBatCounter = 0;
	uint8_t buzzState = 0;

	while(1)
	{
		if(power.logEnabled)
		{
			header.section = SECTION_SYSTEM;
			header.cmd = CMD_SYSTEM_POWER_LOG;

			NetworkSendPacket(&header, (uint8_t*)&power.lastLog, sizeof(SystemPowerLog));
		}

		chThdSleep(MS2ST(100));

		if(power.supplyLog.bat < 10.5f)
		{
			++lowBatCounter;
			if(lowBatCounter == 100)
			{
				CtrlSetController(0);
			}
		}
		else
		{
			lowBatCounter = 0;
		}

#ifdef ENV_RUN
		if(lowBatCounter > 100)	// 10s below 10.5V
		{
			BuzzerEnable(buzzState);
			buzzState ^= 0x01;
		}
#endif
	}

	return 0;
}
