/*
 * spi3.c
 *
 *  Created on: 19.01.2014
 *      Author: AndreR
 */

#include "spi3.h"

#include "stm32f4xx_gpio.h"
#include "stm32f4xx_spi.h"
#include "stm32f4xx_rcc.h"
#include "stm32f4xx_dma.h"
#include "stm32f4xx_tim.h"
#include "misc.h"

#include "util/block_proto.h"
#include "util/log.h"
#include "util/console.h"
#include "util/bit_banding.h"

#define IRQL_SPI3 12

SPI3Global spi3;

#define EVENT_TRANSFER_MASK	0x01FF

#define EVENT_DMA_TX_TE		0x0001
#define EVENT_DMA_TX_DME	0x0002
#define EVENT_DMA_TX_FE		0x0004
#define EVENT_DMA_TX_TC		0x0008
#define EVENT_DMA_RX_TE		0x0010
#define EVENT_DMA_RX_DME	0x0020
#define EVENT_DMA_RX_FE		0x0040
#define EVENT_DMA_RX_TC		0x0080
#define EVENT_SPI_CRCERR	0x0100
#define EVENT_TIMER			0x1000

#define STATE_IDLE			0
#define STATE_ACTIVE		1

#define DMA_Stream_RX		DMA1_Stream0
#define DMA_Stream_TX		DMA1_Stream7

#define DMAClearTCIE(pDMA) 	BBPeriphReset((uint32_t)&pDMA->CR, 4)
#define DMASetTCIE(pDMA)	BBPeriphSet((uint32_t)&pDMA->CR, 4)
#define DMAGetTCIE(pDMA)	BBPeriphGet((uint32_t)&pDMA->CR, 4)

#define DMAClearTEIE(pDMA) 	BBPeriphReset((uint32_t)&pDMA->CR, 2)
#define DMASetTEIE(pDMA)	BBPeriphSet((uint32_t)&pDMA->CR, 2)
#define DMAGetTEIE(pDMA)	BBPeriphGet((uint32_t)&pDMA->CR, 2)

static uint16_t getDMARegs()
{
	uint16_t event = 0;

    if(DMA1->LISR & DMA_FLAG_TEIF0)
    	event |= EVENT_DMA_RX_TE;
    if(DMA1->LISR & DMA_FLAG_DMEIF0)
    	event |= EVENT_DMA_RX_DME;
    if(DMA1->LISR & DMA_FLAG_FEIF0)
    	event |= EVENT_DMA_RX_FE;
    if(DMA1->LISR & DMA_FLAG_TCIF0)
    	event |= EVENT_DMA_RX_TC;

    if(DMA1->HISR & DMA_FLAG_TEIF7)
    	event |= EVENT_DMA_TX_TE;
    if(DMA1->HISR & DMA_FLAG_DMEIF7)
    	event |= EVENT_DMA_TX_DME;
    if(DMA1->HISR & DMA_FLAG_FEIF7)
    	event |= EVENT_DMA_TX_FE;
    if(DMA1->HISR & DMA_FLAG_TCIF7)
    	event |= EVENT_DMA_TX_TC;

    if(DMA_Stream_TX->CR & DMA_SxCR_EN)
    	event |= 0x0200;
    if(DMA_Stream_RX->CR & DMA_SxCR_EN)
    	event |= 0x0400;

    return event;
}

static uint16_t stopDMA()
{
	uint16_t event = 0;

    event = getDMARegs();

    DMA_Cmd(DMA_Stream_TX, DISABLE);
    DMA_Cmd(DMA_Stream_RX, DISABLE);

    while(DMA_Stream_TX->CR & DMA_SxCR_EN);
    while(DMA_Stream_RX->CR & DMA_SxCR_EN);

    SPI3Slave* pSlave = &spi3.slave[spi3.activeSlave];
    pSlave->pCSPort->BSRRL = pSlave->csPin;	// deselect slave

    DMA_ClearFlag(DMA_Stream_RX, DMA_FLAG_TCIF0 | DMA_FLAG_HTIF0 | DMA_FLAG_TEIF0 | DMA_FLAG_DMEIF0 | DMA_FLAG_FEIF0);
    DMA_ClearFlag(DMA_Stream_TX, DMA_FLAG_TCIF7 | DMA_FLAG_HTIF7 | DMA_FLAG_TEIF7 | DMA_FLAG_DMEIF7 | DMA_FLAG_FEIF7);

    if(SPI_I2S_GetFlagStatus(SPI3, SPI_FLAG_CRCERR) == SET)
    	event |= EVENT_SPI_CRCERR;

    SPI_I2S_ClearFlag(SPI3, SPI_FLAG_CRCERR);

    SPI_Cmd(SPI3, DISABLE);
    SPI_CalculateCRC(SPI3, DISABLE);

    // clear OVR (just in case)
    SPI3->DR;
    SPI3->SR;

    return event;
}

void DMA1_Stream0_IRQHandler()
{
	CH_IRQ_PROLOGUE();

	DMAClearTCIE(DMA_Stream_RX);
	DMAClearTEIE(DMA_Stream_RX);

	chSysLockFromIsr();
	chMBPostI(&spi3.eventQueue, EVENT_TRANSFER_MASK);
	chSysUnlockFromIsr();

	CH_IRQ_EPILOGUE();
}

void TIM8_UP_TIM13_IRQHandler()
{
	CH_IRQ_PROLOGUE();

	TIM13->SR = 0;

	msg_t event = EVENT_TIMER;

	chSysLockFromIsr();
	chMBPostI(&spi3.eventQueue, event);
	chSysUnlockFromIsr();

	CH_IRQ_EPILOGUE();
}

void SPI3Init()
{
	// init event queue/variables
	chMBInit(&spi3.eventQueue, spi3.eventQueueData, 4);
	spi3.state = STATE_IDLE;
	spi3.activeSlave = SPI3_SLAVE_LEFT;
	spi3.slave[SPI3_SLAVE_LEFT].pCSPort = GPIOC;
	spi3.slave[SPI3_SLAVE_LEFT].csPin = GPIO_Pin_3;
	spi3.slave[SPI3_SLAVE_RIGHT].pCSPort = GPIOB;
	spi3.slave[SPI3_SLAVE_RIGHT].csPin = GPIO_Pin_5;

	for(uint8_t i = 0; i < 2; i++)
	{
		SPI3Slave* pSlave = &spi3.slave[i];

		pSlave->rxSeq = 0;
		pSlave->txSeq = 0;
		FifoLinInit(&pSlave->txFifo, pSlave->fifoDataTx, SPI3_FIFO_SIZE_TX);
		FifoLinInit(&pSlave->rxFifo, pSlave->fifoDataRx, SPI3_FIFO_SIZE_RX);
	}

#ifdef DEBUG
	spi3.pRxDma = DMA_Stream_RX;
	spi3.pTxDma = DMA_Stream_TX;
	spi3.pSPI = SPI3;
	spi3.pDMA = DMA1;
#endif

	RCC_APB1PeriphClockCmd(RCC_APB1Periph_SPI3, ENABLE);

	// init GPIO
	GPIO_InitTypeDef gpioInit;
	gpioInit.GPIO_Mode = GPIO_Mode_AF;
	gpioInit.GPIO_OType = GPIO_OType_PP;
	gpioInit.GPIO_Speed = GPIO_Speed_50MHz;
	gpioInit.GPIO_PuPd = GPIO_PuPd_DOWN;
	gpioInit.GPIO_Pin = GPIO_Pin_10 | GPIO_Pin_11 | GPIO_Pin_12;	// SCK, MISO, MOSI
	GPIO_Init(GPIOC, &gpioInit);

	gpioInit.GPIO_Mode = GPIO_Mode_OUT;
	gpioInit.GPIO_Pin = GPIO_Pin_5;	// MOTA-NSS
	GPIO_Init(GPIOB, &gpioInit);

	gpioInit.GPIO_Pin = GPIO_Pin_3;	// MOTB-NSS
	GPIO_Init(GPIOC, &gpioInit);

	GPIOB->BSRRL = GPIO_Pin_5;	// deselect slaves
	GPIOC->BSRRL = GPIO_Pin_3;

	GPIO_PinAFConfig(GPIOC, GPIO_PinSource10, GPIO_AF_SPI3);
	GPIO_PinAFConfig(GPIOC, GPIO_PinSource11, GPIO_AF_SPI3);
	GPIO_PinAFConfig(GPIOC, GPIO_PinSource12, GPIO_AF_SPI3);

	// init SPI3
	SPI_InitTypeDef spiInit;
	spiInit.SPI_BaudRatePrescaler = SPI_BaudRatePrescaler_4;
	spiInit.SPI_CPHA = SPI_CPHA_1Edge;
	spiInit.SPI_CPOL = SPI_CPOL_Low;
	spiInit.SPI_CRCPolynomial = 0x1021;	// CRC16-CCITT
	spiInit.SPI_DataSize = SPI_DataSize_16b;
	spiInit.SPI_Direction = SPI_Direction_2Lines_FullDuplex;
	spiInit.SPI_FirstBit = SPI_FirstBit_MSB;
	spiInit.SPI_Mode = SPI_Mode_Master;
	spiInit.SPI_NSS = SPI_NSS_Soft;
	SPI_Init(SPI3, &spiInit);

	SPI_NSSInternalSoftwareConfig(SPI3, SPI_NSSInternalSoft_Set);

	// init DMA
	// TX
	DMA_InitTypeDef dmaInit;
	dmaInit.DMA_Channel = DMA_Channel_0;
	dmaInit.DMA_PeripheralBaseAddr = (uint32_t)&SPI3->DR;
	dmaInit.DMA_Memory0BaseAddr = 0;
	dmaInit.DMA_DIR = DMA_DIR_MemoryToPeripheral;
	dmaInit.DMA_BufferSize = 0;
	dmaInit.DMA_PeripheralInc = DMA_PeripheralInc_Disable;
	dmaInit.DMA_MemoryInc = DMA_MemoryInc_Enable;
	dmaInit.DMA_PeripheralDataSize = DMA_PeripheralDataSize_HalfWord;
	dmaInit.DMA_MemoryDataSize = DMA_MemoryDataSize_HalfWord;
	dmaInit.DMA_Mode = DMA_Mode_Normal;
	dmaInit.DMA_Priority = DMA_Priority_High;
	dmaInit.DMA_FIFOMode = DMA_FIFOMode_Enable;
	dmaInit.DMA_FIFOThreshold = DMA_FIFOThreshold_Full;
	dmaInit.DMA_MemoryBurst = DMA_MemoryBurst_Single;
	dmaInit.DMA_PeripheralBurst = DMA_PeripheralBurst_Single;
	DMA_Init(DMA_Stream_TX, &dmaInit);

	// RX
	dmaInit.DMA_BufferSize = 0;
	dmaInit.DMA_DIR = DMA_DIR_PeripheralToMemory;
	dmaInit.DMA_Memory0BaseAddr = 0;
	DMA_Init(DMA_Stream_RX, &dmaInit);

	DMA_ITConfig(DMA_Stream_RX, DMA_IT_TC | DMA_IT_TE | DMA_IT_DME | DMA_IT_FE, ENABLE);

    DMA_ClearFlag(DMA_Stream_RX, DMA_FLAG_TCIF0 | DMA_FLAG_HTIF0 | DMA_FLAG_TEIF0 | DMA_FLAG_DMEIF0 | DMA_FLAG_FEIF0);
    DMA_ClearFlag(DMA_Stream_TX, DMA_FLAG_TCIF7 | DMA_FLAG_HTIF7 | DMA_FLAG_TEIF7 | DMA_FLAG_DMEIF7 | DMA_FLAG_FEIF7);

	// configure NVIC
	NVIC_InitTypeDef nvicInit;
	nvicInit.NVIC_IRQChannel = DMA1_Stream0_IRQn;
	nvicInit.NVIC_IRQChannelPreemptionPriority = IRQL_SPI3;
	nvicInit.NVIC_IRQChannelSubPriority = 0;
	nvicInit.NVIC_IRQChannelCmd = ENABLE;
	NVIC_Init(&nvicInit);

	SPI_I2S_DMACmd(SPI3, SPI_I2S_DMAReq_Rx, ENABLE);
	SPI_I2S_DMACmd(SPI3, SPI_I2S_DMAReq_Tx, ENABLE);

    // init and start timer
	RCC_APB1PeriphClockCmd(RCC_APB1Periph_TIM13, ENABLE);

	TIM_TimeBaseInitTypeDef timeBaseInit;
	timeBaseInit.TIM_ClockDivision = TIM_CKD_DIV1;
	timeBaseInit.TIM_CounterMode = TIM_CounterMode_Up;
	timeBaseInit.TIM_Period = 500;	// = 1ms
	timeBaseInit.TIM_Prescaler = 83; // clock is now 1MHz
	TIM_TimeBaseInit(TIM13, &timeBaseInit);

	TIM13->SR = 0;
	TIM_ARRPreloadConfig(TIM13, DISABLE);
	TIM_ITConfig(TIM13, TIM_IT_Update, ENABLE);

	nvicInit.NVIC_IRQChannel = TIM8_UP_TIM13_IRQn;
	NVIC_Init(&nvicInit);
}

msg_t SPI3Task(void* params)
{
	(void)params;

	chRegSetThreadName("SPI3");

	msg_t event;

	// TODO: left and right need some time, sync method?
	chThdSleep(MS2ST(1000));

	chMBReset(&spi3.eventQueue);
	TIM_Cmd(TIM13, ENABLE);

	// TODO: alternate select scheme
	// LEFT - pause - RIGHT - pause
	// LEFT - RIGHT - pause

	while(1)
	{
		if(chMBFetch(&spi3.eventQueue, &event, MS2ST(100)) != RDY_OK)
		{
			// no item received. TODO: maybe handle SPI/DMA error condition
			LogWarn("Event reception timed out");
			continue;
		}

		LogDebugC("Received event", event);

		if(event == EVENT_TIMER && spi3.state == STATE_IDLE)
		{
			// select new slave
			++spi3.activeSlave;
			spi3.activeSlave %= 2;

			SPI3Slave* pSlave = &spi3.slave[spi3.activeSlave];

			if(spi3.aboutToSendCb)
				(*spi3.aboutToSendCb)(pSlave, spi3.activeSlave);

			// upload block data
			BlockProtoWriteData(&pSlave->txFifo, pSlave->txData.blockData, sizeof(pSlave->txData.blockData));
//			if(BlockProtoWriteData(&pSlave->txFifo, pSlave->txData.blockData, sizeof(pSlave->txData.blockData)) == 0)
//				ConsolePrint("No block data written\r\n");

			pSlave->pCSPort->BSRRH = pSlave->csPin;

			chMBReset(&spi3.eventQueue);

			// setup data
			DMA_Stream_TX->M0AR = (uint32_t)&pSlave->txData;
			DMA_Stream_RX->M0AR = (uint32_t)&pSlave->rxData;
			DMA_Stream_TX->NDTR = (MOT_EXCHANGE_DATA_MSG_SIZE+0)/2;
			DMA_Stream_RX->NDTR = (MOT_EXCHANGE_DATA_MSG_SIZE+2)/2;

			// Enable the DMA channels
			DMA_Cmd(DMA_Stream_RX, ENABLE);
			DMA_Cmd(DMA_Stream_TX, ENABLE);
			DMASetTCIE(DMA_Stream_RX);
		    DMASetTEIE(DMA_Stream_RX);

		    SPI_CalculateCRC(SPI3, ENABLE);
		    SPI_I2S_ClearFlag(SPI3, SPI_FLAG_CRCERR);
		    SPI_Cmd(SPI3, ENABLE);

			spi3.state = STATE_ACTIVE;

			continue;
		}

		if(event == EVENT_TIMER && spi3.state == STATE_ACTIVE)
		{
			// TODO: missed timer => too much data, ignore
/*			if(DMA_Stream_TX->NDTR == 0 || (DMA1->HISR & DMA_FLAG_FEIF7))
			{
				if((DMA_Stream_TX->CR & DMA_SxCR_EN) == 0)
					LogError("DMA FIFO error");

				stopDMA();
				spi3.state = STATE_IDLE;
			}
			else
			{
				LogWarnC("Too much data", getDMARegs());
			}
*/
			continue;
		}

		if(event & EVENT_TRANSFER_MASK)
		{
			event = stopDMA();

			SPI3Slave* pSlave = &spi3.slave[spi3.activeSlave];
/*
			if(event & (EVENT_DMA_RX_DME | EVENT_DMA_RX_FE | EVENT_DMA_RX_TE | EVENT_DMA_TX_DME | EVENT_DMA_TX_FE | EVENT_DMA_TX_TE))
			{
				ConsolePrint("DMA Error: 0x%0X\r\n", event);
				ConsolePrint("NDTR: %hu/%hu\r\n", DMA_Stream_RX->NDTR, DMA_Stream_TX->NDTR);
				ConsolePrint("CR: 0x%0X/0x%0X\r\n", DMA_Stream_RX->CR, DMA_Stream_TX->CR);
				ConsolePrint("FCR: 0x%0X/0x%0X\r\n", DMA_Stream_RX->FCR, DMA_Stream_TX->FCR);
				ConsolePrint("SPI->SR: 0x%0X\r\n", SPI3->SR);
			}
*/
			if((event & EVENT_SPI_CRCERR) == 0)
			{
				// rx successfull
				if(BlockProtoReadData(&pSlave->rxFifo, pSlave->rxData.blockData, sizeof(pSlave->rxData.blockData)) && spi3.callback)
					(*spi3.callback)(spi3.pUserData);
			}

//			ConsolePrint("Seq: %hhu, RxPkts: %u\r\n", pSlave->rxData.periodic.seqId, pSlave->rxFifo.numPackets);

			spi3.state = STATE_IDLE;
		}
	}

	return 0;
}
