/*
 * touch.c
 *
 *  Created on: 05.07.2014
 *      Author: AndreR
 */

#include "stm32f4xx_gpio.h"
#include "stm32f4xx_rcc.h"
#include "stm32f4xx_syscfg.h"
#include "stm32f4xx_exti.h"
#include "misc.h"

#include "touch.h"
#include "constants.h"
#include "util/bit_banding.h"
#include "GUI.h"
#include "hal/tft.h"

Touch touch;

#define TOUCH_X_MIN		2048	// 204.8*10
#define TOUCH_X_MAX		63488	// 204.8*310
#define TOUCH_Y_MIN		2730	// 273.066*10
#define TOUCH_Y_MAX		62805	// 273.066*230
#define TOUCH_X_RANGE	(TOUCH_X_MAX-TOUCH_X_MIN)
#define TOUCH_Y_RANGE	(TOUCH_Y_MAX-TOUCH_Y_MIN)

#define LCD_WIDTH		240
#define LCD_HEIGHT		320

typedef struct _TSMeas
{
	uint32_t x;
	uint32_t y;
	uint32_t z1;
	uint32_t z2;
} TSMeas;

// touch event active
void EXTI2_IRQHandler()
{
	CH_IRQ_PROLOGUE();

	BBPeriphSet((uint32_t)&EXTI->PR, 2);

	chSysLockFromIsr();
	chBSemSignalI(&touch.touchedSem);
	chSysUnlockFromIsr();

	CH_IRQ_EPILOGUE();
}

void DMA1_Stream2_IRQHandler()
{
	SPISyncDmaRxInterrupt(&touch.spi3);
}

void TouchInit()
{
	chBSemInit(&touch.touchedSem, TRUE);

	RCC_APB1PeriphClockCmd(RCC_APB1Periph_SPI3, ENABLE);

	SPISyncData data3;
	data3.cpha = 0;
	data3.cpol = 0;
	data3.csPin = GPIO_Pin_6;
	data3.pCSPort = GPIOD;
	data3.dmaChRx = 2;
	data3.dmaChTx = 7;
	data3.dmaPrio = 1;
	data3.pDma = DMA1;
	data3.pRegister = SPI3;
	data3.prescaler = SPI_BaudRatePrescaler_2;
	data3.timeoutTicks = MS2ST(100);
	SPISyncInit(&touch.spi3, &data3);

	NVIC_InitTypeDef nvicInit;
	nvicInit.NVIC_IRQChannel = DMA1_Stream2_IRQn;
	nvicInit.NVIC_IRQChannelCmd = ENABLE;
	nvicInit.NVIC_IRQChannelPreemptionPriority = IRQL_SPI3;
	nvicInit.NVIC_IRQChannelSubPriority = 0;
	NVIC_Init(&nvicInit);

	GPIO_InitTypeDef gpioInit;
	gpioInit.GPIO_Mode = GPIO_Mode_AF;
	gpioInit.GPIO_OType = GPIO_OType_PP;
	gpioInit.GPIO_PuPd = GPIO_PuPd_UP;
	gpioInit.GPIO_Pin = GPIO_Pin_10 | GPIO_Pin_11 | GPIO_Pin_12;	// SCK, MISO, MOSI
	gpioInit.GPIO_Speed = GPIO_Speed_25MHz;
	GPIO_Init(GPIOC, &gpioInit);

	GPIO_PinAFConfig(GPIOC, GPIO_PinSource10, GPIO_AF_SPI3);
	GPIO_PinAFConfig(GPIOC, GPIO_PinSource11, GPIO_AF_SPI3);
	GPIO_PinAFConfig(GPIOC, GPIO_PinSource12, GPIO_AF_SPI3);

	// TP-IRQ: E2
	gpioInit.GPIO_Mode = GPIO_Mode_IN;
	gpioInit.GPIO_OType = GPIO_OType_OD;
	gpioInit.GPIO_PuPd = GPIO_PuPd_UP;
	gpioInit.GPIO_Speed = GPIO_Speed_2MHz;
	gpioInit.GPIO_Pin = GPIO_Pin_2;
	GPIO_Init(GPIOE, &gpioInit);

	SYSCFG_EXTILineConfig(EXTI_PortSourceGPIOE, EXTI_PinSource2);

	EXTI_InitTypeDef extiInit;
	EXTI_StructInit(&extiInit);
	extiInit.EXTI_Line = EXTI_Line2;
	extiInit.EXTI_LineCmd = ENABLE;
	extiInit.EXTI_Mode = EXTI_Mode_Interrupt;
	extiInit.EXTI_Trigger = EXTI_Trigger_Rising_Falling;
	EXTI_Init(&extiInit);

	nvicInit.NVIC_IRQChannelCmd = ENABLE;
	nvicInit.NVIC_IRQChannelPreemptionPriority = IRQL_EXTI2;
	nvicInit.NVIC_IRQChannelSubPriority = 0;
	nvicInit.NVIC_IRQChannel = EXTI2_IRQn;
	NVIC_Init(&nvicInit);

	touch.buf[0] = 0x03 << 1;
	touch.buf[1] = 0xFF; // 16 samples, median filter
	touch.buf[2] = 0x07 << 1;
	touch.buf[3] = 0x60; // set PUR to 1ms
	touch.buf[4] = 0x0B << 1; // write Mode Config Register
	touch.buf[5] = 0x06; // avg filter on x,y,z1,z2
	touch.buf[6] = 0x04 << 1;
	touch.buf[7] = 0x55; // ADC sample time 128us on all channels
	touch.buf[8] = 0x05 << 1;
	touch.buf[9] = 0x44; // Panel setup time 1ms
	touch.buf[10]= 0x06 << 1;
	touch.buf[11]= 0x00; // Delayed conversion 1ms
	touch.buf[12]= 0x01 << 1;
	touch.buf[13]= 0xC0;	//

	SPISyncTransfer(&touch.spi3, touch.buf, 0, 12);
}

msg_t TouchTask(void* params)
{
	(void)params;
	TSMeas result;

	uint32_t touchTime = chTimeNow();

	while(1)
	{
		chBSemWaitTimeout(&touch.touchedSem, MS2ST(10));

		// CMD_MEAS_XYZ
		touch.buf[0] = 0x74 << 1;
		SPISyncTransfer(&touch.spi3, touch.buf, 0, 1);

		chThdSleep(MS2ST(10));

		// CMD_READ_XYZ
		touch.buf[0] = 0x52 << 1 | 0x01;
		SPISyncTransfer(&touch.spi3, touch.buf, touch.buf, 9);

		result.x =  touch.buf[1] << 8;
		result.x |= touch.buf[2];
		result.y =  touch.buf[3] << 8;
		result.y |= touch.buf[4];
		result.z1 =  touch.buf[5] << 8;
		result.z1 |= touch.buf[6];
		result.z2 =  touch.buf[7] << 8;
		result.z2 |= touch.buf[8];

		result.y = 65536-result.y;	// Y starts at 0 in lower left corner, so we invert it to match pixel positions

		float x = result.x;
		float z1 = result.z1;
		float z2 = result.z2;

		touch.touchZ1 = result.z1;
		touch.touchZ2 = result.z2;

		touch.touchZ = (x/65536.0f)*((z2/z1)-1.0f);
		if(touch.touchZ < 2.0f && z1 > 1000.0f && z2 < 65000.0f
			&& result.x > TOUCH_X_MIN && result.x < TOUCH_X_MAX && result.y > TOUCH_Y_MIN && result.y < TOUCH_Y_MAX)
		{
			touch.touchY = LCD_HEIGHT-((result.x-TOUCH_X_MIN)*LCD_HEIGHT)/TOUCH_X_RANGE;
			touch.touchX = ((result.y-TOUCH_Y_MIN)*LCD_WIDTH)/TOUCH_Y_RANGE;

			if(touch.touchDown == 0)
				touch.touchPressed = 1;

			touch.touchDown = 1;
		}
		else
		{
			if(touch.touchDown)
				touch.touchReleased = 1;

			touch.touchDown = 0;
		}

		// forward to STemWin
		if(touch.touchDown)
			GUI_TOUCH_StoreState(touch.touchX, touch.touchY);
		else
			GUI_TOUCH_StoreState(-1, -1);

		if(touch.touchPressed)
		{
			touch.touchPressed = 0;
			touchTime = chTimeNow();
			TFTSetBrightness(255);
		}

		if(chTimeElapsedSince(touchTime) > S2ST(5))
		{
			TFTSetBrightness(20);
		}
	}

	return 0;
}
