#include "stm32f4xx.h"
#include "misc.h"
#include "stm32f4xx_rcc.h"
#include "stm32f4xx_exti.h"
#include "stm32f4xx_gpio.h"
#include "stm32f4xx_flash.h"
#include "stm32f4xx_usart.h"
#include "stm32f4xx_iwdg.h"
#include "stm32f4xx_syscfg.h"

#include "core_cm4.h"

#include "hal/led.h"
#include "hal/hw_sync1.h"
#include "hal/usart2.h"
#include "hal/exti.h"
#include "hal/fsmc.h"
#include "hal/sram.h"
#include "hal/audio.h"
#include "hal/touch.h"
#include "hal/ov7725.h"

#include "cli.h"
#include "presenter.h"
#include "intercom1.h"
#include "intercom_constants.h"
#include "usb_host.h"

#include "util/boot.h"
#include "util/bootloader.h"
#include "util/event.h"
#include "util/flash_fs.h"

#include "ch.h"

#include "GUI.h"

#include <stdlib.h>

static void NVIC_Configuration();
extern void SystemClock_Config(void);

#define portNVIC_SYSTICK_CTRL		( ( volatile unsigned long * ) 0xe000e010 )
#define portNVIC_SYSTICK_LOAD		( ( volatile unsigned long * ) 0xe000e014 )
#define portNVIC_SYSTICK_CLK		0x00000004
#define portNVIC_SYSTICK_INT		0x00000002
#define portNVIC_SYSTICK_ENABLE		0x00000001

static WORKING_AREA(waUSART2, 256);
static WORKING_AREA(waLow, 4096);
static WORKING_AREA(waCLI, 8192);
#ifdef ENV_RUN
static WORKING_AREA(waTouch, 256);
#endif

#define EVENTS_LOW_MSGS 10
static msg_t lowBuf[EVENTS_LOW_MSGS];
static EventManager eventsLow;

int main()
{
#ifdef ENV_BOOT
	if(!BootIsBootloaderSelected())
		BootJumpToApplication();

	BootSetBootloaderSelected(0);
#endif

	SystemClock_Config();

	// Enable important clocks
	RCC_AHB1PeriphClockCmd( RCC_AHB1Periph_DMA1 | RCC_AHB1Periph_DMA2, ENABLE);
	RCC_AHB1PeriphClockCmd(	RCC_AHB1Periph_GPIOA | RCC_AHB1Periph_GPIOB | RCC_AHB1Periph_GPIOC | RCC_AHB1Periph_GPIOD |
							RCC_AHB1Periph_GPIOE | RCC_AHB1Periph_GPIOF | RCC_AHB1Periph_GPIOG, ENABLE);
	RCC_APB2PeriphClockCmd(RCC_APB2Periph_SYSCFG, ENABLE);
	RCC_AHB1PeriphClockCmd(RCC_AHB1Periph_CRC, ENABLE);

	SYSCFG_CompensationCellCmd(ENABLE);

	SCB->SHCSR |= SCB_SHCSR_USGFAULTENA_Msk | SCB_SHCSR_BUSFAULTENA_Msk | SCB_SHCSR_MEMFAULTENA_Msk;

	*(portNVIC_SYSTICK_LOAD) = ( 168000000UL / CH_FREQUENCY ) - 1UL;
	*(portNVIC_SYSTICK_CTRL) = portNVIC_SYSTICK_CLK | portNVIC_SYSTICK_INT | portNVIC_SYSTICK_ENABLE;

	chSysInit();

	NVIC_Configuration();

	CPULoadCalibrate(10);
	chThdSleep(14);

	// Init sync mechanism for timed distributed startup
	HWSync1Init();

	// Configure and start Independent watchdog
//	IWDG_WriteAccessCmd(IWDG_WriteAccess_Enable);
//	IWDG_SetPrescaler(IWDG_Prescaler_8);
//	IWDG_SetReload(2500);	// approx. 500ms
//	IWDG_WriteAccessCmd(IWDG_WriteAccess_Disable);

	EventManagerInit(&eventsLow, lowBuf, EVENTS_LOW_MSGS);

	// -- HAL
	LEDInit();
	FlashFSInit(0x08060000, 0x20000);	// 128k
	ExtiInit();
	FSMCInit();
	SRAMInit();
	USART2Init();
	AudioInit();
	TouchInit();
	Intercom1Init(&eventsLow);
	BootloaderSlaveInit(&intercom1);
	CLIInit();
#ifdef ENV_RUN
	PresenterInit(&eventsLow);
	OV7725Init();
	USBHInit();
#endif

	chThdCreateStatic(waUSART2, sizeof(waUSART2), NORMALPRIO, USARTTask, &usart2);
	chThdCreateStatic(waLow, sizeof(waLow), NORMALPRIO-1, EventManagerTask, &eventsLow);
	chThdCreateStatic(waCLI, sizeof(waCLI), NORMALPRIO-2, CLITask, 0);

#ifdef ENV_RUN
	chThdCreateStatic(waTouch, sizeof(waTouch), NORMALPRIO, TouchTask, 0);
#endif

	HWSyncBarrier(&hwSync1);

#ifdef ENV_BOOT
	ConsolePrint("--- Bootloader (MEDIA) ---\r\n");
#else
	ConsolePrint("--- Media ---\r\n");
#endif

	while(1)
	{
//		LEDToggle(LED_MAIN_GREEN);

		chThdSleepMilliseconds(500);
	}

	return 0;
}

static void NVIC_Configuration()
{
#ifdef ENV_BOOT
	NVIC_SetVectorTable(NVIC_VectTab_FLASH, 0x00);
#else
	NVIC_SetVectorTable(NVIC_VectTab_FLASH, BOOT_APP_OFFSET);
#endif
}
