/*
 * cli.c
 *
 *  Created on: 19.11.2012
 *      Author: AndreR
 */

#include "cli.h"

#include "util/console.h"
#include "intercom_constants.h"
#include "hal/led.h"
#include "intercom1.h"
#include "spi1.h"
#include "drive.h"
#include "arm_math.h"
#include "util/sys_time.h"
#include "util/flash.h"
#include "util/flash_fs.h"
#include "cpu_load.h"
#include "dwt.h"
#include "ctrl_foc.h"
#include "ctrl_6step.h"

CLI cli;

static void periodic(void* pUserData);
static void init(void* pUserData);
void processCommand();

#define CALIB_STEPS 100
#define CALIB_NUM_CFGS 6

#define CALIB_RESULT_COLUMnS 9
static float calibResults[CALIB_RESULT_COLUMnS*CALIB_STEPS];

typedef struct _CalibConfig
{
	uint32_t ccer;
	float voltMult[3];
	uint32_t resultIndex[2];	// for i0 and i1
} CalibConfig;

static float m1IMult[CALIB_RESULT_COLUMnS] = {0, 1, 1, -1, -1, 1, 1, -1, -1};
static float m2IMult[CALIB_RESULT_COLUMnS] = {0, -1, -1, 1, 1, -1, -1, 1, 1};
static float m3IMult[CALIB_RESULT_COLUMnS] = {0, -1, -1, 1, 1, -1, -1, 1, 1};
static float m4IMult[CALIB_RESULT_COLUMnS] = {0, -1, -1, 1, 1, -1, -1, 1, 1};

static CalibConfig m1[CALIB_NUM_CFGS] = {
		{ CCER_AB,  {1, 0, 0}, {0, 3} },
		{ CCER_AB,  {0, 1, 0}, {0, 1} },
		{ CCER_BC,  {0, 1, 0}, {7, 2} },
		{ CCER_BC,  {0, 0, 1}, {5, 4} },
		{ CCER_CA,  {0, 0, 1}, {6, 0} },
		{ CCER_CA,  {1, 0, 0}, {8, 0} },
};

static CalibConfig m2[CALIB_NUM_CFGS] = {
		{ CCER_AB,  {1, 0, 0}, {7, 1} },
		{ CCER_AB,  {0, 1, 0}, {5, 3} },
		{ CCER_BC,  {0, 1, 0}, {6, 0} },
		{ CCER_BC,  {0, 0, 1}, {8, 0} },
		{ CCER_CA,  {0, 0, 1}, {0, 4} },
		{ CCER_CA,  {1, 0, 0}, {0, 2} },
};

static CalibConfig m3[CALIB_NUM_CFGS] = {
		{ CCER_AB,  {1, 0, 0}, {3, 0} },
		{ CCER_AB,  {0, 1, 0}, {1, 0} },
		{ CCER_BC,  {0, 1, 0}, {2, 7} },
		{ CCER_BC,  {0, 0, 1}, {4, 5} },
		{ CCER_CA,  {0, 0, 1}, {0, 6} },
		{ CCER_CA,  {1, 0, 0}, {0, 8} },
};

static CalibConfig m4[CALIB_NUM_CFGS] = {
		{ CCER_AB,  {1, 0, 0}, {7, 1} },
		{ CCER_AB,  {0, 1, 0}, {5, 3} },
		{ CCER_BC,  {0, 1, 0}, {6, 0} },
		{ CCER_BC,  {0, 0, 1}, {8, 0} },
		{ CCER_CA,  {0, 0, 1}, {0, 4} },
		{ CCER_CA,  {1, 0, 0}, {0, 2} },
};

static void measDirect(Drive* pDrive, float maxVolt, float offset, CalibConfig* pConfig, float* pIMult, float* pResults)
{
	float inc = maxVolt/CALIB_STEPS;

	for(uint8_t cfg = 0; cfg < CALIB_NUM_CFGS; ++cfg)
	{
		CalibConfig* pCurCfg = &pConfig[cfg];

		ConsolePrint("Cfg %hu\r\n", (uint16_t)cfg);

		DriveSetPhaseVoltagesDirect(pDrive, maxVolt*pCurCfg->voltMult[0]+offset, maxVolt*pCurCfg->voltMult[1]+offset, maxVolt*pCurCfg->voltMult[2]+offset);
		DriveEnablePhase(pDrive, pCurCfg->ccer);
		chThdSleep(MS2ST(500));
		DriveSetPhaseVoltagesDirect(pDrive, 0, 0, 0);
		DriveEnablePhase(pDrive, CCER_ABC);
		chThdSleep(MS2ST(200));

//		int32_t step = 0;
//		for(float volts = 0; volts < maxVolt; volts += inc, ++step)
		int32_t step = CALIB_STEPS;
		for(float volts = maxVolt; volts > 0; volts -= inc, --step)
		{
			float iIdeal[2];
			float iReal[2];

			chThdSleep(MS2ST(500));
			DriveSetPhaseVoltagesDirect(pDrive, volts*pCurCfg->voltMult[0]+offset, volts*pCurCfg->voltMult[1]+offset, volts*pCurCfg->voltMult[2]+offset);
			DriveEnablePhase(pDrive, pCurCfg->ccer);
			chThdSleep(MS2ST(100));
			pDrive->measSample = 0;
			chThdSleep(MS2ST(20));
			DriveSetPhaseVoltagesDirect(pDrive, 0, 0, 0);
			DriveEnablePhase(pDrive, CCER_ABC);

			iReal[0] = 0;
			iReal[1] = 0;

			for(uint16_t n = 0; n < DRIVE_MEAS_SAMPLES; n++)
			{
				iReal[0] += pDrive->meas[0][n];
				iReal[1] += pDrive->meas[1][n];
			}

			iReal[0] /= DRIVE_MEAS_SAMPLES*1000;
			iReal[1] /= DRIVE_MEAS_SAMPLES*1000;

			iIdeal[0] = volts/1.2f*pIMult[pCurCfg->resultIndex[0]];
			iIdeal[1] = volts/1.2f*pIMult[pCurCfg->resultIndex[1]];

//			pResults[step*CALIB_RESULT_COLUMnS+pCurCfg->resultIndex[0]] = iIdeal[0]-iReal[0];
//			pResults[step*CALIB_RESULT_COLUMnS+pCurCfg->resultIndex[1]] = iIdeal[1]-iReal[1];
			pResults[step*CALIB_RESULT_COLUMnS+pCurCfg->resultIndex[0]] = iReal[0];
			pResults[step*CALIB_RESULT_COLUMnS+pCurCfg->resultIndex[1]] = iReal[1];
		}
	}

	for(uint16_t i = 0; i < CALIB_STEPS; i++)
	{
		pResults[i*CALIB_RESULT_COLUMnS+0] = i*inc;
	}
}

static void printCalibResult(float* pResults)
{
	for(uint16_t i = 0; i < CALIB_STEPS; i++)
	{
		ConsolePrint("%d, %d, %d, %d, %d, %d, %d, %d, %d\r\n",
				(int32_t)(pResults[i*CALIB_RESULT_COLUMnS+0]*1000), (int32_t)(pResults[i*CALIB_RESULT_COLUMnS+1]*1000), (int32_t)(pResults[i*CALIB_RESULT_COLUMnS+2]*1000),
				(int32_t)(pResults[i*CALIB_RESULT_COLUMnS+3]*1000), (int32_t)(pResults[i*CALIB_RESULT_COLUMnS+4]*1000),
				(int32_t)(pResults[i*CALIB_RESULT_COLUMnS+5]*1000), (int32_t)(pResults[i*CALIB_RESULT_COLUMnS+6]*1000),
				(int32_t)(pResults[i*CALIB_RESULT_COLUMnS+7]*1000), (int32_t)(pResults[i*CALIB_RESULT_COLUMnS+8]*1000));
		chThdSleep(MS2ST(2));
	}
}

void CLIInit(EventManager* pEventManager)
{
	chMtxInit(&cli.procBufMutex);

	cli.pEventClient = EventClientInit(pEventManager, 0, &periodic, MS2ST(20), 0);

	static uint8_t rxData[CONSOLE_RX_SIZE];
	static uint8_t txData[CONSOLE_TX_SIZE];

	ConsoleData conData;
	conData.pDataTx = txData;
	conData.txSize = CONSOLE_TX_SIZE;
	conData.pDataRx = rxData;
	conData.rxSize = CONSOLE_RX_SIZE;
	conData.printSize = CONSOLE_PRINT_SIZE;
	conData.txCallback = 0;
	conData.pUserData = 0;

	ConsoleInit(conData);

#ifdef MOT_LEFT
	ConsoleIntercomAdapterInit(&cli.adapter, &intercom1, INTERCOM_ADDRESS_LEFT, INTERCOM_ADDRESS_MAIN);
#else
	ConsoleIntercomAdapterInit(&cli.adapter, &intercom1, INTERCOM_ADDRESS_RIGHT, INTERCOM_ADDRESS_MAIN);
#endif

	init(0);
}

static void init(void* pUserData)
{
	(void)pUserData;

	LEDMustOff(LED_MAIN_GREEN);

#ifdef ENV_BOOT
#ifdef MOT_LEFT
	ConsolePrint("--- Bootloader (Left) ---\r\n");
#else
	ConsolePrint("--- Bootloader (Right) ---\r\n");
#endif
#else
#ifdef MOT_LEFT
	ConsolePrint("--- Left ---\r\n");
#else
	ConsolePrint("--- Right ---\r\n");
#endif
#endif
}

static void periodic(void* pUserData)
{
	(void)pUserData;

	if(ConsoleIsCmdReady())
	{
		processCommand();

		ConsoleCmdProcessed();
	}
}

void processCommand()
{
	uint16_t u16;

	if(ConsoleCmpCmd("test"))
	{
		ConsolePrint("Number: %d\r\n", 43);
	}

	if(ConsoleCmpCmd("test2"))
	{
		ConsolePrint("0123456789012345678901234567890123456789012345678901234567890123456789012345678901234567890123456789\r\n");
	}

	if(ConsoleCmpCmd("cpu"))
	{
//		DWT->CYCCNT = 0;

		CPULoad load = cpuLoad;

		ConsolePrint("CPU Load: %3u.%1u%%\r\n", load.load/10, load.load%10);
		ConsolePrint("Max: %3u.%1u%%\r\n", load.maxLoad/10, load.maxLoad%10);
		ConsolePrint("incPerTick: %10u/%u\r\n", load.incPerSysTick, load.maxIncPerSysTick);

//		uint32_t cycs = DWT->CYCCNT;
//
//		ConsolePrint("cyc: %u\r\n", cycs);
		chThdSleep(MS2ST(10));
		cpuLoad.maxLoad = 0;
	}

	if(ConsoleCmpCmd("tasks"))
	{
		ConsolePrint("Name            rPrio\tTime\r\n");

		for(Thread* pThread = chRegFirstThread(); pThread != 0; pThread = chRegNextThread(pThread))
		{
			ConsolePrint("%-16s%u\t%u\r\n", pThread->p_name, pThread->p_realprio, pThread->p_time);
		}
	}

	if(ConsoleCmpCmd("timsr"))
	{
		ConsolePrint("Resetting SR register\r\n");

		TIM1->SR = 0;
		TIM8->SR = 0;
	}

	if(ConsoleCmpCmd("tim"))
	{
		ConsolePrint("TIM1:\r\n");
		ConsolePrint("CR1: 0x%hX, SR: 0x%X, BDTR: 0x%X\r\n", TIM1->CR1, TIM1->SR, TIM1->BDTR);
		ConsolePrint("CCR: %hu, %hu, %hu\r\n", TIM1->CCR1, TIM1->CCR2, TIM1->CCR3);
		ConsolePrint("TIM8:\r\n");
		ConsolePrint("CR1: 0x%hX, SR: 0x%X, BDTR: 0x%X\r\n", TIM8->CR1, TIM8->SR, TIM8->BDTR);
		ConsolePrint("CCR: %hu, %hu, %hu\r\n", TIM8->CCR1, TIM8->CCR2, TIM8->CCR3);
		ConsolePrint("focPID: %hu, %hu\r\n", focRear.pid[0].enabled, focRear.pid[1].enabled, focRear.focOn);
		ConsolePrint("focPID: %hu, %hu\r\n", focFront.pid[0].enabled, focFront.pid[1].enabled, focFront.focOn);
		ConsolePrint("speedPID: %hu, %hu\r\n", focRear.pidSpeed.enabled, focRear.speedOn);
		ConsolePrint("speedPID: %hu, %hu\r\n", focFront.pidSpeed.enabled, focFront.speedOn);
	}

	if(ConsoleCmpCmd("pid"))
	{
		ConsolePrint("front:\r\n");
		ConsolePrint("i: %d, lm: %d, ov: %hu, ou: %d, s: %d\r\n",
				(int32_t)(focFront.pid[0].iTerm*1000), (int32_t)(focFront.pid[0].lastMeasurement*1000),
				(uint16_t)focFront.pid[0].overload, (int32_t)(focFront.pid[0].output*1000), (int32_t)(focFront.pid[0].setpoint*1000));
		ConsolePrint("i: %d, lm: %d, ov: %hu, ou: %d, s: %d\r\n",
				(int32_t)(focFront.pid[1].iTerm*1000), (int32_t)(focFront.pid[1].lastMeasurement*1000),
				(uint16_t)focFront.pid[1].overload, (int32_t)(focFront.pid[1].output*1000), (int32_t)(focFront.pid[1].setpoint*1000));
		ConsolePrint("i: %d, lm: %d, ov: %hu, ou: %d, s: %d\r\n",
				(int32_t)(focFront.pidSpeed.iTerm*1000), (int32_t)(focFront.pidSpeed.lastMeasurement*1000),
				(uint16_t)focFront.pidSpeed.overload, (int32_t)(focFront.pidSpeed.output*1000), (int32_t)(focFront.pidSpeed.setpoint*1000));
		ConsolePrint("inan: %d, onan: %d, iinf: %d, oinf: %d\r\n", isnan(focFront.pidSpeed.iTerm), isnan(focFront.pidSpeed.output),
				isinf(focFront.pidSpeed.iTerm), isinf(focFront.pidSpeed.output));
		ConsolePrint("p: %d, i: %d, d: %d\r\n",
				(int32_t)(focFront.pidSpeed.params.kp*1000), (int32_t)(focFront.pidSpeed.params.ki*1000),
				(int32_t)(focFront.pidSpeed.params.kd*1000));


		ConsolePrint("rear:\r\n");
		ConsolePrint("i: %d, lm: %d, ov: %hu, ou: %d, s: %d\r\n",
				(int32_t)(focRear.pid[0].iTerm*1000), (int32_t)(focRear.pid[0].lastMeasurement*1000),
				(uint16_t)focRear.pid[0].overload, (int32_t)(focRear.pid[0].output*1000), (int32_t)(focRear.pid[0].setpoint*1000));
		ConsolePrint("i: %d, lm: %d, ov: %hu, ou: %d, s: %d\r\n",
				(int32_t)(focRear.pid[1].iTerm*1000), (int32_t)(focRear.pid[1].lastMeasurement*1000),
				(uint16_t)focRear.pid[1].overload, (int32_t)(focRear.pid[1].output*1000), (int32_t)(focRear.pid[1].setpoint*1000));
		ConsolePrint("i: %d, lm: %d, ov: %hu, ou: %d, s: %d\r\n",
				(int32_t)(focRear.pidSpeed.iTerm*1000), (int32_t)(focRear.pidSpeed.lastMeasurement*1000),
				(uint16_t)focRear.pidSpeed.overload, (int32_t)(focRear.pidSpeed.output*1000), (int32_t)(focRear.pidSpeed.setpoint*1000));
		ConsolePrint("inan: %d, onan: %d, iinf: %d, oinf: %d\r\n", isnan(focRear.pidSpeed.iTerm), isnan(focRear.pidSpeed.output),
				isinf(focRear.pidSpeed.iTerm), isinf(focRear.pidSpeed.output));
		ConsolePrint("p: %d, i: %d, d: %d\r\n",
				(int32_t)(focRear.pidSpeed.params.kp*1000), (int32_t)(focRear.pidSpeed.params.ki*1000),
				(int32_t)(focRear.pidSpeed.params.kd*1000));
	}

	if(ConsoleScanCmd("power %hu", &u16) == 1)
	{
		ConsolePrint("Setting rear power %hu\r\n", u16);

		DriveSetPower(pRear, u16);
	}

	if(ConsoleCmpCmd("stubs"))
	{
		EXTI->SWIER = (1 << 3);
	}

	uint8_t mot;
	uint16_t vref, angle;
	if(ConsoleScanCmd("turn %hhu %hu", &mot, &vref) == 2)
	{
		Drive* pDrive = pFront;
		if(mot)
			pDrive = pRear;

		float power = vref/1000.0f;

		DriveSetVoltageAngle(pDrive, power, 0);

		chThdSleep(MS2ST(50));

		ConsolePrint("enc: %hu\r\n", pDrive->pTIMEncoder->CNT);

		for(float a = 0; a <= 2880; a += 5)
		{
			DriveSetVoltageAngle(pDrive, power, a);

//			angle = ((float)a)*PI/180.0;
//
//			DriveSetSpaceVector(&rear, power, angle);

			chThdSleep(MS2ST(20));
		}

		ConsolePrint("enc: %hu\r\n", pDrive->pTIMEncoder->CNT);

		DriveSetVoltageAngle(pDrive, 0, 0);
	}

	if(ConsoleScanCmd("va %hu %hu", &vref, &angle) == 2)
	{
//		float a = ((float)angle)*PI/180.0;
		float r = ((float)vref)/1000.0f;

		DriveSetVoltageAngle(pRear, r, angle);
	}

	if(ConsoleCmpCmd("adc"))
	{
		for(uint8_t a = 0; a < 2; a++)
		{
			ConsolePrint("ADC%hhu: ", a);
			for(uint8_t i = 0; i < DRIVE_ADC_SAMPLES; i++)
			{
				ConsolePrint("%5hu ", pRear->adc[a][i]);
			}
			ConsolePrint("\r\n");
		}
	}

	if(ConsoleCmpCmd("current"))
	{
		ConsolePrint("iA: %d, iB: %d, iC: %d\r\n", (int32_t)(pRear->i[0]*1000), (int32_t)(pRear->i[1]*1000), (int32_t)(pRear->i[2]*1000));
		ConsolePrint("iAlpha: %d, iBeta: %d\r\n", (int32_t)(pRear->iAlpha*1000), (int32_t)(pRear->iBeta*1000));
//		ConsolePrint("iD: %d, iQ: %d\r\n", (int32_t)(rear.pOut->iD*1000), (int32_t)(rear.pOut->iQ*1000));
	}

	if(ConsoleCmpCmd("enc"))
	{
		float angle = (pRear->pTIMEncoder->CNT%256)*1.40625f;
		ConsolePrint("ENC pos: %u\r\n", (uint32_t)(angle*1000));
	}

	int32_t q;
	if(ConsoleScanCmd("foc %d", &q) == 1)
	{
		ConsolePrint("Setting FOC Q: %d\r\n", q);

		focRear.pid[1].setpoint = q*0.001f;
//		rear.setIq = q*0.001f;
//		front.setIq = q*0.001f;

//		DriveSetFOC(&rear, 0, ((float)q)/1000.0f);
	}

	if(ConsoleCmpCmd("focon"))
	{
		ConsolePrint("Enabling Field-Oriented Control!\r\n");

		CtrlFOCEnableFOC(&focRear, 1);
		CtrlFOCEnableFOC(&focFront, 1);
	}

	if(ConsoleCmpCmd("focoff"))
	{
		ConsolePrint("Disabling Field-Oriented Control!\r\n");


		CtrlFOCEnableFOC(&focRear, 0);
		CtrlFOCEnableFOC(&focFront, 0);
	}

	if(ConsoleCmpCmd("speedon"))
	{
		ConsolePrint("Enabling Speed Control!\r\n");

		CtrlFOCEnableSpeedControl(&focRear, 1);
		CtrlFOCEnableSpeedControl(&focFront, 1);
	}

	if(ConsoleCmpCmd("speedoff"))
	{
		ConsolePrint("Disabling Speed Control!\r\n");

		CtrlFOCEnableSpeedControl(&focRear, 0);
		CtrlFOCEnableSpeedControl(&focFront, 0);
	}

/*	if(ConsoleCmpCmd("input"))
	{
		DriveMain2Mot* pFront = &spi1.pRxData->periodic.mot[0];
		DriveMain2Mot* pRear = &spi1.pRxData->periodic.mot[1];

		ConsolePrint("Front:\r\n");
		ConsolePrint("Vel: %d\r\niD:  %d\r\niQ:  %d\r\n",
				(int32_t)(pFront->velocity*1000), (int32_t)(pFront->iD*1000), (int32_t)(pFront->iQ*1000));
		ConsolePrint("Rear:\r\n");
		ConsolePrint("Vel: %d\r\niD:  %d\r\niQ:  %d\r\n",
				(int32_t)(pRear->velocity*1000), (int32_t)(pRear->iD*1000), (int32_t)(pRear->iQ*1000));
	}

	if(ConsoleCmpCmd("output"))
	{
		ConsolePrint("Front:\r\n");
		ConsolePrint("Vel: %d\r\niD:  %d\r\niQ:  %d\r\n",
				(int32_t)(front.pOut->velocity*1000), (int32_t)(front.pOut->iD*1000), (int32_t)(front.pOut->iQ*1000));
		ConsolePrint("Rear:\r\n");
		ConsolePrint("Vel: %d\r\niD:  %d\r\niQ:  %d\r\n",
				(int32_t)(rear.pOut->velocity*1000), (int32_t)(rear.pOut->iD*1000), (int32_t)(rear.pOut->iQ*1000));
	}
*/

	uint32_t p, i, d;
	if(ConsoleScanCmd("focpid %u %u %u", &p, &i, &d) == 3)
	{
		ConsolePrint("Setting FOC PID: %u, %u, %u\r\n", p, i, d);

		CtrlFOCSetFOCPIDParams(&focRear, p*0.001f, i*0.001f, d*0.001f);
		CtrlFOCSetFOCPIDParams(&focFront, p*0.001f, i*0.001f, d*0.001f);
	}

	if(ConsoleScanCmd("speedpid %u %u %u", &p, &i, &d) == 3)
	{
		ConsolePrint("Setting Speed PID: %u, %u, %u\r\n", p, i, d);

		CtrlFOCSetSpeedPIDParams(&focRear, p*0.001f, i*0.001f, d*0.001f);
		CtrlFOCSetSpeedPIDParams(&focFront, p*0.001f, i*0.001f, d*0.001f);
	}

	if(ConsoleCmpCmd("dt"))
	{
		ConsolePrint("dT: %hu\r\n", pRear->encDt);
	}

	if(ConsoleCmpCmd("bat"))
	{
//		ConsolePrint("Bat: %umV\r\n", (uint16_t)(spi1.pRxData->periodic.bat*1000));
	}

	if(ConsoleCmpCmd("state"))
	{
//		ConsolePrint("Speed: %d\r\n", (int32_t)(rear.pOut->velocity*1000));
	}

//	if(ConsoleScanCmd("speed %d", &q) == 1)
//	{
//		ConsolePrint("Setting speed to %d\r\n", q);
//
//		DriveSetSpeed(&rear, q/10.0f);
//	}

	if(ConsoleCmpCmd("show calib"))
	{
		ConsolePrint("Front: %d, %d, %d\r\n", (int32_t)(pFront->tCal.offset[0]*1000),
				(int32_t)(pFront->tCal.offset[1]*1000), (int32_t)(pFront->tCal.offset[2]*1000));
		ConsolePrint("Rear: %d, %d, %d\r\n", (int32_t)(pRear->tCal.offset[0]*1000),
				(int32_t)(pRear->tCal.offset[1]*1000), (int32_t)(pRear->tCal.offset[2]*1000));
	}

	int16_t t;
	if(ConsoleScanCmd("temp calib %hd", &t) == 1)
	{
		float temp = (float)t;

		DriveCalibrateCurrentForTemp(pFront, temp);
		ConsolePrint("Front: %d, %d, %d\r\n", (int32_t)(pFront->tCal.offset[0]*1000),
				(int32_t)(pFront->tCal.offset[1]*1000), (int32_t)(pFront->tCal.offset[2]*1000));

		DriveCalibrateCurrentForTemp(pRear, temp);
		ConsolePrint("Rear: %d, %d, %d\r\n", (int32_t)(pRear->tCal.offset[0]*1000),
				(int32_t)(pRear->tCal.offset[1]*1000), (int32_t)(pRear->tCal.offset[2]*1000));
	}

	if(ConsoleCmpCmd("temp"))
	{
		ConsolePrint("Front: %3d, %3d, %3d\t(%u)\r\n", (int32_t)pFront->t[0], (int32_t)pFront->t[1], (int32_t)pFront->t[2], (SysTimeUSec()-pRear->tMeasTime)/1000);
		ConsolePrint("Rear:  %3d, %3d, %3d\t(%u)\r\n", (int32_t)pRear->t[0], (int32_t)pRear->t[1], (int32_t)pRear->t[2], (SysTimeUSec()-pRear->tMeasTime)/1000);

//		if(DriveTempMeas(pRear))
//		{
//			ConsolePrint("Temp meas interrupted\r\n");
//		}
//		else
//		{
//			ConsolePrint("Temps: %d, %d, %d\r\n", (int32_t)(pRear->t[0]*1000), (int32_t)(pRear->t[1]*1000), (int32_t)(pRear->t[2]*1000));
//		}
	}

	if(ConsoleCmpCmd("time"))
	{
		ConsolePrint("Time: %uus\r\n", SysTimeUSec());
	}

	int16_t cur;
	if(ConsoleScanCmd("meas %hd", &cur) == 1)
	{
		ConsolePrint("Starting measurement with %hdmA\r\n", cur);

		float amps = cur*0.001f;
		for(float amps = 0; amps < 4.0f; amps += 0.05f)
		{
			float iIdealFront[3];
			float iRealFront[3];
			float iIdealRear[3];
			float iRealRear[3];

			DriveSetPhaseVoltagesSVPWM(pFront, amps, 0, 0);
			DriveSetPhaseVoltagesSVPWM(pRear, amps, 0, 0);
			chThdSleep(MS2ST(100));
			pFront->measSample = 0;
			pRear->measSample = 0;
			DriveGetSteadyPhaseCurrents(pFront, iIdealFront);
			DriveGetSteadyPhaseCurrents(pRear, iIdealRear);
			chThdSleep(MS2ST(10));
			DriveSetPhaseVoltagesSVPWM(pFront, 0, 0, 0);
			DriveSetPhaseVoltagesSVPWM(pRear, 0, 0, 0);

			for(uint16_t n = 0; n < DRIVE_MEAS_SAMPLES; n++)
			{
				iRealFront[0] += pFront->meas[0][n];
				iRealFront[1] += pFront->meas[1][n];
				iRealFront[2] += -pFront->meas[0][n]-pFront->meas[1][n];
				iRealRear[0] += pRear->meas[0][n];
				iRealRear[1] += pRear->meas[1][n];
				iRealRear[2] += -pRear->meas[0][n]-pRear->meas[1][n];
			}

			for(uint8_t i = 0; i < 3; i++)
			{
				iRealFront[i] /= DRIVE_MEAS_SAMPLES*1000;
				iRealRear[i] /= DRIVE_MEAS_SAMPLES*1000;
			}

			ConsolePrint("I: %d, %d, %d, %d, %d\r\n", (int32_t)(amps*1000),
					(int32_t)((iIdealFront[1]-iRealFront[1])*1000),
					(int32_t)((iIdealFront[2]-iRealFront[2])*1000),
					(int32_t)((iIdealRear[0]-iRealRear[0])*1000),
					(int32_t)((iIdealRear[1]-iRealRear[1])*1000));
		}

		ConsolePrint("Done\r\n");
	}

	if(ConsoleCmpCmd("get meas"))
	{
		for(uint16_t i = 0; i < DRIVE_MEAS_SAMPLES; i++)
		{
			ConsolePrint("%hd\t%hd\t%hd\t%hd\t%hd\t%hd\r\n",
					pFront->meas[0][i], pFront->meas[1][i], -pFront->meas[0][i]-pFront->meas[1][i],
					pRear->meas[0][i], pRear->meas[1][i], -pRear->meas[0][i]-pRear->meas[1][i]);
			chThdSleep(MS2ST(10));
		}
	}

	if(ConsoleCmpCmd("hall"))
	{
		ConsolePrint("Front: %hu\tRear: %hu\r\n", pFront->halPos, pRear->halPos);
	}

	if(ConsoleCmpCmd("ctrl none"))
	{
		DriveSetController(0, 0);

		ConsolePrint("Controller disabled\r\n");
	}

	if(ConsoleCmpCmd("ctrl foc"))
	{
		DriveSetController(&ctrlFocFront, &ctrlFocRear);

		ConsolePrint("FOC controller active\r\n");
	}

	if(ConsoleCmpCmd("ctrl 6step"))
	{
		DriveSetController(&ctrl6StepFront, &ctrl6StepRear);

		ConsolePrint("6-Step controller active\r\n");
	}

	if(ConsoleScanCmd("step %hu", &u16) == 1)
	{
		ConsolePrint("Setting voltage to %humV\r\n", u16);

		float volt = u16*0.001f;

		stepRear.setVoltage = volt;
		stepFront.setVoltage = volt;
	}

	if(ConsoleScanCmd("ph %hu", &u16) == 1)
	{
		ConsolePrint("Phase: %hu\r\n", u16);

		Ctrl6StepSetPhase(pRear, u16);
	}

	int16_t i16;
	if(ConsoleScanCmd("vel %hd", &i16) == 1)
	{
		stepRear.setVelocity = (float)i16;
//		stepFront.setVelocity = (float)i16;
	}

	uint16_t u162;
	if(ConsoleScanCmd("calib front %hu %hu", &u16, &u162) == 2)
	{
		float volt = u16*0.001f;
		float offset = u162*0.001f;

		ConsolePrint("Calibration...");

#ifdef MOT_RIGHT
		measDirect(pFront, volt, offset, m1, m1IMult, calibResults);
#else
		measDirect(pFront, volt, offset, m2, m2IMult, calibResults);
#endif

		ConsolePrint("Done\r\n");

		printCalibResult(calibResults);
	}

	if(ConsoleScanCmd("calib rear %hu %hu", &u16, &u162) == 2)
	{
		float volt = u16*0.001f;
		float offset = u162*0.001f;

		ConsolePrint("Calibration...");

#ifdef MOT_RIGHT
		measDirect(pRear, volt, offset, m4, m4IMult, calibResults);
#else
		measDirect(pRear, volt, offset, m3, m3IMult, calibResults);
#endif

		ConsolePrint("Done\r\n");

		printCalibResult(calibResults);
	}

	if(ConsoleScanCmd("frontpwr %hu", &u16) == 1)
	{
		ConsolePrint("Front PWR: %humV\r\n", u16);

		float volt = u16*0.001f;

		DriveSetPhaseVoltagesDirect(pFront, 6.0f+volt, 6.0f, 6.0f);
		DriveEnablePhase(pFront, CCER_AB);
		chThdSleep(MS2ST(500));
		pFront->measSample = 0;
	}

	if(ConsoleScanCmd("rearpwr %hu", &u16) == 1)
	{
		ConsolePrint("Rear PWR: %humV\r\n", u16);

		float volt = u16*0.001f;

		DriveSetPhaseVoltagesDirect(pRear, volt, volt, volt);
		DriveEnablePhase(pRear, CCER_AB);
		chThdSleep(MS2ST(500));
		pRear->measSample = 0;
	}

	if(ConsoleCmpCmd("rearsweep"))
	{
		ConsolePrint("Rear voltage sweep\r\n");

		DriveEnablePhase(pRear, CCER_ABC);

		for(float volt = 0.0f; volt < 1.0f; volt += 0.01f)
		{
			DriveSetPhaseVoltagesDirect(pRear, volt, volt, volt);
			chThdSleep(MS2ST(1));
		}

		DriveSetPhaseVoltagesDirect(pRear, 0, 0, 0);
	}

	if(ConsoleCmpCmd("frontsweep"))
	{
		ConsolePrint("Front voltage sweep\r\n");

		DriveEnablePhase(pFront, CCER_ABC);

		for(float volt = 0.0f; volt < 1.0f; volt += 0.01f)
		{
			DriveSetPhaseVoltagesDirect(pFront, volt, volt, volt);
			chThdSleep(MS2ST(1));
		}

		DriveSetPhaseVoltagesDirect(pFront, 0, 0, 0);
	}

	if(ConsoleCmpCmd("frontasweep"))
	{
		ConsolePrint("Front amp sweep\r\n");

		DriveEnablePhase(pFront, CCER_AB);

		for(float volt = 0.0f; volt < 1.0f; volt += 0.01f)
		{
			DriveSetPhaseVoltagesDirect(pFront, volt+2.0f, 2.0f, 2.0f);
			chThdSleep(MS2ST(1));
		}

		DriveSetPhaseVoltagesDirect(pFront, 0, 0, 0);
	}

	if(ConsoleCmpCmd("flash hex"))
	{
		for(uint32_t i = 0; i < flashFS.bytesDel+flashFS.bytesUsed+48; i += 8)
		{
			uint8_t* d = (uint8_t*)(flashFS.flashAddr+i);
			ConsolePrint("0x%04X: %02X %02X %02X %02X %02X %02X %02X %02X\r\n",
					i, d[0], d[1], d[2], d[3], d[4], d[5], d[6], d[7]);
		}
	}

	if(ConsoleCmpCmd("flash list"))
	{
		FlashFSListFiles();
	}

	if(ConsoleCmpCmd("flash erase"))
	{
		ConsolePrint("Erasing flash\r\n");

		FlashErase(flashFS.flashAddr, flashFS.flashAddr+flashFS.flashSize);
	}
}
