/*
 * ctrl_foc.c
 *
 *  Created on: 11.06.2014
 *      Author: AndreR
 */

#include "ctrl_foc.h"

#include "ch.h"

CtrlFOC focFront;
CtrlFOC focRear;

static void onInit(Drive* pDrive);
static void onExit(Drive* pDrive);
static void onCurrentUpdate(Drive* pDrive);
static void onSpeedUpdate(Drive* pDrive);
static void onHallUpdateCb(Drive* pDrive);

DriveCtrl ctrlFocFront = {&onInit, &onExit, &onCurrentUpdate, &onSpeedUpdate, &onHallUpdateCb, &focFront};
DriveCtrl ctrlFocRear  = {&onInit, &onExit, &onCurrentUpdate, &onSpeedUpdate, &onHallUpdateCb, &focRear };

static float cos256[];
static float sin256[];

IN_CCM static void onCurrentUpdate(Drive* pDrive)
{
	CtrlFOC* pFOC = (CtrlFOC*)pDrive->pCtrl->pUserData;

	uint8_t pos = pDrive->pTIMEncoder->CNT%256;
	float sinPos = sin256[pos];
	float cosPos = cos256[pos];

	float iD, iQ;

	iD = pDrive->iAlpha*cosPos + pDrive->iBeta*sinPos;
	iQ = -pDrive->iAlpha*sinPos + pDrive->iBeta*cosPos;

	iD *= 0.81649658092773f;	// sqrt(3/2)
	iQ *= 0.81649658092773f;

	pFOC->curId = iD;
	pFOC->curIq = iQ;

	// 0 = vD, 1 = vQ
	PIDUpdate(&pFOC->pid[0], iD);
	PIDUpdate(&pFOC->pid[1], iQ);

	float vD = pFOC->pid[0].output;
	float vQ = pFOC->pid[1].output;

	float vA = vD*cosPos - vQ*sinPos;
	float vB = vD*sinPos + vQ*cosPos;

	vA *= 1.224744871391589f;	// sqrt(2/3)
	vB *= 1.224744871391589f;

	if(pFOC->focOn)
		DriveSetAlphaBeta(pDrive, vA, vB);
}

IN_CCM static void onSpeedUpdate(Drive* pDrive)
{
	CtrlFOC* pFOC = (CtrlFOC*)pDrive->pCtrl->pUserData;

	if(pFOC->speedOn)
	{
		pFOC->pidSpeed.setpoint = pFOC->setVelocity;
		PIDUpdate(&pFOC->pidSpeed, pDrive->curVelocity);

		pFOC->setIq = pFOC->pidSpeed.output;
		pFOC->pid[1].setpoint = pFOC->setIq;
	}
	else
	{
		pFOC->pid[1].setpoint = pFOC->setIq;
	}
}

IN_CCM static void onHallUpdateCb(Drive* pDrive)
{
	(void)pDrive;
}

static void onInit(Drive* pDrive)
{
	CtrlFOC* pFOC = (CtrlFOC*)pDrive->pCtrl->pUserData;

	PIDParams focPIDParams;
	focPIDParams.kp = 0.5f;
	focPIDParams.ki = 0.1f;
	focPIDParams.kd = 0.0f;
	focPIDParams.outputMin = -2.0f;	// [V]
	focPIDParams.outputMax = 2.0f;	// [V]

	for(uint8_t i = 0; i < 2; i++)
	{
		PIDInit(&pFOC->pid[i], &focPIDParams, 0.00005f, 1);
	}

	PIDParams speedPIDParams;
	speedPIDParams.kp = 0.25f;
	speedPIDParams.ki = 0.025f;
	speedPIDParams.kd = 0.0f;
	speedPIDParams.outputMin = -2.0f;	// [A]
	speedPIDParams.outputMax = 2.0f;	// [A]

	PIDInit(&pFOC->pidSpeed, &speedPIDParams, 0.001f, 1);

	chThdSleep(MS2ST(10));

	DriveSetVoltageAngle(pDrive, 1.0f, 0.0f);
	chThdSleep(MS2ST(500));
	pDrive->pTIMEncoder->CNT = 0;
	DriveSetPower(pDrive, 0);

	CtrlFOCEnableFOC(pFOC, 1);
	CtrlFOCEnableSpeedControl(pFOC, 1);
}

static void onExit(Drive* pDrive)
{
	CtrlFOC* pFOC = (CtrlFOC*)pDrive->pCtrl->pUserData;

	CtrlFOCEnableFOC(pFOC, 0);
	CtrlFOCEnableSpeedControl(pFOC, 0);
}

void CtrlFOCEnableFOC(CtrlFOC* pFOC, uint8_t enable)
{
	PIDEnable(&pFOC->pid[0], enable);
	PIDEnable(&pFOC->pid[1], enable);

	pFOC->focOn = enable;
}

void CtrlFOCEnableSpeedControl(CtrlFOC* pFOC, uint8_t enable)
{
	PIDEnable(&pFOC->pidSpeed, enable);

	pFOC->speedOn = enable;
}

void CtrlFOCSetFOCPIDParams(CtrlFOC* pFOC, float p, float i, float d)
{
	for(uint8_t x = 0; x < 2; x++)
	{
		PIDSetParams(&pFOC->pid[x], p, i, d);
	}
}

void CtrlFOCSetSpeedPIDParams(CtrlFOC* pFOC, float p, float i, float d)
{
	PIDSetParams(&pFOC->pidSpeed, p, i, d);
}

/**
 * DATA Tables
 */
static float sin256[] = {
	0.0000000000f, 0.0245412285f, 0.0490676743f, 0.0735645636f,
	0.0980171403f, 0.1224106752f, 0.1467304745f, 0.1709618888f,
	0.1950903220f, 0.2191012402f, 0.2429801799f, 0.2667127575f,
	0.2902846773f, 0.3136817404f, 0.3368898534f, 0.3598950365f,
	0.3826834324f, 0.4052413140f, 0.4275550934f, 0.4496113297f,
	0.4713967368f, 0.4928981922f, 0.5141027442f, 0.5349976199f,
	0.5555702330f, 0.5758081914f, 0.5956993045f, 0.6152315906f,
	0.6343932842f, 0.6531728430f, 0.6715589548f, 0.6895405447f,
	0.7071067812f, 0.7242470830f, 0.7409511254f, 0.7572088465f,
	0.7730104534f, 0.7883464276f, 0.8032075315f, 0.8175848132f,
	0.8314696123f, 0.8448535652f, 0.8577286100f, 0.8700869911f,
	0.8819212643f, 0.8932243012f, 0.9039892931f, 0.9142097557f,
	0.9238795325f, 0.9329927988f, 0.9415440652f, 0.9495281806f,
	0.9569403357f, 0.9637760658f, 0.9700312532f, 0.9757021300f,
	0.9807852804f, 0.9852776424f, 0.9891765100f, 0.9924795346f,
	0.9951847267f, 0.9972904567f, 0.9987954562f, 0.9996988187f,
	1.0000000000f, 0.9996988187f, 0.9987954562f, 0.9972904567f,
	0.9951847267f, 0.9924795346f, 0.9891765100f, 0.9852776424f,
	0.9807852804f, 0.9757021300f, 0.9700312532f, 0.9637760658f,
	0.9569403357f, 0.9495281806f, 0.9415440652f, 0.9329927988f,
	0.9238795325f, 0.9142097557f, 0.9039892931f, 0.8932243012f,
	0.8819212643f, 0.8700869911f, 0.8577286100f, 0.8448535652f,
	0.8314696123f, 0.8175848132f, 0.8032075315f, 0.7883464276f,
	0.7730104534f, 0.7572088465f, 0.7409511254f, 0.7242470830f,
	0.7071067812f, 0.6895405447f, 0.6715589548f, 0.6531728430f,
	0.6343932842f, 0.6152315906f, 0.5956993045f, 0.5758081914f,
	0.5555702330f, 0.5349976199f, 0.5141027442f, 0.4928981922f,
	0.4713967368f, 0.4496113297f, 0.4275550934f, 0.4052413140f,
	0.3826834324f, 0.3598950365f, 0.3368898534f, 0.3136817404f,
	0.2902846773f, 0.2667127575f, 0.2429801799f, 0.2191012402f,
	0.1950903220f, 0.1709618888f, 0.1467304745f, 0.1224106752f,
	0.0980171403f, 0.0735645636f, 0.0490676743f, 0.0245412285f,
	0.0000000000f, -0.0245412285f, -0.0490676743f, -0.0735645636f,
	-0.0980171403f, -0.1224106752f, -0.1467304745f, -0.1709618888f,
	-0.1950903220f, -0.2191012402f, -0.2429801799f, -0.2667127575f,
	-0.2902846773f, -0.3136817404f, -0.3368898534f, -0.3598950365f,
	-0.3826834324f, -0.4052413140f, -0.4275550934f, -0.4496113297f,
	-0.4713967368f, -0.4928981922f, -0.5141027442f, -0.5349976199f,
	-0.5555702330f, -0.5758081914f, -0.5956993045f, -0.6152315906f,
	-0.6343932842f, -0.6531728430f, -0.6715589548f, -0.6895405447f,
	-0.7071067812f, -0.7242470830f, -0.7409511254f, -0.7572088465f,
	-0.7730104534f, -0.7883464276f, -0.8032075315f, -0.8175848132f,
	-0.8314696123f, -0.8448535652f, -0.8577286100f, -0.8700869911f,
	-0.8819212643f, -0.8932243012f, -0.9039892931f, -0.9142097557f,
	-0.9238795325f, -0.9329927988f, -0.9415440652f, -0.9495281806f,
	-0.9569403357f, -0.9637760658f, -0.9700312532f, -0.9757021300f,
	-0.9807852804f, -0.9852776424f, -0.9891765100f, -0.9924795346f,
	-0.9951847267f, -0.9972904567f, -0.9987954562f, -0.9996988187f,
	-1.0000000000f, -0.9996988187f, -0.9987954562f, -0.9972904567f,
	-0.9951847267f, -0.9924795346f, -0.9891765100f, -0.9852776424f,
	-0.9807852804f, -0.9757021300f, -0.9700312532f, -0.9637760658f,
	-0.9569403357f, -0.9495281806f, -0.9415440652f, -0.9329927988f,
	-0.9238795325f, -0.9142097557f, -0.9039892931f, -0.8932243012f,
	-0.8819212643f, -0.8700869911f, -0.8577286100f, -0.8448535652f,
	-0.8314696123f, -0.8175848132f, -0.8032075315f, -0.7883464276f,
	-0.7730104534f, -0.7572088465f, -0.7409511254f, -0.7242470830f,
	-0.7071067812f, -0.6895405447f, -0.6715589548f, -0.6531728430f,
	-0.6343932842f, -0.6152315906f, -0.5956993045f, -0.5758081914f,
	-0.5555702330f, -0.5349976199f, -0.5141027442f, -0.4928981922f,
	-0.4713967368f, -0.4496113297f, -0.4275550934f, -0.4052413140f,
	-0.3826834324f, -0.3598950365f, -0.3368898534f, -0.3136817404f,
	-0.2902846773f, -0.2667127575f, -0.2429801799f, -0.2191012402f,
	-0.1950903220f, -0.1709618888f, -0.1467304745f, -0.1224106752f,
	-0.0980171403f, -0.0735645636f, -0.0490676743f, -0.0245412285f,
};

static float cos256[] = {
	1.0000000000f, 0.9996988187f, 0.9987954562f, 0.9972904567f,
	0.9951847267f, 0.9924795346f, 0.9891765100f, 0.9852776424f,
	0.9807852804f, 0.9757021300f, 0.9700312532f, 0.9637760658f,
	0.9569403357f, 0.9495281806f, 0.9415440652f, 0.9329927988f,
	0.9238795325f, 0.9142097557f, 0.9039892931f, 0.8932243012f,
	0.8819212643f, 0.8700869911f, 0.8577286100f, 0.8448535652f,
	0.8314696123f, 0.8175848132f, 0.8032075315f, 0.7883464276f,
	0.7730104534f, 0.7572088465f, 0.7409511254f, 0.7242470830f,
	0.7071067812f, 0.6895405447f, 0.6715589548f, 0.6531728430f,
	0.6343932842f, 0.6152315906f, 0.5956993045f, 0.5758081914f,
	0.5555702330f, 0.5349976199f, 0.5141027442f, 0.4928981922f,
	0.4713967368f, 0.4496113297f, 0.4275550934f, 0.4052413140f,
	0.3826834324f, 0.3598950365f, 0.3368898534f, 0.3136817404f,
	0.2902846773f, 0.2667127575f, 0.2429801799f, 0.2191012402f,
	0.1950903220f, 0.1709618888f, 0.1467304745f, 0.1224106752f,
	0.0980171403f, 0.0735645636f, 0.0490676743f, 0.0245412285f,
	0.0000000000f, -0.0245412285f, -0.0490676743f, -0.0735645636f,
	-0.0980171403f, -0.1224106752f, -0.1467304745f, -0.1709618888f,
	-0.1950903220f, -0.2191012402f, -0.2429801799f, -0.2667127575f,
	-0.2902846773f, -0.3136817404f, -0.3368898534f, -0.3598950365f,
	-0.3826834324f, -0.4052413140f, -0.4275550934f, -0.4496113297f,
	-0.4713967368f, -0.4928981922f, -0.5141027442f, -0.5349976199f,
	-0.5555702330f, -0.5758081914f, -0.5956993045f, -0.6152315906f,
	-0.6343932842f, -0.6531728430f, -0.6715589548f, -0.6895405447f,
	-0.7071067812f, -0.7242470830f, -0.7409511254f, -0.7572088465f,
	-0.7730104534f, -0.7883464276f, -0.8032075315f, -0.8175848132f,
	-0.8314696123f, -0.8448535652f, -0.8577286100f, -0.8700869911f,
	-0.8819212643f, -0.8932243012f, -0.9039892931f, -0.9142097557f,
	-0.9238795325f, -0.9329927988f, -0.9415440652f, -0.9495281806f,
	-0.9569403357f, -0.9637760658f, -0.9700312532f, -0.9757021300f,
	-0.9807852804f, -0.9852776424f, -0.9891765100f, -0.9924795346f,
	-0.9951847267f, -0.9972904567f, -0.9987954562f, -0.9996988187f,
	-1.0000000000f, -0.9996988187f, -0.9987954562f, -0.9972904567f,
	-0.9951847267f, -0.9924795346f, -0.9891765100f, -0.9852776424f,
	-0.9807852804f, -0.9757021300f, -0.9700312532f, -0.9637760658f,
	-0.9569403357f, -0.9495281806f, -0.9415440652f, -0.9329927988f,
	-0.9238795325f, -0.9142097557f, -0.9039892931f, -0.8932243012f,
	-0.8819212643f, -0.8700869911f, -0.8577286100f, -0.8448535652f,
	-0.8314696123f, -0.8175848132f, -0.8032075315f, -0.7883464276f,
	-0.7730104534f, -0.7572088465f, -0.7409511254f, -0.7242470830f,
	-0.7071067812f, -0.6895405447f, -0.6715589548f, -0.6531728430f,
	-0.6343932842f, -0.6152315906f, -0.5956993045f, -0.5758081914f,
	-0.5555702330f, -0.5349976199f, -0.5141027442f, -0.4928981922f,
	-0.4713967368f, -0.4496113297f, -0.4275550934f, -0.4052413140f,
	-0.3826834324f, -0.3598950365f, -0.3368898534f, -0.3136817404f,
	-0.2902846773f, -0.2667127575f, -0.2429801799f, -0.2191012402f,
	-0.1950903220f, -0.1709618888f, -0.1467304745f, -0.1224106752f,
	-0.0980171403f, -0.0735645636f, -0.0490676743f, -0.0245412285f,
	-0.0000000000f, 0.0245412285f, 0.0490676743f, 0.0735645636f,
	0.0980171403f, 0.1224106752f, 0.1467304745f, 0.1709618888f,
	0.1950903220f, 0.2191012402f, 0.2429801799f, 0.2667127575f,
	0.2902846773f, 0.3136817404f, 0.3368898534f, 0.3598950365f,
	0.3826834324f, 0.4052413140f, 0.4275550934f, 0.4496113297f,
	0.4713967368f, 0.4928981922f, 0.5141027442f, 0.5349976199f,
	0.5555702330f, 0.5758081914f, 0.5956993045f, 0.6152315906f,
	0.6343932842f, 0.6531728430f, 0.6715589548f, 0.6895405447f,
	0.7071067812f, 0.7242470830f, 0.7409511254f, 0.7572088465f,
	0.7730104534f, 0.7883464276f, 0.8032075315f, 0.8175848132f,
	0.8314696123f, 0.8448535652f, 0.8577286100f, 0.8700869911f,
	0.8819212643f, 0.8932243012f, 0.9039892931f, 0.9142097557f,
	0.9238795325f, 0.9329927988f, 0.9415440652f, 0.9495281806f,
	0.9569403357f, 0.9637760658f, 0.9700312532f, 0.9757021300f,
	0.9807852804f, 0.9852776424f, 0.9891765100f, 0.9924795346f,
	0.9951847267f, 0.9972904567f, 0.9987954562f, 0.9996988187f,
};
