/*
 * drive.c
 *
 *  Created on: 02.02.2014
 *      Author: AndreR
 */

/**	Timer/ADC/Motor assignments
 *
 * 			Right:	Left:
 * 	TIM1:	M1		M3
 * 	TIM2:	ENC4	ENC2
 * 	TIM4:	ENC1	ENC3
 * 	TIM8:	M4		M2
 * 	ADC1:	M4B		M2B
 * 	ADC2:	M4A		M2A
 * 	ADC3:	M1C		M3B
 * 	ADC4:	M1B		M3C
 */

#pragma GCC optimize ("Ofast")

#include "drive.h"
#include "stm32f30x_gpio.h"
#include "stm32f30x_tim.h"
#include "stm32f30x_rcc.h"
#include "stm32f30x_misc.h"
#include "stm32f30x_exti.h"
#include "stm32f30x_syscfg.h"
#include "stm32f30x_adc.h"
#include "stm32f30x_dma.h"
#include "arm_math.h"

#include "spi1.h"
#include "intercom1.h"
#include "constants.h"
#include "util/sys_time.h"
#include "util/flash_fs.h"
#include <math.h>

#define PWM_MAX_VAL 3599
#define PI_3RD 1.047197551196598f
#define MOT_ADC_TO_CURRENT 0.0029296875f
#define MOTOR_RESISTANCE 0.6f	// [Ohm]

Drive drive[2] = {
#ifdef MOT_RIGHT
		{ .currentMapping = {{ 1,  1}, { 0, -1}, {-1,  0}} },	// M1
		{ .currentMapping = {{ 0,  1}, { 1,  0}, {-1, -1}} },	// M4
#else
		{ .currentMapping = {{ 0,  1}, { 1,  0}, {-1, -1}} },	// M2
		{ .currentMapping = {{ 1,  1}, {-1,  0}, { 0, -1}} },	// M3
#endif
};

Drive* pFront = &drive[0];
Drive* pRear = &drive[1];

#define DRIVE_NUM_CORR 72
static const uint16_t ccrCorr[] = {
0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
0, 0, 0, 27, 29, 30, 32, 33, 34, 36,
37, 38, 40, 41, 42, 43, 44, 46, 47, 48,
49, 50, 51, 52, 53, 55, 56, 57, 58, 59,
60, 61, 62, 63, 64, 65, 66, 67, 68, 69,
70, 72 };

#define U_TEMP_MEAS 2.0f

typedef struct _TempMeasConfig
{
	uint32_t ccer;
	float voltMult[3];
} TempMeasConfig;

static const TempMeasConfig tempMeasCfgs[3] = {
		{ CCER_AB, {1, 0, 0} },
		{ CCER_BC, {0, 1, 0} },
		{ CCER_CA, {0, 0, 1} },
};

static WORKING_AREA(waTempMeasFront, 256);
static WORKING_AREA(waTempMeasRear, 256);

static float min3(float a, float b, float c);
static float max3(float a, float b, float c);
static void updateADC(Drive* pDrive);
static void updateSpeed(Drive* pDrive);
static void updateHall(Drive* pDrive, uint16_t hallPos);
static msg_t driveTempMeasTask(void* params);

/**
 * IRQs
 */
// Encoder time measurement
IN_CCM void TIM2_IRQHandler()
{
	static uint16_t last = 0;

	CH_IRQ_PROLOGUE();

	volatile uint16_t now = TIM3->CNT;

	TIM2->SR = 0;

	TIM3->CCR3 = now+64000;

#ifdef MOT_RIGHT
	pRear->encDt = now-last;	// M4
#else
	pFront->encDt = now-last;	// M2
#endif

	last = now;

	CH_IRQ_EPILOGUE();
}

IN_CCM void TIM4_IRQHandler()
{
	static uint16_t last = 0;

	CH_IRQ_PROLOGUE();

	volatile uint16_t now = TIM3->CNT;

	TIM4->SR = 0;

	TIM3->CCR4 = now+64000;

#ifdef MOT_RIGHT
	pFront->encDt = now-last;	// M1
#else
	pRear->encDt = now-last;	// M3
#endif

	last = now;

	CH_IRQ_EPILOGUE();
}

IN_CCM void TIM3_IRQHandler()
{
	CH_IRQ_PROLOGUE();

	volatile uint16_t sr = TIM3->SR;
	TIM3->SR = 0;
	volatile uint16_t now = TIM3->CNT;

	if(sr & TIM_SR_CC1IF)
	{
		TIM3->CCR1 = now+999;
		updateSpeed(pFront);
		updateSpeed(pRear);
	}

	if(sr & TIM_SR_CC3IF)
		TIM2->EGR = TIM_EGR_CC1G;

	if(sr & TIM_SR_CC4IF)
		TIM4->EGR = TIM_EGR_CC1G;

	CH_IRQ_EPILOGUE();
}

IN_CCM void TIM8_UP_IRQHandler()
{
	CH_IRQ_PROLOGUE();

	TIM8->SR = 0;

#ifdef MOT_RIGHT
	updateADC(pRear);	// M4
#else
	updateADC(pFront);	// M2
#endif

	CH_IRQ_EPILOGUE();
}

IN_CCM void TIM1_UP_TIM16_IRQHandler()
{
	CH_IRQ_PROLOGUE();

	TIM1->SR = 0;

#ifdef MOT_RIGHT
	updateADC(pFront);	// M1
#else
	updateADC(pRear);	// M3
#endif

	CH_IRQ_EPILOGUE();
}

#define M13HallMask (EXTI_PR_PR10 | EXTI_PR_PR11 | EXTI_PR_PR12)

IN_CCM void EXTI3_IRQHandler()
{
	CH_IRQ_PROLOGUE();

	EXTI->PR = EXTI_PR_PR3;

	uint16_t pos24 = ((GPIOA->IDR & GPIO_Pin_3) >> 3) | ((GPIOC->IDR & (GPIO_Pin_14 | GPIO_Pin_15)) >> 13);

#ifdef MOT_RIGHT
	updateHall(pRear, pos24);
#else
	updateHall(pFront, pos24);
#endif

	CH_IRQ_EPILOGUE();
}

IN_CCM void EXTI15_10_IRQHandler()
{
	CH_IRQ_PROLOGUE();

	if(EXTI->PR & M13HallMask)
	{
		EXTI->PR = M13HallMask;

		uint16_t pos13 = (GPIOB->IDR & M13HallMask) >> 10;

#ifdef MOT_RIGHT
		updateHall(pFront, pos13);
#else
		updateHall(pRear, pos13);
#endif
	}
	else
	{
		EXTI->PR = (EXTI_PR_PR14 | EXTI_PR_PR15);

		uint16_t pos24 = ((GPIOA->IDR & GPIO_Pin_3) >> 3) | ((GPIOC->IDR & (GPIO_Pin_14 | GPIO_Pin_15)) >> 13);

#ifdef MOT_RIGHT
		updateHall(pRear, pos24);
#else
		updateHall(pFront, pos24);
#endif
	}

	CH_IRQ_EPILOGUE();
}

static void initHw()
{
	GPIO_InitTypeDef gpioInit;
	TIM_TimeBaseInitTypeDef timeBaseInit;
	TIM_OCInitTypeDef ocInit;
	TIM_ICInitTypeDef icInit;
	TIM_BDTRInitTypeDef bdtrInit;
	NVIC_InitTypeDef nvicInit;
	ADC_InitTypeDef adcInit;
	ADC_CommonInitTypeDef adcCommonInit;
	DMA_InitTypeDef dmaInit;
	EXTI_InitTypeDef extiInit;

	// data structures
#ifdef MOT_RIGHT
	pRear->pTIMEncoder = TIM2;
	pRear->pTIMMotor = TIM8;
	pFront->pTIMEncoder = TIM4;
	pFront->pTIMMotor = TIM1;
#else
	pRear->pTIMEncoder = TIM4;
	pRear->pTIMMotor = TIM1;
	pFront->pTIMEncoder = TIM2;
	pFront->pTIMMotor = TIM8;
#endif

	pFront->curBat = 12.0f;
	pRear->curBat = 12.0f;

	// ### DMA configuration (ADC)
	dmaInit.DMA_PeripheralBaseAddr = (uint32_t)&ADC1->DR;
#ifdef MOT_RIGHT
	dmaInit.DMA_MemoryBaseAddr = (uint32_t)pRear->adc[0];
#else
	dmaInit.DMA_MemoryBaseAddr = (uint32_t)pFront->adc[0];
#endif
	dmaInit.DMA_DIR = DMA_DIR_PeripheralSRC;
	dmaInit.DMA_BufferSize = DRIVE_ADC_SAMPLES;
	dmaInit.DMA_PeripheralInc = DMA_PeripheralInc_Disable;
	dmaInit.DMA_MemoryInc = DMA_MemoryInc_Enable;
	dmaInit.DMA_PeripheralDataSize = DMA_PeripheralDataSize_HalfWord;
	dmaInit.DMA_MemoryDataSize = DMA_MemoryDataSize_HalfWord;
	dmaInit.DMA_Mode = DMA_Mode_Circular;
	dmaInit.DMA_Priority = DMA_Priority_Medium;
	dmaInit.DMA_M2M = DMA_M2M_Disable;
	DMA_Init(DMA1_Channel1, &dmaInit);

	dmaInit.DMA_PeripheralBaseAddr = (uint32_t)&ADC2->DR;
#ifdef MOT_RIGHT
	dmaInit.DMA_MemoryBaseAddr = (uint32_t)pRear->adc[1];
#else
	dmaInit.DMA_MemoryBaseAddr = (uint32_t)pFront->adc[1];
#endif
	DMA_Init(DMA2_Channel1, &dmaInit);

	dmaInit.DMA_PeripheralBaseAddr = (uint32_t)&ADC3->DR;
#ifdef MOT_RIGHT
	dmaInit.DMA_MemoryBaseAddr = (uint32_t)pFront->adc[0];
#else
	dmaInit.DMA_MemoryBaseAddr = (uint32_t)pRear->adc[0];
#endif
	DMA_Init(DMA2_Channel5, &dmaInit);

	dmaInit.DMA_PeripheralBaseAddr = (uint32_t)&ADC4->DR;
#ifdef MOT_RIGHT
	dmaInit.DMA_MemoryBaseAddr = (uint32_t)pFront->adc[1];
#else
	dmaInit.DMA_MemoryBaseAddr = (uint32_t)pRear->adc[1];
#endif
	DMA_Init(DMA2_Channel2, &dmaInit);


	// ### ADC
	RCC_AHBPeriphClockCmd(RCC_AHBPeriph_ADC12, ENABLE);
	RCC_AHBPeriphClockCmd(RCC_AHBPeriph_ADC34, ENABLE);

	gpioInit.GPIO_Pin = GPIO_Pin_2;
	gpioInit.GPIO_Mode = GPIO_Mode_AN;
	gpioInit.GPIO_PuPd = GPIO_PuPd_NOPULL;
	GPIO_Init(GPIOA, &gpioInit);	// A2

	gpioInit.GPIO_Pin = GPIO_Pin_2 | GPIO_Pin_13 | GPIO_Pin_14;
	GPIO_Init(GPIOB, &gpioInit);	// B2, B13, B14

	adcCommonInit.ADC_Mode = ADC_Mode_Independent;
	adcCommonInit.ADC_Clock = ADC_Clock_SynClkModeDiv1;
	adcCommonInit.ADC_DMAAccessMode = ADC_DMAAccessMode_Disabled;
	adcCommonInit.ADC_DMAMode = ADC_DMAMode_Circular;
	adcCommonInit.ADC_TwoSamplingDelay = 0;
	ADC_CommonInit(ADC1, &adcCommonInit);
	ADC_CommonInit(ADC2, &adcCommonInit);
	ADC_CommonInit(ADC3, &adcCommonInit);
	ADC_CommonInit(ADC4, &adcCommonInit);

	ADC_VoltageRegulatorCmd(ADC1, ENABLE);
	ADC_VoltageRegulatorCmd(ADC2, ENABLE);
	ADC_VoltageRegulatorCmd(ADC3, ENABLE);
	ADC_VoltageRegulatorCmd(ADC4, ENABLE);
	chThdSleep(MS2ST(4));	// stabilize regulator delay

	ADC_SelectCalibrationMode(ADC1, ADC_CalibrationMode_Single);
	ADC_SelectCalibrationMode(ADC2, ADC_CalibrationMode_Single);
	ADC_SelectCalibrationMode(ADC3, ADC_CalibrationMode_Single);
	ADC_SelectCalibrationMode(ADC4, ADC_CalibrationMode_Single);
	ADC_StartCalibration(ADC1);
	ADC_StartCalibration(ADC2);
	ADC_StartCalibration(ADC3);
	ADC_StartCalibration(ADC4);
	chThdSleep(MS2ST(4));	// calibration delay

#ifdef MOT_RIGHT
	pRear->adcCal[0] = ADC_GetCalibrationValue(ADC1);
	pRear->adcCal[1] = ADC_GetCalibrationValue(ADC2);
	pFront->adcCal[0] = ADC_GetCalibrationValue(ADC3);
	pFront->adcCal[1] = ADC_GetCalibrationValue(ADC4);
#else
	pFront->adcCal[0] = ADC_GetCalibrationValue(ADC1);
	pFront->adcCal[1] = ADC_GetCalibrationValue(ADC2);
	pRear->adcCal[0] = ADC_GetCalibrationValue(ADC3);
	pRear->adcCal[1] = ADC_GetCalibrationValue(ADC4);
#endif

	ADC_StructInit(&adcInit);
	adcInit.ADC_ContinuousConvMode = ADC_ContinuousConvMode_Disable;
	adcInit.ADC_Resolution = ADC_Resolution_12b;
	adcInit.ADC_ExternalTrigConvEvent = ADC_ExternalTrigConvEvent_8;	// TIM8_TRGO2
	adcInit.ADC_ExternalTrigEventEdge = ADC_ExternalTrigEventEdge_RisingEdge;
	adcInit.ADC_DataAlign = ADC_DataAlign_Right;
	adcInit.ADC_OverrunMode = ADC_OverrunMode_Disable;
	adcInit.ADC_AutoInjMode = ADC_AutoInjec_Disable;
	adcInit.ADC_NbrOfRegChannel = DRIVE_ADC_SAMPLES;
	ADC_Init(ADC1, &adcInit);
	ADC_Init(ADC2, &adcInit);

	adcInit.ADC_ExternalTrigConvEvent = ADC_ExternalTrigConvEvent_10;	// TIM1_TRGO2
	ADC_Init(ADC3, &adcInit);
	ADC_Init(ADC4, &adcInit);

	for(uint8_t i = 1; i <= DRIVE_ADC_SAMPLES; i++)
	{
		ADC_RegularChannelConfig(ADC1, ADC_Channel_3,  i, ADC_SampleTime_181Cycles5);
		ADC_RegularChannelConfig(ADC2, ADC_Channel_12, i, ADC_SampleTime_181Cycles5);
		ADC_RegularChannelConfig(ADC3, ADC_Channel_5,  i, ADC_SampleTime_181Cycles5);
		ADC_RegularChannelConfig(ADC4, ADC_Channel_4,  i, ADC_SampleTime_181Cycles5);
	}

	// Configures the ADC DMA
	ADC_DMAConfig(ADC1, ADC_DMAMode_Circular);
	ADC_DMAConfig(ADC2, ADC_DMAMode_Circular);
	ADC_DMAConfig(ADC3, ADC_DMAMode_Circular);
	ADC_DMAConfig(ADC4, ADC_DMAMode_Circular);

	// Enable the ADC DMA
	ADC_DMACmd(ADC1, ENABLE);
	ADC_DMACmd(ADC2, ENABLE);
	ADC_DMACmd(ADC3, ENABLE);
	ADC_DMACmd(ADC4, ENABLE);

	ADC_Cmd(ADC1, ENABLE);
	ADC_Cmd(ADC2, ENABLE);
	ADC_Cmd(ADC3, ENABLE);
	ADC_Cmd(ADC4, ENABLE);
	chThdSleep(MS2ST(4));	// ADC ready delay

	// Enable the DMA channel
	DMA_Cmd(DMA1_Channel1, ENABLE);
	DMA_Cmd(DMA2_Channel1, ENABLE);
	DMA_Cmd(DMA2_Channel5, ENABLE);
	DMA_Cmd(DMA2_Channel2, ENABLE);

	// ### Encoder
	RCC_APB1PeriphClockCmd(RCC_APB1Periph_TIM2, ENABLE);
	RCC_APB1PeriphClockCmd(RCC_APB1Periph_TIM4, ENABLE);

	gpioInit.GPIO_Mode = GPIO_Mode_AF;
	gpioInit.GPIO_OType = GPIO_OType_PP;
	gpioInit.GPIO_Pin = GPIO_Pin_0 | GPIO_Pin_1 | GPIO_Pin_11;
	gpioInit.GPIO_PuPd = GPIO_PuPd_NOPULL;
	gpioInit.GPIO_Speed = GPIO_Speed_Level_1;
	GPIO_Init(GPIOA, &gpioInit);

	gpioInit.GPIO_Pin = GPIO_Pin_7;
	GPIO_Init(GPIOB, &gpioInit);

	GPIO_PinAFConfig(GPIOA, GPIO_PinSource0, GPIO_AF_1);	// TIM2_CH1
	GPIO_PinAFConfig(GPIOA, GPIO_PinSource1, GPIO_AF_1);	// TIM2_CH2
	GPIO_PinAFConfig(GPIOA, GPIO_PinSource11, GPIO_AF_10);	// TIM4_CH1
	GPIO_PinAFConfig(GPIOB, GPIO_PinSource7, GPIO_AF_2);	// TIM4_CH2

	//Timer Init - Encoder inputs
	timeBaseInit.TIM_ClockDivision = 0;
	timeBaseInit.TIM_CounterMode = TIM_CounterMode_Up;
	timeBaseInit.TIM_Period = 2047;	// 65535
	timeBaseInit.TIM_Prescaler = 0;
	timeBaseInit.TIM_RepetitionCounter = 0;
	TIM_TimeBaseInit(TIM2, &timeBaseInit);
	TIM_TimeBaseInit(TIM4, &timeBaseInit);

	icInit.TIM_Channel = TIM_Channel_1;
	icInit.TIM_ICFilter = 0x0F;
	icInit.TIM_ICPolarity = TIM_ICPolarity_Rising;
	icInit.TIM_ICPrescaler = TIM_ICPSC_DIV1;
	icInit.TIM_ICSelection = TIM_ICSelection_DirectTI;
	TIM_ICInit(TIM2, &icInit);
	TIM_ICInit(TIM4, &icInit);

	icInit.TIM_Channel = TIM_Channel_2;
	icInit.TIM_ICSelection = TIM_ICSelection_IndirectTI;
	TIM_ICInit(TIM2, &icInit);
	TIM_ICInit(TIM4, &icInit);

	TIM_ITConfig(TIM2, TIM_IT_CC1, ENABLE);
	TIM_ITConfig(TIM4, TIM_IT_CC1, ENABLE);

	TIM_EncoderInterfaceConfig(TIM2, TIM_EncoderMode_TI12, TIM_ICPolarity_Rising, TIM_ICPolarity_Rising);
	TIM_EncoderInterfaceConfig(TIM4, TIM_EncoderMode_TI12, TIM_ICPolarity_Rising, TIM_ICPolarity_Rising);

	TIM_Cmd(TIM2, ENABLE);
	TIM_Cmd(TIM4, ENABLE);

	nvicInit.NVIC_IRQChannelCmd = ENABLE;
	nvicInit.NVIC_IRQChannelPreemptionPriority = IRQL_MOTOR_UPDATE-1;
	nvicInit.NVIC_IRQChannelSubPriority = 0;
	nvicInit.NVIC_IRQChannel = TIM2_IRQn;
	NVIC_Init(&nvicInit);

	nvicInit.NVIC_IRQChannel = TIM4_IRQn;
	NVIC_Init(&nvicInit);

	// ### Encoder Velocity Measurement Timer
	RCC_APB1PeriphClockCmd(RCC_APB1Periph_TIM3, ENABLE);

	timeBaseInit.TIM_ClockDivision = 0;
	timeBaseInit.TIM_CounterMode = TIM_CounterMode_Up;
	timeBaseInit.TIM_Period = 0xFFFF;	// 65535
	timeBaseInit.TIM_Prescaler = 71;	// = 1us
	timeBaseInit.TIM_RepetitionCounter = 0;
	TIM_TimeBaseInit(TIM3, &timeBaseInit);

	ocInit.TIM_OCMode = TIM_OCMode_Timing;
	ocInit.TIM_Pulse = 1000;
	ocInit.TIM_OutputState = TIM_OutputState_Enable;
	ocInit.TIM_OCPolarity = TIM_OCPolarity_High;
	TIM_OC1Init(TIM3, &ocInit);

	ocInit.TIM_Pulse = 64000;
	TIM_OC3Init(TIM3, &ocInit);
	TIM_OC4Init(TIM3, &ocInit);

	TIM_OC1PreloadConfig(TIM3, TIM_OCPreload_Disable);
	TIM_OC3PreloadConfig(TIM3, TIM_OCPreload_Disable);
	TIM_OC4PreloadConfig(TIM3, TIM_OCPreload_Disable);

	TIM_Cmd(TIM3, ENABLE);

	TIM_ITConfig(TIM3, TIM_IT_CC1 | TIM_IT_CC3 | TIM_IT_CC4, ENABLE);

	nvicInit.NVIC_IRQChannelPreemptionPriority = IRQL_MOTOR_UPDATE+1;
	nvicInit.NVIC_IRQChannel = TIM3_IRQn;
	NVIC_Init(&nvicInit);

	// ### Motors
	nvicInit.NVIC_IRQChannelCmd = ENABLE;
	nvicInit.NVIC_IRQChannelPreemptionPriority = IRQL_MOTOR_UPDATE;
	nvicInit.NVIC_IRQChannelSubPriority = 0;
	nvicInit.NVIC_IRQChannel = TIM8_UP_IRQn;
	NVIC_Init(&nvicInit);

	nvicInit.NVIC_IRQChannel = TIM1_UP_TIM16_IRQn;
	NVIC_Init(&nvicInit);

	RCC_APB2PeriphClockCmd(RCC_APB2Periph_TIM1, ENABLE);
	RCC_APB2PeriphClockCmd(RCC_APB2Periph_TIM8, ENABLE);
	RCC_TIMCLKConfig(RCC_TIM1CLK_PLLCLK);
	RCC_TIMCLKConfig(RCC_TIM8CLK_PLLCLK);	// TIM8 CLK now at 144MHz

	GPIOA->ODR |= GPIO_Pin_7 | GPIO_Pin_8 | GPIO_Pin_9 | GPIO_Pin_10 | GPIO_Pin_12;
	GPIOB->ODR |= GPIO_Pin_0 | GPIO_Pin_1 | GPIO_Pin_6 | GPIO_Pin_8 | GPIO_Pin_9 | GPIO_Pin_15;
	GPIOC->ODR |= GPIO_Pin_13;

	gpioInit.GPIO_Mode = GPIO_Mode_AF;
	gpioInit.GPIO_OType = GPIO_OType_PP;
	gpioInit.GPIO_PuPd = GPIO_PuPd_UP;
	gpioInit.GPIO_Speed = GPIO_Speed_50MHz;
	gpioInit.GPIO_Pin = GPIO_Pin_7 | GPIO_Pin_8 | GPIO_Pin_9 | GPIO_Pin_10 | GPIO_Pin_12;
	GPIO_Init(GPIOA, &gpioInit);

	gpioInit.GPIO_Pin = GPIO_Pin_0 | GPIO_Pin_1 | GPIO_Pin_6 | GPIO_Pin_8 | GPIO_Pin_9 | GPIO_Pin_15;
	GPIO_Init(GPIOB, &gpioInit);

	gpioInit.GPIO_Pin = GPIO_Pin_13;
	GPIO_Init(GPIOC, &gpioInit);

	timeBaseInit.TIM_ClockDivision = 0;
	timeBaseInit.TIM_CounterMode = TIM_CounterMode_CenterAligned1;
	timeBaseInit.TIM_Period = PWM_MAX_VAL;	// = 3599 = 20kHz PWM (2*10kHz due to center aligned mode)
	timeBaseInit.TIM_Prescaler = 0;
	timeBaseInit.TIM_RepetitionCounter = 1;	// update event only every second count (e.g. downcounting) = 20kHz
	TIM_TimeBaseInit(TIM1, &timeBaseInit);
	TIM_TimeBaseInit(TIM8, &timeBaseInit);

	ocInit.TIM_OCMode = TIM_OCMode_PWM2;
	ocInit.TIM_OutputState = TIM_OutputState_Enable;
	ocInit.TIM_OutputNState = TIM_OutputNState_Enable;
	ocInit.TIM_OCPolarity = TIM_OCPolarity_High;
	ocInit.TIM_OCNPolarity = TIM_OCNPolarity_High;
	ocInit.TIM_OCIdleState = TIM_OCIdleState_Set;
	ocInit.TIM_OCNIdleState = TIM_OCNIdleState_Reset;
	TIM_OC1Init(TIM8, &ocInit);
	TIM_OC2Init(TIM8, &ocInit);
	TIM_OC3Init(TIM8, &ocInit);
	TIM_OC1Init(TIM1, &ocInit);
	TIM_OC2Init(TIM1, &ocInit);
	TIM_OC3Init(TIM1, &ocInit);
	ocInit.TIM_Pulse = 864;	// trigger ADC conversion 6us after update event (allow IRQ to sum up previous value)
	TIM_OC4Init(TIM8, &ocInit);
	TIM_OC4Init(TIM1, &ocInit);

	bdtrInit.TIM_AutomaticOutput = TIM_AutomaticOutput_Enable;
	bdtrInit.TIM_Break = TIM_Break_Disable;
	bdtrInit.TIM_BreakPolarity = TIM_BreakPolarity_High;
	bdtrInit.TIM_DeadTime = 0;
	bdtrInit.TIM_LOCKLevel = TIM_LOCKLevel_OFF;
	bdtrInit.TIM_OSSIState = TIM_OSSIState_Disable;
	bdtrInit.TIM_OSSRState = TIM_OSSRState_Enable;
	TIM_BDTRConfig(TIM8, &bdtrInit);
	TIM_BDTRConfig(TIM1, &bdtrInit);

	TIM8->EGR = TIM_EGR_UG;
	TIM1->EGR = TIM_EGR_UG;

	TIM_SelectOutputTrigger2(TIM8, TIM_TRGO2Source_OC4Ref);	// OC4 connected to ADC
	TIM_SelectOutputTrigger2(TIM1, TIM_TRGO2Source_OC4Ref);	// OC4 connected to ADC
//	TIM_SelectOutputTrigger2(TIM8, TIM_TRGO2Source_Update);	// update event connected to ADC

	TIM_ITConfig(TIM8, TIM_IT_Update, ENABLE);
	TIM_ITConfig(TIM1, TIM_IT_Update, ENABLE);

	TIM_Cmd(TIM8, ENABLE);
	TIM_Cmd(TIM1, ENABLE);

	// enable trigger input
	ADC_StartConversion(ADC1);
	ADC_StartConversion(ADC2);
	ADC_StartConversion(ADC3);
	ADC_StartConversion(ADC4);

	TIM_CtrlPWMOutputs(TIM8, ENABLE);
	TIM_CtrlPWMOutputs(TIM1, ENABLE);

	// TIM8
	GPIO_PinAFConfig(GPIOA, GPIO_PinSource7, GPIO_AF_4);
	GPIO_PinAFConfig(GPIOB, GPIO_PinSource0, GPIO_AF_4);
	GPIO_PinAFConfig(GPIOB, GPIO_PinSource1, GPIO_AF_4);
	GPIO_PinAFConfig(GPIOB, GPIO_PinSource6, GPIO_AF_5);
	GPIO_PinAFConfig(GPIOB, GPIO_PinSource8, GPIO_AF_10);
	GPIO_PinAFConfig(GPIOB, GPIO_PinSource9, GPIO_AF_10);

	// TIM1
	GPIO_PinAFConfig(GPIOA, GPIO_PinSource8, GPIO_AF_6);
	GPIO_PinAFConfig(GPIOA, GPIO_PinSource9, GPIO_AF_6);
	GPIO_PinAFConfig(GPIOA, GPIO_PinSource10, GPIO_AF_6);
	GPIO_PinAFConfig(GPIOA, GPIO_PinSource12, GPIO_AF_6);
	GPIO_PinAFConfig(GPIOB, GPIO_PinSource15, GPIO_AF_4);
	GPIO_PinAFConfig(GPIOC, GPIO_PinSource13, GPIO_AF_4);

	// ### Hall Sensors
	gpioInit.GPIO_Mode = GPIO_Mode_IN;
	gpioInit.GPIO_PuPd = GPIO_PuPd_NOPULL;
	gpioInit.GPIO_OType = GPIO_OType_PP;
	gpioInit.GPIO_Speed = GPIO_Speed_2MHz;
	gpioInit.GPIO_Pin = GPIO_Pin_3;
	GPIO_Init(GPIOA, &gpioInit);

	gpioInit.GPIO_Pin = GPIO_Pin_10 | GPIO_Pin_11 | GPIO_Pin_12;
	GPIO_Init(GPIOB, &gpioInit);

	gpioInit.GPIO_Pin = GPIO_Pin_14 | GPIO_Pin_15;
	GPIO_Init(GPIOC, &gpioInit);

	SYSCFG_EXTILineConfig(EXTI_PortSourceGPIOA, EXTI_PinSource3);
	SYSCFG_EXTILineConfig(EXTI_PortSourceGPIOB, EXTI_PinSource10);
	SYSCFG_EXTILineConfig(EXTI_PortSourceGPIOB, EXTI_PinSource11);
	SYSCFG_EXTILineConfig(EXTI_PortSourceGPIOB, EXTI_PinSource12);
	SYSCFG_EXTILineConfig(EXTI_PortSourceGPIOC, EXTI_PinSource14);
	SYSCFG_EXTILineConfig(EXTI_PortSourceGPIOC, EXTI_PinSource15);

	EXTI_StructInit(&extiInit);

	extiInit.EXTI_Line = EXTI_Line3;
	extiInit.EXTI_LineCmd = ENABLE;
	extiInit.EXTI_Mode = EXTI_Mode_Interrupt;
	extiInit.EXTI_Trigger = EXTI_Trigger_Rising_Falling;
	EXTI_Init(&extiInit);

	extiInit.EXTI_Line = EXTI_Line10;
	EXTI_Init(&extiInit);

	extiInit.EXTI_Line = EXTI_Line11;
	EXTI_Init(&extiInit);

	extiInit.EXTI_Line = EXTI_Line12;
	EXTI_Init(&extiInit);

	extiInit.EXTI_Line = EXTI_Line14;
	EXTI_Init(&extiInit);

	extiInit.EXTI_Line = EXTI_Line15;
	EXTI_Init(&extiInit);

	nvicInit.NVIC_IRQChannel = EXTI3_IRQn;
	nvicInit.NVIC_IRQChannelCmd = ENABLE;
	nvicInit.NVIC_IRQChannelPreemptionPriority = IRQL_MOTOR_UPDATE+1;
	nvicInit.NVIC_IRQChannelSubPriority = 0;
	NVIC_Init(&nvicInit);

	nvicInit.NVIC_IRQChannel = EXTI15_10_IRQn;
	NVIC_Init(&nvicInit);
}

/**
 * PRIVATE functions
 */
IN_CCM static float min3(float a, float b, float c)
{
	if(a < b && a < c)
		return a;

	if(b < a && b < c)
		return b;

	return c;
}

IN_CCM static float max3(float a, float b, float c)
{
	if(a > b && a > c)
		return a;

	if(b > a && b > c)
		return b;

	return c;
}

typedef union _u32u16
{
	uint32_t u32;
	uint16_t u16[2];
} u32u16_t;

IN_CCM static void updateADC(Drive* pDrive)
{
	float i0f;
	float i1f;

	uint16_t i016;
	uint16_t i116;

	u32u16_t i1u, i0u;

	i0u.u32 = pDrive->adc32[0][0];
	i1u.u32 = pDrive->adc32[1][0];

	for(uint8_t i = 1; i < DRIVE_ADC_SAMPLES/2; i++)
	{
		i0u.u32 = __UQADD16(i0u.u32, pDrive->adc32[0][i]);
		i1u.u32 = __UQADD16(i1u.u32, pDrive->adc32[1][i]);
	}

	i016 = i0u.u16[0]+i0u.u16[1];
	i116 = i1u.u16[0]+i1u.u16[1];

	i0f = i016*(1.0f/DRIVE_ADC_SAMPLES);
	i1f = i116*(1.0f/DRIVE_ADC_SAMPLES);

	i0f = (i0f-1706.6667f)*-MOT_ADC_TO_CURRENT; // convert to real current
	i1f = (i1f-1706.6667f)*-MOT_ADC_TO_CURRENT; // convert to real current

	pDrive->i[0] = pDrive->currentMapping[0][0]*i0f + pDrive->currentMapping[0][1]*i1f;
	pDrive->i[1] = pDrive->currentMapping[1][0]*i0f + pDrive->currentMapping[1][1]*i1f;
	pDrive->i[2] = pDrive->currentMapping[2][0]*i0f + pDrive->currentMapping[2][1]*i1f;

	pDrive->iAlpha = pDrive->i[0];
	pDrive->iBeta = 0.57735026918963f*pDrive->i[0] + 1.154700538379252f*pDrive->i[1];

	if(pDrive->measSample < DRIVE_MEAS_SAMPLES)
	{
		pDrive->meas[0][pDrive->measSample] = (int16_t)(i0f*1000);
		pDrive->meas[1][pDrive->measSample] = (int16_t)(i1f*1000);
		++pDrive->measSample;
	}

	if(pDrive->pCtrl && pDrive->pCtrl->onCurrentUpdate)
		(*pDrive->pCtrl->onCurrentUpdate)(pDrive);
}

IN_CCM static void updateSpeed(Drive* pDrive)
{
	float vel;

	if(pDrive->encDt > 63000 || pDrive->encDt < 20)
	{
		vel = 0;
	}
	else
	{
		float dt = pDrive->encDt;
		vel = 1e6f/(dt*512);

		if(pDrive->pTIMEncoder->CR1 & TIM_CR1_DIR)
			vel *= -1;
	}

	if(vel < 10000)
		pDrive->curVelocity = vel;

	if(pDrive->pCtrl && pDrive->pCtrl->onSpeedUpdate)
		(*pDrive->pCtrl->onSpeedUpdate)(pDrive);
}

IN_CCM static void updateHall(Drive* pDrive, uint16_t hallPos)
{
	pDrive->halPos = hallPos;

	if(pDrive->pCtrl && pDrive->pCtrl->onHallUpdate)
		(*pDrive->pCtrl->onHallUpdate)(pDrive);
}

/**
 * PUBLIC functions
 */
void DriveInit()
{
	initHw();

	chMtxInit(&pFront->tMeasMutex);
	chMtxInit(&pRear->tMeasMutex);

	int16_t result = FlashFSOpenOrCreate("front/tCal", 0, &pFront->tCal, sizeof(DriveTempCalib), &pFront->pTCalFile);
	if(result == FLASH_FS_CREATED)
	{
		memset(&pFront->tCal, 0, sizeof(DriveTempCalib));
		FlashFSWrite(&pFront->pTCalFile, &pFront->tCal, sizeof(DriveTempCalib));
	}

	result = FlashFSOpenOrCreate("rear/tCal", 0, &pRear->tCal, sizeof(DriveTempCalib), &pRear->pTCalFile);
	if(result == FLASH_FS_CREATED)
	{
		memset(&pRear->tCal, 0, sizeof(DriveTempCalib));
		FlashFSWrite(&pRear->pTCalFile, &pRear->tCal, sizeof(DriveTempCalib));
	}

	chThdCreateStatic(waTempMeasFront, sizeof(waTempMeasFront), NORMALPRIO-5, driveTempMeasTask, pFront);
	chThdCreateStatic(waTempMeasRear, sizeof(waTempMeasRear), NORMALPRIO-5, driveTempMeasTask, pRear);

	pFront->zeroSetpoint = 1;
	pRear->zeroSetpoint = 1;

	pFront->enableTempMeas = 0;
	pRear->enableTempMeas = 0;
}

void DriveSetController(DriveCtrl* pCtrlFront, DriveCtrl* pCtrlRear)
{
	// call exit actions
	if(pFront->pCtrl && pFront->pCtrl->onExit)
		(*pFront->pCtrl->onExit)(pFront);
	if(pRear->pCtrl && pRear->pCtrl->onExit)
		(*pRear->pCtrl->onExit)(pRear);

	DriveSetPower(pFront, 5000);
	DriveSetPower(pRear, 5000);

	pFront->pCtrl = pCtrlFront;
	pRear->pCtrl = pCtrlRear;

	// call init actions
	if(pFront->pCtrl && pFront->pCtrl->onInit)
		(*pFront->pCtrl->onInit)(pFront);
	if(pRear->pCtrl && pRear->pCtrl->onInit)
		(*pRear->pCtrl->onInit)(pRear);
}

void DriveSetPower(Drive* pDrive, uint16_t power)
{
	pDrive->pTIMMotor->CCR1 = power;
	pDrive->pTIMMotor->CCR2 = power;
	pDrive->pTIMMotor->CCR3 = power;
}

void DriveEnablePhase(Drive* pDrive, uint32_t ccer)
{
	uint32_t oldCCER = pDrive->pTIMMotor->CCER & (TIM_CCER_CC1P | TIM_CCER_CC1NP | TIM_CCER_CC2P | TIM_CCER_CC2NP | TIM_CCER_CC3P | TIM_CCER_CC3NP);
	pDrive->pTIMMotor->CCER = oldCCER | ccer;
}

void DriveGetSteadyPhaseCurrents(Drive* pDrive, float* i)
{
	float Vcom = (pDrive->u[0]+pDrive->u[1]+pDrive->u[2])/3;

	i[0] = (pDrive->u[0]-Vcom)/MOTOR_RESISTANCE;
	i[1] = (pDrive->u[1]-Vcom)/MOTOR_RESISTANCE;
	i[2] = (pDrive->u[2]-Vcom)/MOTOR_RESISTANCE;
}

void DriveGetMeasAvg(Drive* pDrive, float* avg)
{
	float iReal[2];

	iReal[0] = 0;
	iReal[1] = 0;

	for(uint16_t n = 0; n < DRIVE_MEAS_SAMPLES; n++)
	{
		iReal[0] += pDrive->meas[0][n];
		iReal[1] += pDrive->meas[1][n];
	}

	iReal[0] *= 1.0f/(DRIVE_MEAS_SAMPLES*1000.0f);
	iReal[1] *= 1.0f/(DRIVE_MEAS_SAMPLES*1000.0f);

	avg[0] = pDrive->currentMapping[0][0]*iReal[0] + pDrive->currentMapping[0][1]*iReal[1];
	avg[1] = pDrive->currentMapping[1][0]*iReal[0] + pDrive->currentMapping[1][1]*iReal[1];
	avg[2] = pDrive->currentMapping[2][0]*iReal[0] + pDrive->currentMapping[2][1]*iReal[1];
}

IN_CCM static void setCCR(Drive* pDrive, uint16_t a, uint16_t b, uint16_t c)
{
	if(a < DRIVE_NUM_CORR)
		a = ccrCorr[a];

	if(b < DRIVE_NUM_CORR)
		b = ccrCorr[b];

	if(c < DRIVE_NUM_CORR)
		c = ccrCorr[c];

	pDrive->pTIMMotor->CCR1 = a;
	pDrive->pTIMMotor->CCR2 = b;
	pDrive->pTIMMotor->CCR3 = c;
}

IN_CCM void DriveSetPhaseVoltagesSVPWM(Drive* pDrive, float Va, float Vb, float Vc)
{
	float Vd = pDrive->curBat;
	if(Vd < 9.0f || Vd > 17.0f)
		Vd = 11.1f;

	float Vcom = Vd/2-0.5f*(max3(Va, Vb, Vc)+min3(Va, Vb, Vc));

	pDrive->u[0] = Va+Vcom;
	pDrive->u[1] = Vb+Vcom;
	pDrive->u[2] = Vc+Vcom;

	float da = 1/Vd*(Va+Vcom);
	float db = 1/Vd*(Vb+Vcom);
	float dc = 1/Vd*(Vc+Vcom);

	if(da > 1.0f)
		da = 1.0f;
	if(da < 0.0f)
		da = 0.0f;

	if(db > 1.0f)
		db = 1.0f;
	if(db < 0.0f)
		db = 0.0f;

	if(dc > 1.0f)
		dc = 1.0f;
	if(dc < 0.0f)
		dc = 0.0f;

	setCCR(pDrive, da*PWM_MAX_VAL, db*PWM_MAX_VAL, dc*PWM_MAX_VAL);
}

IN_CCM void DriveSetPhaseVoltagesDirect(Drive* pDrive, float Va, float Vb, float Vc)
{
	float Vd = pDrive->curBat;
	if(Vd < 9.0f || Vd > 17.0f)
		Vd = 11.1f;

	pDrive->u[0] = Va;
	pDrive->u[1] = Vb;
	pDrive->u[2] = Vc;

	float da = Va/Vd;
	float db = Vb/Vd;
	float dc = Vc/Vd;

	if(da > 1.0f)
		da = 1.0f;
	if(da < 0.0f)
		da = 0.0f;

	if(db > 1.0f)
		db = 1.0f;
	if(db < 0.0f)
		db = 0.0f;

	if(dc > 1.0f)
		dc = 1.0f;
	if(dc < 0.0f)
		dc = 0.0f;

	setCCR(pDrive, da*PWM_MAX_VAL, db*PWM_MAX_VAL, dc*PWM_MAX_VAL);
}

IN_CCM void DriveSetAlphaBeta(Drive* pDrive, float Valpha, float Vbeta)
{
	float Va = Valpha;
	float Vb = -0.5f*Valpha + 0.86602540378444f*Vbeta;
	float Vc = -0.5f*Valpha - 0.86602540378444f*Vbeta;

	DriveSetPhaseVoltagesSVPWM(pDrive, Va, Vb, Vc);
}

IN_CCM void DriveSetVoltageAngle(Drive* pDrive, float Vref, float angle)
{
	float sin, cos;
	while(angle > 180.0f)
		angle -= 360.0f;

	arm_sin_cos_f32(angle, &sin, &cos);

	DriveSetAlphaBeta(pDrive, cos*Vref, sin*Vref);
}

static uint8_t getCurrentForTemp(Drive* pDrive, const TempMeasConfig* pCfg, float* pI)
{
	DriveEnablePhase(pDrive, pCfg->ccer);

	uint16_t n;
	for(n = 1; n < 100; n++)
	{
		if(pDrive->pCtrl)
			break;

		float u = U_TEMP_MEAS*n*0.01f;
		DriveSetPhaseVoltagesDirect(pDrive, u*pCfg->voltMult[0], u*pCfg->voltMult[1], u*pCfg->voltMult[2]);
		chThdSleep(MS2ST(2));
	}

	if(n < 100)
		return 1;	// measurement interrupted

	chThdSleep(MS2ST(50));

	pDrive->measSample = 0;
	chThdSleep(MS2ST(20));
	DriveGetMeasAvg(pDrive, pI);

	for(uint8_t i = 0; i < 3; i++)
		pI[i] *= -1;

	if(pDrive->pCtrl)
		return 1;

	return 0;
}

static uint8_t getCombinedResistance(Drive* pDrive, float* pRcomb)
{
	float iReal[3];

	for(uint8_t i = 0; i < 3; i++)
	{
		if(getCurrentForTemp(pDrive, &tempMeasCfgs[i], iReal))
			return 1;

		iReal[i] += pDrive->tCal.offset[i];

		pRcomb[i] = U_TEMP_MEAS/iReal[i];
	}

	return 0;
}

void DriveCalibrateCurrentForTemp(Drive* pDrive, float tNow)
{
	// disable controller
	const float RNow = MOTOR_RESISTANCE*(1+3.93e-3f*(tNow-25));
	pDrive->pCtrlInactive = pDrive->pCtrl;
	pDrive->pCtrl = 0;

	float iReal[3];

	chMtxLock(&pDrive->tMeasMutex);

	for(uint8_t i = 0; i < 3; i++)
	{
		getCurrentForTemp(pDrive, &tempMeasCfgs[i], iReal);

		pDrive->tCal.offset[i] = (U_TEMP_MEAS*1.0f/(2*RNow))-iReal[i];
	}

	DriveEnablePhase(pDrive, CCER_ABC);
	DriveSetPhaseVoltagesDirect(pDrive, 0, 0, 0);

	chMtxUnlock();

	// re-enable controller
	pDrive->pCtrl = pDrive->pCtrlInactive;
	pDrive->pCtrlInactive = 0;

	FlashFSWrite(&pDrive->pTCalFile, &pDrive->tCal, sizeof(DriveTempCalib));
}


uint8_t DriveTempMeas(Drive* pDrive)
{
	float Rcomb[3]; 	// that is Rab, Rbc, Rca
	float R[3];

	// disable controller
	pDrive->pCtrlInactive = pDrive->pCtrl;
	pDrive->pCtrl = 0;

	chMtxLock(&pDrive->tMeasMutex);

	if(getCombinedResistance(pDrive, Rcomb))
		return 1;

	// check if controller is already back online, measurement has then been interrupted
	// measurement is then invalid
	if(pDrive->pCtrl)
		return 1;

	DriveEnablePhase(pDrive, CCER_ABC);
	DriveSetPhaseVoltagesDirect(pDrive, 0, 0, 0);

	chMtxUnlock();

	// re-enable controller
	pDrive->pCtrl = pDrive->pCtrlInactive;
	pDrive->pCtrlInactive = 0;

	R[2] = (Rcomb[1]-Rcomb[0]+Rcomb[1])*0.5f;
	R[1] = Rcomb[1]-R[2];
	R[0] = Rcomb[2]-R[2];

	for(uint8_t i = 0; i < 3; i++)
		pDrive->t[i] = 25.0f+(R[i]/MOTOR_RESISTANCE-1.0f)/3.93e-3f;

	pDrive->tMeasTime = SysTimeUSec();

	return 0;
}

void DriveSetpointInfo(Drive* pDrive, uint8_t zeroSetpoint)
{
	if(!zeroSetpoint && pDrive->pCtrlInactive)
	{
		pDrive->pCtrl = pDrive->pCtrlInactive;
		pDrive->pCtrlInactive = 0;
	}

	pDrive->zeroSetpoint = zeroSetpoint;
}

static msg_t driveTempMeasTask(void* params)
{
	msg_t msg;
	Drive* pDrive = (Drive*)params;

	chThdSleep(S2ST(5));

	while(1)
	{
		uint8_t doMeas = 1;
		for(uint8_t i = 0; i < 100; i++)
		{
			chThdSleep(MS2ST(5));

			if(!pDrive->zeroSetpoint || pDrive->curVelocity > DRIVE_MOVE_THRESHOLD || pDrive->curVelocity < -DRIVE_MOVE_THRESHOLD)
			{
				doMeas = 0;
				break;
			}
		}

		if(doMeas == 0 || pDrive->enableTempMeas == 0)
			continue;

		DriveTempMeas(pDrive);

		IntercomHeader iHeader;
		iHeader.section = INTERCOM_SECTION_MOTORS;
		iHeader.command = IC_MOTORS_TEMP;
		iHeader.address = INTERCOM_ADDRESS_MAIN | INTERCOM_ADDRESS_MEDIA;

		MotorsTemp temp;
#ifdef MOT_LEFT
		temp.left = 1;
#else
		temp.left = 0;
#endif
		if(pDrive == pRear)
			temp.rear = 1;
		else
			temp.rear = 0;

		memcpy(temp.t, pDrive->t, sizeof(float)*3);

		IntercomSend(&intercom1, &iHeader, sizeof(MotorsTemp), (uint8_t*)&temp);

		chThdSleep(MS2ST(5000));
	}

	return 0;
}
