/*
 * drive.h
 *
 *  Created on: 02.02.2014
 *      Author: AndreR
 */

#ifndef DRIVE_H_
#define DRIVE_H_

#include <stdint.h>

#include "stm32f30x.h"
#include "ch.h"
#include "intercom_constants.h"
#include "util/flash_fs.h"

#define IN_CCM __attribute__((section(".ccmtext")))

#define DRIVE_ADC_SAMPLES 16
#define DRIVE_MEAS_SAMPLES 100
#define DRIVE_MOVE_THRESHOLD 0.25f

// W3  W2  W1
//  5   5   5
#define CCER_A 0x005
#define CCER_B 0x050
#define CCER_C 0x500
#define CCER_AB (CCER_A | CCER_B)
#define CCER_BC (CCER_B | CCER_C)
#define CCER_CA (CCER_C | CCER_A)
#define CCER_ABC (CCER_A | CCER_B | CCER_C)

struct _Drive;

typedef void(*InitCtrlCb)(struct _Drive*);
typedef void(*ExitCtrlCb)(struct _Drive*);
typedef void(*CurrentUpdateCb)(struct _Drive*);
typedef void(*SpeedUpdateCb)(struct _Drive*);
typedef void(*HallUpdateCb)(struct _Drive*);

typedef struct _DriveCtrl
{
	InitCtrlCb onInit;
	ExitCtrlCb onExit;
	CurrentUpdateCb onCurrentUpdate;
	SpeedUpdateCb onSpeedUpdate;
	HallUpdateCb onHallUpdate;

	void* pUserData;
} DriveCtrl;

typedef struct _DriveTempCalib
{
	float offset[3];
} __attribute__((aligned(4),packed)) DriveTempCalib;

typedef struct _Drive
{
	// ### Setup ###
	// Motor and encoder timer
	TIM_TypeDef* pTIMMotor;
	TIM_TypeDef* pTIMEncoder;

	const float currentMapping[3][2];

	// ADC calibration values
	uint32_t adcCal[2];

	// ### Measured values ###
	// ADC DMA storage area
	union
	{
		uint16_t adc[2][DRIVE_ADC_SAMPLES];
		uint32_t adc32[2][DRIVE_ADC_SAMPLES/2];
	};

	uint16_t encDt;	// deltaT between two encoder IRQs
	float curVelocity;	// current velocity [rounds per second]
	float i[3];	// current on each phase [A]
	float iAlpha;	// current on stator-fixed alpha axis [A]
	float iBeta;	// current on stator-fixed beta axis [A]

	DriveTempCalib tCal;	// current sensor offset for temperature measurement
	FlashFile* pTCalFile;
	float t[3];		// temperature of each winding (measurement needs to be explicitly triggered)
	uint32_t tMeasTime;	// system time of temperature measurement
	Mutex tMeasMutex;
	uint8_t enableTempMeas;
	uint16_t halPos;

	// ### control related ###
	DriveCtrl* pCtrl;	// current control instance
	DriveCtrl* pCtrlInactive;

	// ### output/info values ###
	float u[3];

	// ### required input ###
	float curBat;	// [V]
	volatile uint8_t zeroSetpoint;

	// ### data sampling for test purposes ###
	int16_t meas[2][DRIVE_MEAS_SAMPLES];
	uint16_t measSample;
} Drive;

extern Drive drive[2];
extern Drive* pRear;
extern Drive* pFront;

void DriveInit();
void DriveSetController(DriveCtrl* pCtrlFront, DriveCtrl* pCtrlRear);
void DriveSetPower(Drive* pDrive, uint16_t power);
void DriveSetPhaseVoltagesSVPWM(Drive* pDrive, float Va, float Vb, float Vc);
void DriveSetPhaseVoltagesDirect(Drive* pDrive, float Va, float Vb, float Vc);
void DriveSetAlphaBeta(Drive* pDrive, float Valpha, float Vbeta);
void DriveSetVoltageAngle(Drive* pDrive, float Vref, float angle);
void DriveGetSteadyPhaseCurrents(Drive* pDrive, float* i);
void DriveEnablePhase(Drive* pDrive, uint32_t ccer);
void DriveGetMeasAvg(Drive* pDrive, float* avg);
void DriveCalibrateCurrentForTemp(Drive* pDrive, float tNow);
uint8_t DriveTempMeas(Drive* pDrive);
void DriveSetpointInfo(Drive* pDrive, uint8_t zeroSetpoint);

#endif /* DRIVE_H_ */
