#include "stm32f30x.h"
#include "stm32f30x_misc.h"
#include "stm32f30x_rcc.h"
#include "stm32f30x_exti.h"
#include "stm32f30x_gpio.h"
#include "stm32f30x_flash.h"
#include "stm32f30x_usart.h"
#include "stm32f30x_iwdg.h"

#include "core_cm4.h"
#include "cpu_load.h"

#include "hal/led.h"
#include "util/event.h"
#include "util/boot.h"
#include "util/bootloader.h"
#include "util/flash_fs.h"
#include "spi1.h"
#include "intercom1.h"
#include "cli.h"
#include "drive.h"
#include "motors.h"
#include "dwt.h"
#include "ctrl_foc.h"
#include "ctrl_6step.h"

static void NVIC_Configuration();
static void SetSysClock(void);

#define portNVIC_SYSTICK_CTRL		( ( volatile unsigned long * ) 0xe000e010 )
#define portNVIC_SYSTICK_LOAD		( ( volatile unsigned long * ) 0xe000e014 )
#define portNVIC_SYSTICK_CLK		0x00000004
#define portNVIC_SYSTICK_INT		0x00000002
#define portNVIC_SYSTICK_ENABLE		0x00000001

static WORKING_AREA(swaLow, 4096);
static WORKING_AREA(waHigh, 2048);
static WORKING_AREA(waSPI1, 512);

#define EVENTS_LOW_MSGS 10
static msg_t lowBuf[EVENTS_LOW_MSGS];
static EventManager eventsLow;

#define EVENTS_HIGH_MSGS 10
static msg_t highBuf[EVENTS_HIGH_MSGS];
static EventManager eventsHigh;

int main()
{
#ifdef ENV_BOOT
	if(!BootIsBootloaderSelected())
		BootJumpToApplication();

	BootSetBootloaderSelected(0);
#endif

	SetSysClock();

	// Enable important clocks
	RCC_AHBPeriphClockCmd(	RCC_AHBPeriph_DMA1 | RCC_AHBPeriph_DMA2, ENABLE);
	RCC_AHBPeriphClockCmd(	RCC_AHBPeriph_GPIOA | RCC_AHBPeriph_GPIOB | RCC_AHBPeriph_GPIOC | RCC_AHBPeriph_GPIOF, ENABLE);
	RCC_APB2PeriphClockCmd(RCC_APB2Periph_SYSCFG, ENABLE);

	SCB->SHCSR |= SCB_SHCSR_USGFAULTENA_Msk | SCB_SHCSR_BUSFAULTENA_Msk | SCB_SHCSR_MEMFAULTENA_Msk;
	*((volatile uint32_t*)0xE000E008) = 0x02 | 0x08;	// write buffer damaged => disable

	// initialize Data Watch Trace support (instruction cycle counting)
	DWT->CTRL = DWT_CTRL_CYCCNTENA_Msk | DWT_CTRL_NUMCOMP_Msk;

	*(portNVIC_SYSTICK_LOAD) = ( 72000000UL / CH_FREQUENCY ) - 1UL;
	*(portNVIC_SYSTICK_CTRL) = portNVIC_SYSTICK_CLK | portNVIC_SYSTICK_INT | portNVIC_SYSTICK_ENABLE;

	chSysInit();

	NVIC_Configuration();

	CPULoadCalibrate(20);
	chThdSleep(30);

/*	// Configure and start Independent watchdog
	IWDG_WriteAccessCmd(IWDG_WriteAccess_Enable);
	IWDG_SetPrescaler(IWDG_Prescaler_8);
	IWDG_SetReload(2500);	// approx. 500ms
	IWDG_WriteAccessCmd(IWDG_WriteAccess_Disable);
*/

	EventManagerInit(&eventsLow, lowBuf, EVENTS_LOW_MSGS);
	EventManagerInit(&eventsHigh, highBuf, EVENTS_HIGH_MSGS);

	// -- HAL
	LEDInit();
	FlashFSInit(0x0801F800, 0x800);	// 2k
	SPI1Init();
	Intercom1Init(&eventsHigh);
	BootloaderSlaveInit(&intercom1);
#ifdef ENV_RUN
	CLIInit(&eventsLow);
	DriveInit();
//	DriveSetController(&ctrlFocFront, &ctrlFocRear);
	DriveSetController(&ctrl6StepFront, &ctrl6StepRear);
	MotorsInit();
#endif

	// -- Threads
	chThdCreateStatic(waHigh, sizeof(waHigh), NORMALPRIO+1, EventManagerTask, &eventsHigh);
	chThdCreateStatic(swaLow, sizeof(swaLow), NORMALPRIO-1, EventManagerTask, &eventsLow);
	chThdCreateStatic(waSPI1, sizeof(waSPI1), NORMALPRIO+2, SPI1Task, 0);

	// Heartbeat LED
	while(1)
	{
		LEDToggle(LED_MAIN_GREEN);

#ifdef ENV_RUN
		chThdSleep(MS2ST(500));
#else
		chThdSleep(MS2ST(200));
#endif
	}

	return 0;
}

static void NVIC_Configuration()
{
#ifdef ENV_BOOT
	NVIC_SetVectorTable(NVIC_VectTab_FLASH, 0x00);
#else
	NVIC_SetVectorTable(NVIC_VectTab_FLASH, BOOT_APP_OFFSET);
#endif
}

static void SetSysClock(void)
{
	__IO uint32_t HSEStatus = 0;

	/* SYSCLK, HCLK, PCLK2 and PCLK1 configuration -----------*/
	/* Enable HSE */
	RCC->CR |= (RCC_CR_HSEON | RCC_CR_HSEBYP);

	/* Wait till HSE is ready and if Time out is reached exit */
	do
	{
		HSEStatus = RCC->CR & RCC_CR_HSERDY;
	} while (HSEStatus == 0);

	if ((RCC->CR & RCC_CR_HSERDY) != RESET)
	{
		HSEStatus = (uint32_t) 0x01;
	}
	else
	{
		HSEStatus = (uint32_t) 0x00;
	}

	if (HSEStatus == (uint32_t) 0x01)
	{
		/* Enable Prefetch Buffer and set Flash Latency */
		FLASH->ACR = FLASH_ACR_PRFTBE | (uint32_t) FLASH_ACR_LATENCY_1;

		/* HCLK = SYSCLK / 1 */
		RCC->CFGR |= (uint32_t) RCC_CFGR_HPRE_DIV1;

		/* PCLK2 = HCLK / 1 */
		RCC->CFGR |= (uint32_t) RCC_CFGR_PPRE2_DIV1;

		/* PCLK1 = HCLK / 2 */
		RCC->CFGR |= (uint32_t) RCC_CFGR_PPRE1_DIV2;

		/* PLL configuration */
		RCC->CFGR &= (uint32_t) ((uint32_t) ~(RCC_CFGR_PLLSRC
				| RCC_CFGR_PLLXTPRE | RCC_CFGR_PLLMULL));
		RCC->CFGR |= (uint32_t) (RCC_CFGR_PLLSRC_PREDIV1
				| RCC_CFGR_PLLXTPRE_PREDIV1 | RCC_CFGR_PLLMULL12);

		/* Enable PLL */
		RCC->CR |= RCC_CR_PLLON;

		/* Wait till PLL is ready */
		while ((RCC->CR & RCC_CR_PLLRDY) == 0)
		{
		}

		/* Select PLL as system clock source */
		RCC->CFGR &= (uint32_t) ((uint32_t) ~(RCC_CFGR_SW));
		RCC->CFGR |= (uint32_t) RCC_CFGR_SW_PLL;

		/* Wait till PLL is used as system clock source */
		while ((RCC->CFGR & (uint32_t) RCC_CFGR_SWS)
				!= (uint32_t) RCC_CFGR_SWS_PLL)
		{
		}
	}
	else
	{ /* If HSE fails to start-up, the application will have wrong clock
	 configuration. User can add here some code to deal with this error */
	}
}
