/*
 * spi1.c
 *
 *  Created on: 15.01.2014
 *      Author: AndreR
 */

#include "spi1.h"

#include "stm32f30x_gpio.h"
#include "stm32f30x_rcc.h"
#include "stm32f30x_misc.h"
#include "stm32f30x_exti.h"
#include "stm32f30x_syscfg.h"

#include "ch.h"
#include <string.h>

#include "util/bit_banding.h"
#include "util/block_proto.h"
#include "util/log.h"
#include "util/sys_time.h"
#include "intercom1.h"
#include "intercom_constants.h"
#include "motors.h"

SPI1Global spi1;

#define EVENT_RX_DONE	0
#define EVENT_RX_ERROR	1
#define EVENT_SLAVE_DESELECTED 2

#define DMA_Channel_TX DMA1_Channel3
#define DMA_Channel_RX DMA1_Channel2

#define DMAClearTCIE(pDMA) 	BBPeriphReset((uint32_t)&pDMA->CCR, 1)
#define DMASetTCIE(pDMA)	BBPeriphSet((uint32_t)&pDMA->CCR, 1)
#define DMAGetTCIE(pDMA)	BBPeriphGet((uint32_t)&pDMA->CCR, 1)

#define DMAClearTEIE(pDMA) 	BBPeriphReset((uint32_t)&pDMA->CCR, 3)
#define DMASetTEIE(pDMA)	BBPeriphSet((uint32_t)&pDMA->CCR, 3)
#define DMAGetTEIE(pDMA)	BBPeriphGet((uint32_t)&pDMA->CCR, 3)

// DMA RX IRQ
__attribute__((section(".ccmtext"))) void DMA1_Channel2_IRQHandler()
{
	CH_IRQ_PROLOGUE();

	DMAClearTCIE(DMA_Channel_RX);
	DMAClearTEIE(DMA_Channel_RX);

	msg_t event = EVENT_RX_DONE;

    if(DMA1->ISR & DMA1_FLAG_TE2)
    	event = EVENT_RX_ERROR;

    if(SPI1->SR & SPI_FLAG_CRCERR)
    	event = EVENT_RX_ERROR;

    chSysLockFromIsr();
    chMBPostI(&spi1.eventQueue, event);
    chSysUnlockFromIsr();

    CH_IRQ_EPILOGUE();
}

__attribute__((section(".ccmtext"))) void EXTI4_IRQHandler()
{
	CH_IRQ_PROLOGUE();

	EXTI->PR = (1 << 4);

	chSysLockFromIsr();
	chMBPostI(&spi1.eventQueue, EVENT_SLAVE_DESELECTED);
	chSysUnlockFromIsr();

	if(DMA_Channel_RX->CNDTR != 0)
		LogError("desync");

	CH_IRQ_EPILOGUE();
}

void SPI1Init()
{
	// init event queue/variables
	chMBInit(&spi1.eventQueue, spi1.eventQueueData, SPI1_EVENT_QUEUE_SIZE);
	spi1.rxSeq = 0;
	spi1.txSeq = 1;

#ifdef DEBUG
	spi1.pTxDma = DMA1_Channel3;
	spi1.pRxDma = DMA1_Channel2;
	spi1.pSPI = SPI1;
	spi1.pDMA = DMA1;
#endif

	FifoLinInit(&spi1.txFifo, spi1.fifoDataTx, SPI1_FIFO_SIZE_TX);
	FifoLinInit(&spi1.rxFifo, spi1.fifoDataRx, SPI1_FIFO_SIZE_RX);

	RCC_APB2PeriphClockCmd(RCC_APB2Periph_SPI1, ENABLE);

	// init GPIO
	GPIO_InitTypeDef gpioInit;
	gpioInit.GPIO_Mode = GPIO_Mode_AF;
	gpioInit.GPIO_OType = GPIO_OType_PP;
	gpioInit.GPIO_Speed = GPIO_Speed_Level_2;
	gpioInit.GPIO_PuPd = GPIO_PuPd_NOPULL;
	gpioInit.GPIO_Pin = GPIO_Pin_4 | GPIO_Pin_5 | GPIO_Pin_6;
	GPIO_Init(GPIOA, &gpioInit);

	gpioInit.GPIO_Pin = GPIO_Pin_5;
	GPIO_Init(GPIOB, &gpioInit);

	GPIO_PinAFConfig(GPIOA, GPIO_PinSource4, GPIO_AF_5);
	GPIO_PinAFConfig(GPIOA, GPIO_PinSource5, GPIO_AF_5);
	GPIO_PinAFConfig(GPIOA, GPIO_PinSource6, GPIO_AF_5);
	GPIO_PinAFConfig(GPIOB, GPIO_PinSource5, GPIO_AF_5);

	// init SPI1
	SPI_InitTypeDef spiInit;
	spiInit.SPI_BaudRatePrescaler = SPI_BaudRatePrescaler_2;
	spiInit.SPI_CPHA = SPI_CPHA_1Edge;
	spiInit.SPI_CPOL = SPI_CPOL_Low;
	spiInit.SPI_CRCPolynomial = 0x1021;	// CRC16-CCITT
	spiInit.SPI_DataSize = SPI_DataSize_16b;
	spiInit.SPI_Direction = SPI_Direction_2Lines_FullDuplex;
	spiInit.SPI_FirstBit = SPI_FirstBit_MSB;
	spiInit.SPI_Mode = SPI_Mode_Slave;
	spiInit.SPI_NSS = SPI_NSS_Hard;
	SPI_Init(SPI1, &spiInit);

	SPI_CalculateCRC(SPI1, ENABLE);
	SPI_CRCLengthConfig(SPI1, SPI_CRCLength_16b);
	SPI_RxFIFOThresholdConfig(SPI1, SPI_RxFIFOThreshold_HF);

	SPI_I2S_ClearFlag(SPI1, SPI_FLAG_CRCERR);

	// init DMA
	// TX
	// TODO: test circular mode
	DMA_InitTypeDef dmaInit;
	dmaInit.DMA_BufferSize = MOT_EXCHANGE_DATA_MSG_SIZE/2;
	dmaInit.DMA_DIR = DMA_DIR_PeripheralDST;
	dmaInit.DMA_M2M = 0;
	dmaInit.DMA_MemoryBaseAddr = (uint32_t)&spi1.txData;
	dmaInit.DMA_MemoryDataSize = DMA_MemoryDataSize_HalfWord;
	dmaInit.DMA_MemoryInc = DMA_MemoryInc_Enable;
	dmaInit.DMA_Mode = DMA_Mode_Normal;
	dmaInit.DMA_PeripheralBaseAddr = (uint32_t)&SPI1->DR;
	dmaInit.DMA_PeripheralDataSize = DMA_PeripheralDataSize_HalfWord;
	dmaInit.DMA_PeripheralInc = DMA_PeripheralInc_Disable;
	dmaInit.DMA_Priority = DMA_Priority_Low;
	DMA_Init(DMA_Channel_TX, &dmaInit);

	// RX
	dmaInit.DMA_BufferSize = (MOT_EXCHANGE_DATA_MSG_SIZE+2)/2;
	dmaInit.DMA_DIR = DMA_DIR_PeripheralSRC;
	dmaInit.DMA_MemoryBaseAddr = (uint32_t)&spi1.rxData;
	DMA_Init(DMA_Channel_RX, &dmaInit);

	DMA_ITConfig(DMA_Channel_RX, DMA_IT_TC | DMA_IT_TE, ENABLE);

	// configure NVIC
	NVIC_InitTypeDef nvicInit;
	nvicInit.NVIC_IRQChannel = DMA1_Channel2_IRQn;
	nvicInit.NVIC_IRQChannelPreemptionPriority = IRQL_SPI1;
	nvicInit.NVIC_IRQChannelSubPriority = 0;
	nvicInit.NVIC_IRQChannelCmd = ENABLE;
	NVIC_Init(&nvicInit);

	SYSCFG_EXTILineConfig(EXTI_PortSourceGPIOA, EXTI_PinSource4);

	EXTI_InitTypeDef extiInit;
	EXTI_StructInit(&extiInit);
	extiInit.EXTI_Line = EXTI_Line4;
	extiInit.EXTI_LineCmd = ENABLE;
	extiInit.EXTI_Mode = EXTI_Mode_Interrupt;
	extiInit.EXTI_Trigger = EXTI_Trigger_Rising;
	EXTI_Init(&extiInit);

	NVIC_InitTypeDef NVICInit;
	NVICInit.NVIC_IRQChannel = EXTI4_IRQn;
	NVICInit.NVIC_IRQChannelCmd = ENABLE;
	NVICInit.NVIC_IRQChannelPreemptionPriority = IRQL_SPI1;
	NVICInit.NVIC_IRQChannelSubPriority = 0;
	NVIC_Init(&NVICInit);

	SPI_I2S_DMACmd(SPI1, SPI_I2S_DMAReq_Rx, ENABLE);
	SPI_I2S_DMACmd(SPI1, SPI_I2S_DMAReq_Tx, ENABLE);

    // Enable the DMA channels
    DMA_Cmd(DMA_Channel_RX, ENABLE);
    DMA_Cmd(DMA_Channel_TX, ENABLE);

    // Enable the SPI peripheral
    SPI_Cmd(SPI1, ENABLE);
}

msg_t SPI1Task(void* params)
{
	(void)params;

	msg_t event;

	while(1)
	{
		if(chMBFetch(&spi1.eventQueue, &event, MS2ST(500)) != RDY_OK)
			continue;

		spi1.lastEventTime = SysTimeUSec();

		if(event == EVENT_SLAVE_DESELECTED)
		{
			if(spi1.desynced)
			{
				spi1.desynced = 0;
			    SPI_CalculateCRC(SPI1, ENABLE);
			    SPI_I2S_ClearFlag(SPI1, SPI_FLAG_CRCERR);
			    SPI_Cmd(SPI1, ENABLE);
			}

			continue;
		}

		// independent of the event, we need to disable DMA and SPI
		DMA_Cmd(DMA_Channel_RX, DISABLE);
		DMA_Cmd(DMA_Channel_TX, DISABLE);

		while(DMA_Channel_RX->CCR & DMA_CCR_EN);
		while(DMA_Channel_TX->CCR & DMA_CCR_EN);

		DMA_ClearFlag(DMA1_FLAG_GL2 | DMA1_FLAG_TC2 | DMA1_FLAG_HT2 | DMA1_FLAG_TE2);
		DMA_ClearFlag(DMA1_FLAG_GL3 | DMA1_FLAG_TC3 | DMA1_FLAG_HT3 | DMA1_FLAG_TE3);

		SPI_I2S_ClearFlag(SPI1, SPI_FLAG_CRCERR);
		SPI1->DR;
		SPI1->SR;

		SPI_Cmd(SPI1, DISABLE);
		SPI_CalculateCRC(SPI1, DISABLE);

	    while(SPI1->SR & SPI_SR_FRLVL)
	    	SPI1->DR;

		if(event == EVENT_RX_DONE)
		{
			if(*((uint16_t*)spi1.rxData.blockData) == 0)
				LogWarn("blockData == 0");

			uint16_t bytesRead = BlockProtoReadData(&spi1.rxFifo, spi1.rxData.blockData, sizeof(spi1.rxData.blockData));
			if(bytesRead && spi1.callback)
				(*spi1.callback)(spi1.pCbData);

			MotorsRegularIntercom();

			BlockProtoWriteData(&spi1.txFifo, spi1.txData.blockData, sizeof(spi1.txData.blockData));

			// re-enable transfer
			DMA_Channel_RX->CNDTR = (MOT_EXCHANGE_DATA_MSG_SIZE+2)/2;	// include CRC bytes
			DMA_Channel_TX->CNDTR = (MOT_EXCHANGE_DATA_MSG_SIZE+0)/2;
		    DMA_Cmd(DMA_Channel_RX, ENABLE);
		    DMA_Cmd(DMA_Channel_TX, ENABLE);
		    DMASetTCIE(DMA_Channel_RX);
		    DMASetTEIE(DMA_Channel_RX);

		    if(!(GPIOA->IDR & GPIO_Pin_4))
		    {
		    	LogWarn("desync");
		    	spi1.desynced = 1;
		    }
		    else
		    {
				SPI_CalculateCRC(SPI1, ENABLE);
				SPI_I2S_ClearFlag(SPI1, SPI_FLAG_CRCERR);
				SPI_Cmd(SPI1, ENABLE);
		    }
		}

		if(event == EVENT_RX_ERROR)
		{
			LogWarn("EVENT_RX_ERROR");

			// re-enable transfer
		    DMA_Channel_RX->CNDTR = (MOT_EXCHANGE_DATA_MSG_SIZE+2)/2;	// include CRC bytes
			DMA_Channel_TX->CNDTR = (MOT_EXCHANGE_DATA_MSG_SIZE+0)/2;
		    DMA_Cmd(DMA_Channel_RX, ENABLE);
		    DMA_Cmd(DMA_Channel_TX, ENABLE);
		    DMASetTCIE(DMA_Channel_RX);
		    DMASetTEIE(DMA_Channel_RX);

		    if(!(GPIOA->IDR & GPIO_Pin_4))
		    {
		    	LogWarn("desync");
		    	spi1.desynced = 1;
		    }
		    else
		    {
				SPI_CalculateCRC(SPI1, ENABLE);
				SPI_I2S_ClearFlag(SPI1, SPI_FLAG_CRCERR);
				SPI_Cmd(SPI1, ENABLE);
		    }
		}
	}

	return 0;
}
