/*
 * nrf24ex.c
 *
 *  Created on: 29.09.2013
 *      Author: AndreR
 */

#include "nrf24_ex.h"

#include "errors.h"

#include "util/console.h"
#include "util/cobs.h"
#include "util/crc8.h"

#include <string.h>

// commands
#define R_REGISTER			0			// | register address
#define W_REGISTER			0b00100000	// | register address
#define R_RX_PAYLOAD		0b01100001
#define W_TX_PAYLOAD		0b10100000
#define W_TX_PAYLOAD_NO_ACK	0b10110000
#define FLUSH_TX			0b11100001
#define FLUSH_RX			0b11100010
#define REUSE_TX_PL			0b11100011
#define R_RX_PL_WID			0b01100000
#define W_ACK_PAYLOAD		0b10101000	// | PPP = pipe
#define W_TX_PAYLOAD_NO_ACK	0b10110000
#define NOP					0b11111111

// register addresses
#define REG_CONFIG 		0x00
#define REG_EN_AA 		0x01
#define REG_EN_RXADDR	0x02
#define REG_SETUP_AW	0x03
#define REG_SETUP_RETR	0x04
#define REG_RF_CH		0x05
#define REG_RF_SETUP	0x06
#define REG_STATUS		0x07
#define REG_OBSERVE_TX	0x08
#define REG_RPD			0x09
#define REG_ADDR_P0		0x0A
#define REG_ADDR_P1		0x0B
// ...
#define REG_TX_ADDR		0x10
#define REG_RX_PW_P0	0x11
#define REG_RX_PW_P1	0x12
// ...
#define REG_FIFO_STATUS	0x17
#define REG_DYNPD		0x1C
#define REG_FEATURE		0x1D

#define REG_STATUS_RX_DR	0x40
#define REG_STATUS_TX_DS	0x20
#define REG_STATUS_MAX_RT	0x10
#define REG_STATUS_RX_P_NO	0x0E
#define REG_STATUS_TX_FULL	0x01

#define W_CONFIG (W_REGISTER | REG_CONFIG)
#define R_CONFIG (R_REGISTER | REG_CONFIG)
#define W_EN_AA (W_REGISTER | REG_EN_AA)
#define R_EN_AA (R_REGISTER | REG_EN_AA)

// macros
#ifdef STM32F10X_CL
#define CE_LOW(nrf)		(nrf->pCEPort->BRR = nrf->cePin)
#define CE_HIGH(nrf)	(nrf->pCEPort->BSRR = nrf->cePin)
#else
#define CE_LOW(nrf)		(nrf->pCEPort->BSRRH = nrf->cePin)
#define CE_HIGH(nrf)	(nrf->pCEPort->BSRRL = nrf->cePin)
#endif

#define COBS_SIZE COBSMaxStuffedSize(NRF24_MAX_MSG_SIZE)+1

static const uint8_t addresses[] = {
		0xF1, 0xF5, 0xE2, 0xEA,
		0xD3, 0xDC, 0xC4, 0xCF,
		0xB5, 0xB1, 0xA6, 0xAE,
		0x97, 0x92, 0x88, 0x8F,
		0x79, 0x71, 0x6A, 0x63,
		0x5B, 0x54, 0x4C, 0x47,
		0x3D, 0x31, 0x2E, 0x29,
		0x1F, 0x13, 0xFF, 0x00};
static uint8_t tx[36] __attribute__((aligned(4)));
static uint8_t rx[36] __attribute__((aligned(4)));
static uint8_t cobsTx[COBS_SIZE];
static uint8_t cobsRx[COBS_SIZE];

static __inline__ void enable(NRF24Ex* pNRF24)
{
	pNRF24->enabled = 1;
	CE_HIGH(pNRF24);
}

static __inline__ void disable(NRF24Ex* pNRF24)
{
	pNRF24->enabled = 0;
	CE_LOW(pNRF24);
}

static int16_t readRxLength(NRF24Ex* pNRF24, uint8_t* pValue)
{
	tx[0] = R_RX_PL_WID;
	tx[1] = 0;

	int16_t result = SPISyncTransfer(&pNRF24->spi, tx, rx, 2);
	if(result)
		return result;

	pNRF24->lastStatus = rx[0];
	*pValue = rx[1];

	return 0;
}

static int16_t writeSpecialCmd(NRF24Ex* pNRF24, uint8_t cmd)
{
	tx[0] = cmd;

	int16_t result = SPISyncTransfer(&pNRF24->spi, tx, rx, 1);
	if(result)
		return result;

	pNRF24->lastStatus = rx[0];

	return 0;
}

static int16_t writeConfig(NRF24Ex* pNRF24, uint8_t powerUp, uint8_t rx)
{
	return NRF24ExWriteRegister(pNRF24, REG_CONFIG, (powerUp << 1) | rx | 0x08);
}

static int16_t setId(NRF24Ex* pNRF24, uint8_t id)
{
	if(id > NRF24_MAX_BOT_ID)
		id = NRF24_MAX_BOT_ID+1;

	uint8_t address = addresses[id];

	tx[0] = W_REGISTER | REG_ADDR_P0;

	// address is repeated 3 times
	for(uint8_t i = 0; i < 3; i++)
		tx[i+1] = address;

	// set RX address
	int16_t result = SPISyncTransfer(&pNRF24->spi, tx, rx, 4);
	if(result)
		return result;

	// set TX address
	tx[0] = W_REGISTER | REG_TX_ADDR;
	result = SPISyncTransfer(&pNRF24->spi, tx, rx, 4);
	if(result)
		return result;

	pNRF24->lastStatus = rx[0];

	return 0;
}

int16_t NRF24ExWaitForIRQ(NRF24Ex* pNRF24, uint32_t waitTicks)
{
	if(chBSemWaitTimeout(&pNRF24->irqSem, waitTicks) != RDY_OK)
		return ERROR_NRF24EX_TIMEOUT;

	return 0;
}

int16_t NRF24ExChangeId(NRF24Ex* pNRF24, uint8_t id)
{
	if(pNRF24->pQueue)
		pNRF24->pQueue->id = id;

	return setId(pNRF24, id);
}

int16_t NRF24ExInit(NRF24Ex* pNRF24, NRF24ExData* pData)
{
	pNRF24->cePin = pData->cePin;
	pNRF24->pCEPort = pData->pCEPort;
	pNRF24->enabled = 0;
	pNRF24->rx = pData->rx;
	pNRF24->channel = pData->channel;
	pNRF24->maxRt = pData->maxRt;
	pNRF24->ackDelay = ((pData->ackDelayUs/250)-1) << 4;
	pNRF24->speed = pData->speed;
	chBSemInit(&pNRF24->irqSem, FALSE);

	// init CE pin
	GPIO_InitTypeDef gpioInit;
#ifdef STM32F10X_CL
	gpioInit.GPIO_Mode = GPIO_Mode_Out_PP;
#else
	gpioInit.GPIO_Mode = GPIO_Mode_OUT;
	gpioInit.GPIO_OType = GPIO_OType_PP;
	gpioInit.GPIO_PuPd = GPIO_PuPd_NOPULL;
#endif
	gpioInit.GPIO_Speed = GPIO_Speed_2MHz;
	gpioInit.GPIO_Pin = pData->cePin;
	GPIO_Init(pData->pCEPort, &gpioInit);

	CE_LOW(pNRF24);

	// init SPI
	pData->spiData.dmaPrio = 3;	// very high
	pData->spiData.prescaler = SPI_BaudRatePrescaler_8;
	pData->spiData.cpha = 0;
	pData->spiData.cpol = 0;

	SPISyncInit(&pNRF24->spi, &pData->spiData);

	return 0;
}

int16_t NRF24ExSetup(NRF24Ex* pNRF24)
{
	int16_t result;

	// first write is lost on STM32F4, don't know why
	result = NRF24ExWriteRegister(pNRF24, REG_STATUS, 0x70);
	if(result)
		return result;

	result = NRF24ExWriteRegister(pNRF24, REG_FEATURE, 0x07); // enable dynamic packet payload and ack payload and no-ack
	if(result)
		return result;

	result = NRF24ExWriteRegister(pNRF24, REG_EN_RXADDR, 0x01); // enable data pipe 0
	if(result)
		return result;

	result = NRF24ExWriteRegister(pNRF24, REG_DYNPD, 0x01);	// enable dynamic payload length on pipe 0
	if(result)
		return result;

	result = NRF24ExWriteRegister(pNRF24, REG_SETUP_AW, 0x01);
	if(result)
		return result;

	result = setId(pNRF24, 0);
	if(result)
		return result;

	result = NRF24ExWriteRegister(pNRF24, REG_RF_SETUP, 0b00000110 | pNRF24->speed);	// speed, 0dBm output power
	if(result)
		return result;

	result = NRF24ExWriteRegister(pNRF24, REG_RF_CH, pNRF24->channel);
	if(result)
		return result;

	result = NRF24ExWriteRegister(pNRF24, REG_SETUP_RETR, pNRF24->ackDelay | pNRF24->maxRt);	// x ACK wait, y auto-retransmit
	if(result)
		return result;

	result = writeSpecialCmd(pNRF24, FLUSH_RX);
	if(result)
		return result;

	result = writeSpecialCmd(pNRF24, FLUSH_TX);
	if(result)
		return result;

	result = NRF24ExWriteRegister(pNRF24, REG_STATUS, 0x70);
	if(result)
		return result;

	// finally power up
	writeConfig(pNRF24, 1, pNRF24->rx);

	if(pNRF24->rx)
		enable(pNRF24);

	return 0;
}

int16_t NRF24ExCheckChannel(NRF24Ex* pNRF24, uint8_t* pRPD)
{
	int16_t result;
	uint8_t wasEnabled = pNRF24->enabled;

	if(pNRF24->enabled)
		disable(pNRF24);

	result = writeConfig(pNRF24, 0, pNRF24->rx);	// power down
	if(result)
		return result;

	chThdSleep(MS2ST(2));

	result = writeConfig(pNRF24, 1, 1);	// power up as rx
	if(result)
		return result;

	uint8_t rpd;
	*pRPD = 0;

	for(uint8_t i = 0; i < 100; i++)
	{
		enable(pNRF24);

		chThdSleep(MS2ST(2));

		disable(pNRF24);

		result = NRF24ExReadRegister(pNRF24, REG_RPD, &rpd);
		if(result)
			return result;

		if(rpd & 0x01)
			(*pRPD)++;
	}

	result = writeConfig(pNRF24, 0, pNRF24->rx);	// power down
	if(result)
		return result;

	// power up
	result = writeConfig(pNRF24, 1, pNRF24->rx);
	if(result)
		return result;

	// power up delay
	chThdSleep(MS2ST(20));

	if(wasEnabled && pNRF24->rx)
		enable(pNRF24);

	return 0;
}

int16_t NRF24ExChangeChannel(NRF24Ex* pNRF24, uint8_t channel)
{
	int16_t result = 0;

	if(channel > 127)
	{
		return ERROR_NRF24EX_INVALID_CHANNEL;
	}

	pNRF24->channel = channel;

	// set channel
	result = NRF24ExWriteRegister(pNRF24, REG_RF_CH, channel);
	if(result)
		return result;

	return result;
}

int16_t NRF24ExChangeSpeed(NRF24Ex* pNRF24, uint8_t speed)
{
	int16_t result = 0;

	pNRF24->speed = speed;

	// set speed
	result = NRF24ExWriteRegister(pNRF24, REG_RF_SETUP, 0b00000110 | speed);
	if(result)
		return result;

	return result;
}

int16_t NRF24ExChangeRetransmission(NRF24Ex* pNRF24, uint8_t maxRt, uint16_t delayUs)
{
	int16_t result = 0;

	if(delayUs%250 != 0)
		return ERROR_INVALID_PARAMETER;

	pNRF24->ackDelay = ((delayUs/250)-1) << 4;
	pNRF24->maxRt = maxRt;

	// set max rt and ack delay
	result = NRF24ExWriteRegister(pNRF24, REG_SETUP_RETR, pNRF24->ackDelay | pNRF24->maxRt);
	if(result)
		return result;

	return result;

}

// IRQL: wifi handler
void NRF24ExIRQ(NRF24Ex* pNRF24)
{
	if(!pNRF24->enabled)
		return;

	chSysLockFromIsr();
	chBSemSignalI(&pNRF24->irqSem);
	chSysUnlockFromIsr();
}

int16_t NRF24ExReadRegister(NRF24Ex* pNRF24, uint8_t reg, uint8_t* pValue)
{
	tx[0] = R_REGISTER | reg;
	tx[1] = 0;

	int16_t result = SPISyncTransfer(&pNRF24->spi, tx, rx, 2);
	if(result)
		return result;

	pNRF24->lastStatus = rx[0];
	*pValue = rx[1];

	return 0;
}

int16_t NRF24ExWriteRegister(NRF24Ex* pNRF24, uint8_t reg, uint8_t value)
{
	tx[0] = W_REGISTER | reg;
	tx[1] = value;

	int16_t result = SPISyncTransfer(&pNRF24->spi, tx, rx, 2);
	if(result)
		return result;

	pNRF24->lastStatus = rx[0];

	return 0;
}

int16_t NRF24ExRxDownload(NRF24Ex* pNRF24)
{
	uint8_t rxLen;
	int16_t result;
	uint8_t status;

	result = NRF24ExWriteRegister(pNRF24, REG_STATUS, 0x70);	// clear IRQ sources
	if(result)
		return result;

	while(1)
	{
		result = readRxLength(pNRF24, &rxLen);
		if(result)
		{
			ConsolePrint("readRxLength error'd: %hd\r\n", result);
			return result;
		}
		status = pNRF24->lastStatus;

		// RX fifo empty?
		if((status & REG_STATUS_RX_P_NO) == REG_STATUS_RX_P_NO)
			break;

		if(rxLen > 32)
		{
			result = writeSpecialCmd(pNRF24, FLUSH_RX);
			ConsolePrint("rxLen>32 -> transmission error\r\n");
			if(result)
				return result;
		}
		else if(rxLen > 0)
		{
			tx[0] = R_RX_PAYLOAD;
			result = SPISyncTransfer(&pNRF24->spi, tx, rx, rxLen+1);
			if(result)
				return result;

			chMtxLock(&pNRF24->pQueue->rxMutex);
			{
//				if(rxLen > 1)
//				{
//					ConsolePrint("Recv(%hhu/%hhu): ", rxLen, pNRF24->pQueue->id);
//					for (uint8_t* d = rx + 1; d < rx + 1 + rxLen; d++) {
//						ConsolePrint("0x%02hhx ", *d);
//					}
//					ConsolePrint("\r\n");
//				}
				result = FifoDelimPush(&pNRF24->pQueue->rxFifo, rx + 1, rxLen);
				if(result)
					return result;
			}
			chMtxUnlock();
		}
	}

	return 0;
}

int16_t NRF24ExRxUpload(NRF24Ex* pNRF24)
{
	uint32_t size = 0;
	int16_t result = 0;

	while((pNRF24->lastStatus & REG_STATUS_TX_FULL) == 0)
	{
		chMtxLock(&pNRF24->pQueue->txMutex);
		{
//			size = FifoGetBytesAvailable(&pNRF24->pQueue->txFifo.fifo);
			result = FifoDelimFindNextDelimiter(&pNRF24->pQueue->txFifo, &size);
			size++;
//			if(size == 0)
			if(result == ERROR_FIFO_CHAR_NOT_FOUND)
			{
				chMtxUnlock();
				return 0;
			}
//			if(result)
//			{
//				ConsolePrint("FifoDelimFindNextDelimiter error'd: %hX\r\n", result);
//				return result;
//			}
			if(size > NRF24_MAX_LENGTH)
				size = NRF24_MAX_LENGTH;

			result = FifoDelimPop(&pNRF24->pQueue->txFifo, tx+1, size);
			if(result) 
			{
				chMtxUnlock();
				return result;
			}
		}
		chMtxUnlock();

//		ConsolePrint("Send(%u/%hhu): ", size, pNRF24->pQueue->id);
//		for (uint8_t* d = tx + 1; d < tx + 1 + size; d++) {
//			ConsolePrint("0x%02hhx ", *d);
//		}
//		ConsolePrint("\r\n");

		tx[0] = W_ACK_PAYLOAD;
		result = SPISyncTransfer(&pNRF24->spi, tx, rx, size+1);
		if(result)
			return result;

		tx[0] = NOP;
		result = SPISyncTransfer(&pNRF24->spi, tx, rx, 1);
		if(result)
			return result;

		pNRF24->lastStatus = rx[0];
	}

	return 0;
}

int16_t NRF24ExTxProcessOne(NRF24Ex* pNRF24)
{
	int16_t retResult = 0;
	uint8_t rxLen;
	uint32_t size;
	int16_t result;

	chMtxLock(&pNRF24->pQueue->txMutex);
	tx[0] = W_TX_PAYLOAD;

	size = FifoGetBytesAvailable(&pNRF24->pQueue->txFifo.fifo);
	if(size == 0)
	{
		// transmit empty package
		chMtxUnlock();	// txMutex
		tx[1] = 0;
		result = SPISyncTransfer(&pNRF24->spi, tx, 0, 2);
		if (result)
			return result;
	}
	else
	{
		if (size > NRF24_MAX_LENGTH)
			size = NRF24_MAX_LENGTH;

		result = FifoDelimPop(&pNRF24->pQueue->txFifo, tx + 1, size);
		chMtxUnlock();	// txMutex
		if (result)
			return result;

		result = SPISyncTransfer(&pNRF24->spi, tx, 0, size + 1);
		if (result)
			return result;
	}

	chBSemWaitTimeout(&pNRF24->irqSem, TIME_IMMEDIATE);	// clear semaphore

	enable(pNRF24);

	if(NRF24ExWaitForIRQ(pNRF24, MS2ST(20)))
	{
		disable(pNRF24);
		return ERROR_NRF24EX_IRQ_MALFUNCTION;
	}

	disable(pNRF24);

	result = readRxLength(pNRF24, &rxLen);
	if(result)
		return result;

	uint8_t status = pNRF24->lastStatus;

	if (status & REG_STATUS_RX_DR) {
		if (rxLen > 32) {
			ConsolePrint("rxLen>32 -> transmission error!\r\n");
			result = writeSpecialCmd(pNRF24, FLUSH_RX);
			if(result)
				return result;
		}
		else if(rxLen > 0)
		{
			tx[0] = R_RX_PAYLOAD;
			result = SPISyncTransfer(&pNRF24->spi, tx, rx, rxLen+1);
			if(result)
				return result;

			chMtxLock(&pNRF24->pQueue->rxMutex);
			{
				result = FifoDelimPush(&pNRF24->pQueue->rxFifo, rx + 1, rxLen);
			}
			chMtxUnlock();

			if (result) {
				pNRF24->pQueue->stats.rxPacketsOverrun++;
			}
			else
			{
				if(pNRF24->pNewDataSem)
					chBSemSignal(pNRF24->pNewDataSem);
			}
		}
	}

	if(status & REG_STATUS_TX_DS)
	{
		pNRF24->pQueue->stats.txPacketsAcked++;
	}

	if(status & REG_STATUS_MAX_RT)
	{
		pNRF24->pQueue->stats.txPacketsMaxRT++;

		result = writeSpecialCmd(pNRF24, FLUSH_TX);
		if(result)
			return result;

		retResult = ERROR_NRF24EX_TIMEOUT;
	}

	result = NRF24ExWriteRegister(pNRF24, REG_STATUS, 0x70);// clear IRQ sources
	if (result)
		return result;

	return retResult;
}

void NRF24ExSetEventSink(NRF24Ex* pNRF24, BinarySemaphore* pEventSink)
{
	pNRF24->pNewDataSem = pEventSink;
}

void NRF24ExUseQueue(NRF24Ex* pNRF24, NRF24Queue* pQueue)
{
	pNRF24->pQueue = pQueue;

	setId(pNRF24, pQueue->id);
}

void NRF24QueueInit(NRF24Queue* pQueue, uint8_t* pDataTx, uint16_t txSize, uint8_t* pDataRx, uint16_t rxSize)
{
	FifoDelimInit(&pQueue->rxFifo, pDataRx, rxSize, 0);
	FifoDelimInit(&pQueue->txFifo, pDataTx, txSize, 0);

	chMtxInit(&pQueue->txMutex);
	chMtxInit(&pQueue->rxMutex);

	memset(&pQueue->stats, 0, sizeof(pQueue->stats));
}

int16_t NRF24QueueGetPacket(NRF24Queue* pQueue, uint8_t* pDst, uint32_t dstSize, uint32_t* pPktSize)
{
	int16_t result;
	uint32_t pktSize;
	*pPktSize = 0;

	chMtxLock(&pQueue->rxMutex);
	{
		if(pQueue->rxFifo.numDelimiters == 0)
		{
			chMtxUnlock();	// rxMutex
			return ERROR_NRF24EX_QUEUE_EMPTY;
		}

		result = FifoDelimFindNextDelimiter(&pQueue->rxFifo, &pktSize);
		if(result == ERROR_FIFO_CHAR_NOT_FOUND)
		{
			return 0;
		}
		if(result)
		{
			chMtxUnlock();	// rxMutex
			ConsolePrint("FifoDelimFindNextDelimiter error'd: %hX\r\n", result);
			return result;
		}

		if(pktSize+1 > COBS_SIZE)
		{
			FifoDelimDrop(&pQueue->rxFifo, pktSize+1, 1);
			chMtxUnlock();	// rxMutex

			ConsolePrint("Not enough memory: %u > %d\r\n", (pktSize + 1), (int32_t)(COBS_SIZE));
			pQueue->stats.rxPacketsMalformed++;
			return ERROR_NOT_ENOUGH_MEMORY;
		}

		result = FifoDelimPop(&pQueue->rxFifo, cobsRx, pktSize+1);
	}
	chMtxUnlock();	// rxMutex

	if(pktSize == 0)
	{
		// got an empty package, nothing to do
		return 0;
	}

	result = COBSDecode(cobsRx, pktSize, pDst, dstSize, pPktSize);
	if(result)
	{
		ConsolePrint("COBSDecode error'd: %hX\r\n", result);
		return result;
	}

	uint8_t crcResult = Crc8(pDst, *pPktSize);
	if(crcResult != 0xF9)
	{
		ConsolePrint("CRC8 mismatch 0x%02hhX\r\n", crcResult);
		ConsolePrint("pktSize (COBS/orig): %u/%u\r\n", pktSize, *pPktSize);
		ConsolePrint("COBS: ");
		for(uint32_t i = 0; i < pktSize; i++)
			ConsolePrint("0x%02hhX ", cobsRx[i]);

		ConsolePrint("\r\nOrig: ");
		for(uint32_t i = 0; i < *pPktSize; i++)
			ConsolePrint("0x%02hhX ", pDst[i]);
		ConsolePrint("\r\n\r\n");

		--*pPktSize;

		pQueue->stats.rxPacketsMalformed++;
		return ERROR_NRF24EX_BAD_CHECKSUM;
	}

	--*pPktSize;

	pQueue->stats.rxPacketsRecv++;
	pQueue->stats.rxBytesRecvCOBS += pktSize+1;
	pQueue->stats.rxBytesRecvRaw += *pPktSize;

	return 0;
}

/**
 * Put data in an NRF24 queue.
 *
 * @param pQueue the queue
 * @param pData data to send
 * @param dataSize datasize
 * @return 0 on success, errorcode otherwise
 */
int16_t NRF24QueueSend(NRF24Queue* pQueue, uint8_t* pData, uint32_t dataSize)
{
	uint32_t encodedSize;

	pData[dataSize] = Crc8(pData, dataSize);
	++dataSize;

	int16_t result = COBSEncode(pData, dataSize, cobsTx, COBS_SIZE, &encodedSize);

//	COBSState state;
//	int16_t result = COBSStartEncode(&state, dataSize+1, cobsTx, COBS_SIZE);	// +1 for CRC8
	if(result)
	{
		pQueue->stats.txPacketsOverrun++;
		return result;
	}

//	for(uint32_t i = 0; i < dataSize; i++)
//	{
//		COBSFeedEncode(&state, pData[i]);
//	}
//
//	uint8_t crc8 = Crc8(pData, dataSize);
//	COBSFeedEncode(&state, crc8);
//
//	COBSFinalizeEncode(&state, &encodedSize);
//
//	cobsTx[encodedSize++] = 0;	// packet delimiter
	cobsTx[encodedSize] = 0;	// packet delimiter

	if(encodedSize > COBS_SIZE)
	{
		pQueue->stats.txPacketsOverrun++;
		return ERROR_NOT_ENOUGH_MEMORY;
	}

	chMtxLock(&pQueue->txMutex);
	{
		result = FifoDelimPush(&pQueue->txFifo, cobsTx, encodedSize+1);
		if(result)
		{
			pQueue->stats.txPacketsOverrun++;
			chMtxUnlock();	// txMutex
			return result;
		}
	}
	chMtxUnlock();	// txMutex

	pQueue->stats.txPacketsSent++;
	pQueue->stats.txBytesSentRaw += dataSize;
	pQueue->stats.txBytesSentCOBS += encodedSize+1;

	return 0;
}

void NRF24QueueClear(NRF24Queue* pQueue)
{
	chMtxLock(&pQueue->txMutex);
	{
		FifoDelimClear(&pQueue->txFifo);
	}
	chMtxUnlock();	// txMutex

	chMtxLock(&pQueue->rxMutex);
	{
		FifoDelimClear(&pQueue->rxFifo);
	}
	chMtxUnlock();	// rxMutex
}
