/*
 * nrf24_ex.h
 *
 *  Created on: 29.09.2013
 *      Author: AndreR
 *
 * NRF24 extended (v2) interface.
 */

#ifndef NRF24_EX_H_
#define NRF24_EX_H_

#ifdef STM32F10X_CL
#include "stm32f10x.h"
#include "stm32f10x_gpio.h"
#endif
#ifdef STM32F4XX
#include "stm32f4xx.h"
#include "stm32f4xx_gpio.h"
#endif
#ifdef STM32F30X
#include "stm32f30x.h"
#include "stm32f30x_gpio.h"
#endif

#include "util/spi_sync.h"
#include "util/fifo.h"
#include "util/fifo_delim.h"

#include "ch.h"

#define NRF24_MAX_LENGTH	32
#define NRF24_MAX_BOT_ID	23
#define NRF24_MAX_MSG_SIZE	280

#define NRF24_SPEED_250K	0x20
#define NRF24_SPEED_1M		0x00
#define NRF24_SPEED_2M		0x08

struct _NRF24EX;

typedef struct _NRF24ExStats
{
	uint32_t rxPacketsRecv;			// correctly received user packets
	uint32_t rxPacketsMalformed;	// malformed packets (not lost due to buffer orverrun)

	uint32_t rxBytesRecvCOBS;
	uint32_t rxBytesRecvRaw;

	uint32_t txPacketsSent;
	uint32_t txPacketsOverrun;		// packets too big or tx buffer overrun

	uint32_t txBytesSentCOBS;
	uint32_t txBytesSentRaw;

	// nRF24 packet related
	uint32_t rxPacketsOverrun;		// packets lost due to rx buffer overrun (nRF24 packets)

	// only used by TX nRF24 (= base station)
	uint32_t txPacketsMaxRT;
	uint32_t txPacketsAcked;
} NRF24QueueStats;

typedef struct _NRF24Queue
{
	FifoDelim rxFifo;
	FifoDelim txFifo;

	Mutex txMutex;
	Mutex rxMutex;

	NRF24QueueStats stats;

	uint8_t id;
} NRF24Queue;

void NRF24QueueInit(NRF24Queue* pQueue, uint8_t* pDataTx, uint16_t txSize, uint8_t* pDataRx, uint16_t rxSize);

typedef struct _NRF24Ex
{
	GPIO_TypeDef* pCEPort; // one of GPIOx
	uint16_t cePin; // one of GPIO_Pin_x
	SPISync spi;

	uint8_t enabled;
	uint8_t rx;
	uint8_t channel;
	uint8_t maxRt;
	uint8_t speed;
	uint8_t ackDelay;

	volatile uint8_t lastStatus;			// status response of any command

	NRF24Queue* pQueue;

	BinarySemaphore irqSem;
	BinarySemaphore* pNewDataSem;
} NRF24Ex;

typedef struct _NRF24ExData
{
	// SPI exclusive config
	SPISyncData spiData;

	GPIO_TypeDef* pCEPort;
	uint16_t cePin;

	// RF
	uint8_t channel;
	uint8_t maxRt;
	uint8_t speed;
	uint16_t ackDelayUs;

	uint8_t rx;
} NRF24ExData;

/**
 * Channel and address are initialized to defaults.
 * Use ChangeAddress and ChangeChannel after init.
 */
int16_t NRF24ExInit(NRF24Ex* pNRF24, NRF24ExData* pData);
int16_t NRF24ExSetup(NRF24Ex* pNRF24);

void NRF24ExIRQ(NRF24Ex* pNRF24);
int16_t NRF24ExWaitForIRQ(NRF24Ex* pNRF24, uint32_t waitTicks);
void NRF24ExSetEventSink(NRF24Ex* pNRF24, BinarySemaphore* pEventSink);

int16_t NRF24ExRxDownload(NRF24Ex* pNRF24);
int16_t NRF24ExRxUpload(NRF24Ex* pNRF24);
int16_t NRF24ExTxProcessOne(NRF24Ex* pNRF24);

int16_t NRF24ExChangeId(NRF24Ex* pNRF24, uint8_t id);
int16_t NRF24ExChangeChannel(NRF24Ex* pNRF24, uint8_t channel);
int16_t NRF24ExChangeSpeed(NRF24Ex* pNRF24, uint8_t speed);
int16_t NRF24ExChangeRetransmission(NRF24Ex* pNRF24, uint8_t maxRt, uint16_t delayUs);
int16_t NRF24ExCheckChannel(NRF24Ex* pNRF24, uint8_t* pRPD);

int16_t NRF24ExReadRegister(NRF24Ex* pNRF24, uint8_t reg, uint8_t* pValue);
int16_t NRF24ExWriteRegister(NRF24Ex* pNRF24, uint8_t reg, uint8_t value);
void NRF24ExUseQueue(NRF24Ex* pNRF24, NRF24Queue* pQueue);

void NRF24QueueInit(NRF24Queue* pQueue, uint8_t* pDataTx, uint16_t txSize, uint8_t* pDataRx, uint16_t rxSize);
/**
 * Put data to tx queue.
 * Data is not transmitted immediately. Call NRF24Process to send data.
 */
int16_t NRF24QueueSend(NRF24Queue* pNRF24, uint8_t* pData, uint32_t dataSize);
int16_t NRF24QueueGetPacket(NRF24Queue* pNRF24, uint8_t* pDst, uint32_t dstSize, uint32_t* pPktSize);
void NRF24QueueClear(NRF24Queue* pQueue);

#endif /* NRF24_H_ */
