/*
 * flash_fs.c
 *
 *  Created on: 25.06.2014
 *      Author: AndreR
 */

#include "flash_fs.h"
#include "util/flash.h"
#include "util/console.h"
#include <string.h>

FlashFS flashFS;

void FlashFSInit(uint32_t flashAddr, uint32_t flashSize)
{
	chMtxInit(&flashFS.writeMutex);

	flashFS.bytesDel = 0;
	flashFS.bytesUsed = 0;
	flashFS.flashAddr = flashAddr;
	flashFS.flashSize = flashSize;
	flashFS.pWrite = (uint8_t*)flashAddr;

	FlashFile* pFH = (FlashFile*)flashFS.pWrite;

	while(pFH->dataSize != 0xFFFF)
	{
		uint32_t blockSize = pFH->dataSize + sizeof(FlashFile);

		if(pFH->del != 0xFFFFFFFF)
			flashFS.bytesDel += blockSize;
		else
			flashFS.bytesUsed += blockSize;

		flashFS.pWrite += blockSize;

		pFH = (FlashFile*)flashFS.pWrite;
	}

	FlashFSCompact();
}

FlashFile* FlashFSOpen(const char* pFilename, uint16_t version)
{
	uint8_t* pRead = (uint8_t*)flashFS.flashAddr;
	FlashFile* pFH = (FlashFile*)pRead;

	while(pFH->dataSize != 0xFFFF)
	{
		uint32_t blockSize = pFH->dataSize + sizeof(FlashFile);

		if(pFH->del == 0xFFFFFFFF && strncmp(pFilename, pFH->filename, FLASH_FILENAME_LENGTH) == 0 && pFH->version == version)
		{
			return pFH;
		}

		pRead += blockSize;
		pFH = (FlashFile*)pRead;
	}

	return 0;
}

FlashFile* FlashFSCreate(const char* pFilename, uint16_t version, uint16_t size)
{
	uint32_t nul = 0;
	uint8_t* pRead = (uint8_t*)flashFS.flashAddr;
	FlashFile* pFH = (FlashFile*)pRead;

	chMtxLock(&flashFS.writeMutex);

	while(pFH->dataSize != 0xFFFF)
	{
		uint32_t blockSize = pFH->dataSize + sizeof(FlashFile);

		if(pFH->del == 0xFFFFFFFF && strncmp(pFilename, pFH->filename, FLASH_FILENAME_LENGTH) == 0)
		{
			FlashProgram((uint32_t)&pFH->del, &nul, 1);
			flashFS.bytesDel += blockSize;
			flashFS.bytesUsed -= blockSize;
		}

		pRead += blockSize;
		pFH = (FlashFile*)pRead;
	}

	FlashFile header;
	header.dataSize = size;
	header.del = 0xFFFFFFFF;
	memcpy(header.filename, pFilename, strlen(pFilename)+1);
	header.version = version;

	FlashProgram((uint32_t)flashFS.pWrite, (uint32_t*)&header, sizeof(FlashFile)/4);

	pFH = (FlashFile*)flashFS.pWrite;

	flashFS.pWrite += size + sizeof(FlashFile);
	flashFS.bytesUsed += size + sizeof(FlashFile);

	chMtxUnlock();

	return pFH;
}

void FlashFSRead(FlashFile* pFH, void* pDst)
{
	uint8_t* pRead = (uint8_t*)pFH;
	pRead += sizeof(FlashFile);

	memcpy(pDst, pRead, pFH->dataSize);
}

int16_t FlashFSOpenOrCreate(const char* pFilename, uint16_t version, void* pData, uint16_t size, FlashFile** ppFH)
{
	*ppFH = FlashFSOpen(pFilename, version);
	if(!*ppFH)
	{
		*ppFH = FlashFSCreate(pFilename, version, size);
		if(!*ppFH)
			return FLASH_FS_NOMEM;

		return FLASH_FS_CREATED;
	}
	else
	{
		FlashFSRead(*ppFH, pData);
		return FLASH_FS_OPEN;
	}
}

void FlashFSWrite(FlashFile** ppFH, void* pData, uint16_t size)
{
	uint32_t nul = 0;
	FlashFile* pFH = *ppFH;
	if(pFH->dataSize != size || size%4 != 0)
		return;

	if(sizeof(FlashFile) + size + flashFS.bytesDel + flashFS.bytesUsed > flashFS.flashSize)
		FlashFSCompact();

	chMtxLock(&flashFS.writeMutex);

	*ppFH = (FlashFile*)flashFS.pWrite;

	FlashProgram((uint32_t)flashFS.pWrite, (uint32_t*)pFH, sizeof(FlashFile)/4);
	flashFS.pWrite += sizeof(FlashFile);

	FlashProgram((uint32_t)&pFH->del, &nul, 1);

	FlashProgram((uint32_t)flashFS.pWrite, (uint32_t*)pData, size/4);
	flashFS.pWrite += size;

	flashFS.bytesDel += sizeof(FlashFile)+size;

	chMtxUnlock();
}

void FlashFSDelete(const char* pFilename)
{
	uint32_t nul = 0;
	uint8_t* pRead = (uint8_t*)flashFS.flashAddr;
	FlashFile* pFH = (FlashFile*)pRead;

	chMtxLock(&flashFS.writeMutex);

	while(pFH->dataSize != 0xFFFF)
	{
		uint32_t blockSize = pFH->dataSize + sizeof(FlashFile);

		if(pFH->del == 0xFFFFFFFF && strncmp(pFilename, pFH->filename, FLASH_FILENAME_LENGTH) == 0)
		{
			FlashProgram((uint32_t)&pFH->del, &nul, 1);
			flashFS.bytesDel += blockSize;
			flashFS.bytesUsed -= blockSize;
		}

		pRead += blockSize;
		pFH = (FlashFile*)pRead;
	}

	chMtxUnlock();
}

int16_t FlashFSCompact()
{
	if(flashFS.bytesUsed > FLASH_TEMP_SIZE)
		return 1;

	if(flashFS.bytesDel + flashFS.bytesUsed < flashFS.flashSize/2)
		return 0;

	chMtxLock(&flashFS.writeMutex);

	uint8_t* pTempWrite = flashFS.temp;

	uint8_t* pRead = (uint8_t*)flashFS.flashAddr;
	FlashFile* pFH = (FlashFile*)pRead;

	while(pFH->dataSize != 0xFFFF)
	{
		uint32_t blockSize = pFH->dataSize + sizeof(FlashFile);

		if(pFH->del == 0xFFFFFFFF)
		{
			memcpy(pTempWrite, pRead, blockSize);
			pTempWrite += blockSize;
		}

		pRead += blockSize;
		pFH = (FlashFile*)pRead;
	}

	FlashErase(flashFS.flashAddr, flashFS.flashAddr+flashFS.flashSize);

	FlashProgram(flashFS.flashAddr, (uint32_t*)flashFS.temp, flashFS.bytesUsed/4);

	flashFS.bytesDel = 0;
	flashFS.pWrite = (uint8_t*)(flashFS.flashAddr+flashFS.bytesUsed);

	chMtxUnlock();

	return 0;
}

void FlashFSListFiles()
{
	uint8_t* pRead = (uint8_t*)flashFS.flashAddr;
	FlashFile* pFH = (FlashFile*)pRead;

	ConsolePrint("Flash FS\r\n");
	ConsolePrint("Bytes used: %u\r\n", flashFS.bytesUsed);
	ConsolePrint("Bytes del: %u\r\n", flashFS.bytesDel);
	ConsolePrint("Size: %u\r\n", flashFS.flashSize);
	ConsolePrint("Address: 0x%08X\r\n", flashFS.flashAddr);
	ConsolePrint("pWrite:  0x%08X\r\n\r\n", (uint32_t)flashFS.pWrite);

	ConsolePrint("File\t\tVersion\tSize\r\n");

	chThdSleep(MS2ST(5));

	while(pFH->dataSize != 0xFFFF)
	{
		uint32_t blockSize = pFH->dataSize + sizeof(FlashFile);

		if(pFH->del == 0xFFFFFFFF)
		{
			ConsolePrint("%-16s\t%3hu\t%8hu\t%X\r\n", pFH->filename, pFH->version, pFH->dataSize, pFH->del);
		}

		pRead += blockSize;
		pFH = (FlashFile*)pRead;
	}
}
